/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* PUGH_makestack.c */
/* changed and extended by Bernd Bruegmann, 9/97, from: */ 

 /*@@
   @file      PPMG_MakeStack.c
   @date      Sun Aug 24 13:24:33 1997
   @author    Paul Walker
   @desc 
   Makes the multi grid "stack", eg, collection of nested GH's.
   @enddesc 
 @@*/

#include <string.h>

#include "cctk_Flesh.h"
#include "cctk_GHExtensions.h"
#include "bbmg.h"
#include "CactusPUGH/PUGH/src/include/pughi.h"




 /*@@
   @routine    PPMG_MakeStack
   @date       Sun Aug 24 13:24:51 1997
   @author     Paul Walker
   @desc 
   Makes the multigrid "stack", eg, collection of nested GH's. It
   makes them as small as it can get unless ppmg_nlevs is set then
   it makes them as small as it can get or ppmg_nlevs whichever is
   the larger. You know what I mean. There is a limit on smallness
   placed by the requirement that data is processor block aligned.
   @enddesc 
   @calledby   PPMG_Main
   @calls      MakeGridFuncs, PUGH_EnableGADataStorage
@@*/

#include "bbmg.h"


int fieldstorage[MGFMAX];
void EnableGFDS(pGH *GH, int n, int alwayssetzero);




/* Make multigrid stack
   In PUGH, the GH defines multigrid level 0.
   Positive mglevels are introduced for the coarsenings.
   Negative mglevels for finer levels in the BIB stack. 
   Only GHs that have the same physical time as GH are used. 
   On exit, all mglevels are set back to 0.
   In BAM (see PUGH_to_BAM.c), the GH defines the level "ltop". 
   1 to ltop-1 are added as multigrid stack with coarser levels.
   ltop+1 to lmax are added from BIB, where lmax is
   the finest level that is aligned in time with GH.
   Here is an example with 4 BIB levels, 4 MG coarsenings, ltop=5, lmax=8 
   PUGH   4  3  2  1  0 -1 -2 -3
   BAM    1  2  3  4  5  6  7  8
*/

void PUGH_makestack(cGH *cctkGH, pGH *GH, int *nf, int n, int nws) 
{
  DECLARE_CCTK_PARAMETERS
  cGroup pgroup;
  int group;
  pGH *GHc=NULL, *GHl=NULL;                /* Current and lower */
  int fac, i, nst;
  int nxl, nyl, nzl, nl[3];
  int min;
  int l, cl;
  int bam_wsmin_index = 0;
  pGExtras *extras;
  pConnectivity *connectivity;
  int perme [3];

  if (0) CCTK_INFO("PUGH_makestack"); 
  if (octant && vb) 
    CCTK_INFO("octant?! are you sure you don't want true 3d?");

  /* tempary fix: fill in coordinate information that Cactus 4.0 no 
     longer puts here 
  */
#if 0
  /* FIXME: origin is gone in pGH */
  CCTK_CoordRange(cctkGH, &GH->cx0, &dummy, -1, "x", "cart3d");
  CCTK_CoordRange(cctkGH, &GH->cy0, &dummy, -1, "y", "cart3d");
  CCTK_CoordRange(cctkGH, &GH->cz0, &dummy, -1, "z", "cart3d");
#endif
  GH->dx0 = cctkGH->cctk_delta_space[0];
  GH->dy0 = cctkGH->cctk_delta_space[1];
  GH->dz0 = cctkGH->cctk_delta_space[2];
  /* lx0, ... appears to be unused */

  /* enable missing fields */
  cl = GH->convlevel;
  for (l = GH->level;; l++) {
    GHc = GHbyLevel(cl, l, 0);
    if (!GHc) break;
    /*if (!dequal(GHc->phys_time, GH->phys_time)) break;*/
    if (0) 
      printf("GHc = %d %d %d = %p, dx = %6.3f\n", cl, l, 0, GHc, GHc->dx0);

    /* additional problem dependent fields may not be there */
    /* remember entry status so that we can restore it when stack destroyed */
    if (n > MGFMAX) errorexit("PUGH_makestack(): make static array bigger");
    for (i = 0; i < n; i++) {
      fieldstorage[i] = GFByIndex(GHc, nf[i])->storage;
      EnableGFDS(GHc, nf[i], 0);
    }

    /* top level always needs those fields */
    bam_wsmin_index =  CCTK_VarIndex("bam_elliptic::bam_wsmin");
    for (i = 0; i < nws; i++) 
      EnableGFDS(GHc, bam_wsmin_index+i, 1);

    /* count down for higher refinements */
    GHc->mglevel = GH->level - l;
  }

  /* init top level pars */
  min = gridminsize;
  if (0) {
    /* FIXME: we assume 3D grid functions here */
    extras = GH->GFExtras [2];
    printf ("    finest grid %3dx%3dx%3d  (min=%d)\n", 
            extras->nsize[0], extras->nsize[1], extras->nsize[2], min);
  }

  nst = gridmaxlevels;
  GHc = GH;
  fac = 1;
  /* FIXME: we assume 3D grid functions here */
  extras = GHc->GFExtras [2];
  connectivity = GHc->Connectivity [2];

  /* coarsen */
  while (GHc && 
        (((extras->nsize[0]-ox) % 2 == 0 && (extras->nsize[0]-ox)/2+ox >= min) &&
         ((extras->nsize[1]-oy) % 2 == 0 && (extras->nsize[1]-oy)/2+oy >= min)
          || cartoon) &&
         ((extras->nsize[2]-oz) % 2 == 0 && (extras->nsize[2]-oz)/2+oz >= min) && 
         (nst < 0 || GHc->mglevel < nst)) {

    if (0) printf ("oxyz = %d %d %d, nsizes = %d %d %d\n", ox, oy, oz, 
		   extras->nsize[0], extras->nsize[1], extras->nsize[2]);
    nxl = (extras->nsize[0]-ox)/2 + ox;
    /* FIXME: we assume 3D grid functions here */
    nyl = (extras->nsize[1]-oy)/2 + oy; 
    if (cartoon) nyl = GH->GFExtras [2]->nsize[1];
    nzl = (extras->nsize[2]-oz)/2 + oz;

    if (nxl-ox < connectivity->nprocs[0] * 2 ||
        (nyl-oy < connectivity->nprocs[1] * 2 && !cartoon) ||
        nzl-oz < connectivity->nprocs[2] * 2 ||
        (nxl-ox) % (1*connectivity->nprocs[0]) ||
        ((nyl-oy) % (1*connectivity->nprocs[1]) && !cartoon) ||
        (nzl-oz) % (1*connectivity->nprocs[2])) {
      GHc = NULL;
    } else {

      /* do nothing if GH alread exists */
      GHl = GHbyLevel(GH->convlevel, GH->level, GHc->mglevel + 1);
      if (GHl) {
        GHc = GHl;
        /* FIXME: we assume 3D grid functions here */
        extras = GHc->GFExtras [2];

        if (0)
          printf("     found grid %3dx%3dx%3d\n", 
                 extras->nsize[0], extras->nsize[1], extras->nsize[2]);

        continue;
      }

      /* create GH */
      if (0) printf("  creating grid %3dx%3dx%3d\n", nxl,nyl,nzl);

      /* set up pGH */
      nl[0] = nxl;
      nl[1] = nyl;
      nl[2] = nzl;
      PUGH_GFPeriodic (3, perme);
      BAM_SetPartitionInfo (3, nl);
      GHl = PUGH_SetupPGH(cctkGH, 3, nl, extras->nghostzones, 
			  CCTK_StaggerVars() ? PUGH_STAGGER : PUGH_NO_STAGGER,
			  perme);

      /* FIXME: we assume 3D grid functions here */
      extras = GHl->GFExtras [2];

      /* cactus 4 casualty */
      addtoGHList(GHl);

      /* set up the MG Level */
      GHl->level = GH->level;
      GHl->convlevel = GH->convlevel;
      GHl->mglevel = GHc->mglevel + 1;
      fac = fac * 2;

      /* used to be: MakeGridFuncs(GHl); */
      /* later: only setup those groups that are actually needed: 
                bam_ws, but also metric etc 
      */

      for(group = 0; group < CCTK_NumGroups(); group++)
      {
        CCTK_GroupData(group,&pgroup);
        if (pgroup.grouptype == CCTK_SCALAR || pgroup.grouptype == CCTK_GF)
        {
          PUGH_SetupGroup(GHl,  
                          NULL,
                          NULL,
                          pgroup.grouptype, 
                          pgroup.vartype, 
                          pgroup.dim, 
                          pgroup.numvars, 
                          1,
                          pgroup.stagtype,
                          pgroup.numtimelevels,
			  0);
        }
        else
        {
          int *cctk_groupsize, *cctk_ghostsize;
          CCTK_INT **cctk_groupsize_p, **cctk_ghostsize_p;


          /* we have to convert CCTK_INT into int,
             otherwise we run into problems on T3E */
          cctk_groupsize = (int *) malloc (2 * pgroup.dim * sizeof (int));
          cctk_ghostsize = cctk_groupsize + pgroup.dim;
          cctk_groupsize_p = CCTK_GroupSizesI (group);
          cctk_ghostsize_p = CCTK_GroupGhostsizesI (group);
          for (i = 0; i < pgroup.dim; i++)
          {
            cctk_groupsize[i] = (*cctk_groupsize_p)[i];

            /* note that CCTK_GroupGhostsizesI() returns NULL
               if no ghostzones were specified for this CCTK_ARRAY */
            cctk_ghostsize[i] = cctk_ghostsize_p ? (*cctk_ghostsize_p)[i] : 0;
          }

          PUGH_SetupGroup(GHl,
                          cctk_groupsize,
                          cctk_ghostsize,
                          pgroup.grouptype,
                          pgroup.vartype,
                          pgroup.dim,
                          pgroup.numvars,
                          1,
                          pgroup.stagtype,
                          pgroup.numtimelevels,
			  0);

          free (cctk_groupsize);
        }
      }

      /* set up coordinate values */
      GHl->dt0 = GHc->dt0;
      /*GHl->phys_time = GHc->phys_time;*/

      GHl->dx0 = GHc->dx0 * 2;
      GHl->dy0 = GHc->dy0 * 2;
      GHl->dz0 = GHc->dz0 * 2;

#if 0
      /* FIXME: origin is gone in pGH */
      GHl->cx0 = GHc->cx0;
      GHl->cy0 = GHc->cy0;
      GHl->cz0 = GHc->cz0;
 
      GHl->cx0 -= GHc->dx0 * oxdo;
      GHl->cy0 -= GHc->dy0 * oydo;
      GHl->cz0 -= GHc->dz0 * ozdo;
#endif

      /* multigrid levels need all these fields */
      for (i = 0; i < n; i++)
        EnableGFDS(GHl, nf[i], 1);
      for (i = 0; i < nws; i++)
        EnableGFDS(GHl, bam_wsmin_index+i, 1);

      /* cactus 3 had those on by default */
      /* check which are really needed! */
      for (i = 0; i < 1; i++) EnableGFDS(GHl, GFI_alp + i, 1);
      for (i = 0; i < 6; i++) EnableGFDS(GHl, GFI_gxx + i, 1);
      for (i = 0; i < 6; i++) EnableGFDS(GHl, GFI_kxx + i, 1);
      for (i = 0; i < 1; i++) EnableGFDS(GHl, GFI_psi + i, 1);
      if (conformal_metric) {
        for (i = 0; i < 3; i++) EnableGFDS(GHl, GFI_psix + i, 1);
        for (i = 0; i < 6; i++) EnableGFDS(GHl, GFI_psixx + i, 1);
      }
      if (veclaplacebssn) {
        for (i = 0; i < 18; i++) EnableGFDS(GHl, GFI_vl_bssn + i, 1);
      }
      if (problem == TSD) {
        for (i = 0; i < 8; i++) EnableGFDS(GHl, GFI_lbeta11 + i, 1);
      }
      if (problem == KSD) {
	for (i=0; i<6; i++) EnableGFDS(GHl, GFI_igxx + i, 1);
	EnableGFDS(GHl, GFI_K, 1);
	for (i=0; i<3; i++) EnableGFDS(GHl, GFI_Kx + i, 1);
	for (i=0; i<3; i++) EnableGFDS(GHl, GFI_Ax + i, 1);
	EnableGFDS(GHl, GFI_R, 1);
      } 
      if (problem == YORK) {
	for (i=0; i<13; i++) EnableGFDS(GHl, GFI_Ygxx + i, 1);
	for (i=0; i<10; i++) EnableGFDS(GHl, GFI_YABYxx + i, 1);
	EnableGFDS(GHl, GFI_K, 1);
      }	

      /* next */
      GHc = GHl;
    }
  }
}




void PUGH_destroystack(cGH *cctkGH, pGH *GH, int *nf, int n, int nws) {
  DECLARE_CCTK_PARAMETERS
  pGHList *tGH, *next;
  pGH *GHc;
  int i, l, cl = GH->convlevel;
  int bam_wsmin_index = CCTK_VarIndex("bam_elliptic::bam_wsmin");

  /* persist? */
  if (bam_persist) return;

  /* remove all GHs that only appear in multigrid */
  for (tGH = GHList; tGH; tGH = next) {
    next = tGH->next;

    if (0) 
      printf("PUGH_destroystack: convlevel %d  mglevel %d  biblevel %d\n", 
             tGH->GH->convlevel, tGH->GH->mglevel, tGH->GH->level);
    
    if (tGH->GH->mglevel > 0 && tGH->GH->convlevel == GH->convlevel) {
      /* DestroyPGH() sets the passed pGH pointer to NULL
         so we have save it for removefromGHList() */
      pGH *destroyGH = tGH->GH;

      PUGH_DestroyPGH (&destroyGH);
      removefromGHList (tGH->GH);
    }
  } 

  /* disable all multigrid storage in top levels */
  for (l = GH->level;; l++) {
    GHc = GHbyLevel(cl, l, GH->level - l);
    if (!GHc) break;

#ifdef FIXME
    /* output */
    for (i = bam_wsmin_index; i <= GFI_BAM_WSMAX; i++)
      _outputGF(GHc, ((pGF ***)GHc->gridFuncs[i]);
#endif

    if (0) 
      printf("GHc = %3d %3d %3d = %p, dx = %6.3f disabled\n", 
             cl, l, GH->level-l, GHc, GHc->dx0);
    for (i = nws-1; i >= 0; i--)
      PUGH_DisableGArrayDataStorage(GFByIndex(GHc, bam_wsmin_index+i));
    for (i = n-1; i >= 0; i--)
      if (!fieldstorage[i])
        PUGH_DisableGArrayDataStorage(GFByIndex(GHc, nf[i]));

    /* remove entry in multigrid hierarchy */ 
    GHc->mglevel = 0;
  }
}





#define dec_forface(GH, extras) \
  int i0 = 0; \
  int j0 = 0; \
  int k0 = 0; \
  int i1 = extras->lnsize[0]-1; \
  int j1 = extras->lnsize[1]-1; \
  int k1 = extras->lnsize[2]-1; \
  int fi0 = (extras->lb[GH->myproc][0] == 0); \
  int fj0 = (extras->lb[GH->myproc][1] == 0); \
  int fk0 = (extras->lb[GH->myproc][2] == 0); \
  int fi1 = (extras->ub[GH->myproc][0] == extras->nsize[0]-1); \
  int fj1 = (extras->ub[GH->myproc][1] == extras->nsize[1]-1); \
  int fk1 = (extras->ub[GH->myproc][2] == extras->nsize[2]-1); \


#define forfacei0(GH) \
   if (fi0) \
   for (k = k0; k <= k1; k++) \
   for (j = j0; j <= j1; j++) \
   for (i = i0, ijk = DATINDEX(GH,i,j,k); i <= i0; i++, ijk++) \

#define forfacei1(GH) \
   if (fi1) \
   for (k = k0; k <= k1; k++) \
   for (j = j0; j <= j1; j++) \
   for (i = i1, ijk = DATINDEX(GH,i,j,k); i <= i1; i++, ijk++) \


#define forfacej0(GH) \
   if (fj0) \
   for (k = k0; k <= k1; k++) \
   for (j = j0; j <= j0; j++) \
   for (i = i0, ijk = DATINDEX(GH,i,j,k); i <= i1; i++, ijk++) \

#define forfacej1(GH) \
   if (fj1) \
   for (k = k0; k <= k1; k++) \
   for (j = j1; j <= j1; j++) \
   for (i = i0, ijk = DATINDEX(GH,i,j,k); i <= i1; i++, ijk++) \


#define forfacek0(GH) \
   if (fk0) \
   for (k = k0; k <= k0; k++) \
   for (j = j0; j <= j1; j++) \
   for (i = i0, ijk = DATINDEX(GH,i,j,k); i <= i1; i++, ijk++) \

#define forfacek1(GH) \
   if (fk1) \
   for (k = k1; k <= k1; k++) \
   for (j = j0; j <= j1; j++) \
   for (i = i0, ijk = DATINDEX(GH,i,j,k); i <= i1; i++, ijk++) \



/* have to initialize storage at boundary
   some coefficient fields are only needed in the interior and are
   only initialized there, but a nan at the boundary shows in verbose
   mode and creates a problem on e.g. t3e and dec
*/
void EnableGFDS(pGH *GH, int n, int alwayssetzero) 
{
  CCTK_REAL *a;
  int i, j, k, ijk;
  pGA *GF = GFByIndex(GH, n);
  /* FIXME: This should come from a parameter */
  int padding_active=0;
  int padding_cacheline_bits=0;
  int padding_size=0;
  int padding_address_spacing=0;
  const char *initialize_memory = "zero";
  dec_forface(GH, GF->extras);

  /* if there is no storage, allocate and set to zero */
  if (!GF->storage) {
    PUGH_EnableGArrayDataStorage(GF, GH->myproc,
                                 initialize_memory,
                                 padding_active,
                                 padding_cacheline_bits,
                                 padding_size,
                                 padding_address_spacing);
    return;
  }

  if (!alwayssetzero)
    return;

  a = ((CCTK_REAL *)GF->data);

  /* initialize only the ghostzones, except for BAM's workspace arrays */
  /* FIXME: nghostzones is an array now */
  if (GF->extras->nghostzones[0] <= 1 &&
      strncmp (CCTK_VarName (n), "bam_ws", 6))
  {
    forfacei0(GF->extras) a[ijk] = 0;
    forfacei1(GF->extras) a[ijk] = 0;
    forfacej0(GF->extras) a[ijk] = 0;
    forfacej1(GF->extras) a[ijk] = 0;
    forfacek0(GF->extras) a[ijk] = 0;
    forfacek1(GF->extras) a[ijk] = 0;
  } else {
    for (ijk = 0; ijk < GF->extras->npoints; ijk++)
      a[ijk] = 0;
  }

  if (0) printf("EnableGFDS cleared GH %p, GF %p, name %s\n", 
                GH, GF, CCTK_VarName(n));
}
