/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* PUGH_to_BAM.c */
/* Bernd Bruegmann, 4/97, 10/97 */

#include "bbmg.h"
#include "CactusPUGH/PUGH/src/include/pughi.h"

pGH *GHbyLevel(int,int,int);


/* given SOMETHING describing a spatial refinement,
   make a multigrid stack, refinement factor 2

   for now assumes that there is a single outermost box
   if there are several disconnected regions, as happens
   on intermediate levels of time evolution, we need one more loop
*/
void BAM_translatestack(cGH *cctkGH, int ltop, int lmax, struct mgstack **s, 
			int *nf, int n, int nws)
{
  int pr = vb && 1;
  int u, v, w, f;
  int i = 0, j, l, lpugh;
  int nc, ncomps = 0;
  struct box *g;
  pGH *h;
  int lbo = (*s)->level;
  int lconv = (*s)->convlevel;
  int bam_wsmin_index = CCTK_VarIndex("bam_elliptic::bam_wsmin");
  pGExtras *extras;

  /* cactus 4 pGA fix: now even cx0 has been removed from GH so that the
     driver is stunningly independent of the coordinate system, so this sits
     here rather in makestack where it belongs */
  /* set outer boundary with proper offsets for centering
     the actual range of coordinate is equal to that or bigger */
  CCTK_CoordRange(cctkGH, &outxmin, &outxmax, -1, "x", "cart3d");
  CCTK_CoordRange(cctkGH, &outymin, &outymax, -1, "y", "cart3d");
  CCTK_CoordRange(cctkGH, &outzmin, &outzmax, -1, "z", "cart3d");
  outxmin += oxdo * cctkGH->cctk_delta_space[0];
  outymin += oydo * cctkGH->cctk_delta_space[1];
  outzmin += ozdo * cctkGH->cctk_delta_space[2];
  outxmax -= oxup * cctkGH->cctk_delta_space[0];
  outymax -= oyup * cctkGH->cctk_delta_space[1];
  outzmax -= ozup * cctkGH->cctk_delta_space[2];

  /* allocate level heads */
  for (l = 0; l < MGLMAX; l++) {
    level[l] = (struct box *) malloc(sizeof(struct box));
    level[l]->next = 0;
  }

  /* for all levels */
  for (l = 1; l <= lmax; l++) {
    lpugh = ltop - l;

    /* for all boxes on this level */
    /* in cactus, so far there is at most one */
    g = level[l];
    if (l <= ltop) 
      h = GHbyLevel(lconv, lbo, lpugh);
    else
      h = GHbyLevel(lconv, lbo+l-ltop, lpugh); 
    if (!h) errorexit("BAM_translatestack(): missing level");

    /* FIXME: we assume 3D grid functions here */
    extras = h->GFExtras[2];
    {
      /* make linked list of boxes */
      g->next = (struct box *) malloc(sizeof(struct box));
      g = g->next;
      g->next = 0;
      g->l = l;
      g->ltop = ltop;
      g->lmax = lmax;
      g->reffac = pow((double)2.0, (double)(lmax-l)) + 0.5;

      if (0) {
	printf("l %d  lpugh %d  reffac %d  lconv %d  ", 
	       g->l, lpugh, g->reffac, h->convlevel);
	printf("nx ny nz     %3d %3d %3d   ", 
	       extras->nsize[0], extras->nsize[1], extras->nsize[2]);
	printf("lnx lny lnz  %3d %3d %3d   ", 
	       extras->lnsize[0], extras->lnsize[1], extras->lnsize[2]);
	printf("\n");
      }

      /* parameter translation */
      if (!dequal(h->dx0,h->dy0) || !dequal(h->dx0,h->dz0))
	errorexit("implement dx, dy, dz for PPMG");
      g->m = extras->nsize[0];
      g->n = extras->nsize[1];
      g->o = extras->nsize[2];
      g->d = h->dx0;
      g->dx = h->dx0;
      g->dy = h->dy0;
      g->dz = h->dz0;

      g->x = outxmin - oxdo * g->dx;
      g->y = outymin - oydo * g->dy;
      g->z = outzmin - ozdo * g->dz;

      g->xmax = g->x + (g->m-1)*g->d;
      g->ymax = g->y + (g->n-1)*g->d;
      g->zmax = g->z + (g->o-1)*g->d;

      /* fix indexing convention */
      /* BAM: C ordered
	 g->di = g->n * g->o;
	 g->dj = g->o;
	 g->dk = 1;
	 reorder loops, too, for optimization!
	 note that FORTRAN interface "guarantees" that di, dj, dk works
      */
      g->di = 1;
      g->dj = extras->lnsize[0];
      g->dk = extras->lnsize[0]*extras->lnsize[1];

      /* offset into local space, see #define ijkof() */
      g->lijk = extras->lb[h->myproc][0] * g->di +
                extras->lb[h->myproc][1] * g->dj +
                extras->lb[h->myproc][2] * g->dk;

      /* more local stuff */
      g->lm = extras->lnsize[0];
      g->ln = extras->lnsize[1];
      g->lo = extras->lnsize[2];

      /* make pointers point to dynamic fields 
	 for vector fields, use contiguous range, e.g. u1, u2, u3
      */
      g->cctkGH = cctkGH;
      g->GH = h;
      i = 0;      


      /* scalar equations */
      if (!(eqn_type == VECLAPLACE || eqn_type == VEC4 
	    || eqn_type == TSD || eqn_type == KSD || eqn_type == YORK)) {
	ncomps = 1;

	/* u */
	g->GF[i++] = GFByIndex(h, nf[0]); 
	
	/* workspace v, w, f */
	for (j = 0; j < MGF-1; j++)
	  g->GF[i++] = GFByIndex(h, bam_wsmin_index+j);
	
	/* M and N */
	g->GF[i++] = GFByIndex(h, nf[1]);
	g->GF[i++] = GFByIndex(h, nf[2]);

	/* metric coefficients gi and gg */
	for (j = MGF-1; j < nws; j++)
	  g->GF[i++] = GFByIndex(h, bam_wsmin_index+j);
	
	/* coefficients in power series */
	for (j = 3; j < n; j++) {
	  g->GF[i++] = GFByIndex(h, nf[j]);
	  g->ppow[j-3] = ppowers[j];
	}
	g->ppow[j-3] = 0;
      }


      /* vector laplace equation: 3 components */
      if (eqn_type == VECLAPLACE) {
	ncomps = 3;

	/* u */
	for (nc = 0; nc < ncomps; nc++)  
	  g->GF[i++] = GFByIndex(h, nf[nc]); 

	/* workspace v, w, f */
	for (j = 0; j < MGF-1; j++)
	  for (nc = 0; nc < ncomps; nc++)  
	    g->GF[i++] = GFByIndex(h, bam_wsmin_index+j*ncomps+nc);

	/* N */
	for (nc = 0; nc < ncomps; nc++)
	  g->GF[i++] = GFByIndex(h, nf[3+nc]);
      }

      /* vector equation with a variable number of components */
      if (eqn_type == VEC4 || eqn_type == TSD || eqn_type == KSD || 
	  eqn_type == YORK) {
	if (eqn_type == VEC4) ncomps = 4;
	if (eqn_type == KSD)  ncomps = 4;
	if (eqn_type == YORK) ncomps = 4;
	if (eqn_type == TSD)  ncomps = 5;

	/* u */
	for (nc = 0; nc < ncomps; nc++)  
	  g->GF[i++] = GFByIndex(h, nf[nc]); 

	/* workspace v, w, f */
	for (j = 0; j < MGF-1; j++)
	  for (nc = 0; nc < ncomps; nc++)  
	    g->GF[i++] = GFByIndex(h, bam_wsmin_index+j*ncomps+nc);
      }


      /* make storage pointers */
      for (j = 0; j < i; j++)
	g->store[j] = g->GF[j]->data;

      /* number of fields and components that we just produced */
      g->nfields = i;
      g->ncomps = ncomps;
      if (ncomps > MGCMAX)
	CCTK_WARN(0, "number of components greater than MGCMAX");

      /* debug */
      if (0)
      printf(
	"%2d %7.3f %7.3f %7.3f %7.3f %7.3f %7.3f  %10.3e  %3d %3d %3d\n",
	l, g->x, g->y, g->z, g->xmax, g->ymax, g->zmax, g->d, 
	g->m, g->n, g->o);
    }
  }

  /* set local boundary flags for PUGH */
  if (0) printf("outer box:  %6.3f %6.3f %6.3f  %6.3f %6.3f %6.3f\n",
		 outxmin, outymin, outzmin, outxmax, outymax, outzmax);
  for (l = 1; l <= lmax; l++) {
    forlevel(l, g) {         
      setboundflags(g);
      if (0&&pr) prboxpar(g);
    }
  }

  /* print info */
  if (0) {
    printf("made multigrid stack for %d scalar fields,\n  ", i);
    for (j = 0; j < i; j++)
      printf("%s ", fieldtitle[j]);
    printf("\n");
    fflush(stdout);
  }
}





/* free level structures */
void BAM_freelevels() 
{
  struct box *g, *gnext;
  int l;

  for (l = MGLMAX-1; l >= 0; l--) {
    for (g = level[l]->next; g; g = gnext) {
      gnext = g->next;
      free(g);
    }
    free(level[l]);
  }
}




/* primitive print */
void prprimGF(char *s, struct box *g)
{
  double *tmp;

  /* FIXME Timelevels */
#if 0
  int tl=0;
#endif

  tmp = g->store[0];
  g->store[0] = GFDataByName(g->GH, s);
  if (!g->store[0])
    printf("prprimGF(): no data for %s\n", s);
  else {
    printf("prprimGF ");
    prprim(s, g, 0);
  }
  g->store[0] = tmp;
}




/* set range for local processor sweeps 
   talk about screwy: first version used GH->ownership,
   but for simple onlyowned flag we now use GH->nghostzones, see SetupPGH.c
   onlyowned = 0: all
             = 1: remove ghosts
	     = 2: remove ghosts, move to next inner even index
	     = 3: remove ghosts, move to next outer even index
   next rewrite: introduce g->lbi0, ...
*/
void setrangePUGH(struct box *g, int onlyowned)
{
  pGH *GH = g->GH;
  int p = GH->myproc;
  /* FIXME: we assume 3D grid functions here */
  pGExtras *extras = GH->GFExtras [2];
  int w = (onlyowned >= 1) ? extras->nghostzones[0] : 0;

#if 0
  /* offset at lower bound: sort out inner grids versus outer boundary
     only the parity of o is used 
  */
  int ox, oy, oz;
  ox = oy = oz = 0; 
  if (neumann) ox = oy = oz = 1; 
  if (GH->mglevel < 0) ox = oy = oz = 0;
  if (octant) {ox = oxdo; oy = oydo; oz = ozdo;}
  if (bitant) oz = ozdo;
  if (cartoon) {ox = oxdo; oy = oydo;}
#endif

  /* for debugging only:
     loop over all processors 
  */
  if (0) {
    for (p = 0; p < GH->nprocs; p++) {

      int lnx = extras->ub[p][0] - extras->lb[p][0] + 1 ;
      int lny = extras->ub[p][1] - extras->lb[p][1] + 1 ;
      int lnz = extras->ub[p][2] - extras->lb[p][2] + 1 ;

      g->limin = (extras->lb[p][0] == 0) ? g->imin : w;
      g->ljmin = (extras->lb[p][1] == 0) ? g->jmin : w;
      g->lkmin = (extras->lb[p][2] == 0) ? g->kmin : w;
      
      g->limax = (extras->ub[p][0] == extras->nsize[0]-1) ? 
	lnx-1 - (g->m-1-g->imax) : lnx-1 - w;
      g->ljmax = (extras->ub[p][1] == extras->nsize[1]-1) ? 
	lny-1 - (g->n-1-g->jmax) : lny-1 - w;
      g->lkmax = (extras->ub[p][2] == extras->nsize[2]-1) ? 
	lnz-1 - (g->o-1-g->kmax) : lnz-1 - w;
      
      g->limin += extras->lb[p][0];
      g->ljmin += extras->lb[p][1];
      g->lkmin += extras->lb[p][2];
      
      g->limax += extras->lb[p][0];
      g->ljmax += extras->lb[p][1];
      g->lkmax += extras->lb[p][2];
      
      if (onlyowned == 2) {
	g->limin += (g->limin + oxdo) % 2;
	g->ljmin += (g->ljmin + oydo) % 2;
	g->lkmin += (g->lkmin + ozdo) % 2;
	
	g->limax -= (g->limax + oxdo) % 2;
	g->ljmax -= (g->ljmax + oydo) % 2;
	g->lkmax -= (g->lkmax + ozdo) % 2;
      }
      
      if (onlyowned == 3) {
	if (extras->lb[p][0] != 0) g->limin -= (g->limin + oxdo) % 2;
	if (extras->lb[p][1] != 0) g->ljmin -= (g->ljmin + oydo) % 2;
	if (extras->lb[p][2] != 0) g->lkmin -= (g->lkmin + ozdo) % 2;
	
	if (extras->ub[p][0] != extras->nsize[0]-1) 
	  g->limax += (g->limax + oxdo) % 2;
	if (extras->ub[p][1] != extras->nsize[1]-1) 
	  g->ljmax += (g->ljmax + oydo) % 2;
	if (extras->ub[p][2] != extras->nsize[2]-1) 
	  g->lkmax += (g->lkmax + ozdo) % 2;
      }

      if (0) {
	printf("setrange p=%3d owned=%d w=%d  ", p, onlyowned, w);
	printf("   %3d %3d   %3d %3d   %3d %3d ", 
	       g->limin, g->limax, g->ljmin, g->ljmax, g->lkmin, g->lkmax);
	printf("   %3d %3d   %3d %3d   %3d %3d ", 
	       g->imin, g->imax, g->jmin, g->jmax, g->kmin, g->kmax);
	printf("   %3d %3d   %3d %3d   %3d %3d\n", 
	       extras->lb[p][0], extras->ub[p][0],
	       extras->lb[p][1], extras->ub[p][1],
	       extras->lb[p][2], extras->ub[p][2]);
	printf("\n");
      } 
    }

    p = GH->myproc;
    
  }
  /* end debuggging loop */


  /* local index with or without ghosts */
  g->limin = (extras->lb[p][0] == 0) ? g->imin : w;
  g->ljmin = (extras->lb[p][1] == 0) ? g->jmin : w;
  g->lkmin = (extras->lb[p][2] == 0) ? g->kmin : w;

  g->limax = (extras->ub[p][0] == extras->nsize[0]-1) ? 
              extras->lnsize[0]-1 - (g->m-1-g->imax) : extras->lnsize[0]-1 - w;
  g->ljmax = (extras->ub[p][1] == extras->nsize[1]-1) ? 
              extras->lnsize[1]-1 - (g->n-1-g->jmax) : extras->lnsize[1]-1 - w;
  g->lkmax = (extras->ub[p][2] == extras->nsize[2]-1) ? 
              extras->lnsize[2]-1 - (g->o-1-g->kmax) : extras->lnsize[2]-1 - w;

  /* absolute index */
  g->limin += extras->lb[p][0];
  g->ljmin += extras->lb[p][1];
  g->lkmin += extras->lb[p][2];

  g->limax += extras->lb[p][0];
  g->ljmax += extras->lb[p][1];
  g->lkmax += extras->lb[p][2];


  /* for two-level operations, align to coarse (reffac = 2, absolute!) */
  /* if there is a ghost at lower bound, the odd-even property is reversed */

  /* exclusive, e.g. restrict */
  if (onlyowned == 2) {
    g->limin += (g->limin + oxdo) % 2;
    g->ljmin += (g->ljmin + oydo) % 2;
    g->lkmin += (g->lkmin + ozdo) % 2;
    
    g->limax -= (g->limax + oxdo) % 2;
    g->ljmax -= (g->ljmax + oydo) % 2;
    g->lkmax -= (g->lkmax + ozdo) % 2;
  }

  /* inclusive, e.g. ipol, watch out for DIRICHLET treatment in prolong.c */
  if (onlyowned == 3) {
    if (extras->lb[p][0] != 0) g->limin -= (g->limin + oxdo) % 2;
    if (extras->lb[p][1] != 0) g->ljmin -= (g->ljmin + oydo) % 2;
    if (extras->lb[p][2] != 0) g->lkmin -= (g->lkmin + ozdo) % 2;
    
    if (extras->ub[p][0] != extras->nsize[0]-1) g->limax += (g->limax + oxdo) % 2;
    if (extras->ub[p][1] != extras->nsize[1]-1) g->ljmax += (g->ljmax + oydo) % 2;
    if (extras->ub[p][2] != extras->nsize[2]-1) g->lkmax += (g->lkmax + ozdo) % 2;
  }

  if (0) {
    printf("setrange p=%3d owned=%d w=%d  ", p, onlyowned, w);
    printf("   %3d %3d   %3d %3d   %3d %3d ", 
	   g->limin, g->limax, g->ljmin, g->ljmax, g->lkmin, g->lkmax);
    printf("   %3d %3d   %3d %3d   %3d %3d\n", 
	   g->imin, g->imax, g->jmin, g->jmax, g->kmin, g->kmax);
  } 

}




/* PUGH wrapper for forced synchronization
   of a group of 'n_vars' variables starting with index 'first_var' */
static void PUGHsyncVars (pGH *pughGH, int first_var, int n_vars)
{
  int i, commflag, *commflags;
  /* FIXME: timelevel */
  const int timelevel = 0;
  pGA ***GFs = ((pGA ***) pughGH->variables) + first_var;

  /* save the original communication flags */
  commflags = n_vars <= 1 ? &commflag : (int *) malloc (n_vars * sizeof (int));
  for (i = 0; i < n_vars; i++) {
    if (GFs [i][timelevel]->comm == NULL)
      GFs [i][timelevel]->comm =
        PUGH_SetupGArrayComm (pughGH,
                              GFs [i][timelevel]->extras->dim,
                              GFs [i][timelevel]->id,
                              timelevel,
                              GFs [i][timelevel]->vtype,
                              GFs [i][timelevel]->extras);
    commflags [i] = GFs [i][timelevel]->comm->commflag;
    if (commflags [i] != PUGH_ALLCOMM) {
      PUGH_EnableGArrayComm (GFs [i][timelevel], PUGH_ALLCOMM);
      commflags [i] = PUGH_ALLCOMM;
    }
  }

  /* do the synchronization */
  for (i = 0; i < n_vars; i++)
    PUGH_SyncGArray (GFs [i][timelevel]);

  /* restore original communication flags */
  for (i = 0; i < n_vars; i++)
    if (commflags [i] != PUGH_ALLCOMM)
      PUGH_EnableGArrayComm (GFs [i][timelevel], commflags [i]);
  if (n_vars > 1)
    free (commflags);
}



void PUGHsyncd(double *u)
{
  pGA *GF = locateGFbyDataPtr(u);
  pGH *pughGH = (pGH *) GF->parent;

  PUGHsyncVars (pughGH, GF->id, 1);
}




/* wrapper for PUGH synchronize to make sure we communicate */
void PUGHsyncn(struct box *g, int nf, int n)
{
  int pr = 0;

  if (pr) prprim("bef sync", g, nf);
  PUGHsyncVars (g->GH, g->GF[nf]->id, n);
  if (pr) prprim("aft sync", g, nf+n-1);
}




/* wrapper for PUGH synchronize for grid functions */
void PUGHsyncGFn(struct box *g, int nGF, int n)
{
  PUGHsyncVars (g->GH, GFByIndex(g->GH, nGF)->id, n);
}




/* wrapper for PUGH synchronize, automatic for vectors */
void PUGHsyncvec(struct box *g, int nf)
{
  int n = (nf < MGF*g->ncomps) ? g->ncomps : 1;

  PUGHsyncn(g, nf, n);
}




/* for debugging, sync all overriding commflag */
void PUGHsyncall(struct box *g) 
{
  PUGHsyncn(g, 0, g->nfields);
  PUGHsyncGFn(g, 0, g->GH->nvariables);
}
