/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* algebra.c */
/* Bernd Bruegmann, 4/97 */

#include "bbmg.h"

/* transparent vector implementation:
   pass fields by index numbers that can be locally incremented
*/


/**************************************************************************/
/* vector algebra */


/* set constant */
void setconst(struct box *g, int nfa, double c)
{
  int nf;

  for (nf = 0; nf < g->ncomps; nf++, nfa++)
    setconst_scalar(g, nfa, c);
}




/* copy */
void copy(struct box *g, int nfa, int nfr)
{
  int nf;

  for (nf = 0; nf < g->ncomps; nf++, nfa++, nfr++)
    copy_scalar(g, nfa, nfr);
}



/* add */
void add(struct box *g, int nfa, int nfb, int nfr)
{
  int nf;

  for (nf = 0; nf < g->ncomps; nf++, nfa++, nfb++, nfr++)
    add_scalar(g, nfa, nfb, nfr);
}



/* subtract */
void subtract(struct box *g, int nfa, int nfb, int nfr)
{
  int nf;

  for (nf = 0; nf < g->ncomps; nf++, nfa++, nfb++, nfr++)
    subtract_scalar(g, nfa, nfb, nfr);
}



/* average of the l2 norms of the components of a vector field */
double l2norm(struct box *g, int nfa)
{
  int nf;
  double sum = 0.0;
  int o = (problem == VEC4) ? 1 : 0;
  nfa += o;
  for (nf = 0+o; nf < g->ncomps; nf++, nfa++)
    sum += l2norm_scalar(g, nfa);
  return sum/(g->ncomps-o);
}





/* l infinity norm of the components of a vector field */
double linfnorm(struct box *g, int nfa)
{
  int nf;
  double x, max = -1;

  for (nf = 0; nf < g->ncomps; nf++, nfa++) {
    x = linfnorm_scalar(g, nfa);
    if (max < x) max = x;
  }
  return max;
}






/*************************************************************************/
/* scalar algebra */


/* set scalar field to constant */
void setconst_scalar(struct box *g, int nfa, double c)
{
  double *a = g->store[nfa];
  int i, j, k, ijk;

  setrangePUGH(g, 0);
  forrange(g)
    a[ijk] = c;
}



/* copy scalar field */
void copy_scalar(struct box *g, int nfa, int nfr)
{
  double *a = g->store[nfa];
  double *r = g->store[nfr];
  int i, j, k, ijk;

  setrangePUGH(g, 0);
  forrange(g)
    r[ijk] = a[ijk];
}



/* add scalar fields */
void add_scalar(struct box *g, int nfa, int nfb, int nfr)
{
  double *a = g->store[nfa];
  double *b = g->store[nfb];
  double *r = g->store[nfr];
  int i, j, k, ijk;

  setrangePUGH(g, 0);
  forrange(g)
    r[ijk] = a[ijk] + b[ijk];
}



/* subtract scalar fields */
void subtract_scalar(struct box *g, int nfa, int nfb, int nfr)
{
  double *a = g->store[nfa];
  double *b = g->store[nfb];
  double *r = g->store[nfr];
  int i, j, k, ijk;

  setrangePUGH(g, 0);
  forrange(g)
    r[ijk] = a[ijk] - b[ijk];
}



/* l2 norm of a scalar field (not counting the physical boundary) */
double l2norm_scalar(struct box *g, int nfa)
{
  double *a = g->store[nfa];
  int i, j, k, ijk;
  double sum = 0;
  int n = 0;
#ifdef CCTK_MPI
  CCTK_REAL locals[2], globals[2];
#endif

  setrangePUGH(g, 1);
  forrange(g) {
    sum += a[ijk] * a[ijk];
    n++;
  }

#ifdef CCTK_MPI
  if (0) printf("sum %e  n %d\n", sum, n);

  locals[0] = sum;
  locals[1] = n;
  MPI_Allreduce(locals, globals, 2, 
		PUGH_MPI_REAL, MPI_SUM, g->GH->PUGH_COMM_WORLD);
  sum = globals[0];
  n   = globals[1] + 0.5;

  if (0) printf("sum %e  n %d\n", sum, n);
#endif

  return (n) ? sqrt(sum/n) : sum;
}



/* linf norm of a scalar field (not counting the physical boundary) */
double linfnorm_scalar(struct box *g, int nfa)
{
  double *a = g->store[nfa];
  int i, j, k, ijk;
  CCTK_REAL max = -1;
#ifdef CCTK_MPI
  CCTK_REAL lmax;
#endif 

  setrangePUGH(g, 1);
  forrange(g) {
    if (max < fabs(a[ijk])) max = fabs(a[ijk]);
  }

#ifdef CCTK_MPI
  lmax = max;
  MPI_Allreduce(&lmax, &max, 1,
		PUGH_MPI_REAL, MPI_MAX, g->GH->PUGH_COMM_WORLD);
#endif

  return max;
}



#if 0
/* this is how one does l2 norms in Cactus 4 (from TR), but we don't
   want the physical boundaries */
double l2norm_scalar(struct box *g, int nfa)
{
  int reduction_handle;
  CCTK_REAL reduction_value;

  /* Should we really return 0.0 in case of an error ?? */
  reduction_value = 0.0;

  reduction_handle = CCTK_ReductionHandle ("norm2");

  if (reduction_handle >= 0)
  {
    if (CCTK_Reduce (g->cctkGH, -1, reduction_handle, 1, CCTK_VARIABLE_REAL,
                     &reduction_value, 1, g->GF[nfa]->id) < 0)
    {
      CCTK_VWarn (1, __LINE__, __FILE__, CCTK_THORNSTRING,
                  "l2norm_scalar: Failed to get l2norm of '%s'",
                  CCTK_VarName (g->GF[nfa]->id));
    }
  }
  else
  {
    CCTK_WARN (1, "l2norm_scalar: Couldn't get handle "
                  "for reduction operator 'norm2'");
  }

  printf("l2norm for %s: %e\n", 
	 CCTK_VarName (g->GF[nfa]->id), reduction_value);

  return ((double) reduction_value);
}
#endif

