/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* brbr.c */
/* Bernd Bruegmann, 11/97 */

#include "bbmg.h"



#define EPSZERO (dequaleps/1000.0)

double MBL[N_MWBL], CBL[N_MWBL][3], PBL[N_MWBL][3], SBL[N_MWBL][3];
int dalota;

void mwblpsiofxyz(double x, double y, double z, double *psi,
		  double *dp1op, double *dp2op, double *dp3op,
		  double *ddp11op, double *ddp12op, double *ddp13op,
		  double *ddp22op, double *ddp23op, double *ddp33op);
double mwblpsi(double x, double y, double z);
void bbalphaofxyz(double x, double y, double z, double *alpha);

void byKofxyz(double x, double y, double z,
	      double *K11, double *K12, double *K13, 
	      double *K22, double *K23, double *K33);
void dltKofxyz(double x, double y, double z,
	      double *K11, double *K12, double *K13, 
	      double *K22, double *K23, double *K33);

void set_byK_GH(pGH *GH);
void set_byK_box(struct box *g);





/**************************************************************************/

void setnames_brbr() 
{
  int i = 0;

  fieldtitle[i++] = "u";
  fieldtitle[i++] = "v";
  fieldtitle[i++] = "w";
  fieldtitle[i++] = "f";
  fieldtitle[i++] = "M";
  fieldtitle[i++] = "N";
  if (fisheye) {
    fieldtitle[i++] = "gi11";
    fieldtitle[i++] = "gi12";
    fieldtitle[i++] = "gi13";
    fieldtitle[i++] = "gi22";
    fieldtitle[i++] = "gi23";
    fieldtitle[i++] = "gi33";
    fieldtitle[i++] = "gg1";
    fieldtitle[i++] = "gg2";
    fieldtitle[i++] = "gg3";
  }
}




void setfieldpointers_brbr(struct box *g, double **M, double **N)
{
  int i = MGF*g->ncomps;

  *M = g->store[i++];
  *N = g->store[i++];
}



/* Misner-Wheeler-Brill-Lindquist conformal factor psi and derivatives over psi

   psi(x,y,z) = 1 + 1/2 Sum_{i} m_i / |r - c_i|

*/
double mwblpsi(double x, double y, double z)
{
  int i;
  double r, xi, yi, zi;
  double p = 1;

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  for (i = 0; MBL[i] != 0.0 && i < N_MWBL; i++) {
    xi = x-CBL[i][0];
    yi = y-CBL[i][1];
    zi = z-CBL[i][2];

    r  = sqrt(xi*xi+yi*yi+zi*zi);
    if (dequal(r, 0.0)) r = EPSZERO;

    p += 0.5 * MBL[i] / r;
  }
  return p;
}





/* psi and its derivatives, analytic calculation */
void mwblpsiofxyz(double x, double y, double z, double *psi,
		  double *dp1op, double *dp2op, double *dp3op,
		  double *ddp11op, double *ddp12op, double *ddp13op,
		  double *ddp22op, double *ddp23op, double *ddp33op)
{
  int i;
  double p, dp1, dp2, dp3, ddp11, ddp12, ddp13, ddp22, ddp23, ddp33;
  double s1, s3, s5;
  double r, ri, xi, yi, zi;
  double x_fish = x, y_fish = y, z_fish = z;


  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  p = dp1 = dp2 = dp3 = ddp11 = ddp12 = ddp13 = ddp22 = ddp23 = ddp33 = 0;

  for (i = 0; MBL[i] != 0.0 && i < N_MWBL; i++) {
    xi = x-CBL[i][0];
    yi = y-CBL[i][1];
    zi = z-CBL[i][2];

#if 1
    r  = sqrt(xi*xi+yi*yi+zi*zi);
    if (dequal(r, 0.0)) r = EPSZERO;
    ri = 1/r;
#else
    ri = 1.0/sqrt(xi*xi+yi*yi+zi*zi);
#endif

    s1 = MBL[i]*ri*(0.5);
    s3 = s1*ri*ri*(-1.0);
    s5 = s3*ri*ri*(-3.0);

    p += s1;

    dp1 += xi*s3; 
    dp2 += yi*s3; 
    dp3 += zi*s3; 

    ddp11 += xi*xi*s5 + s3;
    ddp12 += xi*yi*s5;
    ddp13 += xi*zi*s5;
    ddp22 += yi*yi*s5 + s3;
    ddp23 += yi*zi*s5;
    ddp33 += zi*zi*s5 + s3;
  }
  p += 1.0;

  /* transform derivatives to fisheye coordinates */
  if (fisheye) {
    FishEye_StdToFish_ddPsi_dPsi(x_fish, y_fish, z_fish, 
                                 &ddp11, &ddp12, &ddp13, 
                                 &ddp22, &ddp23, &ddp33, 
                                 &dp1, &dp2, &dp3);
  }

  /* this is for cactus, so we don't use psi4 as in BAM or G-code, 
     and we divide derivatives by psi; don't ask why 
  */
  *psi = p;
  *dp1op = dp1/p;
  *dp2op = dp2/p;
  *dp3op = dp3/p;
  *ddp11op = ddp11/p;
  *ddp12op = ddp12/p;
  *ddp13op = ddp13/p;
  *ddp22op = ddp22/p;
  *ddp23op = ddp23/p;
  *ddp33op = ddp33/p;
}





/* set MWBL, conformal metric: gab = dab, Kab = 0, psi and derivs */
void set_mwbl_box(struct box *g)
{
  int i, j, k, ijk;
  double x, y, z;
  double *p[10];
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;
  void setfieldpointers_psi10(struct box *g, double **p);

  setfieldpointers_psi10(g, p);
  setfieldpointers_gab(g, &g11, &g12, &g13, &g22, &g23, &g33);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);

  setrangePUGH(g, 0);
  forrange(g) {
    x = g->x + i*g->d;
    y = g->y + j*g->d;
    z = g->z + k*g->d;

    mwblpsiofxyz(x, y, z, p[0]+ijk, p[1]+ijk, p[2]+ijk, p[3]+ijk, p[4]+ijk,
		 p[5]+ijk, p[6]+ijk, p[7]+ijk, p[8]+ijk, p[9]+ijk);

    g11[ijk] = g22[ijk] = g33[ijk] = 1.0;
    g12[ijk] = g13[ijk] = g23[ijk] = 0.0;

    K11[ijk] = K22[ijk] = K33[ijk] = 0.0;
    K12[ijk] = K13[ijk] = K23[ijk] = 0.0;

    if (fisheye) 
      FishEye_IndexTransDDSym(x, y, z,
		       g11+ijk, g12+ijk, g13+ijk, g22+ijk, g23+ijk, g33+ijk);
  }
}



/* same with cactus loop */
void set_mwbl_GH(pGH *GH)
{
  int ijk, n = 0;
  double *x, *y, *z;
  double *p[10];
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;

  x = GFDataByName(GH, "grid::x"); 
  y = GFDataByName(GH, "grid::y"); 
  z = GFDataByName(GH, "grid::z");
  p[n++] = GFDataByName(GH, "staticconformal::psi");
  p[n++] = GFDataByName(GH, "staticconformal::psix");
  p[n++] = GFDataByName(GH, "staticconformal::psiy");
  p[n++] = GFDataByName(GH, "staticconformal::psiz");
  p[n++] = GFDataByName(GH, "staticconformal::psixx");
  p[n++] = GFDataByName(GH, "staticconformal::psixy");
  p[n++] = GFDataByName(GH, "staticconformal::psixz");
  p[n++] = GFDataByName(GH, "staticconformal::psiyy");
  p[n++] = GFDataByName(GH, "staticconformal::psiyz");
  p[n++] = GFDataByName(GH, "staticconformal::psizz");
  g11 = GFDataByName(GH, "admbase::gxx");
  g12 = GFDataByName(GH, "admbase::gxy");
  g13 = GFDataByName(GH, "admbase::gxz");
  g22 = GFDataByName(GH, "admbase::gyy");
  g23 = GFDataByName(GH, "admbase::gyz");
  g33 = GFDataByName(GH, "admbase::gzz");
  K11 = GFDataByName(GH, "admbase::kxx");
  K12 = GFDataByName(GH, "admbase::kxy");
  K13 = GFDataByName(GH, "admbase::kxz");
  K22 = GFDataByName(GH, "admbase::kyy");
  K23 = GFDataByName(GH, "admbase::kyz");
  K33 = GFDataByName(GH, "admbase::kzz");

  /* FIXME: we assume 3D grid functions here */
  for (ijk = 0; ijk < GH->GFExtras [2]->npoints; ijk++) {

    mwblpsiofxyz(x[ijk], y[ijk], z[ijk], 
		 p[0]+ijk, p[1]+ijk, p[2]+ijk, p[3]+ijk, p[4]+ijk,
		 p[5]+ijk, p[6]+ijk, p[7]+ijk, p[8]+ijk, p[9]+ijk);

    g11[ijk] = g22[ijk] = g33[ijk] = 1.0;
    g12[ijk] = g13[ijk] = g23[ijk] = 0.0;

    K11[ijk] = K22[ijk] = K33[ijk] = 0.0;
    K12[ijk] = K13[ijk] = K23[ijk] = 0.0;

    if (fisheye) 
      FishEye_IndexTransDDSym(x[ijk], y[ijk], z[ijk],
		       g11+ijk, g12+ijk, g13+ijk, g22+ijk, g23+ijk, g33+ijk);
  }
}




/* set lapse for BrBr as done in BAM (init_maximal.c)
   idea is to get boundary approximately static
*/
void bbalphaofxyz(double x, double y, double z, double *alpha)
{
  int i;
  double xi, yi, zi, ri;
  double p = 0.0;

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  for (i = 0; MBL[i] != 0.0 && i < N_MWBL; i++) {
    xi = x-CBL[i][0];
    yi = y-CBL[i][1];
    zi = z-CBL[i][2];
    ri = sqrt(xi*xi+yi*yi+zi*zi);
    if (dequal(ri,0.0)) ri = EPSZERO;
    p += MBL[i]/ri;
  }
  p *= 0.5;
  *alpha = (1-p)/(1+p);
}




/* cactus loop */
void set_bbalpha_GH(pGH *GH)
{
  int ijk;
  double *x, *y, *z;
  double *alpha;

  x = GFDataByName(GH, "grid::x"); 
  y = GFDataByName(GH, "grid::y"); 
  z = GFDataByName(GH, "grid::z"); 
  alpha = GFDataByName(GH, "admbase::alp");

  /* FIXME: we assume 3D grid functions here */
  for (ijk = 0; ijk < GH->GFExtras [2]->npoints; ijk++) {
    bbalphaofxyz(x[ijk], y[ijk], z[ijk], alpha + ijk);
  }
}




/* cactus loops for stencils */
#define forallpoints(GH) \
  for (k = 0, di = 1, dj = GH->lnx, dk = GH->lnx*GH->lny; k < GH->lnz; k++) \
  for (j = 0; j < GH->lny; j++) \
  for (i = 0, ijk = DATINDEX(GH,i,j,k); i < GH->lnx; i++, ijk++)

#define forinnerpoints(GH) \
  for (k = 1, di = 1, dj = GH->lnx, dk = GH->lnx*GH->lny; k < GH->lnz-1; k++)\
  for (j = 1; j < GH->lny-1; j++) \
  for (i = 1, ijk = DATINDEX(GH,i,j,k); i < GH->lnx-1; i++, ijk++)





/* Brill-Lindquist data + Bowen-York momentum and spin 
   = Brandt-Bruegmann data (gr-qc/9703066)

   gphys_ij = psi^4 delta_ij
   Kphys_ij = psi^-2 K_ij,    K_ij = Sum K_ij(P,S)

   1/a = Sum m/(2r)
   psi = u + 1/a    -> u is determined numerically for solution to
                       Hamiltonian constraint
   phi = 1 + 1/a    -> used for time independent conformal rescaling,
                       identical to Brill-Lindquist

   In terms of u, the Hamiltonian constraint becomes

   H = Laplace(flat) u + b (1 + a u)^-7,   b = 1/8 a^7 K_ij K^ij

   Note that a, b, and u are regular on R^3.
*/


#define fori     for (i = 1; i <= 3; i++)
#define forsymij fori for (j = i; j <= 3; j++)
#define delta(i,j) (((i)==(j))?1.0:0.0)



/* Bowen-York extrinsic curvature */
void byKofxyz(double x, double y, double z,
	      double *K11, double *K12, double *K13, 
	      double *K22, double *K23, double *K33)
{
  double xold = x, yold = y, zold = z;
  int i, j, n;
  double xc, yc, zc, r;
  double *K[4][4], P[4], S[4], N[4], NP, eNS[4];

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  K[1][1] = K11;
  K[1][2] = K12;
  K[1][3] = K13;
  K[2][2] = K22;
  K[2][3] = K23;
  K[3][3] = K33;
  forsymij
    *K[i][j] = 0;

  for (n = 0; MBL[n] != 0.0 && n < N_MWBL; n++) {
    xc = x-CBL[n][0];
    yc = y-CBL[n][1];
    zc = z-CBL[n][2];
    r  = sqrt(xc*xc+yc*yc+zc*zc);
    if (dequal(r, 0.0)) r = EPSZERO;

    N[1] = xc/r;
    N[2] = yc/r;
    N[3] = zc/r;
    
    fori P[i] = PBL[n][i-1];
    fori S[i] = SBL[n][i-1];
    
    NP = N[1]*P[1] + N[2]*P[2] + N[3]*P[3];
    
    eNS[1] = N[2]*S[3] - N[3]*S[2];
    eNS[2] = N[3]*S[1] - N[1]*S[3];
    eNS[3] = N[1]*S[2] - N[2]*S[1];
    
    forsymij 
      *K[i][j] +=
      1.5/(r*r) * (N[i]*P[j] + N[j]*P[i] - (delta(i,j) - N[i]*N[j])*NP) -
      3.0/(r*r*r) * (eNS[i]*N[j] + eNS[j]*N[i]);
  }

#if 0
  if (KK);
    *KK = K11[ijk]*K11[ijk] + K22[ijk]*K22[ijk] + K33[ijk]*K33[ijk] + 
          2*K12[ijk]*K12[ijk] + 2*K13[ijk]*K13[ijk] + 2*K23[ijk]*K23[ijk];
#endif

  if (fisheye)
    FishEye_IndexTransDDSym(xold, yold, zold, K11, K12, K13, K22, K23, K33);
}




void set_byK_GH(pGH *GH)
{
  int ijk;
  double *xp, *yp, *zp, *K11, *K12, *K13, *K22, *K23, *K33;

  xp = GFDataByName(GH, "grid::x"); 
  yp = GFDataByName(GH, "grid::y"); 
  zp = GFDataByName(GH, "grid::z"); 
  K11 = GFDataByName(GH, "admbase::kxx");
  K12 = GFDataByName(GH, "admbase::kxy");
  K13 = GFDataByName(GH, "admbase::kxz");
  K22 = GFDataByName(GH, "admbase::kyy");
  K23 = GFDataByName(GH, "admbase::kyz");
  K33 = GFDataByName(GH, "admbase::kzz");

  /* FIXME: we assume 3D grid functions here */
  if (dalota) {
    for (ijk = 0; ijk < GH->GFExtras [2]->npoints; ijk++) {
      dltKofxyz(xp[ijk], yp[ijk], zp[ijk], 
	       K11+ijk, K12+ijk, K13+ijk, K22+ijk, K23+ijk, K33+ijk);
    }
  } else {
    for (ijk = 0; ijk < GH->GFExtras [2]->npoints; ijk++) {
      byKofxyz(xp[ijk], yp[ijk], zp[ijk], 
	       K11+ijk, K12+ijk, K13+ijk, K22+ijk, K23+ijk, K33+ijk);
    }
  }
}




void set_byK_box(struct box *g)
{
  int i, j, k, ijk;
  double x, y, z;
  double *K11, *K12, *K13, *K22, *K23, *K33;

  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);

  setrangePUGH(g, 0);
  if (dalota) {
    forrange(g) {
      x = g->x + i*g->d;
      y = g->y + j*g->d;
      z = g->z + k*g->d;
      dltKofxyz(x, y, z, K11+ijk, K12+ijk, K13+ijk, K22+ijk, K23+ijk, K33+ijk);
    }
  } else {
    forrange(g) {
      x = g->x + i*g->d;
      y = g->y + j*g->d;
      z = g->z + k*g->d;
      byKofxyz(x, y, z, K11+ijk, K12+ijk, K13+ijk, K22+ijk, K23+ijk, K33+ijk);
    }
  }
}





/* L(u) = Delta u + N (1 + M u)^-7 
   M = (Sum m/(2r))^-1
   N = 1/8 M^7 K_ij K^ij
*/
void init_brbr_box(struct box *g)
{
  int i, j, k, ijk;
  int u, v, w, f;
  double *M, *N;
  double *phi, *dphi;
  double *K11, *K12, *K13, *K22, *K23, *K33;
  double K2;
  
  setmgind(g, &u, &v, &w, &f);
  setfieldpointers_brbr(g, &M, &N);
  setfieldpointers_psi(g, &phi, &dphi, &dphi, &dphi);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);

  if (fisheye) {
    if (0) printf("--> calling set_laplace()\n");
    set_laplace(g);
    setrangeall(g);
    if (0) printf("--> calling set_K2()\n");
    /* from standalone maximal: N = - K_ij K^ij 
       call this utility function since fisheye has a nontrivial g^ij
    */
    set_K2(g, N, M);
  }

  setrangePUGH(g, 0);
  forrange(g) {

    M[ijk] = 1/(phi[ijk]-1);

    if (fisheye) {
      N[ijk] = 0.125 * pow(M[ijk], 7.0) * (-N[ijk]);
    } else {
      K2 = K11[ijk]*K11[ijk] + K22[ijk]*K22[ijk] + K33[ijk]*K33[ijk] + 
	   2*K12[ijk]*K12[ijk] + 2*K13[ijk]*K13[ijk] + 2*K23[ijk]*K23[ijk];
      N[ijk] = 0.125 * pow(M[ijk], 7.0) * K2;
    }

    g->store[u][ijk] = 1;
    g->store[v][ijk] = 0;
    g->store[w][ijk] = 0;
    g->store[f][ijk] = 0;
  }
}




void set_brbr_box(struct box *g)
{
  int i, j, k, ijk;
  int u, v, w, f;
  double *M, *N;
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;
  double psi, psi2, phi, psi4ophi4;

  setfieldpointers_brbr(g, &M, &N);
  setfieldpointers_gab(g, &g11, &g12, &g13, &g22, &g23, &g33);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);
  setmgind(g, &u, &v, &w, &f);

  setrangePUGH(g, 0);
  forrange(g) {

    psi = g->store[u][ijk] + 1/M[ijk];
    psi2 = psi*psi;
    phi = 1 + 1/M[ijk];
    psi4ophi4 = pow(psi/phi, (double) 4);

    g11[ijk] *= psi4ophi4;
    g12[ijk] *= psi4ophi4;
    g13[ijk] *= psi4ophi4;
    g22[ijk] *= psi4ophi4;
    g23[ijk] *= psi4ophi4;
    g33[ijk] *= psi4ophi4;

    K11[ijk] /= psi2;
    K12[ijk] /= psi2;
    K13[ijk] /= psi2;
    K22[ijk] /= psi2;
    K23[ijk] /= psi2;
    K33[ijk] /= psi2;
    
    if (0) {
      shrinkrange(g);
      lop(g, u, v);
      setrangeall(g);
    }
  }
}





/* initialize
   we don't set one level and then inject because for FMR we can't
*/
void init_brbr(int init, int ltop, int lmax, struct mgstack **s) 
{
  int l;
  int u, v, w, f;
  struct box *g;
  int pr = vb;

  if (0) printf("BAM: init_brbr\n");
  if (!init) return;

  for (l = 1; l <= lmax; l++)
    forlevel(l, g) {

      /* time invariant conformal factor for evolution */
      set_mwbl_box(g);

      /* Bowen-York extrinsic curvature */
      set_byK_box(g);

      /* initialize coefficients M, N, and multigrid fields */
      init_brbr_box(g);
    }

  /* print */
  for (l = lmax; l >= 1; l--)
    forlevel(l, g) {
      if (0) prprim("init brbr", g, MGF+0);
      if (0) prprim("init brbr", g, MGF+1);
    }

  if (pr) printf("BAM: initialized fields, starting solve\n\n");
}




/* fill in final gab and Kab */
void set_brbr(int ltop, int lmax, struct mgstack **s) 
{
  int l;
  int u, v, w, f;
  struct box *g;
  int pr = 0;

  /* set final data */
  if (vb) printf("BAM: given conformal factor, set brbr\n");

  for (l = ltop; l <= lmax; l++)
    forlevel(l, g) {

      /* scale g and K by appropriate conformal factors */
      set_brbr_box(g);
      setBM(g->GH);
    }

  if (pr) printf("BAM: computed Brandt-Bruegmann data\n\n");
}














