/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* brill.c */
/* Bernd Bruegmann, 9/96, 6/97 */

/* brillofxyz() computes metric components of a "Brill wave" in 
   Cartesian coordinates at a given point

   Brill (1959):

     ds^2 = e^(2q) (drho^2 + dz^2) + rho^2 dphi^2

   We use A PARTICULAR choice for the function q, which may be
   changed maintaining regularity at the center and proper fall-off.

   NOTE: The physical metric is psi^4 times the Brill metric
         Brill plus Schwarzschild not finished
*/


#include "bbmg.h"


double brill_a, brill_b, brill_c, brill_BHmass;
double brill_rho0, brill_r0, brill_srho, brill_sr, brill_sz, brill_dfindif;
int brill_q;
double dfindif;

double q_brillofxyz(double x, double y, double z);
double q_brillofrhoz(double rho, double z);
void set_admcon_box(struct box *g);




/**************************************************************************/
/* Brill q function */
double q_brillofrhoz(double rho, double z)
{
  double drho, dr2;
  double q = 0.0;
  
  if (brill_q == 2) {

    dr2 = (rho*rho + z*z - brill_r0*brill_r0)/(brill_sr*brill_sr);
    q = brill_a * pow(rho/brill_srho, brill_b) / exp(pow(dr2, brill_c/2));

  } else if (brill_q == 1) {

    dr2 = (rho*rho + z*z - brill_r0*brill_r0)/(brill_sr*brill_sr);
    q = brill_a * pow(rho/brill_srho, brill_b) / (1 + pow(dr2, brill_c/2));  

  } else if (brill_q == 0) {

    drho = (rho - brill_rho0)/brill_srho;
    q = (rho < dequaleps) ? 0 : 
      brill_a * pow(rho, 2.0+brill_b) * exp(-drho*drho) / (rho*rho + z*z);
    if (brill_sz != 0) 
      q *= exp(-z*z/(brill_sz*brill_sz)); 
  } else 
    errorexit("BAM q_brillofrhoz(): unknown q type");

  return q;
}





void prinfo_brill()
{
  printf("q type = %d\n", brill_q);
  
  if (brill_q == 0) {
    printf(
      "q = a rho^b sin^2(theta) / exp((rho-rho0)^2/srho^2) / exp(z^2/sz^2)\n");
    printf("a     = %10.3e\n", brill_a);
    printf("b     = %10.3e\n", brill_b);
    printf("srho  = %10.3e\n", brill_srho);
    printf("rho0  = %10.3e\n", brill_rho0);
    printf("sz    = %10.3e\n", brill_sz);
  }
  
  if (brill_q == 1)
    printf("q = a (rho/srho)^b / (1 + [(r^2-r0^2)/sr^2]^(c/2)) \n");

  if (brill_q == 2)
    printf("q = a (rho/srho)^b / exp([(r^2-r0^2)/sr^2]^(c/2)) \n");

  if (brill_q == 1 || brill_q == 2) {
    printf("a     = %10.3e\n", brill_a);
    printf("b     = %10.3e\n", brill_b);
    printf("c     = %10.3e\n", brill_c);
    printf("srho  = %10.3e\n", brill_srho);
    printf("sr    = %10.3e\n", brill_sr);
    printf("r0    = %10.3e\n", brill_r0);
  }

  if (vb) printf("dfindif = %10.3e\n", brill_dfindif);
  printf("\n");

  if (brill_BHmass != 0.0) {
    printf("BHmass  = %10.3e\n", MBL[0]); 
    printf("BHx     = %10.3e\n", CBL[0][0]); 
    printf("BHy     = %10.3e\n", CBL[0][1]); 
    printf("BHz     = %10.3e\n", CBL[0][2]); 
    printf("\n");
  }
}

/**************************************************************************/




void setnames_brill() 
{
  int i = 0;

  fieldtitle[i++] = "u";
  fieldtitle[i++] = "v";
  fieldtitle[i++] = "w";
  fieldtitle[i++] = "f";
  fieldtitle[i++] = "M";
  fieldtitle[i++] = "N";
}



void setfieldpointers_brill(struct box *g, double **M, double **N)
{
  int i = MGF*g->ncomps;

  *M = g->store[i++];
  *N = g->store[i++];
}



/* tansform a diagonal matrix in cylindrical coordinates to a matrix in 
   cartesian coordinates
   not valid at rho = 0, because limit needed for lp1 * gpp,
   e.g. gpp = rho^2 cancels divergence
*/ 
void cyltocar(double x, double y, double z,
         double grr, double gpp, double gzz,
         double *g11, double *g12, double *g13, 
         double *g22, double *g23, double *g33)
{
  double lr1, lr2, lr3, lz1, lz2, lz3, lp1, lp2, lp3;
  double rho2 = x*x+y*y;
  double rho = sqrt(rho2);

  if (rho < dequaleps) {
    rho = dequaleps;
    rho2 = rho*rho;
    x = 0;
    y = dequaleps;
  }

  if (rho2 == 0.0)
    errorexit("cyltocar(): rho == 0.0");

  lr1 = x/rho;
  lr2 = y/rho;
  lr3 = 0;
  
  lp1 = - y/rho2;
  lp2 = x/rho2;
  lp3 = 0;

  lz1 = 0;
  lz2 = 0;
  lz3 = 1;

  *g11 = lr1*lr1*grr + lz1*lz1*gzz + lp1*lp1*gpp; 
  *g12 = lr1*lr2*grr + lz1*lz2*gzz + lp1*lp2*gpp; 
  *g13 = lr1*lr3*grr + lz1*lz3*gzz + lp1*lp3*gpp; 
  *g22 = lr2*lr2*grr + lz2*lz2*gzz + lp2*lp2*gpp; 
  *g23 = lr2*lr3*grr + lz2*lz3*gzz + lp2*lp3*gpp; 
  *g33 = lr3*lr3*grr + lz3*lz3*gzz + lp3*lp3*gpp; 
}





/* Brill metric coefficients for external parameters */
void brillofrhoz(double rho, double z, double *grr, double *gpp, double *gzz)
{
  *grr = *gzz = exp(2.0*q_brillofrhoz(rho, z));
  *gpp = rho*rho;
}





/* Brill metric coefficients mapped onto Cartesian coordinates */
void brillofxyz(double x, double y, double z,          
		double *g11, double *g12, double *g13, 
		double *g22, double *g23, double *g33)
{
  double rho = sqrt(x*x + y*y);
  double grr, gpp, gzz;

  if (0 && rho < dequaleps) rho = dequaleps;
  if (1 && rho < dequaleps) {
    *g11 = *g22 = *g33 = 1.0;
    *g12 = *g13 = *g23 = 0.0;
  } else {
    brillofrhoz(rho, z, &grr, &gpp, &gzz);

    /* might want cyclic rotation of x, y, z */
    if (1)
      cyltocar(x, y, z, grr, gpp, gzz, g11, g12, g13, g22, g23, g33);
    else
      cyltocar(x, y, z, grr, gpp, gzz, g22, g23, g12, g33, g13, g11);
  }
}




/* Brill q function: Cartesian */
double q_brillofxyz(double x, double y, double z)
{
  return q_brillofrhoz(sqrt(x*x + y*y), z);
}




/* compare grt/brill.nb */
void MN_brillofxyz(double x, double y, double z, double *M, double *N)
{
  double q, qzz, qrhorho;
  double d, uc, up, um;
  double drho, drho2, r, r2, r4, rho, rho2, rho2b, rho4, z2;
  double b = brill_b, var = brill_srho*brill_srho;

  /* z-axis is special */
  rho = sqrt(x*x + y*y);
  if (1 && rho < dequaleps) rho = dequaleps; 
  if (0 && rho < dequaleps) {
    *M = 0;
    *N = 0;
    return;
  }

  /* finite differencing for derivatives */
  if (1) {
    d = dfindif;
    uc = q_brillofrhoz(rho, z);
    up = q_brillofrhoz(rho+d, z);
    um = q_brillofrhoz(rho-d, z);
    qrhorho = (up - 2*uc + um)/(d*d);
    uc = q_brillofrhoz(rho, z);
    up = q_brillofrhoz(rho, z+d);
    um = q_brillofrhoz(rho, z-d);
    qzz = (up - 2*uc + um)/(d*d);
  }

  /* analytic available? */
  if (0) {
    /* helpers */
    z2 = z*z;
    rho2 = rho*rho;
    rho4 = rho2*rho2;
    rho2b = pow(rho, 2.0+brill_b);
    r = sqrt(x*x + y*y + z*z);
    r2 = r*r;
    r4 = r2*r2;
    drho = rho - brill_rho0;
    drho2 = drho*drho;
    
    /* Gaussian */
    q = brill_a * rho2b/r2 * exp(-drho2/var);
    
    /* derivatives */
    qzz = q * 2 * (3*z2 - rho2)/r4;
    qrhorho = q * (4*drho2*r4*rho2 - 4*drho*r2*rho*((2 + b)*r2 - 2*rho2)*var + 
		   var*(-2*(5 + 2*b)*r2*rho2*var + 8*rho4*var + 
			r4*(-2*rho2 + 2*var + 3*b*var + b*b*var))
		   )/(r4*rho2*var*var);
  }

  /* M and N with hopefully the correct sign and factor conventions */
  *M = 0.25 * (qzz + qrhorho);
  if (brill_BHmass != 0.0) {
    x -= CBL[0][0];
    y -= CBL[0][1];
    z -= CBL[0][2];
    *N = (*M) * brill_BHmass/sqrt(x*x+y*y+z*z)/2.0;
  } else
    *N = 0.0;
}





void init_brill_box(struct box *g)
{
  int i, j, k, ijk;
  int u, v, w, f;
  double x, y, z;
  double *M, *N;
  
  setmgind(g, &u, &v, &w, &f);
  setfieldpointers_brill(g, &M, &N);
  dfindif = (brill_dfindif < 0) ? g->d*fabs(brill_dfindif) : brill_dfindif;  

  setrangePUGH(g, 0);
  forrange(g) {
    x = g->x + i*g->d;
    y = g->y + j*g->d;
    z = g->z + k*g->d;

    MN_brillofxyz(x, y, z, &M[ijk], &N[ijk]);

    g->store[u][ijk] = 1;
    g->store[v][ijk] = 0;
    g->store[w][ijk] = 0;
    g->store[f][ijk] = 0;

    if (0) g->store[u][ijk] = sqrt(x*x + y*y + z*z);
  }
}




void set_brill_box(struct box *g)
{
  DECLARE_CCTK_PARAMETERS

  int pr = 0;
  int i, j, k, ijk;
  int u, v, w, f;
  double x, y, z;
  double *psi, psi4, *p[10];
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;
  void setfieldpointers_psi10(struct box *g, double **p);

  setfieldpointers_psi10(g, p);
  setfieldpointers_gab(g, &g11, &g12, &g13, &g22, &g23, &g33);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);
  setmgind(g, &u, &v, &w, &f);
  psi = g->store[u];

  setrangePUGH(g, 0);
  forrange(g) {
    x = g->x + i*g->d;
    y = g->y + j*g->d;
    z = g->z + k*g->d;
    brillofxyz(x, y, z, g11+ijk, g12+ijk, g13+ijk, g22+ijk, g23+ijk, g33+ijk);

    psi4 = psi[ijk]; if (0) psi4 = 1;

    if (brill_BHmass != 0) 
      psi4 = 1 + (psi4 - 1)/p[0][ijk];

    psi4 *= psi4;
    psi4 *= psi4;

    g11[ijk] *= psi4;
    g12[ijk] *= psi4;
    g13[ijk] *= psi4;
    g22[ijk] *= psi4;
    g23[ijk] *= psi4;
    g33[ijk] *= psi4;

    if (pr) g->store[v][ijk] = pow(psi[ijk],4.0);
    if (pr) g->store[w][ijk] = g33[ijk];
  }
  if (pr) prprim("psi^1", g, u);
  if (pr) prprim("psi^4", g, v);
  if (pr) prprim("g33", g, w);
  if (pr) prprimGF("gzz", g);

  forrange(g) K11[ijk] = 0;
  forrange(g) K12[ijk] = 0;
  forrange(g) K13[ijk] = 0;
  forrange(g) K22[ijk] = 0;
  forrange(g) K23[ijk] = 0;
  forrange(g) K33[ijk] = 0;

  if (cctk_itlast == 0 && brill_savepsi) {
    if (brill_BHmass != 0) {
      printf("BAM: saving psi/psi0 = 1 + (u-1)/psi0 in psi\n");
      forrange(g) p[0][ijk] = 1 + (psi[ijk] -1)/p[0][ijk];
    } else {
      printf("BAM: saving u in psi\n");
      forrange(g) p[0][ijk] = psi[ijk];
    }
    printf("BAM: saving L(u) in psix\n");
    shrinkrange(g);
    lop(g, u, v);
    forrange(g) p[1][ijk] = g->store[v][ijk];
  }

  /* symmetry boundaries */
  if (g->GH->GFExtras [g->GH->dim - 1]->nghostzones[0] > 1) {
    setob_2s(g->GH, CCTK_VarIndex("admbase::gxx"));
    setob_2s(g->GH, CCTK_VarIndex("admbase::kxx"));
  }
}




/* adm constraints */
void set_admcon_box(struct box *g)
{ 
  int pr = 0;
  int u, v, w, f;
  double *hamil, *diffeo1, *diffeo2, *diffeo3;
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;

  setfieldpointers_gab(g, &g11, &g12, &g13, &g22, &g23, &g33);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);
  setmgind(g, &u, &v, &w, &f);
  hamil = g->store[w];
  diffeo1 = g->store[f];
  diffeo2 = g->store[f];
  diffeo3 = g->store[f];

  setrangeall(g);
  subtract(g, w, w, w);
  shrinkrange(g);
  admcon(g, g11, g12, g13, g22, g23, g33, 
	    K11, K12, K13, K22, K23, K33, 
	    hamil, diffeo1, diffeo2, diffeo3);
  PUGHsyncvec(g, w);
  setrangeall(g);
  if (pr) prprim("hamil", g, w);

  /*  if (0) IO_Write1D(g->GH, g->GF[u]);*/

}





/* for testing: Schwarzschild */
void set_ss_box(struct box *g)
{
  int i, j, k, ijk;
  int u, v, w, f;
  double x, y, z, r;
  double psi4;
  double *g11, *g12, *g13, *g22, *g23, *g33;
  double *K11, *K12, *K13, *K22, *K23, *K33;

  setfieldpointers_gab(g, &g11, &g12, &g13, &g22, &g23, &g33);
  setfieldpointers_Kab(g, &K11, &K12, &K13, &K22, &K23, &K33);
  setmgind(g, &u, &v, &w, &f);

  setrangePUGH(g, 0);
  forrange(g) {
    x = g->x + i*g->d;
    y = g->y + j*g->d;
    z = g->z + k*g->d;

    x += 4;
    y += 5;
    z += 6;

    r = sqrt(x*x+y*y+z*z);
    if (r < dequaleps) r = dequaleps;
    psi4 = 1 + 0.5/r;

    x += -3;
    y += -5;
    z +=  4;

    r = sqrt(x*x+y*y+z*z);
    if (r < dequaleps) r = dequaleps;
    psi4 += 1/r;
    psi4 *= psi4;
    psi4 *= psi4;

    g11[ijk] = g22[ijk] = g33[ijk] = psi4;
    g12[ijk] = g13[ijk] = g23[ijk] = 0;
    K11[ijk] = K12[ijk] = K13[ijk] = K22[ijk] = K23[ijk] = K33[ijk] = 0;
  }
}





/* initialize
   we don't set one level and then inject because for FMR we can't
*/
void init_brill(int init, int ltop, int lmax, struct mgstack **s) 
{
  int l;
  int u, v, w, f;
  struct box *g;
  int pr = vb;

  if (pr) printf("BAM: init_brill\n");
  if (!init) return;

  for (l = 1; l <= lmax; l++)
    forlevel(l, g) {
      init_brill_box(g);
      if (0&&pr) prprim("init brill", g, MGF+0);
      if (0&&pr) prprim("init brill", g, MGF+1);
    }

  if (pr) printf("BAM: initialized fields, starting solve\n\n");
}




/* fill in final gab and Kab */
void set_brill(int ltop, int lmax, struct mgstack **s) 
{
  int l;
  int u, v, w, f;
  struct box *g;
  int pr = 0;


  /* set final data */
  if (vb) printf("BAM: given conformal factor, set brill\n");

  for (l = ltop; l <= lmax; l++)
    forlevel(l, g) {
      set_mwbl_box(g);
      set_brill_box(g);
      if (0) set_ss_box(g);
      if (0) set_admcon_box(g);
      setBM(g->GH);
    }

  if (pr) printf("BAM: computed Brill wave data\n\n");

  /* check numerics */
  if (0) {
    double g11, g12, g13, g22, g23, g33, M, N;
    double d, rho, z, qrhorho, qzz, uc, up, um;

    printf("\nchecking brill:\n");
    /* don't change those numbers, done by hand for cross check */
    brill_a    = -0.5;
    brill_b    = 2;
    brill_rho0 = 1;
    brill_srho  = sqrt(0.2);
    printf("should have q = -0.28271118, is %e\n",
	   q_brillofxyz(1.0, 1.0, 1.0));
    printf("  gxx = gyy = 0.78406008, gzz = 0.56812015, gxy = -0.21593992\n");
    brillofxyz(1.0, 1.0, 1.0, &g11, &g12, &g13, &g22, &g23, &g33);
    printf("%10.3e %10.3e %10.3e %10.3e %10.3e %10.3e\n",
	   g11, g12, g13, g22, g23, g33); 
    rho = sqrt(2.0);
    z = 1.0;
    d = 1e-6;
    uc = q_brillofrhoz(rho, z);
    up = q_brillofrhoz(rho+d, z);
    um = q_brillofrhoz(rho-d, z);
    qrhorho = (up - 2*uc + um)/(d*d);
    uc = q_brillofrhoz(rho, z);
    up = q_brillofrhoz(rho, z+d);
    um = q_brillofrhoz(rho, z-d);
    qzz = (up - 2*uc + um)/(d*d);
    M = 0.25 * (qrhorho + qzz);
    printf("from finite diff:   M = %e, qrhorho %e, qzz %e\n",M, qrhorho, qzz);
    MN_brillofxyz(1.0, 1.0, 1.0, &M, &N);
    printf("from MN_brillofxyz: M = %e\n", M);
/* 
from finite diff:   M = -2.284006e-01, qrhorho 1.890099e+00, qzz -6.289413e-02
from MN_brillofxyz: M = -2.284195e-01
=> probably not worth the trouble to derive analytic expression, but 
   d cannot be made much smaller
*/
  }
}





