/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* fasmg.c */
/* Bernd Bruegmann, 4/97 */

#include "bbmg.h"

struct box *level[MGLMAX];
char *fieldtitle[MGFMAX];
int wcycle_nmin, wcycle_nmax;
double wcycle_factol;

#define PR bbmgPR


/* memory usage is close to optimal in that only one auxiliary field is used
   a single scratch field should always be available anyway
   makes rbGS much simpler
   however, the algorithm would look simpler if auxiliary fields were
   available for the truncation error and the residuum
   ... already added one more field :)
*/

/* optimization:
   - make special versions of all restrictors and prolongators that
     add their result to a given field, subtract from and morf ...
*/




/* the one place where we define the indices for the MG fields */
void setmgind(struct box *g, int *u, int *v, int *w, int *f)
{
  if (!g) return;
  *u = g->ncomps*0;
  *v = g->ncomps*1;
  *w = g->ncomps*2;
  *f = g->ncomps*3;
}




/* free multigrid stack, should be problem independent */
void free_mgstack() 
{
  struct box *g, *gnext;
  int l;

  for (l = 1; l <= MGLMAX; l++)
    for (g = level[l]->next; g; g = gnext) {
      gnext = g->next;
      free(g);
    }
}





/* solve 
   recursive V or W cycles
*/
void solve(int l, int lmax, int npre, int npost) 
{
  int ressmallenough;
  int careaboutres = 0;
  int careaboutreslmax = 0;
  double res2, resi;
  int nwcycle = 0;
  int pr = PR;

  if (pr) printf(".............. entering solve level %d ............\n", l);

  if (l == 1) {

    exactsolve(l);

  } else {

    relax(l, npre);

    ressmallenough = finetocoarse(l, l-1);

    /* W cycles */
    while (1) {
      
      /* obsolete */
      if (0 && ressmallenough && careaboutres && 
	  !(careaboutreslmax && l != lmax))
	return;

      solve(l-1, lmax, npre, npost);

      /* 0 wcycles if we don't want W cycles at all (the default) */
      if (nwcycle >= wcycle_nmax) break;

      /* perform minimal number of cycles without any check */
      if (nwcycle < wcycle_nmin) continue;

      /* check errors to find out whether we are done with W cycling */
      res2 = resi = 0;
      restau(l-1, (tolres2)?&res2:0, (tolresi)?&resi:0, 0, 0, 0);
      if (vb) {
	printf(". . . . . . .  W-cycle %d  . . . . . . .\n", nwcycle);
        printf("l %d   res2 %e   tolres2 %e\n\n", l-1, res2, tolres2);
      }
      if (0) printf("resi %e   tolresi %e\n", resi, tolresi);

      if (res2 <= tolres2 && resi <= tolresi)
	break;
      
      nwcycle++;
    }

    coarsetofine(l-1, l);

    relax(l, npost);
  }
  if (pr) printf("..............  leaving solve level %d ............\n\n",l);
}




/* prepare coarse grid equation:  
     LH uH = fH
     uH = R uh
     fH = R fh + tauHh,  tauHh = L R uh - R L uh
   control: 
     no sense to make residuum smaller than estimate of
     true truncation error, which for refinement factor 2 is
     approximated by tauHh/3
   return 1 if |dh| < |tauHh|/3 
   return 0 else
*/
int finetocoarse(int lh, int lH) 
{
  struct box *gh, *gH;
  int u, v, w, f;
  double normres, normtau;
  double epsrestau = -1;
  int pr = PR;

  /* operate on fine seperately because this may be expensive */ 
  forlevel(lh, gh) {
                                 if (pr) prprim("uh", gh, u);
    shrinkrange(gh);
                                 if (pr) prprim("uh", gh, u);
    lop(gh, u, v);
                                 if (pr) prprim("vh = L(uh)", gh, v);
  }    

  /* for all the overlap between fine and coarse */
  foroverlap(lh, gh, lH, gH) {

    setrangeftoc(gh, gH);
    mgrid_restrict(gh, v, gH, v);
                                 if (pr) prprim("vH = R(vh)", gH, v);    

    setrangeall(gh);
    shrinkrange(gh);
    subtract(gh, v, f, v);
                                 if (pr) prprim("dh = vh = vh - fh", gh, v);
    if (pr) normres = l2norm(gh, v);

                                 if (pr) prprim("fh", gh, f);
                                 if (pr) prprim("uh", gh, u);
    setrangeftoc(gh, gH);
    mgrid_restrict(gh, u, gH, u);
                                 if (pr) prprim("uH = R(uh)", gH, u);
    lop(gH, u, f);
                                 if (pr) prprim("fH = LR(uh)", gH, f);
    subtract(gH, f, v, f);
                                 if (pr) prprim("tau = LRu-RLu = fH-vH", gH,f);
    if (pr) normtau = l2norm(gH, f);

    if (pr) printrestau(lh, normres, normtau);
    if (0 && normres < normtau * epsrestau) 
      return 1;

    mgrid_restrict(gh, f, gH, v);
                                 if (pr) prprim("fH", gH, v);
    add(gH, v, f, f);
                                 if (pr) prprim("fH = R(fh) + tau", gH, f);
    setrangeall(gH);
    copy(gH, u, w);
                                 if (pr) prprim("uHold = uH", gH, w);
  }

  return 0;
}






/* perform coarse grid correction
     uh = uh + P (uH - R uh)
   one of several beautiful instances where doing something harmless
     like setting P R uh = uh spoils everything
*/
void coarsetofine(int lH, int lh)
{
  struct box *gH, *gh;
  int u, v, w, f;
  int pr = PR;

  /* for all the overlap between fine and coarse */
  foroverlap(lh, gh, lH, gH) {

#if 0
    /* reconstruct uHold, or use one more auxiliary field for uHold */
    injectftoc(gh, u, gH, v); /* DIR compatible if DIR by injection anyway */
    setrangeftoc(gh, gH);
    mgrid_restrict(gh, u, gH, v); 
                                     if (pr) prprim("uoldH = R(uh)", gH, v);
#endif
    /* correct fine u */
    /* (!) don't set vH and vh to zero to allow update of ref bound */
    setrangectof(gH, gh);
    subtract(gH, u, w, v);
                                     if (pr) prprim("vH = uH - uoldH", gH, v);
    prolong(gH, v, gh, v);
                                     if (pr) prprim("vh = P(vH)", gh, v);
    add(gh, u, v, u);
                                     if (pr) prprim("uh = uh + vh", gh, u);
  }
}




/* prolongate u */
void prolong_u(int lH, int lh)
{
  struct box *gH, *gh;
  int u, v, w, f;
  int pr = PR;

  if (vb || pr) 
    printf(".............. prolongating u %d -> %d .............\n", lH, lh);

  /* for all the overlap between fine and coarse */
  foroverlap(lh, gh, lH, gH) {

    /* prolongate */
                               if (pr) prprim("prolong_uH", gH, u);
                               if (0) prprim("prolong_uh", gh, u);
    setrangectof(gH, gh);
    prolong(gH, u, gh, u);
    setrangeall(gh);
                               if (pr) prprim("prolong_u: uh after PHh", gh,u);
    setboundmask(gh, u, 1);
                               if (pr) prprim("prolong_u: uh bound set", gh,u);
  }
}




/* find residual for interior of a given box, uses v as temporary storage */
double findres(int p, struct box *gh)
{
  double normres;
  int u, v, w, f;

  setrangeall(gh);
  shrinkrange(gh);
  setmgind(gh, &u, &v, &w, &f);
  lop(gh, u, v);
  if (p) prprim("\nfindres uh", gh, u);
  if (p) prprim("findres fh", gh, f);
  if (p) prprim("findres luh", gh, v);
  subtract(gh, v, f, v);
  normres = l2norm(gh, v);
  if (p) prprim("findres res", gh, v);
  if (p) printf(" %2dx%2dx%d  %6.3e\n", gh->m, gh->n, gh->o, normres);
  return normres;
}  





/* for diagnostics, compute residuals and truncation errors for one level 
   does not leave fields unchanged:  vh = resh, uH = Ruh, wH = tauH,
*/
void restau(int lh, double *res2, double *resi, double *tau2, double *taui, 
	    int p) 
{
  int lH = lh-1;
  struct box *gh, *gH;
  int u, v, w, f;
  int n;
  double x;
  int pr = PR;

  if (pr) printf(".............. prnormres l%d ............\n", lh);

  forlevel(lh, gh) {
    shrinkrange(gh);

    /* residuum, only good for printing since not cumulative */
    if (resi || res2) {
                                 if (pr) prprim("uh", gh, u);    
      lop(gh, u, v);
                                 if (pr) prprim("vh = L(uh)", gh, v);    
      subtract(gh, v, f, v);
                                 if (pr) prprim("dh = vh = vh - fh", gh, v);
      if (resi) *resi = linfnorm(gh, v);
      if (res2) *res2 = l2norm(gh, v);
    }

    /* truncation error
       to be independent of nesting property, average over boxes */
    if (!(taui || tau2)) continue;
    if (taui) *taui = 0;
    if (tau2) *tau2 = 0;
    n = 0;
    forlevel(lH, gH) {
      if (setoverlap(gh,gH)) continue; 
      setrangeftoc(gh, gH);

      mgrid_restrict(gh, v, gH, v);
                                 if (pr) prprim("vH = R(vh)", gH, v);    

                                 if (pr) prprim("uh", gh, u);
      mgrid_restrict(gh, u, gH, u);
                                 if (pr) prprim("uH = R(uh)", gH, u);
      lop(gH, u, w);
                                 if (pr) prprim("wH = LR(uh)", gH, w);
      subtract(gH, w, v, w);
                                 if (pr) prprim("tau = LRu-RLu = fH-vH", gH,w);
      if (taui) {			 
	x = linfnorm(gH, w);
	if (*taui < x) *taui = x;
      }
      if (tau2) {
	*tau2 += l2norm(gH, w);
	n++; /* unused */
      }
    }
    
    /* turn tau^H_h into tau^H, the relative local truncation error */
    if (taui) *taui /= 3;
    if (tau2) *tau2 /= 3;

    /* print */
    if (p) {
      printf("  l%-2d  %3dx%3dx%3d   ", lh, gh->m, gh->n, gh->o);
      if (0) {
	printf(" %.9e %.9e  ", *res2, *resi);
	if (n) printf(" %.9e %.9e", *tau2, *taui);
      } else {
	printf(" %.3e %.3e  ", *res2, *resi);
	if (n) printf(" %.3e %.3e", *tau2, *taui);
      }
    }
  }
  if (p) printf("\n");
}




/* for diagnostics, compute and print residuals per grid for one level 
   does not leave fields unchanged:  vh = resh, uH = Ruh, wH = tauH,
*/
void prnormres(int lh) 
{
  double res2, resi, tau2, taui;

  restau(lh, &res2, &resi, &tau2, &taui, 1);
}




/* for diagnostics, compute and print residuals per grid for one level 
   does not leave fields unchanged:  vh = resh, uH = Ruh, wH = tauH,
*/
void prnormresold(int lh) 
{
  int lH = lh-1;
  struct box *gh, *gH;
  int u, v, w, f;
  int n;
  double res2, resi;
  double tau2, taui, x;
  int pr = 0;

  if (pr) printf(".............. prnormres l%d ............\n", lh);

  forlevel(lh, gh) {
    shrinkrange(gh);

    /* residuum */
    lop(gh, u, v);
                                 if (pr) prprim("vh = L(uh)", gh, v);    
    subtract(gh, v, f, v);
                                 if (pr) prprim("dh = vh = vh - fh", gh, v);
    resi = linfnorm(gh, v);
    res2 = l2norm(gh, v);

    /* truncation error
       to be independent of nesting property, average over boxes */
    tau2 = taui = 0;
    n = 0;
    forlevel(lH, gH) {
      if (setoverlap(gh,gH)) continue; 
      setrangeftoc(gh, gH);

      mgrid_restrict(gh, v, gH, v);
                                 if (pr) prprim("vH = R(vh)", gH, v);    

                                 if (pr) prprim("uh", gh, u);
      mgrid_restrict(gh, u, gH, u);
                                 if (pr) prprim("uH = R(uh)", gH, u);
      lop(gH, u, w);
                                 if (pr) prprim("wH = LR(uh)", gH, w);
      subtract(gH, w, v, w);
                                 if (pr) prprim("tau = LRu-RLu = fH-vH", gH,w);
      x = linfnorm(gH, w);
      if (taui < x) taui = x;
      tau2 += l2norm(gH, w);
      n++;
    }

    /* print */
    printf("  l%-2d  %3dx%3dx%3d   ", lh, gh->m, gh->n, gh->o);
    if (0) {
      printf(" %.9e %.9e  ", res2, resi);
      if (n) printf(" %.9e %.9e", tau2, taui);
    } else {
      printf(" %.3e %.3e  ", res2, resi);
      if (n) printf(" %.3e %.3e", tau2, taui);
    }
  }
  printf("\n");
}




/* for diagnostics, print residuals for all grids on all levels 
   does not leave u and f on l=1..lmax-1 unchanged
*/
void prresiduals(int lmax)
{
  int l;  

  printf("level      size        l2(res)   li(res)     l2(tau)   li(tau)\n");
  for (l = 1; l <= lmax; l++)
    prnormres(l);
  printf("\n");
  fflush(stdout);
}





