/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* misner.c */
/* Bernd Bruegmann, 05/00 */

/* version of misner data using coordinate transformations the same way
   the puncture data does (e.g. finite differencing for derivs of psi)
   compare Misner_standard.c for actual formulas
*/


#include "bbmg.h"


/* set to 0 for compatibility with Misner_standard.c
          1 avoid singularities
*/
#define CHECK0 0

#define EPSZERO (dequaleps/1000.0)



static int local_nmax;
static double local_mu;




/* Cadez lapse for Misner data */
double misneralp(double x, double y, double z)
{
  int n;
  double csch, coth, xx, yy, r1, r2;
  double a = 1, p = 1, s = 1;

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  for (n = local_nmax; n >= 1; n--) {
    csch = 1/sinh(local_mu*n);
    coth = 1/tanh(local_mu*n);
    xx = x*x;
    yy = y*y;
    r1 = sqrt(xx + yy + (z+coth)*(z+coth));
    r2 = sqrt(xx + yy + (z-coth)*(z-coth));
#if CHECK0
    if (dequal(r1, 0.0)) r1 = EPSZERO;
    if (dequal(r2, 0.0)) r2 = EPSZERO;
#endif
    s *= -1;
    a += s * csch * (1/r1 + 1/r2);
    p +=     csch * (1/r1 + 1/r2);
  }
  return a / p;
}




/* conformal factor of Misner data */
double misnerpsi(double x, double y, double z)
{
  int n;
  double csch, coth, xx, yy, r1, r2;
  double p = 1;

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  for (n = local_nmax; n >= 1; n--) {
    csch = 1/sinh(local_mu*n);
    coth = 1/tanh(local_mu*n);
    xx = x*x;
    yy = y*y;
    r1 = sqrt(xx + yy + (z+coth)*(z+coth));
    r2 = sqrt(xx + yy + (z-coth)*(z-coth));
#if CHECK0
    if (dequal(r1, 0.0)) r1 = EPSZERO;
    if (dequal(r2, 0.0)) r2 = EPSZERO;
#endif
    p += csch * (1/r1 + 1/r2);
  }

  return p;
}


double dmisnerpsi(double x, double y, double z,
		       int index )
{
  int n;
  double csch, coth, xx, yy, r1, r2;
  double pi=0;
  double x1i,x2i;


  if (fisheye) FishEye_CoordTrans(&x, &y, &z);



  for (n = local_nmax; n >= 1; n--) {
    csch = 1/sinh(local_mu*n);
    coth = 1/tanh(local_mu*n);
    xx = x*x;
    yy = y*y;

    if(index==1)x1i=x2i=x;
    else if (index==2)x1i=x2i=y;
    else if (index==3){
      x1i=z+coth;
      x2i=z-coth;
    } 

    r1 = sqrt(xx + yy + (z+coth)*(z+coth));
    r2 = sqrt(xx + yy + (z-coth)*(z-coth));

#if CHECK0
    if (dequal(r1, 0.0)) r1 = EPSZERO;
    if (dequal(r2, 0.0)) r2 = EPSZERO;
#endif
    pi += -csch * (x1i/r1/r1/r1 + x2i/r2/r2/r2);
    
  }
  
  return pi;
}

double ddmisnerpsi(double x, double y, double z,
		       int index1, int index2 )
{
  int n;
  double csch, coth, xx, yy, r1, r2;
  double pij=0;
  double x1i,x2i,x1j,x2j;

  if (fisheye) FishEye_CoordTrans(&x, &y, &z);

  for (n = local_nmax; n >= 1; n--) {
    csch = 1/sinh(local_mu*n);
    coth = 1/tanh(local_mu*n);
    xx = x*x;
    yy = y*y;

    if(index1==1)x1i=x2i=x;
    else if (index1==2)x1i=x2i=y;
    else if (index1==3){
      x1i=z+coth;
      x2i=z-coth;
    } 
    if(index2==1)x1j=x2j=x;
    else if (index2==2)x1j=x2j=y;
    else if (index2==3){
      x1j=z+coth;
      x2j=z-coth;
    } 

    r1 = sqrt(xx + yy + (z+coth)*(z+coth));
    r2 = sqrt(xx + yy + (z-coth)*(z-coth));

#if CHECK0
    if (dequal(r1, 0.0)) r1 = EPSZERO;
    if (dequal(r2, 0.0)) r2 = EPSZERO;
#endif
    pij += 3*csch * (x1i*x1j/r1/r1/r1/r1/r1 + x2i*x2j/r2/r2/r2/r2/r2);
    if(index1==index2)pij -= csch * (1/r1/r1/r1 + 1/r2/r2/r2);
  }
  
  return pij;
}


/* psi and its derivatives, finite difference calculation */
void edmisnerpsi(double sx, double sy, double sz, 
		 double *psi,
		 double *dpsidx, double *dpsidy, double *dpsidz,
		 double *ddpsidxdx, double *ddpsidxdy, double *ddpsidxdz, 
		 double *ddpsidydy, double *ddpsidydz, double *ddpsidzdz)
{
  double x=sx,y=sy,z=sz;

  *dpsidx=dmisnerpsi(x,y,z,1);
  *dpsidy=dmisnerpsi(x,y,z,2);
  *dpsidz=dmisnerpsi(x,y,z,3);
  *ddpsidxdx=ddmisnerpsi(x,y,z,1,1);
  *ddpsidxdy=ddmisnerpsi(x,y,z,1,2);
  *ddpsidxdz=ddmisnerpsi(x,y,z,1,3);
  *ddpsidydy=ddmisnerpsi(x,y,z,2,2);
  *ddpsidydz=ddmisnerpsi(x,y,z,2,3);
  *ddpsidzdz=ddmisnerpsi(x,y,z,3,3);
    
  if(fisheye){
    FishEye_StdToFish_ddPsi_dPsi(x, y, z, 
			 ddpsidxdx, ddpsidxdy, ddpsidxdz, 
			 ddpsidydy, ddpsidydz, ddpsidzdz, 
			 dpsidx, dpsidy, dpsidz);
  }

  *psi = misnerpsi(x,y,z);

  *dpsidx/=*psi; 
  *dpsidy/=*psi; 
  *dpsidz/=*psi; 

  *ddpsidxdx/=*psi; 
  *ddpsidxdy/=*psi; 
  *ddpsidxdz/=*psi; 
  *ddpsidydy/=*psi; 
  *ddpsidydz/=*psi; 
  *ddpsidzdz/=*psi; 
}



/* set up misner data */
void set_misner(cGH *cctkGH)
{
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  char s[1000];
  double mass;
  int i, n, npoints = cctk_lsh[0] * cctk_lsh[1] * cctk_lsh[2];

  /* info */
  sprintf(s, "Setting Misner data with mu = %f, nmax = %d, BHs on z-axis\n", 
	  misner_mu, misner_nmax);
  CCTK_INFO(s);

  /* set local parameters */
  local_mu = misner_mu;
  local_nmax = misner_nmax;

  /* conformal factor psi and its derivatives */
  for (i = 0; i < npoints; i++)
    edmisnerpsi(x[i], y[i], z[i],
		psi+i, psix+i, psiy+i, psiz+i, 
		psixx+i, psixy+i, psixz+i, psiyy+i, psiyz+i, psizz+i);

  /* ADM mass */
  for (mass = 0, n = 1; n <= misner_nmax; n++)
    mass += 4/sinh(n*misner_mu);
  sprintf(s, "ADM mass is %f", mass);
  CCTK_INFO(s);

  /* Cadez lapse */
  if (CCTK_Equals(initial_lapse,"misner_cadez")) { 
    CCTK_INFO("Initialise with cadez lapse");
    for (i = 0; i < npoints; i++)
      alp[i] = misneralp(x[i], y[i], z[i]);
  }

  /* metric */
  CCTK_INFO("Uses conformal metric and conformal derivatives");

  if (conformal_metric) {
    for (i = 0; i < npoints; i++) {
      gxx[i] = gyy[i] = gzz[i] = 1;
      gxy[i] = gxz[i] = gyz[i] = 0;
    }
  } else {
    for (i = 0; i < npoints; i++) {
      gxx[i] = gyy[i] = gzz[i] = pow(psi[i], 4.0);
      gxy[i] = gxz[i] = gyz[i] = 0;
    }
  }
  if (fisheye) {
    for (i = 0; i < npoints; i++) {
      FishEye_IndexTransDDSym(x[i], y[i], z[i],
		       gxx+i, gxy+i, gxz+i, gyy+i, gyz+i, gzz+i);
    }
  }
  
  /* extrinsic curvature */
  for (i = 0; i < npoints; i++) {
    kxx[i] = kyy[i] = kzz[i] = 0;
    kxy[i] = kxz[i] = kyz[i] = 0;
  }
  
}
