/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann 
*/
/* prolong.c */
/* Bernd Bruegmann, 4/97 */

#include "bbmg.h"


void injectctof_scalar(struct box *rc, struct box *r, double *uc, double *u);
void ipolr2o1_scalar(struct box *r, double *u);
void BIB_injectctof(struct box *rc, struct box *r, double *uc, double *u);


/* vector prolongation operator, refinement factor 2 */
void prolong(struct box *rc, int uc, struct box *r, int u)
{
  int pr = 0;
  int prolongorder = 1;
  int nc;

  /* inject */
  for (nc = 0; nc < r->ncomps; nc++)
    injectctof_scalar(rc, r, rc->store[uc+nc], r->store[u+nc]);
  PUGHsyncvec(r, u);
                               if (pr) prprim("P: after inject", r, u);

  /* interpolate in place */ 
  for (nc = 0; nc < r->ncomps; nc++)
    if (prolongorder == 1) ipolr2o1_scalar(r, r->store[u+nc]);
  PUGHsyncvec(r, u);

  /* leave with consistent range for fine interior */
  if (r->bi0 == DIRICHLET) r->imin-=1;
  if (r->bi1 == DIRICHLET) r->imax+=1;
  if (r->bj0 == DIRICHLET) r->jmin-=1;
  if (r->bj1 == DIRICHLET) r->jmax+=1;
  if (r->bk0 == DIRICHLET) r->kmin-=1;
  if (r->bk1 == DIRICHLET) r->kmax+=1;
  setrangePUGH(r, 3);
}




/* inject coarse into fine, refinement factor 2 */
void injectctof_scalar(struct box *rc, struct box *r, double *uc, double *u)
{
  int i, j, k, ic, jc, kc;

  if (rc->l >= rc->ltop) {
    BIB_injectctof(rc, r, uc, u);
    return;
  }

  setrangePUGH(r, 2);
  setrangePUGH(rc, 1);
  forrangefc(r, rc)
    u[gijkof(r,i,j,k)] = uc[gijkof(rc,ic,jc,kc)];
}





/* interpolate, refinement factor 2, order 1, trilinear
   u must contain already the coarse grid points
   awkward to do it here, but since the sweeps are a mixture between
     coarse and fine, protect Dirichlet boundaries by setting coarse
     interior, expanding to fine interior after coarse sweep
     next rewrite: split into three sweeps (range, incri, incrj, incrk)
*/
void ipolr2o1_scalar(struct box *r, double *u) 
{
  int i, j, k, ijk;
  int di = r->di;
  int dj = r->dj;
  int dk = r->dk;
  int lijk = r->lijk;
  int imin = r->imin;
  int imax = r->imax;
  int jmin = r->jmin;
  int jmax = r->jmax;
  int kmin = r->kmin;
  int kmax = r->kmax;

  /* for the available coarse data, interpolate in k direction */
  if (r->bk0 == DIRICHLET) r->kmin-=2;
  if (r->bk1 == DIRICHLET) r->kmax+=2;
  setrangePUGH(r, 3);
  for (i = r->limin; i <= r->limax; i+=2)
    for (j = r->ljmin; j <= r->ljmax; j+=2)
      for (k = r->lkmin+1; k < r->lkmax; k+=2) {
	ijk = ijkof(i,j,k);
	u[ijk] = 0.5*(u[ijk+dk] + u[ijk-dk]);
      }

  /* for the now available data, interpolate in j direction */
  if (r->bj0 == DIRICHLET) r->jmin-=2;
  if (r->bj1 == DIRICHLET) r->jmax+=2;
  if (r->bk0 == DIRICHLET) r->kmin+=1;
  if (r->bk1 == DIRICHLET) r->kmax-=1;
  setrangePUGH(r, 3);
  for (i = r->limin; i <= r->limax; i+=2)
    for (j = r->ljmin+1; j < r->ljmax; j+=2)
      for (k = r->lkmin; k <= r->lkmax; k++) {
	ijk = ijkof(i,j,k);
	u[ijk] = 0.5*(u[ijk+dj] + u[ijk-dj]);
      }
  
  /* for the now available data, interpolate in i direction */
  if (r->bi0 == DIRICHLET) r->imin-=2;
  if (r->bi1 == DIRICHLET) r->imax+=2;
  if (r->bj0 == DIRICHLET) r->jmin+=1;
  if (r->bj1 == DIRICHLET) r->jmax-=1;
  setrangePUGH(r, 3);
  for (i = r->limin+1; i < r->limax; i+=2)
    for (j = r->ljmin; j <= r->ljmax; j++)
      for (k = r->lkmin; k <= r->lkmax; k++) {
	ijk = ijkof(i,j,k);
	u[ijk] = 0.5*(u[ijk+di] + u[ijk-di]);
      }

  /* restore original range */
  r->imin = imin;
  r->imax = imax;
  r->jmin = jmin;
  r->jmax = jmax;
  r->kmin = kmin;
  r->kmax = kmax;
}





#if 0
/* this is probably not needed here, but just to show that we can make
   it available ... */

/* unchanged from BAM */
/* interpolate, refinement factor 2, order 3, tricubic 
   u must contain already the coarse grid points
   gives same numerical answer as double version of Num Rec's polint
   note shifted stencil at boundary
*/
void ipolr2o3(double ***u, 
	      int imin, int imax, int jmin, int jmax, int kmin, int kmax)
{
  int i, j, k;
  int di, dj, dk;

  di = imax - imin;
  dj = jmax - jmin;
  dk = kmax - kmin;
  if (di && di < 6 || dj && dj < 6 || dk && dk < 6)  
    errorexit("ipolr2o3(): need width 0 or >= 6"); 

  /* for the available coarse data, interpolate in k direction */
  for (i = imin; i <= imax; i+=2) {
    for (j = jmin; j <= jmax; j+=2) {

      k = kmin+1;
      if (k-1 >= kmin && k+5 <= kmax)
	u[i][j][k] = 0.3125*(u[i][j][k-1] - u[i][j][k+3]) +
	             0.0625*(u[i][j][k+5] + u[i][j][k+1]*15.0);

      for (k+=2; k < kmax-2; k+=2)
	u[i][j][k] = 0.5625*(u[i][j][k+1] + u[i][j][k-1]) -
                     0.0625*(u[i][j][k+3] + u[i][j][k-3]);

      if (k-5 >= kmin && k+1 <= kmax)
	u[i][j][k] = 0.3125*(u[i][j][k+1] - u[i][j][k-3]) +
	             0.0625*(u[i][j][k-5] + u[i][j][k-1]*15.0);
    }
  }

  /* for the now available data, interpolate in j direction */
  for (i = imin; i <= imax; i+=2) {

    j = jmin+1;
    if (j-1 >= jmin && j+5 <= jmax)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.3125*(u[i][j-1][k] - u[i][j+3][k]) +
	             0.0625*(u[i][j+5][k] + u[i][j+1][k]*15.0);

    for (j+=2; j < jmax-2; j+=2)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.5625*(u[i][j+1][k] + u[i][j-1][k]) -
                     0.0625*(u[i][j+3][k] + u[i][j-3][k]);

    if (j-5 >= jmin && j+1 <= jmax)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.3125*(u[i][j+1][k] - u[i][j-3][k]) +
	             0.0625*(u[i][j-5][k] + u[i][j-1][k]*15.0);
  }

  /* for the now available data, interpolate in i direction */
  i = imin+1;
  if (i-1 >= imin && i+5 <= imax)
    for (j = jmin; j <= jmax; j++)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.3125*(u[i-1][j][k] - u[i+3][j][k]) +
	             0.0625*(u[i+5][j][k] + u[i+1][j][k]*15.0);

  for (i+=2; i < imax-2; i+=2)
    for (j = jmin; j <= jmax; j++)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.5625*(u[i+1][j][k] + u[i-1][j][k]) -
                     0.0625*(u[i+3][j][k] + u[i-3][j][k]);

  if (i-5 >= imin && i+1 <= imax)
    for (j = jmin; j <= jmax; j++)
      for (k = kmin; k <= kmax; k++)  
	u[i][j][k] = 0.3125*(u[i+1][j][k] - u[i-3][j][k]) +
	             0.0625*(u[i-5][j][k] + u[i-1][j][k]*15.0);
}


#endif



