/* thorn_BAM_Elliptic: Multigrid Elliptic Solver for Cactus
   Copyright (C) 1998 Bernd Bruegmann
*/
/* stencil.c */
/* Bernd Bruegmann, 4/97 */

#include "bbmg.h"


void stencil_flatlinear(struct box *g, int nfu, int ijk,
			double *lu, double *lii);
void stencil_linear(struct box *g, int nfu, int ijk,
		    double *lu, double *lii);
void stencil_linear_sym(struct box *g, int nfu, int ijk,
			double *lu, double *lii);
void stencil_brbr(struct box *g, int nfu, int ijk,
		  double *lu, double *lii);
void stencil_brbr_fe(struct box *g, int nfu, int ijk,
		     double *lu, double *lii);
void stencil_powers(struct box *g, int nfu, int ijk,
		  double *lu, double *lii);
void stencil_vecL(struct box *g, int nfu, int ijk,
		  double *lu, double *lii);
void stencil_vec4(struct box *g, int nfu, int ijk,
		  double *lu, double *lii);
void stencil_vecn(struct box *g, int nfu, int ijk,
		  double *lu, double *lii);




/* switch to different stencils
   to be efficient it is essential to have
     1. inlining of both stencil and stencil_xxx
     2. triple loop hoisting of if, reordering of if (see ellops.c)
*/
void stencil(struct box *g, int u, int ijk, double *lu, double *lii)
{
  if (eqn_type == LINEAR) {
    if (stencil_symmetric)
      stencil_linear_sym(g, u, ijk, lu, lii);
    else
      stencil_linear(g, u, ijk, lu, lii);
  }
  else if (eqn_type == BRBR) {
    if (fisheye)
      stencil_brbr_fe(g, u, ijk, lu, lii);
    else
      stencil_brbr(g, u, ijk, lu, lii);
  } else if (eqn_type == FLATLINEAR)
    stencil_flatlinear(g, u, ijk, lu, lii);
  else if (eqn_type == POWERS)
    stencil_powers(g, u, ijk, lu, lii);
  else if (eqn_type == VECLAPLACE)
    stencil_vecL(g, u, ijk, lu, lii);
  else if (eqn_type == VEC4)
    stencil_vec4(g, u, ijk, lu, lii);
  else if ((eqn_type == TSD) || (eqn_type == KSD) || (eqn_type==YORK))
     stencil_vecn(g, u, ijk, lu, lii);
  else
    errorexit("stencil(): stencil type not known\n");
}




/* flat, linear:
   L u  =  Laplace(delta) u  +  M u  + N
*/
void stencil_flatlinear(struct box *g, int nfu, int ijk,
			double *lu, double *lii)
{
  double oodx2 = 1.0/(g->d*g->d);
  int di = g->di, dj = g->dj, dk = g->dk;
  int i = MGF*g->ncomps;
  double *u = g->store[nfu];
  double *M = g->store[i++];
  double *N = g->store[i];

  /* Lu  =  delta d d u  + M u + N */
  *lu = (u[ijk+di] + u[ijk-di] +
	 u[ijk+dj] + u[ijk-dj] +
	 u[ijk+dk] + u[ijk-dk] - 6.0*u[ijk])*oodx2 + M[ijk]*u[ijk] + N[ijk];

  if (cartoon_stencil) {
    int j, k, lijk = g->lijk;
    double r, x, d1, d2;

    ijkto(i, j, k);
    x = g->x + i*g->d;
    if (!dequal(x,0))
      d1 = (u[ijk+di] - u[ijk-di])/g->d/2/x;
    else
      d1 = (u[ijk+di] + u[ijk-di] - 2.0*u[ijk])*oodx2;

    r =  d1 + (u[ijk+di] + u[ijk-di] +
	       u[ijk+dk] + u[ijk-dk] - 4.0*u[ijk])*oodx2
         + M[ijk]*u[ijk] + N[ijk];

    if (0) {
      d2 = (u[ijk+dj] + u[ijk-dj] - 2.0*u[ijk])*oodx2;
      printf("d1 %19.12e  d2 %19.12e\n", d1, d2);
    }
    if (0)
      printf("i %3d  x %6.3f  lu %19.12e  r %19.12e\n", i, x, *lu, r);

    *lu = r;
  }

  /* Lii = del L_i(u) / del u_i  */
  if (!lii) return;
  *lii =  - 6.0*oodx2 + M[ijk];
  if (cartoon_stencil) {
#if 0
    /* catch point exactly on axis: negligible effect */
    int j, k, lijk = g->lijk;
    double r, x, d1, d2;

    ijkto(i, j, k);
    x = g->x + i*g->d;
    if (dequal(x,0)) *lii =  - 6.0*oodx2 + M[ijk];
    else             *lii =  - 4.0*oodx2 + M[ijk];
#else
    *lii =  - 4.0*oodx2 + M[ijk];
#endif
  }
}




/* linear:
   L u  =  gi del del u  +  gg del u  +  M u  +  N
*/
void stencil_linear(struct box *g, int nfu, int ijk,
		    double *lu, double *lii)
{
  int di = g->di, dj = g->dj, dk = g->dk;
  double dx = g->d, dy = g->d, dz = g->d;
  double *u = g->store[nfu];

  /* inlined by hand, just to be sure = setfieldpointers_linear */
  int i = MGF*g->ncomps;
  double  *M = g->store[i++];
  double  *N = g->store[i++];
  double  *gi11 = g->store[i++];
  double  *gi12 = g->store[i++];
  double  *gi13 = g->store[i++];
  double  *gi22 = g->store[i++];
  double  *gi23 = g->store[i++];
  double  *gi33 = g->store[i++];
  double  *gg1  = g->store[i++];
  double  *gg2  = g->store[i++];
  double  *gg3  = g->store[i++];

  if (cartoon_stencil) {
    int j, k, lijk = g->lijk;
    double x, dduddy;

    ijkto(i, j, k);
    x = g->x + i*dx;
    if (!dequal(x,0))
      dduddy = (u[ijk+di] - u[ijk-di])/dx/2/x;
    else
      dduddy = (u[ijk+di] + u[ijk-di] - 2.0*u[ijk])/(dx*dx);

*lu
=
M[ijk]*u[ijk] + N[ijk] +
   gg1[ijk]*(-u[-di + ijk] + u[di + ijk])/(2*dx) +
   gg3[ijk]*(-u[-dk + ijk] + u[dk + ijk])/(2*dz) +
   gi13[ijk]*(u[-di - dk + ijk] - u[di - dk + ijk] - u[-di + dk + ijk] +
       u[di + dk + ijk])/(2*dx*dz) +
   gi11[ijk]*(-2*u[ijk] + u[-di + ijk] + u[di + ijk])/(dx*dx) +
   gi22[ijk]*dduddy                                           +
   gi33[ijk]*(-2*u[ijk] + u[-dk + ijk] + u[dk + ijk])/(dz*dz)
;

if (!lii) return;

*lii
=
M[ijk] - 2*(gi11[ijk]/(dx*dx) + gi33[ijk]/(dz*dz))
;

  } else {

*lu
=
M[ijk]*u[ijk] + N[ijk] +
   gg1[ijk]*(-u[-di + ijk] + u[di + ijk])/(2*dx) +
   gg2[ijk]*(-u[-dj + ijk] + u[dj + ijk])/(2*dy) +
   gg3[ijk]*(-u[-dk + ijk] + u[dk + ijk])/(2*dz) +
   gi12[ijk]*(u[-di - dj + ijk] - u[di - dj + ijk] - u[-di + dj + ijk] +
       u[di + dj + ijk])/(2*dx*dy) +
   gi13[ijk]*(u[-di - dk + ijk] - u[di - dk + ijk] - u[-di + dk + ijk] +
       u[di + dk + ijk])/(2*dx*dz) +
   gi23[ijk]*(u[-dj - dk + ijk] - u[dj - dk + ijk] - u[-dj + dk + ijk] +
       u[dj + dk + ijk])/(2*dy*dz) +
   gi11[ijk]*(-2*u[ijk] + u[-di + ijk] + u[di + ijk])/(dx*dx) +
   gi22[ijk]*(-2*u[ijk] + u[-dj + ijk] + u[dj + ijk])/(dy*dy) +
   gi33[ijk]*(-2*u[ijk] + u[-dk + ijk] + u[dk + ijk])/(dz*dz)
;

if (!lii) return;

*lii
=
M[ijk] - 2*(gi11[ijk]/(dx*dx) + gi22[ijk]/(dy*dy) + gi33[ijk]/(dz*dz))
;

  }
}





/* black hole puncture data
   L(u) = Laplace(delta) u + N (1 + M u)^-7
*/
#define BHNONLIN 1
void stencil_brbr(struct box *g, int nfu, int ijk,
		  double *lu, double *lii)
{
  double oodx2 = 1.0/(g->d*g->d);
  int di = g->di, dj = g->dj, dk = g->dk;
  int i = MGF*g->ncomps;
  double *u = g->store[nfu];
  double *M = g->store[i++];
  double *N = g->store[i];

  /* Lu  =  delta d d u  + N (1 + M u)^-7 */
  *lu = (u[ijk+di] + u[ijk-di] +
	 u[ijk+dj] + u[ijk-dj] +
	 u[ijk+dk] + u[ijk-dk] - 6.0*u[ijk])*oodx2 +
         N[ijk] * pow(1.0 + M[ijk]*(1+BHNONLIN*(u[ijk]-1)), -7.0);

  /* Lii = del L_i(u) / del u_i  */
  if (!lii) return;
  *lii =  - 6.0*oodx2 - BHNONLIN * 7.0*M[ijk]*N[ijk] *
       pow(1.0 + M[ijk]*u[ijk], -8.0);
}





/* black hole puncture data with fisheye, copied from "linear"
   L(u) = Laplace(g) u + N (1 + M u)^-7
*/
void stencil_brbr_fe(struct box *g, int nfu, int ijk,
		  double *lu, double *lii)
{
  int di = g->di, dj = g->dj, dk = g->dk;
  double dx = g->d, dy = g->d, dz = g->d;
  double *u = g->store[nfu];
  int i = MGF*g->ncomps;
  double  *M = g->store[i++];
  double  *N = g->store[i++];
  double  *gi11 = g->store[i++];
  double  *gi12 = g->store[i++];
  double  *gi13 = g->store[i++];
  double  *gi22 = g->store[i++];
  double  *gi23 = g->store[i++];
  double  *gi33 = g->store[i++];
  double  *gg1  = g->store[i++];
  double  *gg2  = g->store[i++];
  double  *gg3  = g->store[i++];

  /* Lu  =  gi d d u + gg d u + N (1 + M u)^-7 */
  *lu = N[ijk] * pow(1.0 + M[ijk]*(1+BHNONLIN*(u[ijk]-1)), -7.0) +
        gg1[ijk]*(-u[-di + ijk] + u[di + ijk])/(2*dx) +
        gg2[ijk]*(-u[-dj + ijk] + u[dj + ijk])/(2*dy) +
        gg3[ijk]*(-u[-dk + ijk] + u[dk + ijk])/(2*dz) +
        gi12[ijk]*(u[-di - dj + ijk] - u[di - dj + ijk] - u[-di + dj + ijk] +
		   u[di + dj + ijk])/(2*dx*dy) +
        gi13[ijk]*(u[-di - dk + ijk] - u[di - dk + ijk] - u[-di + dk + ijk] +
		   u[di + dk + ijk])/(2*dx*dz) +
        gi23[ijk]*(u[-dj - dk + ijk] - u[dj - dk + ijk] - u[-dj + dk + ijk] +
		   u[dj + dk + ijk])/(2*dy*dz) +
        gi11[ijk]*(-2*u[ijk] + u[-di + ijk] + u[di + ijk])/(dx*dx) +
        gi22[ijk]*(-2*u[ijk] + u[-dj + ijk] + u[dj + ijk])/(dy*dy) +
        gi33[ijk]*(-2*u[ijk] + u[-dk + ijk] + u[dk + ijk])/(dz*dz);

  /* Lii = del L_i(u) / del u_i  */
  if (!lii) return;
  *lii = - BHNONLIN * 7.0*M[ijk]*N[ijk] * pow(1.0 + M[ijk]*u[ijk], -8.0)
         - 2*(gi11[ijk]/(dx*dx) + gi22[ijk]/(dy*dy) + gi33[ijk]/(dz*dz));
}





/* flat, linear:
   L u  =  Laplace(delta) u  +  M u  + N
   catches octant
*/
#define cen2(u, i, di, ds) \
  ((u[i+di] - 2.0*u[i] + u[i-di])/(ds*ds))
#define ocen2(u, i, di, dm, dp) \
  (2.0*((u[i+di]-u[i])/dp - (u[i]-u[i-di])/dm)/(dp+dm))

void stencil_flatlinear_o(struct box *g, int nfu,
			  int i, int j, int k, int ijk,
			  double *lu, double *lii)
{
  double dxm, dym, dzm;
  double dx = g->d, dy = g->d, dz = g->d;
  int di = g->di, dj = g->dj, dk = g->dk;
  int n = MGF*g->ncomps;
  double *u = g->store[nfu];
  double *M = g->store[n++];
  double *N = g->store[n];

  /* check for reflective lower boundary */
  if (octant) {
    dxm = (i == 1) ? g->dxm : dx;
    dym = (j == 1) ? g->dym : dy;
    dzm = (k == 1) ? g->dzm : dz;

    *lu = M[ijk]*u[ijk] + N[ijk]
      + ocen2(u, ijk, di, dxm, dx)
      + ocen2(u, ijk, dj, dym, dy)
      + ocen2(u, ijk, dk, dzm, dz);

    if (!lii) return;
    *lii = M[ijk] - 2*(1/(dxm*dx) + 1/(dym*dy) + 1/(dzm*dz));

    return;
  }

  /* Lu  =  delta d d u  + M u + N */
  *lu = M[ijk]*u[ijk] + N[ijk]
    + cen2(u, ijk, di, dx)
    + cen2(u, ijk, dj, dy)
    + cen2(u, ijk, dk, dz);

  /* Lii = del L_i(u) / del u_i  */
  if (!lii) return;
  *lii =  M[ijk] - 2*(1/(dx*dx) + 1/(dy*dy) + 1/(dz*dz));
}





/* power series:
   L u  =  gi del del u  +  gg del u  +  M u + N + Sum P u^p
*/
void stencil_powers(struct box *g, int nfu, int ijk,
		    double *lu, double *lii)
{
  int di = g->di, dj = g->dj, dk = g->dk;
  double dx = g->d, dy = g->d, dz = g->d;
  double *u = g->store[nfu];
  double uijk = u[ijk];
  int j;

  /* inlined by hand, just to be sure = setfieldpointers_linear */
  int i = MGF*g->ncomps;
  double  *M = g->store[i++];
  double  *N = g->store[i++];
  double  *gi11 = g->store[i++];
  double  *gi12 = g->store[i++];
  double  *gi13 = g->store[i++];
  double  *gi22 = g->store[i++];
  double  *gi23 = g->store[i++];
  double  *gi33 = g->store[i++];
  double  *gg1  = g->store[i++];
  double  *gg2  = g->store[i++];
  double  *gg3  = g->store[i++];

*lu
=
M[ijk]*uijk + N[ijk] +
   gg1[ijk]*(-u[-di + ijk] + u[di + ijk])/(2*dx) +
   gg2[ijk]*(-u[-dj + ijk] + u[dj + ijk])/(2*dy) +
   gg3[ijk]*(-u[-dk + ijk] + u[dk + ijk])/(2*dz) +
   gi12[ijk]*(u[-di - dj + ijk] - u[di - dj + ijk] - u[-di + dj + ijk] +
       u[di + dj + ijk])/(2*dx*dy) +
   gi13[ijk]*(u[-di - dk + ijk] - u[di - dk + ijk] - u[-di + dk + ijk] +
       u[di + dk + ijk])/(2*dx*dz) +
   gi23[ijk]*(u[-dj - dk + ijk] - u[dj - dk + ijk] - u[-dj + dk + ijk] +
       u[dj + dk + ijk])/(2*dy*dz) +
   gi11[ijk]*(-2*uijk + u[-di + ijk] + u[di + ijk])/(dx*dx) +
   gi22[ijk]*(-2*uijk + u[-dj + ijk] + u[dj + ijk])/(dy*dy) +
   gi33[ijk]*(-2*uijk + u[-dk + ijk] + u[dk + ijk])/(dz*dz)
;


for (j = 0; g->ppow[j] != 0; j++)
  *lu += g->store[i+j][ijk] * pow(uijk, g->ppow[j]);


if (!lii) return;


*lii
=
M[ijk] - 2*(gi11[ijk]/(dx*dx) + gi22[ijk]/(dy*dy) + gi33[ijk]/(dz*dz))
;

for (j = 0; g->ppow[j] != 0; j++)
  *lii += g->store[i+j][ijk] * pow(uijk, g->ppow[j]-1.0) * g->ppow[j];
}





/* vector Laplacian:  exactly three components
   L u  =  vecL u  +  N
   interface to thorn_vectorlaplace
*/
void stencil_vecL(struct box *g, int nfu, int ijk,
		  double *lu, double *lii)
{
#define f(n) GF[n][0]->data
  double *u1 = g->store[nfu];
  double *u2 = g->store[nfu+1];
  double *u3 = g->store[nfu+2];
  int i = MGF*g->ncomps;
  double *N1 = g->store[i++];
  double *N2 = g->store[i++];
  double *N3 = g->store[i++];
  pGA ***GF = (pGA ***) g->GH->variables;

  if (veclaplacebssn)
  {
    if (CCTK_IsFunctionAliased ("BAM_BSSN_Stencil"))
    {
      BAM_BSSN_Stencil(ijk,
		   g->di, g->dj, g->dk,
		   g->dx, g->dy, g->dz,
		   f(GFI_vl_bssn+0), f(GFI_vl_bssn+1), f(GFI_vl_bssn+2),
		   f(GFI_vl_bssn+3), f(GFI_vl_bssn+4), f(GFI_vl_bssn+5),
		   f(GFI_vl_bssn+6), f(GFI_vl_bssn+7), f(GFI_vl_bssn+8),
		   f(GFI_vl_bssn+9), f(GFI_vl_bssn+10), f(GFI_vl_bssn+11),
		   u1, u2, u3, N1, N2, N3,
		   lu, lu+1, lu+2, lii, lii+1, lii+2);
    }
    else
    {
      errorexit("stencil_vecL(): aliased routine 'BAM_BSSN_Stencil' "
                "is not provided by another thorn.");
    }
  }
  else
  {
    if (conformal_metric)
    {
      if (CCTK_IsFunctionAliased ("BAM_VecLaplace"))
      {
        BAM_VecLaplace(ijk,
		    g->di, g->dj, g->dk,
		    g->dx, g->dy, g->dz,
		    f(GFI_psi),
		    f(GFI_psix), f(GFI_psix+1), f(GFI_psix+2),
		    f(GFI_psixx+0), f(GFI_psixx+1), f(GFI_psixx+2),
		    f(GFI_psixx+3), f(GFI_psixx+4), f(GFI_psixx+5),
		    f(GFI_alp),
		    f(GFI_gxx+0), f(GFI_gxx+1), f(GFI_gxx+2),
		    f(GFI_gxx+3), f(GFI_gxx+4), f(GFI_gxx+5),
		    f(GFI_kxx+0), f(GFI_kxx+1), f(GFI_kxx+2),
		    f(GFI_kxx+3), f(GFI_kxx+4), f(GFI_kxx+5),
		    u1, u2, u3, N1, N2, N3,
		    lu, lu+1, lu+2, lii, lii+1, lii+2);
      }
      else
      {
        errorexit("stencil_vecL(): aliased routine 'BAM_VecLaplace' "
                  "is not provided by another thorn.");
      }
    }
    else
    {
      if (CCTK_IsFunctionAliased ("BAM_VecLaplace_nconf"))
      {
        BAM_VecLaplace_nconf(ijk,
			  g->di, g->dj, g->dk,
			  g->dx, g->dy, g->dz,
			  f(GFI_psi),
			  f(GFI_psix), f(GFI_psix+1), f(GFI_psix+2),
			  f(GFI_psixx+0), f(GFI_psixx+1), f(GFI_psixx+2),
			  f(GFI_psixx+3), f(GFI_psixx+4), f(GFI_psixx+5),
			  f(GFI_alp),
			  f(GFI_gxx+0), f(GFI_gxx+1), f(GFI_gxx+2),
			  f(GFI_gxx+3), f(GFI_gxx+4), f(GFI_gxx+5),
			  f(GFI_kxx+0), f(GFI_kxx+1), f(GFI_kxx+2),
			  f(GFI_kxx+3), f(GFI_kxx+4), f(GFI_kxx+5),
			  u1, u2, u3, N1, N2, N3,
			  lu, lu+1, lu+2, lii, lii+1, lii+2);
      }
      else
      {
        errorexit("stencil_vecL(): aliased routine 'BAM_VecLaplace_nconf' "
                  "is not provided by another thorn.");
      }
    }
  }
}





/* four vector stencil
   L u  =  something external
   currently tied to thorn_vectorlaplace
*/
void stencil_vec4(struct box *g, int nfu, int ijk,
		  double *lu, double *lii)
{
#define f(n) GF[n][0]->data
  double *u0 = g->store[nfu];
  double *u1 = g->store[nfu+1];
  double *u2 = g->store[nfu+2];
  double *u3 = g->store[nfu+3];
  pGA ***GF = (pGA ***) g->GH->variables;

  if (problem == VEC4)
  {
    if (CCTK_IsFunctionAliased ("BAM_BSSN_Stencil4"))
    {
      BAM_BSSN_Stencil4 (ijk,
		    g->di, g->dj, g->dk,
		    g->dx, g->dy, g->dz,
		    f(GFI_vl_bssn+0), f(GFI_vl_bssn+1), f(GFI_vl_bssn+2),
		    f(GFI_vl_bssn+3), f(GFI_vl_bssn+4), f(GFI_vl_bssn+5),
		    f(GFI_vl_bssn+6), f(GFI_vl_bssn+7), f(GFI_vl_bssn+8),
		    f(GFI_vl_bssn+9), f(GFI_vl_bssn+10), f(GFI_vl_bssn+11),
		    f(GFI_vl_bssn+12), f(GFI_vl_bssn+13), f(GFI_vl_bssn+14),
		    f(GFI_vl_bssn+15), f(GFI_vl_bssn+16), f(GFI_vl_bssn+17),
		    f(GFI_vl_N1), f(GFI_vl_N1+1), f(GFI_vl_N1+2),
		    u0, u1, u2, u3,
		    lu, lu+1, lu+2, lu+3,
		    lii, lii+1, lii+2, lii+3);
    }
    else
    {
      errorexit("stencil_vecL(): aliased routine 'BAM_BSSN_Stencil4' "
                "is not provided by another thorn.");
    }
  }
}





/* n vector stencil
   L u  =  something external
   currently tied to thorn BAM_IniDat
*/
void stencil_vecn(struct box *g, int nfu, int ijk,
		  double *lu, double *lii)
{
#define f(n) GF[n][0]->data

  pGA ***GF = (pGA ***) g->GH->variables;

  if (problem == KSD) {
    double gxx, gxy, gxz, gyy, gyz, gzz;
    double igxx, igxy, igxz, igyy, igyz, igzz;
    double kxx, kxy, kxz, kyy, kyz, kzz;
    double Kx, Ky, Kz, Ax, Ay, Az, K, R;
    double *phi = g->store[nfu];
    double *w1 = g->store[nfu+1];
    double *w2 = g->store[nfu+2];
    double *w3 = g->store[nfu+3];
    double x, y, z, d;
    int i, j, k, di, dj, dk, lijk;

    di = g->di;
    dj = g->dj;
    dk = g->dk;

    lijk = g->lijk;
    ijkto(i, j, k);

    d = g->d;

    x = g->x + i * d;
    y = g->y + j * d;
    z = g->z + k * d;

    gxx = ((double *)GF[GFI_gxx+0][0]->data)[ijk];
    gxy = ((double *)GF[GFI_gxx+1][0]->data)[ijk];
    gxz = ((double *)GF[GFI_gxx+2][0]->data)[ijk];
    gyy = ((double *)GF[GFI_gxx+3][0]->data)[ijk];
    gyz = ((double *)GF[GFI_gxx+4][0]->data)[ijk];
    gzz = ((double *)GF[GFI_gxx+5][0]->data)[ijk];

    igxx = ((double *)GF[GFI_igxx+0][0]->data)[ijk];
    igxy = ((double *)GF[GFI_igxx+1][0]->data)[ijk];
    igxz = ((double *)GF[GFI_igxx+2][0]->data)[ijk];
    igyy = ((double *)GF[GFI_igxx+3][0]->data)[ijk];
    igyz = ((double *)GF[GFI_igxx+4][0]->data)[ijk];
    igzz = ((double *)GF[GFI_igxx+5][0]->data)[ijk];

    kxx = ((double *)GF[GFI_kxx+0][0]->data)[ijk];
    kxy = ((double *)GF[GFI_kxx+1][0]->data)[ijk];
    kxz = ((double *)GF[GFI_kxx+2][0]->data)[ijk];
    kyy = ((double *)GF[GFI_kxx+3][0]->data)[ijk];
    kyz = ((double *)GF[GFI_kxx+4][0]->data)[ijk];
    kzz = ((double *)GF[GFI_kxx+5][0]->data)[ijk];

    Kx = ((double *)GF[GFI_Kx+0][0]->data)[ijk];
    Ky = ((double *)GF[GFI_Kx+1][0]->data)[ijk];
    Kz = ((double *)GF[GFI_Kx+2][0]->data)[ijk];

    Ax = ((double *)GF[GFI_Ax+0][0]->data)[ijk];
    Ay = ((double *)GF[GFI_Ax+1][0]->data)[ijk];
    Az = ((double *)GF[GFI_Ax+2][0]->data)[ijk];

    K = ((double *)GF[GFI_K][0]->data)[ijk];
    R = ((double *)GF[GFI_R][0]->data)[ijk];

    if (CCTK_IsFunctionAliased ("BAM_ksd_stencil"))
    {
      BAM_ksd_stencil(x, y, z, d, d, d, ijk, di, dj, dk,
		  gxx, gxy, gxz, gyy, gyz, gzz,
		  igxx, igxy, igxz, igyy, igyz, igzz,
		  kxx, kxy, kxz, kyy, kyz, kzz,
		  Kx, Ky, Kz, Ax, Ay, Az, K, R,
		  phi, w1, w2, w3,
		  lu, lu+1, lu+2, lu+3, lii, lii+1, lii+2, lii+3);
    }
    else
    {
      errorexit("stencil_vecn(): aliased routine 'BAM_ksd_stencil' "
                "is not provided by another thorn (eg. BAM_IniDat).");
    }
  }

  else if (problem == YORK) {
    double *phi = g->store[nfu];
    double *w1 = g->store[nfu+1];
    double *w2 = g->store[nfu+2];
    double *w3 = g->store[nfu+3];

    if (CCTK_IsFunctionAliased ("BAM_yorkpuncstencil"))
    {
      BAM_yorkpuncstencil(ijk, g->di, g->dj, g->dk, g->dx, g->dy, g->dz,
	         f(GFI_psi),
                 f(GFI_psix), f(GFI_psix+1), f(GFI_psix+2),
	         f(GFI_psixx+0), f(GFI_psixx+1), f(GFI_psixx+2),
                 f(GFI_psixx+3), f(GFI_psixx+4), f(GFI_psixx+5),
	         f(GFI_YABYxx+9),
	         f(GFI_Ygxx+0), f(GFI_Ygxx+1), f(GFI_Ygxx+2),
                 f(GFI_Ygxx+3), f(GFI_Ygxx+4), f(GFI_Ygxx+5),
	         f(GFI_Ygxx+6), f(GFI_Ygxx+7), f(GFI_Ygxx+8),
                 f(GFI_Ygxx+9), f(GFI_Ygxx+10), f(GFI_Ygxx+11),
	         f(GFI_YABYxx+0), f(GFI_YABYxx+1), f(GFI_YABYxx+2),
                 f(GFI_YABYxx+3), f(GFI_YABYxx+4), f(GFI_YABYxx+5),
                 f(GFI_YABYxx+6), f(GFI_YABYxx+7), f(GFI_YABYxx+8),
                 f(GFI_Ygxx+12),
	         phi, w1, w2, w3,
	         lu, lu+1, lu+2, lu+3, lii, lii+1, lii+2, lii+3);
    }
    else
    {
      errorexit("stencil_vecn(): aliased routine 'BAM_yorkpuncstencil' "
                "is not provided by another thorn (eg. BAM_IniDat).");
    }
  }

  else if (problem == TSD) {
    double *u0 = g->store[nfu];
    double *u1 = g->store[nfu+1];
    double *u2 = g->store[nfu+2];
    double *u3 = g->store[nfu+3];
    double *u4 = g->store[nfu+4];

    if (CCTK_IsFunctionAliased ("BAM_tsdstencil"))
    {
      BAM_tsdstencil(ijk,
	         g->di, g->dj, g->dk,
	         g->dx, g->dy, g->dz,
	         f(GFI_psi), f(GFI_psix), f(GFI_psix+1), f(GFI_psix+2),
	         u0, u1, u2, u3, u4,
 	         f(GFI_lbeta11), f(GFI_lbeta11+1), f(GFI_lbeta11+2),
	         f(GFI_lbeta11+3), f(GFI_lbeta11+4), f(GFI_lbeta11+5),
	         f(GFI_lbeta11+6), f(GFI_lbeta11+7),
	         lu, lu+1, lu+2, lu+3, lu+4,
	         lii, lii+1, lii+2, lii+3, lii+4);
    }
    else
    {
      errorexit("stencil_vecn(): aliased routine 'BAM_tsdstencil' "
                "is not provided by another thorn (eg. BAM_IniDat).");
    }
  }
}





/* linear, symmetric, see Anninos et al, PRD 52 (1995) 2059
   L u  =  1/g del(g gi del u)  +  M u  +  N
   symmetric finite differences
*/
void stencil_linear_sym(struct box *g, int nfu, int ijk,
			double *lu, double *lii)
{
  int di = g->di, dj = g->dj, dk = g->dk;
  double dx = g->d, dy = g->d, dz = g->d;
  double *u = g->store[nfu];

  int i = MGF*g->ncomps;
  double  *M = g->store[i++];
  double  *N = g->store[i++];
  double  *gi11 = g->store[i++];
  double  *gi12 = g->store[i++];
  double  *gi13 = g->store[i++];
  double  *gi22 = g->store[i++];
  double  *gi23 = g->store[i++];
  double  *gi33 = g->store[i++];
  double  *gg   = g->store[i++];

  *lu =

    (u[ijk+di]*(gi11[ijk+di]+gi11[ijk]) +
     u[ijk-di]*(gi11[ijk-di]+gi11[ijk]) -
     u[ijk]*(gi11[ijk+di]+2*gi11[ijk]+gi11[ijk-di]))/(2*dx*dx) +

    (u[ijk+dj]*(gi22[ijk+dj]+gi22[ijk]) +
     u[ijk-dj]*(gi22[ijk-dj]+gi22[ijk]) -
     u[ijk]*(gi22[ijk+dj]+2*gi22[ijk]+gi22[ijk-dj]))/(2*dy*dy) +

    (u[ijk+dk]*(gi33[ijk+dk]+gi33[ijk]) +
     u[ijk-dk]*(gi33[ijk-dk]+gi33[ijk]) -
     u[ijk]*(gi33[ijk+dk]+2*gi33[ijk]+gi33[ijk-dk]))/(2*dz*dz) +

    (u[ijk+di+dj]*(gi12[ijk+di]+gi12[ijk+dj]) -
     u[ijk-di+dj]*(gi12[ijk-di]+gi12[ijk+dj]) -
     u[ijk+di-dj]*(gi12[ijk+di]+gi12[ijk-dj]) +
     u[ijk-di-dj]*(gi12[ijk-di]+gi12[ijk-dj]))/(4*dx*dy) +

    (u[ijk+di+dk]*(gi13[ijk+di]+gi13[ijk+dk]) -
     u[ijk-di+dk]*(gi13[ijk-di]+gi13[ijk+dk]) -
     u[ijk+di-dk]*(gi13[ijk+di]+gi13[ijk-dk]) +
     u[ijk-di-dk]*(gi13[ijk-di]+gi13[ijk-dk]))/(4*dx*dz) +

    (u[ijk+dj+dk]*(gi23[ijk+dj]+gi23[ijk+dk]) -
     u[ijk-dj+dk]*(gi23[ijk-dj]+gi23[ijk+dk]) -
     u[ijk+dj-dk]*(gi23[ijk+dj]+gi23[ijk-dk]) +
     u[ijk-dj-dk]*(gi23[ijk-dj]+gi23[ijk-dk]))/(4*dy*dz);

  *lu = (*lu)/gg[ijk] + M[ijk]*u[ijk] + N[ijk];

  if (!lii) return;

  *lii = M[ijk] -
    ((gi11[ijk+di]+2*gi11[ijk]+gi11[ijk-di])/(dx*dx) +
     (gi22[ijk+dj]+2*gi22[ijk]+gi22[ijk-dj])/(dy*dy) +
     (gi33[ijk+dk]+2*gi33[ijk]+gi33[ijk-dk])/(dz*dz))/(2*gg[ijk]);
}
