/* $Header: /numrelcvs/AEIThorns/FiniteDifferencing/src/diff.c,v 1.4 2005/02/27 12:53:01 schnetter Exp $ */

#include <assert.h>

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"



void
FD_Diff_gv (CCTK_POINTER_TO_CONST restrict const cctkGH_,
            CCTK_INT const dir,
            CCTK_REAL const * restrict const var,
            CCTK_REAL * restrict const dvar,
            CCTK_INT const table_handle)
{
  cGH const * restrict const cctkGH = cctkGH_;
  DECLARE_CCTK_ARGUMENTS;
  DECLARE_CCTK_PARAMETERS;
  
  int imin[3], imax[3];         /* extent of the interior of the grid */
  
  CCTK_REAL invdx;              /* inverse grid spacing */
  int dind;                     /* indexing offset */
  
  int i, j, k;
  int d, e, f;
  
  assert (cctkGH);
  assert (cctk_dim == 3);
  assert (var);
  assert (dvar);
  
  if (dir<0 || dir>=3) {
    CCTK_VWarn
      (0, __LINE__, __FILE__, CCTK_THORNSTRING,
       "A derivate in direction %d does not make sense -- the direction must be between 0 and 2",
       (int)dir);
  }
  
  /* Calculate the extent of the interior of the grid */
  for (d=0; d<3; ++d) {
    imin[d] = 0;
    imax[d] = cctk_lsh[d];
  }
  
  if (order/2 > cctk_nghostzones[dir]) {
    CCTK_VWarn
      (0, __LINE__, __FILE__, CCTK_THORNSTRING,
       "A finite differencing order of %d requires at least %d ghost zones",
       (int)order, (int)order/2);
  }
  
  imin[dir] += cctk_bbox[2*dir  ] ? order/2 : cctk_nghostzones[dir];
  imax[dir] -= cctk_bbox[2*dir+1] ? order/2 : cctk_nghostzones[dir];
  
  /* Calculate inverse grid spacing */
  invdx = 1.0 / CCTK_DELTA_SPACE(dir);
  
  /* Calculate indexing offset */
  switch (dir) {
  case 0:
    dind = CCTK_GFINDEX3D (cctkGH, 1,0,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 1:
    dind = CCTK_GFINDEX3D (cctkGH, 0,1,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 2:
    dind = CCTK_GFINDEX3D (cctkGH, 0,0,1) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  default:
    assert (0);
  }
  
  /* Loop over the interior and calculate the derivative */
  for (k=imin[2]; k<imax[2]; ++k) {
    for (j=imin[1]; j<imax[1]; ++j) {
      for (i=imin[0]; i<imax[0]; ++i) {
        int const ind = CCTK_GFINDEX3D (cctkGH, i, j, k);
        switch (order) {
        case 2:
          dvar[ind] = (+ (-1.0/2.0) * var[ind - dind]
                       + ( 1.0/2.0) * var[ind + dind]) * invdx;
          break;
        case 4:
          dvar[ind] = (+ ( 1.0/12.0) * var[ind - 2*dind]
                       + (-8.0/12.0) * var[ind -   dind]
                       + ( 8.0/12.0) * var[ind +   dind]
                       + (-1.0/12.0) * var[ind + 2*dind]) * invdx;
          break;
        default:
          assert (0);
        }
      }
    }
  }
  
  /* Loop over the boundaries and set the derivative to zero */
  for (d=0; d<3; ++d) {
    for (f=0; f<2; ++f) {
      int bmin[3], bmax[3];     /* extent of the current boundary */
      for (e=0; e<3; ++e) {
        bmin[e] = 0;
        bmax[e] = cctk_lsh[e];
      }
      if (f==0) {
        bmax[d] = imin[d];
      } else {
        bmin[d] = imax[d];
      }
      for (k=bmin[2]; k<bmax[2]; ++k) {
        for (j=bmin[1]; j<bmax[1]; ++j) {
          for (i=bmin[0]; i<bmax[0]; ++i) {
            int const ind = CCTK_GFINDEX3D (cctkGH, i, j, k);
            dvar[ind] = 0;
          }
        }
      }
    }
  }
}



void
FD_Diff2_gv (CCTK_POINTER_TO_CONST restrict const cctkGH_,
             CCTK_INT const dir1,
             CCTK_INT const dir2,
             CCTK_REAL const * restrict const var,
             CCTK_REAL * restrict const dvar,
             CCTK_INT const table_handle)
{
  cGH const * restrict const cctkGH = cctkGH_;
  DECLARE_CCTK_ARGUMENTS;
  DECLARE_CCTK_PARAMETERS;
  
  int imin[3], imax[3];         /* extent of the interior of the grid */
  
  CCTK_REAL invdx2;             /* inverse grid spacing */
  int dind1, dind2;             /* indexing offsets */
  int samedir;                  /* are the directions equal? */
  
  int i, j, k;
  int d, e, f;
  
  assert (cctkGH);
  assert (cctk_dim == 3);
  assert (var);
  assert (dvar);
  
  if (dir1<0 || dir1>=3) {
    CCTK_VWarn
      (0, __LINE__, __FILE__, CCTK_THORNSTRING,
       "A derivate in direction %d does not make sense -- the direction must be between 0 and 2",
       (int)dir1);
  }
  if (dir2<0 || dir2>=3) {
    CCTK_VWarn
      (0, __LINE__, __FILE__, CCTK_THORNSTRING,
       "A derivate in direction %d does not make sense -- the direction must be between 0 and 2",
       (int)dir2);
  }
  
  /* Calculate the extent of the interior of the grid */
  for (d=0; d<3; ++d) {
    imin[d] = 0;
    imax[d] = cctk_lsh[d];
  }
  
  if (order/2 > cctk_nghostzones[dir1] || order/2 > cctk_nghostzones[dir2]) {
    CCTK_VWarn
      (0, __LINE__, __FILE__, CCTK_THORNSTRING,
       "A finite differencing order of %d requires at least %d ghost zones",
       (int)order, (int)order/2);
  }
  
  samedir = dir1 == dir2;
  
  imin[dir1] += cctk_bbox[2*dir1  ] ? order/2 : cctk_nghostzones[dir1];
  imax[dir1] -= cctk_bbox[2*dir1+1] ? order/2 : cctk_nghostzones[dir1];
  if (! samedir) {
    imin[dir2] += cctk_bbox[2*dir2  ] ? order/2 : cctk_nghostzones[dir2];
    imax[dir2] -= cctk_bbox[2*dir2+1] ? order/2 : cctk_nghostzones[dir2];
  }
  
  /* Calculate inverse grid spacings */
  invdx2 = 1.0 / (CCTK_DELTA_SPACE(dir1) * CCTK_DELTA_SPACE(dir2));
  
  /* Calculate indexing offsets */
  switch (dir1) {
  case 0:
    dind1 = CCTK_GFINDEX3D (cctkGH, 1,0,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 1:
    dind1 = CCTK_GFINDEX3D (cctkGH, 0,1,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 2:
    dind1 = CCTK_GFINDEX3D (cctkGH, 0,0,1) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  default:
    assert (0);
  }
  switch (dir2) {
  case 0:
    dind2 = CCTK_GFINDEX3D (cctkGH, 1,0,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 1:
    dind2 = CCTK_GFINDEX3D (cctkGH, 0,1,0) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  case 2:
    dind2 = CCTK_GFINDEX3D (cctkGH, 0,0,1) - CCTK_GFINDEX3D (cctkGH, 0,0,0);
    break;
  default:
    assert (0);
  }
  
  /* Loop over the interior and calculate the derivative */
  for (k=imin[2]; k<imax[2]; ++k) {
    for (j=imin[1]; j<imax[1]; ++j) {
      for (i=imin[0]; i<imax[0]; ++i) {
        int const ind = CCTK_GFINDEX3D (cctkGH, i, j, k);
        switch (order) {
        case 2:
          if (samedir) {
            dvar[ind] = (+ ( 1.0/1.0) * var[ind - dind1]
                         + (-2.0/1.0) * var[ind        ]
                         + ( 1.0/1.0) * var[ind + dind1]) * invdx2;
          } else {
            dvar[ind] = (+ ( 1.0/4.0) * var[ind - dind1 - dind2]
                         + (-1.0/4.0) * var[ind - dind1 + dind2]
                         + (-1.0/4.0) * var[ind + dind1 - dind2]
                         + ( 1.0/4.0) * var[ind + dind1 + dind2]) * invdx2;
          }
          break;
        case 4:
          if (samedir) {
            dvar[ind] = (+ (- 1.0/12.0) * var[ind - 2*dind1]
                         + ( 16.0/12.0) * var[ind -   dind1]
                         + (-30.0/12.0) * var[ind          ]
                         + ( 16.0/12.0) * var[ind +   dind1]
                         + (- 1.0/12.0) * var[ind + 2*dind1]) * invdx2;
          } else {
            dvar[ind] = (+ (  1.0/144.0) * var[ind - 2*dind1 - 2*dind2]
                         + (- 8.0/144.0) * var[ind - 2*dind1 -   dind2]
                         + (  8.0/144.0) * var[ind - 2*dind1 +   dind2]
                         + (- 1.0/144.0) * var[ind - 2*dind1 + 2*dind2]
                         + (- 8.0/144.0) * var[ind -   dind1 - 2*dind2]
                         + ( 64.0/144.0) * var[ind -   dind1 -   dind2]
                         + (-64.0/144.0) * var[ind -   dind1 +   dind2]
                         + (  8.0/144.0) * var[ind -   dind1 + 2*dind2]
                         + (  8.0/144.0) * var[ind +   dind1 - 2*dind2]
                         + (-64.0/144.0) * var[ind +   dind1 -   dind2]
                         + ( 64.0/144.0) * var[ind +   dind1 +   dind2]
                         + (- 8.0/144.0) * var[ind +   dind1 + 2*dind2]
                         + (- 1.0/144.0) * var[ind + 2*dind1 - 2*dind2]
                         + (  8.0/144.0) * var[ind + 2*dind1 -   dind2]
                         + (- 8.0/144.0) * var[ind + 2*dind1 +   dind2]
                         + (  1.0/144.0) * var[ind + 2*dind1 + 2*dind2]) * invdx2;
          }
          break;
        default:
          assert (0);
        }
      }
    }
  }
  
  /* Loop over the boundaries and set the derivative to zero */
  for (d=0; d<3; ++d) {
    for (f=0; f<2; ++f) {
      int bmin[3], bmax[3];     /* extent of the current boundary */
      for (e=0; e<3; ++e) {
        bmin[e] = 0;
        bmax[e] = cctk_lsh[e];
      }
      if (f==0) {
        bmax[d] = imin[d];
      } else {
        bmin[d] = imax[d];
      }
      for (k=bmin[2]; k<bmax[2]; ++k) {
        for (j=bmin[1]; j<bmax[1]; ++j) {
          for (i=bmin[0]; i<bmax[0]; ++i) {
            int const ind = CCTK_GFINDEX3D (cctkGH, i, j, k);
            dvar[ind] = 0;
          }
        }
      }
    }
  }
}
