 /*@@
   @file      Output.c
   @date      Sun 5 Aug 2001
   @author    Thomas Radke
   @desc
              Functions to deal with AHFinder HDF5 output of variables
   @enddesc
   @version   $Id: Output.c,v 1.27 2004/06/23 18:04:40 tradke Exp $
 @@*/

#include <stdlib.h>
#include <string.h>

#include "cctk.h"
#include "cctk_Parameters.h"
#include "CactusBase/IOUtil/src/ioGH.h"
#include "CactusBase/IOUtil/src/ioutil_AdvertisedFiles.h"
#include "CactusPUGHIO/IOHDF5Util/src/ioHDF5UtilGH.h"
#include "ioAHFinderHDF5GH.h"

/* the rcs ID and its dummy function to use it */
static const char *rcsid = "$Header: /numrelcvs/AEIThorns/IOAHFinderHDF5/src/Output.c,v 1.27 2004/06/23 18:04:40 tradke Exp $";
CCTK_FILEVERSION(AEIThorns_IOAHFinderHDF5_Output_c)


/********************************************************************
 ********************    Macro Definitions   ************************
 ********************************************************************/
/* define this if you want debugging output */
/* #define IO_DEBUG 1 */

/* filename of the AHFinder HDF5 data outpu file */
#define DISKFILENAME  "AHFinderData.h5"


/********************************************************************
 ********************    Internal Routines   ************************
 ********************************************************************/
static void WriteScalars (const cGH *GH, hid_t time_group, int status,
                          int mtype, int mfind, int horizon);
static int WriteExpansionCoefficients (const cGH *GH, hid_t geom_group,
                                       int ahf_lmax);
static int WriteGaussianCurvature (const cGH *GH, hid_t group, int ntheta,
                                   int nphi);


/*@@
   @routine    IOAHFinderHDF5_OutputVarAs
   @date       Sun 5 Aug 2001
   @author     Thomas Radke
   @desc
               Unconditional output of AHFinder data into HDF5 files.
               So far only unchunked output by processor 0 is supported.
   @enddesc

   @var        GH
   @vdesc      Pointer to CCTK GH
   @vtype      const cGH *
   @vio        in
   @endvar
   @var        fullname
   @vdesc      complete name of variable to output
   @vtype      const char *
   @vio        in
   @endvar
   @var        alias
   @vdesc      alias name of variable to output
               (used to generate output filename)
   @vtype      const char *
   @vio        in
   @endvar

   @returntype int
   @returndesc
                0 for success, or<BR>
               -1 if alias argument couldn't be parsed
   @endreturndesc
@@*/
int IOAHFinderHDF5_OutputVarAs (const cGH *GH, const char *fullname,
                                const char *alias)
{
  int ioproc_save, ioproc_every_save, unchunked_save;
  int i, myproc, status, mtype, mfind, horizon;
  char buffer[64], *filename;
  CCTK_INT ahf_find3;
  ioAHFinderHDF5GH *myGH;
  ioHDF5UtilGH *utilGH;
  ioGH *ioUtilGH;
#ifdef H5_HAVE_STREAM
  hid_t plist;
  H5FD_stream_fapl_t fapl;
#endif /* H5_HAVE_STREAM */
  ioAdvertisedFileDesc advertised_file;
  DECLARE_CCTK_PARAMETERS


  /* suppress compiler warning about unused parameter */
  (void) (fullname + 0);

  /* check whether output is enabled */
  if (! (write_to_file || write_to_stream))
  {
    return (0);
  }

  /* get processor ID */
  myproc = CCTK_MyProc (GH);

  /* parse options string */
  if (sscanf (alias, "mfind=%d mtype=%d status=%d horizon=%d",
              &mfind, &mtype, &status, &horizon) != 4)
  {
    CCTK_WARN (1, "IOAHFinderHDF5_OutputVarAs: failed to parse options "
                  "argument");
    return (-1);
  }

  /* get IOAHFinderHDF5 and IOHDF5Util GH extensions */
  myGH = CCTK_GHExtension (GH, "IOAHFinderHDF5");
  utilGH = CCTK_GHExtension (GH, "IOHDF5Util");
  ioUtilGH = CCTK_GHExtension (GH, "IO");

  /* get AHFinder parameter values */
  ahf_find3 = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_find3",
                                                     "AHFinder", NULL);

  /* open the file at every timestep on the first horizon */
  if (! ahf_find3 || mfind == 1)
  {
    if (myproc == 0)
    {
      /* FIXME: AHFinder doesn't call this I/O method if mfind == 3 but
                not further horizon was found. This keeps old files open
                which is bad. So we close them here if neccessary. */
      if (myGH->stream >= 0)
      {
        HDF5_ERROR (H5Gclose (myGH->time_group_stream));
        HDF5_ERROR (H5Fclose (myGH->stream));
      }
      if (myGH->file >= 0)
      {
        HDF5_ERROR (H5Gclose (myGH->time_group_file));
        HDF5_ERROR (H5Fclose (myGH->file));
      }

#ifdef H5_HAVE_STREAM
      /* open the file to be streamed */
      if (write_to_stream && myGH->socket != INVALID_SOCKET)
      {
        if (CCTK_Equals (verbose, "full"))
        {
          CCTK_VInfo (CCTK_THORNSTRING, "Opening AHFinder HDF5 output file "
                                        "on port %u", myGH->port);
        }

        /* set file access property list to use the Stream VFD
           and open the file */
        fapl.increment = 0;
        fapl.socket = myGH->socket;
        fapl.do_socket_io = 1;
        fapl.backlog = 5;
        fapl.broadcast_fn = NULL;
        fapl.broadcast_arg = NULL;

        HDF5_ERROR (plist = H5Pcreate (H5P_FILE_ACCESS));
        HDF5_ERROR (H5Pset_fapl_stream (plist, &fapl));

        /* filename is not used if we pass a plist
           but it must not be NULL or an empty string */
        HDF5_ERROR (myGH->stream = H5Fcreate ("unused", H5F_ACC_TRUNC,
                                              H5P_DEFAULT, plist));
        HDF5_ERROR (H5Pclose (plist));
      }
#endif /* H5_HAVE_STREAM */

      /* open the file to be written to disk */
      if (write_to_file)
      {
        /* build the filename for the file to be written to disk */
        filename = malloc (strlen (out_dir) + 64);
        if (out_timesteps_per_file > 0)
        {
          /* skip the './' prefix if output goes into current directory */
          if (strcmp (out_dir, ".") == 0)
          {
            sprintf (filename, "AHFinderData_it%04d.h5", GH->cctk_iteration);
          }
          else
          {
            sprintf (filename, "%s/AHFinderData_it%04d.h5", out_dir,
                     GH->cctk_iteration);
          }
        }
        else
        {
          /* skip the './' prefix if output goes into current directory */
          if (strcmp (out_dir, ".") == 0)
          {
            strcpy (filename, DISKFILENAME);
          }
          else
          {
            sprintf (filename, "%s/%s", out_dir, DISKFILENAME);
          }
        }

        /* create the disk file the first time thru */
        if (out_timesteps_per_file > 0 || myGH->ntimes_called++ == 0)
        {
          HDF5_ERROR (myGH->file = H5Fcreate (filename, H5F_ACC_TRUNC,
                                              H5P_DEFAULT, H5P_DEFAULT));

          /* advertise the file for downloading */
          advertised_file.slice = "";
          advertised_file.thorn = CCTK_THORNSTRING;
          advertised_file.varname = "AHFinder data";
          advertised_file.description = "AHFinder HDF5 data";
          advertised_file.mimetype = "data/x-ahfinder-hdf5";

          IOUtil_AdvertiseFile (GH, filename, &advertised_file);
        }
        else
        {
          HDF5_ERROR (myGH->file = H5Fopen (filename,H5F_ACC_RDWR,H5P_DEFAULT));
        }
        free (filename);
      }

      sprintf (buffer, "Physical Time %g", (double) GH->cctk_time);
      if (write_to_file)
      {
        HDF5_ERROR (myGH->time_group_file = H5Gcreate (myGH->file, buffer,0));
        WRITE_ATTRIBUTE ("Physical Time", &GH->cctk_time, myGH->time_group_file,
                         utilGH, 0, HDF5_REAL);
        WRITE_ATTRIBUTE ("Iteration", &GH->cctk_iteration,
                         myGH->time_group_file, utilGH, 0, H5T_NATIVE_INT);
      }
      if (write_to_stream)
      {
        HDF5_ERROR (myGH->time_group_stream = H5Gcreate (myGH->stream,
                                                         buffer, 0));
        WRITE_ATTRIBUTE ("Physical Time", &GH->cctk_time,
                         myGH->time_group_stream, utilGH, 0, HDF5_REAL);
        WRITE_ATTRIBUTE ("Iteration", &GH->cctk_iteration,
                         myGH->time_group_stream, utilGH, 0, H5T_NATIVE_INT);
      }
    }

    /* re-parse the 'IOAHFinderHDF5::out_vars' parameter if it was changed */
    i = CCTK_ParameterQueryTimesSet ("out_vars", CCTK_THORNSTRING);
    if (i != myGH->out_vars_lastset)
    {
      IOUtil_ParseVarsForOutput (GH, CCTK_THORNSTRING,
                                 "IOAHFinderHDF5::out_vars",
                                 0, out_vars, 1, myGH->requests);

      /* Save the last setting of 'out_vars' parameter */
      myGH->out_vars_lastset = i;
    }

    /* set the I/O mode to unchunked output by processor 0 */
    ioproc_save = ioUtilGH->ioproc;
    ioproc_every_save = ioUtilGH->ioproc_every;
    unchunked_save = ioUtilGH->unchunked;
    ioUtilGH->ioproc = 0;
    ioUtilGH->ioproc_every = CCTK_nProcs (GH);
    ioUtilGH->unchunked = 1;

    /* append variables from 'IOAHFinderHDF5::out_vars' list */
    for (i = CCTK_NumVars () - 1; i >= 0; i--)
    {
      if (myGH->requests[i])
      {
        if (write_to_file)
        {
          IOHDF5Util_DumpVar (GH, myGH->requests[i], myGH->time_group_file);
        }
        if (write_to_stream)
        {
          IOHDF5Util_DumpVar (GH, myGH->requests[i], myGH->time_group_stream);
        }
      }
    }

    /*** FIXME: restore the original I/O mode ***/
    ioUtilGH->ioproc = ioproc_save;
    ioUtilGH->ioproc_every = ioproc_every_save;
    ioUtilGH->unchunked = unchunked_save;
  }

  /* the rest is output by processor 0 only */
  if (myproc == 0)
  {
    if (write_to_stream)
    {
      WriteScalars (GH, myGH->time_group_stream, status, mtype, mfind, horizon);
    }
    if (write_to_file)
    {
      WriteScalars (GH, myGH->time_group_file, status, mtype, mfind, horizon);
    }

    if (! ahf_find3 || mfind == 3)
    {
      if (write_to_stream)
      {
        HDF5_ERROR (H5Gclose (myGH->time_group_stream));
        HDF5_ERROR (H5Fclose (myGH->stream));
        myGH->time_group_stream = myGH->stream = -1;
      }
      if (write_to_file)
      {
        HDF5_ERROR (H5Gclose (myGH->time_group_file));
        HDF5_ERROR (H5Fclose (myGH->file));
        myGH->time_group_file = myGH->file = -1;
      }
    }
  }

  return (0);
}


static void WriteScalars (const cGH *GH, hid_t time_group, int status,
                          int mtype, int mfind, int horizon)
{
  char buffer[64];
  CCTK_REAL attr[3];
  ioHDF5UtilGH *utilGH;
  hid_t ah_group, geom_group, points_group;
  CCTK_INT ahf_ReportAlways, ahf_find3, ahf_lmax, ahf_ntheta, ahf_nphi;


  /* get IOAHFinderHDF5 and IOHDF5Util GH extensions */
  utilGH = CCTK_GHExtension (GH, "IOHDF5Util");

  /* get AHFinder parameter values */
  ahf_find3 = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_find3",
                                                     "AHFinder", NULL);
  ahf_ntheta = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_ntheta",
                                                      "AHFinder", NULL);
  ahf_nphi = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_nphi",
                                                    "AHFinder", NULL);
  ahf_ReportAlways = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_ReportAlways",
                                                            "AHFinder", NULL);

  if (ahf_find3)
  {
    sprintf (buffer, "Apparent Horizon %d", mfind - 1);
  }
  else
  {
    strcpy (buffer, "Apparent Horizon");
  }
  HDF5_ERROR (ah_group = H5Gcreate (time_group, buffer, 0));

  if (status && (ahf_ReportAlways || mtype == 1))
  {
    HDF5_ERROR (geom_group = H5Gcreate (ah_group, "Geometry", 0));

    ahf_lmax = *(const CCTK_INT *) CCTK_ParameterGet ("ahf_lmax", "AHFinder",
                                                      NULL);
    if (ahf_lmax > 0)
    {
      WriteExpansionCoefficients (GH, geom_group, ahf_lmax);
    }

    /* attach geometry attributes */
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_mass");
    WRITE_ATTRIBUTE ("Mass", attr, geom_group, utilGH, 0, HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_radius");
    WRITE_ATTRIBUTE ("Radius", attr, geom_group, utilGH, 0, HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_perimeter");
    WRITE_ATTRIBUTE ("Equatorial Circumference", attr, geom_group, utilGH, 0,
                     HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_meridian1");
    attr[1] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_meridian2");
    WRITE_ATTRIBUTE ("Length of Meridian, phi=0,pi/2", attr, geom_group, utilGH,
                     2, HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_area");
    WRITE_ATTRIBUTE ("Horizon Area", attr, geom_group, utilGH, 0, HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_asymx");
    attr[1] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_asymy");
    attr[2] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_asymz");
    WRITE_ATTRIBUTE ("Asymmetries on xyz Reflection", attr, geom_group, utilGH,
                     3, HDF5_REAL);
    attr[0] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_centerx");
    attr[1] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_centery");
    attr[2] = *(CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_centerz");
    WRITE_ATTRIBUTE ("Surface Center", attr, geom_group, utilGH, 3, HDF5_REAL);

    if (mtype == 1)
    {
      strcpy (buffer, horizon ? "Outer Horizon" : "Outer Horizon?");
    }
    else if (mtype == 2)
    {
      strcpy (buffer, horizon ? "Inner Horizon" : "Inner Horizon?");
    }
    else if (mtype == 3)
    {
      strcpy (buffer, horizon ? "Marginal Horizon?" : "Trapped Surface");
    }
    else if (mtype == 4)
    {
      strcpy (buffer, horizon ? "Marginal Horizon?" : "Not a Horizon");
    }
    else
    {
      CCTK_WARN (1, "IOAHFinderHDF5_OutputVarAs: Invalid mtype");
      buffer[0] = 0;
    }
    WRITE_ATTRIBUTE ("Surface Found", buffer, geom_group, utilGH, 0,
                     utilGH->HDF5_STRING);

    HDF5_ERROR (H5Gclose (geom_group));
  }

  HDF5_ERROR (points_group = H5Gcreate (ah_group, "Points", 0));
  WriteGaussianCurvature (GH, points_group, ahf_ntheta, ahf_nphi);

  HDF5_ERROR (H5Gclose (points_group));

  HDF5_ERROR (H5Gclose (ah_group));
}


static int WriteExpansionCoefficients (const cGH *GH, hid_t geom_group,
                                       int ahf_lmax)
{
  hid_t filespace, dataset;
  CCTK_REAL *buffer;
  hsize_t total_elements;
  const CCTK_REAL *radius, *c0, *cc, *cs;
  int i, j;


  if (! CCTK_QueryGroupStorage (GH, "AHFinder::out_1d_legen") ||
      ! CCTK_QueryGroupStorage (GH, "AHFinder::out_2d_legen"))
  {
    CCTK_WARN (1, "WriteExpansionCoefficients: No storage enabled for "
                  "AHFinder expansion coefficients");
    return (-1);
  }

  radius = (const CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_radius");
  c0 = (const CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_c0");
  cc = (const CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_cc");
  cs = (const CCTK_REAL *) CCTK_VarDataPtr (GH, 0, "AHFinder::out_cs");

  for (i = total_elements = 1; i <= ahf_lmax; i++)
  {
    total_elements += 2*i + 1;
  }

  buffer = malloc (total_elements * sizeof (CCTK_REAL));

  *buffer++ = *radius;
  for (i = 0; i < ahf_lmax; i++)
  {
    for (j = i; j >= 0; j--)
    {
      *buffer++ = cs[j*ahf_lmax + i];
    }
    *buffer++ = c0[i];
    for (j = 0; j <= i; j++)
    {
      *buffer++ = cc[j*ahf_lmax + i];
    }
  }
  buffer -= total_elements;

  /* create the dataset with a 1D filespace */
  HDF5_ERROR (filespace = H5Screate_simple (1, &total_elements, NULL));
  HDF5_ERROR (dataset = H5Dcreate (geom_group, "R", HDF5_REAL, filespace,
                                     H5P_DEFAULT));
  HDF5_ERROR (H5Dwrite (dataset, HDF5_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT,
                          buffer));
  HDF5_ERROR (H5Dclose (dataset));
  HDF5_ERROR (H5Sclose (filespace));

  free (buffer);

  return (0);
}


static int WriteGaussianCurvature (const cGH *GH, hid_t group, int ntheta,
                                   int nphi)
{
  hid_t filespace, dataset;
  const void *buffer;
  hsize_t dims[2];
  const char *const *domain;
  const ioHDF5UtilGH *utilGH;


  if (CCTK_QueryGroupStorage (GH, "AHFinder::ahfinder_gaussian") <= 0)
  {
    CCTK_WARN (1, "WriteGaussianCurvature: No storage enabled for "
                  "AHFinder Gaussian curvature");
    return (-1);
  }

  /* get IOAHFinderHDF5 and IOHDF5Util GH extensions */
  utilGH = CCTK_GHExtension (GH, "IOHDF5Util");

  dims[0] = ntheta + 1;
  dims[1] = nphi + 1;
  buffer = CCTK_VarDataPtr (GH, 0, "AHFinder::gaussian");

  /* create the dataset with a 1D filespace */
  HDF5_ERROR (filespace = H5Screate_simple (2, dims, NULL));
  HDF5_ERROR (dataset = H5Dcreate (group, "Gaussian curvature", HDF5_REAL,
                                   filespace, H5P_DEFAULT));
  HDF5_ERROR (H5Dwrite (dataset, HDF5_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT,
                        buffer));
  domain = (const char *const *)
           CCTK_ParameterGet ("domain", CCTK_ImplementationThorn ("Grid"),NULL);
  WRITE_ATTRIBUTE ("Domain", domain ? *domain : "unknown", dataset, utilGH, 0,
                   utilGH->HDF5_STRING);
  HDF5_ERROR (H5Dclose (dataset));
  HDF5_ERROR (H5Sclose (filespace));

  return (0);
}
