 /*@@
   @file      Startup.c
   @date      Sun 5 Aug 2001
   @author    Thomas Radke
   @desc
              Startup routines for IOAHFinderHDF5
   @enddesc
   @version   $Id: Startup.c,v 1.12 2004/06/21 16:26:33 tradke Exp $
 @@*/


#include <stdlib.h>
#include <string.h>
#include <hdf5.h>

#include "cctk.h"
#include "cctk_Parameters.h"
#include "util_Network.h"
#include "CactusBase/IOUtil/src/ioutil_AdvertisedFiles.h"
#include "ioAHFinderHDF5GH.h"

/* the rcs ID and its dummy function to use it */
static const char *rcsid = "$Header: /numrelcvs/AEIThorns/IOAHFinderHDF5/src/Startup.c,v 1.12 2004/06/21 16:26:33 tradke Exp $";
CCTK_FILEVERSION(AEIThorns_IOAHFinderHDF5_Startup_c)


/********************************************************************
 ********************    Macro Definitions   ************************
 ********************************************************************/
#define FILENAME_TEMPLATE "fileXXXXXX"


/********************************************************************
 ********************    External Routines   ************************
 ********************************************************************/
void IOAHFinderHDF5_Startup (void);
void IOAHFinderHDF5_Terminate (const cGH *GH);


/********************************************************************
 ********************    Internal Routines   ************************
 ********************************************************************/
static void *SetupGH (tFleshConfig *config, int conv_level, cGH *GH);

 /*@@
   @routine   IOAHFinderHDF5_Startup
   @date      Sun 5 Aug 2001
   @author    Thomas Radke
   @desc
              The startup registration routine for IOAHFinderHDF5.
              Registers the GH extensions needed for IOAHFinderHDF5
              along with its setup routine.
   @enddesc
   @calls     CCTK_RegisterGHExtension
              CCTK_RegisterGHExtensionSetupGH
@@*/
void IOAHFinderHDF5_Startup (void)
{
  DECLARE_CCTK_PARAMETERS


  /* check that thorn AHFinder was activated */
  if (CCTK_IsThornActive ("AHFinder"))
  {
    /* check that IOAHFinder output (to file and/or stream) was requested */
    if (write_to_file || write_to_stream)
    {
      CCTK_RegisterGHExtensionSetupGH (
        CCTK_RegisterGHExtension ("IOAHFinderHDF5"), SetupGH);
    }
    else
    {
      CCTK_INFO ("No IOAHFinderHDF5 output requested.");
    }
  }
  else
  {
    CCTK_WARN (1, "Thorn AHFinder was not activated. "
                  "No IOAHFinderHDF5 I/O methods will be enabled.");
  }
}


 /*@@
   @routine   IOAHFinderHDF5_Terminate
   @date      Sun 09 Sep 2001
   @author    Thomas Radke
   @desc
              IOAHFinderHDF5's termination routine
              Closes the open data socket and removes the temporary file.
   @enddesc

   @var       GH
   @vdesc     Pointer to CCTK grid hierarchy
   @vtype     const cGH *
   @vio       in
   @endvar
@@*/
void IOAHFinderHDF5_Terminate (const cGH *GH)
{
  ioAHFinderHDF5GH *myGH;


  myGH = CCTK_GHExtension (GH, "IOAHFinderHDF5");
  if (myGH)
  {
    /* close the data output socket */
    if (myGH->socket != INVALID_SOCKET)
    {
      Socket_CloseSocket (myGH->socket);
    }

    /* remove advertised file */
    if (strcmp (myGH->advertised_filename, FILENAME_TEMPLATE))
    {
      remove (myGH->advertised_filename);
    }
  }
}


/********************************************************************
 ********************    Internal Routines   ************************
 ********************************************************************/
 /*@@
   @routine   SetupGH
   @date      Sun 5 Aug 2001
   @author    Thomas Radke
   @desc
              Allocates and sets up IOAHFinderHDF5's GH extension structure
   @enddesc

   @calls     CCTK_RegisterIOMethod
              CCTK_RegisterIOMethodOutputVarAs

   @var       config
   @vdesc     the CCTK configuration as provided by the flesh
   @vtype     tFleshConfig *
   @vio       unused
   @endvar
   @var       conv_level
   @vdesc     the convergence level
   @vtype     int
   @vio       unused
   @endvar
   @var       GH
   @vdesc     Pointer to CCTK grid hierarchy
   @vtype     cGH *
   @vio       unused
   @endvar

   @returntype void *
   @returndesc
               pointer to the allocated GH extension structure
   @endreturndesc
@@*/
static void *SetupGH (tFleshConfig *config, int conv_level, cGH *GH)
{
  ioAHFinderHDF5GH *myGH;
#ifdef H5_HAVE_STREAM
  char hostname[256];
  FILE *advertised_file_fd;
  ioAdvertisedFileDesc advertised_file;
#endif /* H5_HAVE_STREAM */
  DECLARE_CCTK_PARAMETERS


  /* suppress compiler warnings about unused variables */
  (void) (config + 0);
  (void) (conv_level + 0);
  (void) (GH + 0);

  /* allocate the GH extension and its components */
  myGH = calloc (1, sizeof (ioAHFinderHDF5GH));

  if (myGH)
  {
    /* register the IOAHFinderHDF5 routine as output method  */
    CCTK_RegisterIOMethodOutputVarAs (CCTK_RegisterIOMethod ("IOAHFinderHDF5"),
                                      IOAHFinderHDF5_OutputVarAs);

    /* mark the HDF5 handles for open files and toplevel groups as invalid */
    myGH->stream = myGH->file = -1;
    myGH->time_group_stream = myGH->time_group_file = -1;
    strcpy (myGH->advertised_filename, FILENAME_TEMPLATE);

    /* reset number of iterations counter */
    myGH->ntimes_called = 0;

    /* only processor 0 is doing socket I/O */
    myGH->socket = INVALID_SOCKET;
    if (write_to_stream && CCTK_MyProc (GH) == 0)
    {
#ifdef H5_HAVE_STREAM
      myGH->socket = Socket_TCPOpenServerSocket (port, &myGH->port, 1);
      if (myGH->socket == INVALID_SOCKET)
      {
        CCTK_VWarn (1, __LINE__, __FILE__, CCTK_THORNSTRING,
                    "Couldn't open TCP server socket on output port %d. No "
                    "AHFinder HDF5 streaming output will be available !", port);
      }
      else if (Socket_SetNonBlocking (myGH->socket) < 0)
      {
        CCTK_WARN (1, "Couldn't set output socket into non-blocking mode. "
                      "No AHFinder HDF5 streaming output will be available !");
        Socket_CloseSocket (myGH->socket);
        myGH->socket = INVALID_SOCKET;
      }
      else
      {
        Util_GetHostName (hostname, sizeof (hostname));
        CCTK_VInfo (CCTK_THORNSTRING,
                    "AHFinder HDF5 data streaming service started on\n"
                    "                         %s:%u",
                    hostname, myGH->port);

        /* write the hostname/portnumber information in a temporary file
           which gets advertised and then can be downloaded */
#ifdef HAVE_MKSTEMP
        advertised_file_fd = fdopen (mkstemp (myGH->advertised_filename), "w");
#else
        advertised_file_fd = tmpnam (myGH->advertised_filename) ?
                             fopen (myGH->advertised_filename, "w") : NULL;
#endif
        if (advertised_file_fd)
        {
          fprintf (advertised_file_fd, "Hostname:  %s\n", hostname);
          fprintf (advertised_file_fd, "Data port: %u", myGH->port);
          fclose (advertised_file_fd);

          advertised_file.slice = "";
          advertised_file.thorn = CCTK_THORNSTRING;
          advertised_file.varname = "AHFinder data";
          advertised_file.description = "Streamed AHFinder HDF5 data";
          advertised_file.mimetype = "data/x-streamed-ahfinder-hdf5";

          IOUtil_AdvertiseFile (GH, myGH->advertised_filename,
                                &advertised_file);
        }
        else
        {
          CCTK_WARN (2, "Couldn't create unique temporary file ! "
                        "AHFinder HDF5 data streaming was not advertised.");
        }
      }
#else
      CCTK_WARN (1, "Cactus was compiled with an HDF5 installation that didn't "
                    "include the Stream Virtual File Driver !\n"
                    "No AHFinder HDF5 streaming output will be available !");
#endif /* H5_HAVE_STREAM */
    }

    /* initialize to -1 so that the 'IOAHFinderHDF5::out_vars' parameter
       will be parsed the first time thru */
    myGH->out_vars_lastset = -1;

    /* allocate the I/O request description list */
    myGH->requests = calloc (CCTK_NumVars (), sizeof (ioRequest *));

    if (! CCTK_Equals (verbose, "none"))
    {
      CCTK_INFO ("I/O Method 'IOAHFinderHDF5' registered: output of AHFinder "
                 "data to HDF5 files");
    }
  }
  else
  {
    CCTK_WARN (0, "Unable to allocate memory for GH in IOAHFinderHDF5");
  }

  return (myGH);
}
