 /*@@
   @file      POWERClock.c
   @date      Aug 16 2004
   @author    Steve White
   @desc
              A Cactus timer clock based on the POWER Real-Time Clock
	      (RTC) or PowerPC Time Base Register on a AIX-based system.
	      see
http://nscp.upenn.edu/aix4.3html/aixbman/prftungd/accproctimer.htm
http://www.nersc.gov/vendor_docs/ibm/libs/basetrf2/read_real_time.htm
   @enddesc
   @version   $Id: POWERClock.c,v 1.1.1.1 2004/08/19 11:32:40 swhite Exp $
 @@*/

#include <stdlib.h>
#include <sys/time.h>

#include "cctk.h"

#ifdef TIMEBASE_SZ

static const char *rcsid = "$Header: /numrelcvs/AEIThorns/POWERClock/src/POWERClock.c,v 1.1.1.1 2004/08/19 11:32:40 swhite Exp $";

CCTK_FILEVERSION(CactusUtils_POWERClock_c)

static const char * ClockHeading = "POWER RTC";
static char *ClockUnits = "secs";
#if 0
typedef struct timebasestruct {
        int             flag;    /* indicats time base or real time */
        unsigned int    tb_high; /* high 32 bits, or seconds */
        unsigned int    tb_low;  /* low 32 bits, or nanoseconds */
        } timebasestruct_t;

The time_base_to_time subroutine converts time base information to real time,
if necessary.

If t->flag is RTC_POWER, the subroutine simply returns (the data is already in real time format).

If t->flag is RTC_POWER_PPC, the time base information in t->tb_high and t->tb_low is converted to seconds and nanoseconds; t->tb_high is replaced by the seconds; t->tb_low is replaced by the nanoseconds; and t->flag is changed to RTC_POWER.

#endif
/* The read_real_time subroutine reads either the processor real time clock
 * (for POWER or PowerPC 601) or the time base register
 * (in the case of PowerPC processors other than the PowerPC 601).
 * The time_base_to_time converts the structure values to "real time":
 * tb_high = sec, tb_low = nanosec
 * */
static void get_real_time( timebasestruct_t* tb )
{
  read_real_time( tb, TIMEBASE_SZ);
  time_base_to_time( tb, TIMEBASE_SZ);
}

/*
const timebasestruct_t TIMEBASE_ZERO = { RTC_POWER, 0, 0 };
*/
const timebasestruct_t TIMEBASE_ZERO = { 0, 0, 0 };

typedef struct
{
  double           total;
  timebasestruct_t last;
} t_Clock;

static double
time_interval( timebasestruct_t* finish, timebasestruct_t* start )
{
  double time = 0;
  int w1 = finish->tb_high - start->tb_high;
  int w2 = finish->tb_low - start->tb_low; 
       
  /* if there was a carry from low-order to high-order during  */
  /* the measurement, we may have to undo it.                  */
  if( w2 < 0)
  {    
    w1--;
    w2 += 1000000000;
  }
  time = ((double) w2) * 1.e-9;  /* usec -> sec */
  if( w1 > 0)      
    time += (double) w1;
  return time;
}

/* --------------  Clock functions -------------------- */
static void *ClockCreate(int timernum)
{
  t_Clock *data = (t_Clock *)calloc (1, sizeof (t_Clock));

  (void) (timernum + 0);

  get_real_time( &data->last );
  return data;
}

static void ClockDestroy(int timernum, void *data)
{
  (void) (timernum + 0);

  if( data )
  {
    free( data );
  }
}

static void ClockStart(int timernum, void *idata)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  get_real_time( &data->last );
  data->total = 0.0;
}

static void ClockStop(int timernum, void *idata)
{
  timebasestruct_t tc, diff;
  t_Clock *data = ( t_Clock *)idata;
  tc = data->last;
  (void) (timernum + 0);

  get_real_time( &data->last );
  data->total += time_interval( &data->last, &tc );
}

static void ClockReset(int timernum, void *idata)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  get_real_time( &data->last );
  data->total = 0.0;
}

static void ClockGet(int timernum, void *idata, cTimerVal *vals)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  vals[0].type    = val_double;
  vals[0].heading = ClockHeading;
  vals[0].units   = ClockUnits;
  vals[0].val.d   = data->total;
  vals[0].seconds = vals[0].val.d;
  vals[0].resolution = 1.e-9;
}

static void ClockSet(int timernum, void *idata, cTimerVal *vals)
{
  t_Clock *data = ( t_Clock *)idata;
  unsigned long	carry;
  double	val = vals[0].val.d;

  (void) (timernum + 0);

  get_real_time( &data->last );
  data->total = 0.0;
}

#else
#error "Thorn POWERClock requires an AIX 4 development environment."
#endif /*TIMEBASE_SZ*/

 /*@@
   @routine    POWERClock_Register
   @date       Aug 16 2004
   @author     Steve White
   @desc       Register the POWER clock with the Flesh
   @enddesc
   @calls      
@@*/
void POWERClock_Register(void)
{
  cClockFuncs functions;
  functions.n_vals  = 1;
  functions.create  = ClockCreate;
  functions.destroy = ClockDestroy;
  functions.start   = ClockStart;
  functions.stop    = ClockStop;
  functions.reset   = ClockReset;
  functions.get     = ClockGet;
  functions.set     = ClockSet;

  CCTK_ClockRegister( "POWER", &functions );
}
