 /*@@
   @file      TSCClock.c
   @date      Aug 16 2004
   @author    Steve White
   @desc
              A Cactus timer clock based on the Pentium Time-Stamp Counter
	      (TSC) .
	      see
http://www.cs.rochester.edu/~cding/Teaching/252Spring2003/LectureNotes/Time.ppt
              http://www.math.uwaterloo.ca/~jamuir/rdtscpm1.pdf
	      FOR ALPHA RPCC:
              http://www.mostang.com/~davidm/papers/expo97/paper/doc004.html
	Note: these docs say more can be done to make timer counts more
	accurate

   @enddesc
   @version   $Id: TSCClock.c,v 1.6 2004/12/29 22:49:01 schnetter Exp $
 @@*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cctk.h"

#include <asm/cpufeature.h>	/* more portable means? */
#ifdef X86_FEATURE_TSC

static const char *rcsid = "$Header: /numrelcvs/AEIThorns/TSCClock/src/TSCClock.c,v 1.6 2004/12/29 22:49:01 schnetter Exp $";

CCTK_FILEVERSION(CactusUtils_TSCClock_c)

static const char * ClockHeading = "TSC";
static char *ClockUnits = "secs";

/* The Pentium Time Stamp clock is a 64 bit unsigned integer composed
 * of two 32-bit unsigned integers */
typedef struct
{
  unsigned long lo;
  unsigned long hi;
} t_ClockCounter;

static double
time_interval( const t_ClockCounter * a, const t_ClockCounter * b )
{
  static double  double32bit = 4294967296.0;	/* 2^32 */

  t_ClockCounter tc = { a->lo - b->lo, a->hi - b->hi };
  unsigned long  borrow = tc.lo > a->lo;
  tc.hi -= borrow;
  return (double)tc.hi * double32bit + tc.lo;
}

typedef struct
{
  double total;
  double rate;
  t_ClockCounter last;
} t_Clock;

/*
 * Here's your CPU and assembler-specific bit.
 * I got this off a web page (see above)
 */
static void get_counter( t_ClockCounter *tc )
{
 /* rdtsc means (ReaD Time-Stamp Counter) */
  asm( "rdtsc; movl %%edx,%0; movl %%eax,%1"
       : "=r" (tc->hi), "=r" (tc->lo)
       : /* no input */
       : "%edx", "%eax" );
}

static double readCPURate( void );

/* --------------  Clock functions -------------------- */
static void *ClockCreate(int timernum)
{
  t_Clock *data = ( t_Clock *)calloc (1, sizeof (t_Clock));

  (void) (timernum + 0);

  get_counter( &data->last );
  data->rate = readCPURate();
  return data;
}

static void ClockDestroy(int timernum, void *data)
{
  (void) (timernum + 0);

  if( data )
  {
    free( data );
  }
}

static void ClockStart(int timernum, void *idata)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  get_counter( &data->last );
}

static void ClockStop(int timernum, void *idata)
{
  t_Clock        *data = ( t_Clock *)idata;
  t_ClockCounter tc = data->last;

  (void) (timernum + 0);

  get_counter( &data->last );
  data->total += time_interval( &data->last, &tc ) / data->rate;
}

static void ClockReset(int timernum, void *idata)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  get_counter( &data->last );
  data->total = 0.0;
}

static void ClockGet(int timernum, void *idata, cTimerVal *vals)
{
  t_Clock *data = ( t_Clock *)idata;
  (void) (timernum + 0);

  vals[0].type    = val_double;
  vals[0].heading = ClockHeading;
  vals[0].units   = ClockUnits;
  vals[0].val.d   = data->total;
  vals[0].seconds = vals[0].val.d;
  vals[0].resolution = data->rate;
}

static void ClockSet(int timernum, void *idata, cTimerVal *vals)
{
  t_Clock        *data = ( t_Clock *)idata;

  (void) (timernum + 0);

  get_counter( &data->last );

  data->total = vals[0].val.d;
}

static const size_t BUFSIZE = 80;

/* Read off the CPU rate from the Linux /proc filesystem.
 * A) Oh, I wish I had a nice string library.
 * B) I don't believe this is robust.  The formatting of /proc/cpuinfo alone...
 * */
static double
readCPURate( void )
{
  double rate = 0.0;
  FILE   *cpuinfo = fopen( "/proc/cpuinfo", "r" );

  if( cpuinfo != NULL )
  {
    char buf[BUFSIZE];
    int  nextchar = 0;
    int  found_item = 0;

    buf[0] = '\0';

    while( ( nextchar = fgetc( cpuinfo ) ) != EOF )
    {
      if( nextchar == '\n' )
      {
        if( found_item )
        {
          rate = atof( buf ) * 1.e6; /* that's MHz */
	  break;
        }
	buf[0] = '\0';
      }
      else
      {
        size_t len = strlen( buf );
	if( len + 1 < BUFSIZE )	/* unless line doesn't fit buffer */
	{
          buf[len+1] = '\0';
	  buf[len] = nextchar;
	  if( strcmp( buf, "cpu MHz\t\t: " ) == 0 )
	  {
            found_item = 1;
	    buf[0] = '\0';
	  }
	}
      }
    }
    fclose( cpuinfo );
  }
  else
  {
    CCTK_VWarn(CCTK_WARN_ABORT, __LINE__ , __FILE__, "TSCClock",
              "No file '/proc/cpuinfo'." );
  }
  return rate;
}
 /*@@
   @routine    TSCClock_Register
   @date       Aug 16 2004
   @author     Steve White
   @desc       Register the TSC clock with the Flesh
   @enddesc
   @calls
@@*/
void TSCClock_Register(void)
{
  cClockFuncs functions;
  double rate = readCPURate();

  if( rate > 0.0 )
  {
    functions.n_vals  = 1;
    functions.create  = ClockCreate;
    functions.destroy = ClockDestroy;
    functions.start   = ClockStart;
    functions.stop    = ClockStop;
    functions.reset   = ClockReset;
    functions.get     = ClockGet;
    functions.set     = ClockSet;

    CCTK_ClockRegister( "TSC", &functions );
  }
  else
    CCTK_VWarn( CCTK_WARN_ABORT, __LINE__ , __FILE__, "TSCClock",
              "Could not determine CPU clock rate." );
}

#else
#error "The TSCClock thorn requires a Pentium architecture environment"
#endif /* cpu_has_tsc */
