/* $Header: /numrelcvs/AEIThorns/TensorTypes/src/tensortypes.c,v 1.1 2004/08/31 20:18:49 schnetter Exp $ */

#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "cctk.h"

#include "tensortypes.h"



/* Calculate the number of elements in an array */
#define ARRSIZE(T) (sizeof(T) / sizeof(*(T)))



/* A scalar */
static int const scalar_vars[] = {
  0
};
static int const scalar_parity[] = {
  +1
};
static int const scalar_comps[] = {
  0
};
struct tensor const TT_scalar = {
  "scalar", 3, 0,
  ARRSIZE(scalar_vars), scalar_vars, scalar_parity,
  ARRSIZE(scalar_comps), scalar_comps
};



/* A vector */
static int const vector_vars[] = {
  0,1,2
};
static int const vector_parity[] = {
  +1,+1,+1
};
static int const vector_comps[] = {
  0,1,2
};
struct tensor const TT_vector = {
  "vector", 3, 1,
  ARRSIZE(vector_vars), vector_vars, vector_parity,
  ARRSIZE(vector_comps), vector_comps
};



/* A second rank tensor without symmetries */
static int const tensor_vars[] = {
  0,1,2,   3,4,5,   6,7,8
};
static int const tensor_parity[] = {
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1
};
static int const tensor_comps[] = {
  0,1,2,   3,4,5,   6,7,8
};
struct tensor const TT_tensor = {
  "tensor", 3, 2,
  ARRSIZE(tensor_vars), tensor_vars, tensor_parity,
  ARRSIZE(tensor_comps), tensor_comps
};



/* A symmetric second rank tensor */
static int const symmtensor_vars[] = {
  0,1,2,   1,3,4,   2,4,5
};
static int const symmtensor_parity[] = {
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1
};
static int const symmtensor_comps[] = {
  0,1,2,   4,5,   8
};
struct tensor const TT_symmtensor = {
  "symmetric tensor T_(ij)", 3, 2,
  ARRSIZE(symmtensor_vars), symmtensor_vars, symmtensor_parity,
  ARRSIZE(symmtensor_comps), symmtensor_comps
};



/* A third rank tensor with symmetry in the first two indices */
static int const symmtensor3a_vars[] = {
   0, 1, 2,
   3, 4, 5,
   6, 7, 8,
  
   3, 4, 5,
   9,10,11,
  12,13,14,
  
   6, 7, 8,
  12,13,14,
  15,16,17
};
static int const symmtensor3a_parity[] = {
  +1,+1,+1,
  +1,+1,+1,
  +1,+1,+1,
  
  +1,+1,+1,
  +1,+1,+1,
  +1,+1,+1,
  
  +1,+1,+1,
  +1,+1,+1,
  +1,+1,+1,
};
static int const symmtensor3a_comps[] = {
   0, 1, 2,
   3, 4, 5,
   6, 7, 8,
  
  12,13,14,
  15,16,17,
  
  24,25,26
};
struct tensor const TT_symmtensor3a = {
  "symmetric tensor T_(ij)k", 3, 3,
  ARRSIZE(symmtensor3a_vars), symmtensor3a_vars, symmtensor3a_parity,
  ARRSIZE(symmtensor3a_comps), symmtensor3a_comps
};



/* A third rank tensor with symmetry in the last two indices */
static int const symmtensor3b_vars[] = {
   0, 1, 2,    1, 3, 4,    2, 4, 5,
   6, 7, 8,    7, 9,10,    8,10,11,
  12,13,14,   13,15,16,   14,16,17
};
static int const symmtensor3b_parity[] = {
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1
};
static int const symmtensor3b_comps[] = {
   0, 1, 2,    4, 5,    8,
   9,10,11,   13,14,   17,
  18,19,20,   22,23,   26
};
struct tensor const TT_symmtensor3b = {
  "symmetric tensor T_i(jk)", 3, 3,
  ARRSIZE(symmtensor3b_vars), symmtensor3b_vars, symmtensor3b_parity,
  ARRSIZE(symmtensor3b_comps), symmtensor3b_comps
};



/* A fourth rank tensor with symmetries both in its first and last two
   indices */
static int const symmtensor4_vars[] = {
   0, 1, 2,    1, 3, 4,    2, 4, 5,
   6, 7, 8,    7, 9,10,    8,10,11,
  12,13,14,   13,15,16,   14,16,17,
  
   6, 7, 8,    7, 9,10,    8,10,11,
  18,19,20,   19,21,22,   20,22,23,
  24,25,26,   25,27,28,   26,28,29,
  
  12,13,14,   13,15,16,   14,16,17,
  24,25,26,   25,27,28,   26,28,29,
  30,31,32,   31,33,34,   32,34,35
};
static int const symmtensor4_parity[] = {
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1,
  +1,+1,+1,   +1,+1,+1,   +1,+1,+1
};
static int const symmtensor4_comps[] = {
   0, 1, 2,    4, 5,    8,
   9,10,11,   13,14,   17,
  18,19,20,   22,23,   26,
  
  36,37,38,   40,41,   44,
  45,46,47,   49,50,   53,
  
  72,73,74,   76,77,   80
};
struct tensor const TT_symmtensor4 = {
  "symmetric tensor T_(ij)(kl)", 3, 4,
  ARRSIZE(symmtensor4_vars), symmtensor4_vars, symmtensor4_parity,
  ARRSIZE(symmtensor4_comps), symmtensor4_comps
};



struct tensor const * const TT_alltensors[] = {
  &TT_scalar,
  &TT_vector,
  &TT_tensor,
  &TT_symmtensor,
  &TT_symmtensor3a,
  &TT_symmtensor3b,
  &TT_symmtensor4
};

int const TT_numtensors = ARRSIZE(TT_alltensors);



/* Ensure that all tensor declarations are internally consistent */
void
CheckTensorType (struct tensor const * restrict const atensor)
{
  int i, n;
  assert (atensor->name);
  assert (atensor->dim>=0);
  assert (atensor->rank>=0);
  assert (atensor->ncomps>=0);
  assert (atensor->ncomps == floor(pow(atensor->dim, atensor->rank) + 0.5));
  assert (atensor->nvars>=0 && atensor->nvars<=atensor->ncomps);
  assert (atensor->vars);
  for (i=0; i<atensor->ncomps; ++i) {
    assert (atensor->vars[i]>=0 && atensor->vars[i]<atensor->nvars);
    assert (abs(atensor->parity[i]) <= 1);
  }
  assert (atensor->comps);
  for (n=0; n<atensor->nvars; ++n) {
    assert (atensor->comps[n]>=0 && atensor->comps[n]<atensor->ncomps);
    assert (atensor->vars[atensor->comps[n]] == n);
  }
}
