// InterpGV.cc -- subdriver for CCTK_InterpGV()
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/Attic/InterpGV.cc,v 1.2 2003/01/17 15:23:13 jthorn Exp $
//
// do_InterpGV
//

#include <assert.h>
#include <math.h>

#include "cctk.h"
#include "util_ErrorCodes.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "axis_list.hh"
#include "input_grid.hh"
#include "interp_grid.hh"

//******************************************************************************

//
// This function is a subdriver to do the actual interpolation using
// CCTK_InterpGV().
//
// This only supports the contiguous-in-memory axes specifiers, i.e.
// {x, xy, xyz}.
//
// interp_operator[] is ignored; the interpolation operator is selected
// via  interp_order .  
//
// No derivative operations are supported.
// 
void do_InterpGV(const cGH *GH, int timer_handle,
		 const char coord_system[],
		 const input_grid& input_grid, interp_grid& interp_grid,
		 const char interp_operator[], int interp_order,
					       const char interp_pars[],
		 int N_operations, const int opcodes[max_N_operations])
{
const int coord_system_handle = CCTK_CoordSystemHandle(coord_system);
const int interp_handle = get_interp_handle_from_order(interp_order);      

//
// check that this interpolator can handle this test case
//
input_grid.input_axes.verify_contiguous();
if (! ((N_operations == 1) && (opcodes[0] == DERIV(0))) )
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"***** do_InterpGV(): derivatives aren't supported!\n");	/*NOTREACHED*/
const int iop = 0;

//
// do the actual interpolation
//
int status = -999999;	// dummy value; should always be overwritten in switch
switch	(interp_grid.N_dims)
	{
case 1:
	CCTK_INFO("do_InterpGV(): about to call interpolator for 1-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpGV
		(const_cast<cGH*>(GH),
		 interp_handle, coord_system_handle,
		 interp_grid.N_points,
		 N_input_arrays, N_output_arrays,
		 static_cast<const void*>(interp_grid.interp_x),
		 CCTK_VARIABLE_REAL,    // data type for interp_x
		 input_grid.varindex_real, input_grid.varindex_real4,
		 			   input_grid.varindex_real8,
		 input_grid.varindex_complex, input_grid.varindex_complex8,
		 			      input_grid.varindex_complex16,
		 interp_grid.output_arrays[iop].ptr_real,
		 interp_grid.output_arrays[iop].ptr_real4,
		 interp_grid.output_arrays[iop].ptr_real8,
		 interp_grid.output_arrays[iop].ptr_complex,
		 interp_grid.output_arrays[iop].ptr_complex8,
		 interp_grid.output_arrays[iop].ptr_complex16,
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

case 2:
	CCTK_INFO("do_InterpGV(): about to call interpolator for 2-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpGV
		(const_cast<cGH*>(GH),
		 interp_handle, coord_system_handle,
		 interp_grid.N_points,
		 N_input_arrays, N_output_arrays,
		 static_cast<const void*>(interp_grid.interp_x),
		 static_cast<const void*>(interp_grid.interp_y),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for interp_[xy]
		 input_grid.varindex_real, input_grid.varindex_real4,
		 			   input_grid.varindex_real8,
		 input_grid.varindex_complex, input_grid.varindex_complex8,
		 			      input_grid.varindex_complex16,
		 interp_grid.output_arrays[iop].ptr_real,
		 interp_grid.output_arrays[iop].ptr_real4,
		 interp_grid.output_arrays[iop].ptr_real8,
		 interp_grid.output_arrays[iop].ptr_complex,
		 interp_grid.output_arrays[iop].ptr_complex8,
		 interp_grid.output_arrays[iop].ptr_complex16,
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

case 3:
	CCTK_INFO("do_InterpGV(): about to call interpolator for 3-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpGV
		(const_cast<cGH*>(GH),
		 interp_handle, coord_system_handle,
		 interp_grid.N_points,
		 N_input_arrays, N_output_arrays,
		 static_cast<const void*>(interp_grid.interp_x),
		 static_cast<const void*>(interp_grid.interp_y),
		 static_cast<const void*>(interp_grid.interp_z),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for interp_[xyz]
		 input_grid.varindex_real, input_grid.varindex_real4,
		 			   input_grid.varindex_real8,
		 input_grid.varindex_complex, input_grid.varindex_complex8,
		 			      input_grid.varindex_complex16,
		 interp_grid.output_arrays[iop].ptr_real,
		 interp_grid.output_arrays[iop].ptr_real4,
		 interp_grid.output_arrays[iop].ptr_real8,
		 interp_grid.output_arrays[iop].ptr_complex,
		 interp_grid.output_arrays[iop].ptr_complex8,
		 interp_grid.output_arrays[iop].ptr_complex16,
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "***** N_dims=%d is not supported!",
		   interp_grid.N_dims);				/*NOTREACHED*/
	}

if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "***** error return %d from InterpGV()!",
		   status);					/*NOTREACHED*/
CCTK_VInfo(CCTK_THORNSTRING,
	   "do_InterpGV(): status=%d return from interpolator",
	   status);
}
