// InterpLocal.cc -- subdriver for CCTK_InterpLocal()
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/Attic/InterpLocal.cc,v 1.3 2005/09/23 04:20:40 schnetter Exp $
//
// do_InterpLocal
//

#include <assert.h>
#include <math.h>
#include <stddef.h>

#include "cctk.h"
#include "util_ErrorCodes.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "axis_list.hh"
#include "input_grid.hh"
#include "interp_grid.hh"

//******************************************************************************

// prototypes for functions local to this file
namespace {
fp *setup_coord_array(char axis, int N, fp origin, fp delta);
	  };

//******************************************************************************

//
// This function is a subdriver to do the actual interpolation using
// CCTK_InterpLocal().
//
// This only supports the contiguous-in-memory axes specifiers, i.e.
// {x, xy, xyz}.
//
// interp_operator[] is ignored; the interpolation operator is selected
// via  interp_order .
//
// No derivative operations are supported.
// 
void do_InterpLocal(const cGH *GH, int timer_handle,
		    const char coord_system[],
		    const input_grid& input_grid, interp_grid& interp_grid,
		    const char interp_operator[], int interp_order,
						  const char interp_pars[],
		    int N_operations, const int opcodes[max_N_operations])
{
const int iop = 0;
const int N_dims = interp_grid.N_dims;
const int interp_handle = get_interp_handle_from_order(interp_order);

//
// check that this interpolator can handle this test case
//
input_grid.input_axes.verify_contiguous();
if (! ((N_operations == 1) && (opcodes[0] == DERIV(0))) )
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"***** do_InterpGV(): derivatives aren't supported!\n");	/*NOTREACHED*/


//
// create the coordinate arrays
//
fp *coords_x = (N_dims >= 1) ? setup_coord_array('x',
						 input_grid.array_dims[0],
						 input_grid.xyz_origin[0],
						 input_grid.xyz_delta[0])
			     : NULL;
fp *coords_y = (N_dims >= 2) ? setup_coord_array('y',
						 input_grid.array_dims[1],
						 input_grid.xyz_origin[1],
						 input_grid.xyz_delta[1])
			     : NULL;
fp *coords_z = (N_dims >= 3) ? setup_coord_array('z',
						 input_grid.array_dims[2],
						 input_grid.xyz_origin[2],
						 input_grid.xyz_delta[2])
			     : NULL;

//
// do the actual interpolation
//
int status = -999999;	// dummy value; should always be overwritten in switch
switch	(N_dims)
	{
case 1:
	CCTK_INFO("do_InterpLocal(): about to call interpolator for 1-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpLocal
		(const_cast<cGH*>(GH),
		 interp_handle,
		 interp_grid.N_points,
		 N_dims,
		 N_input_arrays, N_output_arrays,
		 input_grid.array_dims[0],
		 coords_x,
		 CCTK_VARIABLE_REAL,    // data type for coords_x
		 static_cast<const void*>(interp_grid.interp_x),
		 CCTK_VARIABLE_REAL,    // data type for interp_x
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real4),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16,
					// data types for input_grid.array_*
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real4),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

case 2:
	CCTK_INFO("do_InterpLocal(): about to call interpolator for 2-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpLocal
		(const_cast<cGH*>(GH),
		 interp_handle,
		 interp_grid.N_points,
		 N_dims,
		 N_input_arrays, N_output_arrays,
		 input_grid.array_dims[0], input_grid.array_dims[1],
		 coords_x, coords_y,
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for coords_[xy]
		 static_cast<const void*>(interp_grid.interp_x),
		 static_cast<const void*>(interp_grid.interp_y),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for interp_[xy]
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real4),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16,
					// data types for input_grid.array_*
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real4),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

case 3:
	CCTK_INFO("do_InterpLocal(): about to call interpolator for 3-D");
	CCTK_TimerStartI(timer_handle);
	status
	   = CCTK_InterpLocal
		(const_cast<cGH*>(GH),
		 interp_handle,
		 interp_grid.N_points,
		 N_dims,
		 N_input_arrays, N_output_arrays,
		 input_grid.array_dims[0], input_grid.array_dims[1],
					   input_grid.array_dims[2],
		 coords_x, coords_y, coords_z,
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for coords_[xyz]
		 static_cast<const void*>(interp_grid.interp_x),
		 static_cast<const void*>(interp_grid.interp_y),
		 static_cast<const void*>(interp_grid.interp_z),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL,
					// data types for interp_[xyz]
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real4),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_real8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex8),
		 static_cast<const void*>(input_grid.data_ptrs.ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16,
					// data types for input_grid.array_*
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real4),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_real8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex8),
		 static_cast<void*>(interp_grid.output_arrays[iop].ptr_complex16),
		 CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
		 CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8,
					CCTK_VARIABLE_COMPLEX16);
					// data types for interp_grid.array_*
	CCTK_TimerStopI(timer_handle);
	break;

default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "***** N_dims=%d is not supported!",
		   N_dims);					/*NOTREACHED*/
	}

if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "***** error return %d from InterpLocal()!",
		   status);					/*NOTREACHED*/
CCTK_VInfo(CCTK_THORNSTRING,
	   "do_InterpLocal(): status=%d return from interpolator",
	   status);

if (N_dims >= 3) then delete[] coords_z;
if (N_dims >= 2) then delete[] coords_y;
if (N_dims >= 1) then delete[] coords_x;
}

//******************************************************************************

//
// This function new[]-allocates and sets up a coordinate array for
// InterpLocal().  It returns a pointer to the coordinate array.
//
// Arguments:
// axis = 'x', 'y', or 'z' to specify which axis this is.  This argument
//	  is (only) used for formatting messages.
// N = The number of points in the coordinate array (numbered 0 to N-1).
// origin,delta = These specify the coordinates.
//
namespace {
fp *setup_coord_array(char axis, int N, fp origin, fp delta)
{
fp *const coords = new fp[N];
	for (int i = 0 ; i < N ; ++i)
	{
	coords[i] = origin + i*delta;
	}

CCTK_VInfo(CCTK_THORNSTRING,
	   "setup_coord_array('%c'): coords[0,%d] = [%g,%g]",
	   axis, N-1, double(coords[0]), double(coords[N-1]));

return coords;
}
	  }
