// InterpLocalUniform.cc -- subdriver for CCTK_InterpLocalUniform()
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/InterpLocalUniform.cc,v 1.3 2003/01/18 17:55:13 jthorn Exp $
//
// do_InterpLocalUniform
/// setup_input_array_subscripting
//

#include <assert.h>
#include <math.h>

#include "cctk.h"
#include "util_ErrorCodes.h"
#include "util_Table.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "axis_list.hh"
#include "input_grid.hh"
#include "interp_grid.hh"

#undef DEBUG_ARRAY_INDEXING

//******************************************************************************

// prototypes for functions local to this file
namespace {
void setup_input_array_subscripting(int param_table_handle,
				    const input_grid& input_grid);
	  }

//******************************************************************************

//
// This function is a subdriver to do the actual interpolation using
// CCTK_InterpLocalUniform().
//
void do_InterpLocalUniform(const cGH *GH, int timer_handle,
			   const char coord_system[],
			   const input_grid& input_grid,
			   interp_grid& interp_grid,
			   const char interp_operator[], int interp_order,
							 const char interp_pars[],
			   int N_operations, const int opcodes[max_N_operations])
{
int status, status1, status2;

const int N_output = N_output_arrays * N_operations;

const int param_table_handle
	= setup_interp_param_table(interp_pars, interp_order);
const int interp_handle = get_interp_handle_from_name(interp_operator);
const int coord_system_handle = get_coord_system_handle(coord_system);

if (! input_grid.input_axes.is_prefix_of_xyz())
   then setup_input_array_subscripting(param_table_handle,
				       input_grid);

setup_derivs_in_param_table(param_table_handle,
			    N_operations, opcodes);



//
// set up the interpolation coordinates
//
const void *const interp_coords[max_N_dims]
	= { static_cast<const void *>(interp_grid.interp_x),
	    static_cast<const void *>(interp_grid.interp_y),
	    static_cast<const void *>(interp_grid.interp_z) };

//
// set up the input arrays
//
const CCTK_INT input_array_type_codes[N_input_arrays]
   = {
     CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
     CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8, CCTK_VARIABLE_COMPLEX16,
     };
const void* input_arrays[N_input_arrays]
	= {
	  static_cast<const void*>(input_grid.data_ptrs.ptr_real),
	  static_cast<const void*>(input_grid.data_ptrs.ptr_real4),
	  static_cast<const void*>(input_grid.data_ptrs.ptr_real8),
	  static_cast<const void*>(input_grid.data_ptrs.ptr_complex),
	  static_cast<const void*>(input_grid.data_ptrs.ptr_complex8),
	  static_cast<const void*>(input_grid.data_ptrs.ptr_complex16),
	  };

//
// set up the output arrays
//
CCTK_INT output_array_type_codes[max_N_output];
void*    output_arrays          [max_N_output];
setup_output_arrays(interp_grid,
		    output_array_type_codes, output_arrays);

#ifdef DEBUG_ARRAY_INDEXING
printf("### input_grid.data_ptrs.ptr_real[0] = %g\n",
       input_grid.data_ptrs.ptr_real[0]);
printf("### input_grid.data_ptrs.ptr_real[1] = %g\n",
       input_grid.data_ptrs.ptr_real[1]);

printf("### interp_grid.interp_xyz[0] = %g %g %g\n",
       interp_grid.interp_x[0], interp_grid.interp_y[0], interp_grid.interp_z[0]);
printf("### interp_grid.interp_xyz[1] = %g %g %g\n",
       interp_grid.interp_x[1], interp_grid.interp_y[1], interp_grid.interp_z[1]);
printf("### interp_grid.interp_xyz[10] = %g %g %g\n",
       interp_grid.interp_x[10], interp_grid.interp_y[10], interp_grid.interp_z[10]);
#endif

//
// do the actual interpolation
//
CCTK_VInfo(CCTK_THORNSTRING,
	   "do_InterpLocalUniform(): about to call interpolator (%d-D)",
	   interp_grid.N_dims);

CCTK_TimerStartI(timer_handle);
status = CCTK_InterpLocalUniform(interp_grid.N_dims,
				 interp_handle,
				 param_table_handle,
				 input_grid.xyz_origin, input_grid.xyz_delta,
				 interp_grid.N_points,
				    CCTK_VARIABLE_REAL,
				    interp_coords,
				 N_input_arrays,
				    input_grid.array_dims,
				    input_array_type_codes,
				    input_arrays,
				 N_output,
				    output_array_type_codes,
				    output_arrays);
CCTK_TimerStopI(timer_handle);

#ifdef DEBUG_ARRAY_INDEXING
printf("### interp_grid.output_arrays[0].ptr_real[0] = %g\n",
       interp_grid.output_arrays[0].ptr_real[0]);
printf("### interp_grid.output_arrays[0].ptr_real[1] = %g\n",
       interp_grid.output_arrays[0].ptr_real[1]);
printf("### interp_grid.output_arrays[0].ptr_real[10] = %g\n",
       interp_grid.output_arrays[0].ptr_real[10]);
#endif

if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"do_InterpLocalUniform(): status=%d return from interpolator!",
		   status);					/*NOTREACHED*/
CCTK_VInfo(CCTK_THORNSTRING,
	   "do_InterpLocalUniform(): status=%d return from interpolator",
	   status);
}

//******************************************************************************

//
// This function sets up the input array subscripting in the parameter
// table.
//
namespace {
void setup_input_array_subscripting(int param_table_handle,
				    const input_grid& input_grid)
{
int status;

status = Util_TableSetIntArray(param_table_handle,
			       input_grid.N_dims, input_grid.array_strides,
			       "input_array_strides");
if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   setup_input_array_subscripting(): error when trying to set input array\n"
"                                     subscripting info in parameter table!\n"
"                                     error status=%d"
		   ,
		   status);					/*NOTREACHED*/
}
	  }
