// Jacobian.cc -- storage for Jacobian matrix
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/Jacobian.cc,v 1.2 2003/01/17 15:23:13 jthorn Exp $
//
// Jacobian::Jacobian
// Jacobian::~Jacobian
// Jacobian::setup_interp_query
// Jacobian::delete_interp_query
// Jacobian::hash
// Jacobian::same_molecule_size
//

#include <assert.h>
#include <math.h>

#include "cctk.h"
#include "util_ErrorCodes.h"
#include "util_Table.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "axis_list.hh"
#include "input_grid.hh"
#include "interp_grid.hh"

//******************************************************************************

//
// This function constructs a Jacobian object, setting
//	molecule_min_m[]
//	molecule_max_m[]
// appropriately (based on a preliminary query of the interpolator),
// and all the internal pointers to point to new[]-allocated arrays of
// the appropriate sizes and types.
//
// Note that we don't want the preliminary interpolator query (to
// check the Jacobian type and to determine the molecule min/max m
// bounds) to do any actual interpolation.  We ensure this by passing
// N_input_arrays = N_output_arrays = 0 to the interpolator for this
// query.  This means (requires) that this object be constructed
// *before* putting input array strides or other optional inputs into
// the interpolator parameter table which would cause error conditions
// with no input/output arrays specified.
//
// Bugs:
// We leak memory if we encounter an error condition. :(
//
Jacobian::Jacobian(const grid& input_grid, const grid& interp_grid,
		   int interp_handle, int param_table_handle)
	: N_dims(input_grid.N_dims)
{
int status, status1, status2, status3, status4;
int axis;

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): interp_grid.N_points=%d",
	   interp_grid.N_points);

//
// preliminary interpolator query to check Jacobian type
// and get molecule min/max m bounds
//
status1 = Util_TableSetInt(param_table_handle,
			   0, "molecule_min_m");
status2 = Util_TableSetInt(param_table_handle,
			   0, "molecule_max_m");
if ((status1 < 0) || (status2 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): error when trying to set molecule min/max m\n"
"                         query flags in parameter table!\n"
"                         error status1=%d status2=%d\n"
		   ,
		   status1, status2);			/*NOTREACHED*/

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): making preliminary interpolator query");
status = CCTK_InterpLocalUniform(N_dims,
				 interp_handle, param_table_handle,
				 input_grid.origin, input_grid.delta,
				 0,		// no interp points
				 CCTK_VARIABLE_REAL, NULL,
				 0,		// no input arrays
				 input_grid.array_dims,
				 NULL, NULL,
				 0,		// no output_arrays,
				 NULL, NULL);
if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): error return %d from InterpLocalUniform()!\n"
"                         when doing molecule min/max m query\n"
		   ,
		   status);				/*NOTREACHED*/

//
// get Jacobian type
//
status1 = Util_TableGetInt(param_table_handle,
			   &MSS_is_fn_of_interp_coords,
			   "MSS_is_fn_of_interp_coords");
status2 = Util_TableGetInt(param_table_handle,
			   &MSS_is_fn_of_which_operation,
			   "MSS_is_fn_of_which_operation");
status3 = Util_TableGetInt(param_table_handle,
			   &MSS_is_fn_of_input_array_values,
			   "MSS_is_fn_of_input_array_values");
status4 = Util_TableGetInt(param_table_handle,
			   &Jacobian_is_fn_of_input_array_values,
			   "Jacobian_is_fn_of_input_array_values");
if ((status1 < 0) || (status2 < 0) || (status3 < 0) || (status4 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): error when trying to get MSS/Jacobian type flags\n"
"                         from parameter table!\n"
"                         error status1=%d status2=%d status3=%d status4=%d\n"
		   ,
		   status1, status2, status3, status4);		/*NOTREACHED*/

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): MSS_is_fn_of_interp_coords=%d",
	   MSS_is_fn_of_interp_coords);
CCTK_VInfo(CCTK_THORNSTRING,
	   "                      MSS_is_fn_of_which_operation=%d",
	   MSS_is_fn_of_which_operation);
CCTK_VInfo(CCTK_THORNSTRING,
	   "                      MSS_is_fn_of_input_array_values=%d",
	   MSS_is_fn_of_input_array_values);
CCTK_VInfo(CCTK_THORNSTRING,
	   "                      Jacobian_is_fn_of_input_array_values=%d",
	   Jacobian_is_fn_of_input_array_values);

//
// check that we grok this Jacobian type
//
if (! (    (MSS_is_fn_of_interp_coords == 0)
	&& (MSS_is_fn_of_which_operation == 0)
	&& (MSS_is_fn_of_input_array_values == 0)
	&& (Jacobian_is_fn_of_input_array_values == 0)    ) )
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): we don't grok this Jacobian type!\n"
"                         MSS_is_fn_of_interp_coords=%d\n"
"                         MSS_is_fn_of_which_operation=%d\n"
"                         MSS_is_fn_of_input_array_values=%d\n"
"                         Jacobian_is_fn_of_input_array_values=%d\n"
		   ,
		   MSS_is_fn_of_interp_coords,
		   MSS_is_fn_of_which_operation,
		   MSS_is_fn_of_input_array_values,
		   Jacobian_is_fn_of_input_array_values);	/*NOTREACHED*/

//
// get the molecule min/max m bounds
// ... we first zero the whole arrays so unused entries will stay 0
//     (this is needed by subscript())
//
	for (axis = 0 ; axis < max_N_dims ; ++axis)
	{
	molecule_min_m[axis] = 0;
	molecule_max_m[axis] = 0;
	}
status1 = Util_TableGetIntArray(param_table_handle,
				N_dims, molecule_min_m,
				"molecule_min_m");
status2 = Util_TableGetIntArray(param_table_handle,
				N_dims, molecule_max_m,
				"molecule_max_m");
if ((status1 != N_dims) || (status2 != N_dims))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): error or wrong-sized array(s) when trying to get\n"
"                         molecule min/max m bounds from parameter table!\n"
"                         error status1=%d status2=%d\n"
		   ,
		   status1, status2);				/*NOTREACHED*/
	for (axis = 0 ; axis < N_dims ; ++axis)
	{
	CCTK_VInfo(CCTK_THORNSTRING,
		   "Jacobian::Jacobian(): axis=%d molecule_[min,max]_m=[%d,%d]",
		   axis, int(molecule_min_m[axis]), int(molecule_max_m[axis]));
	}

//
// delete the molecule min/max m table entries
// so future interpolator calls don't unnecessarily redo this query
//
status1 = Util_TableDeleteKey(param_table_handle, "molecule_min_m");
status2 = Util_TableDeleteKey(param_table_handle, "molecule_max_m");
if ((status1 < 0) || (status2 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::Jacobian(): error when trying to delete molecule min/max m\n"
"                         entries from parameter table!\n"
"                         error status1=%d status2=%d\n"
		   ,
		   status1, status2);				/*NOTREACHED*/

//
// set up the storage for the molecule position arrays
//
CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): setting up molecule position arrays");
	for (axis = 0 ; axis < N_dims ; ++axis)
	{
	molecule_position_pointers[axis] = new CCTK_INT[interp_grid.N_points];
	molecule_position_voidptrs[axis] = static_cast<CCTK_POINTER>(
						molecule_position_pointers[axis]
								    );
	}

//
// now do the actual storage setup for the Jacobian:
// * we only store the Jacobian for the 0th gridfn, which is of type
//   CCTK_REAL, so we don't have to worry about complex values
// * we keep each molecule contiguous in Fortran storage order
// * unused elements of the molecule_{min,max}_m[] arrays were zeroed earlier
// * we zero unused elements of the Jacobian_m_strides[] array
//
	for (axis = 0 ; axis < max_N_dims ; ++axis)
	{
	Jacobian_m_strides[axis] = 0;
	}

int stride = 1;
	for (axis = 0 ; axis < N_dims ; ++axis)
	{
	const int N = jtutil::how_many_in_range(molecule_min_m[axis],
						molecule_max_m[axis]);
	Jacobian_m_strides[axis] = stride;
	CCTK_VInfo(CCTK_THORNSTRING,
		   "Jacobian::Jacobian(): axis=%d N=%d Jacobian_m_strides=%d",
		   axis, N, int(Jacobian_m_strides[axis]));
	stride *= N;
	}
Jacobian_interp_point_stride = stride;
CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): Jacobian_interp_point_stride=%d",
	   int(Jacobian_interp_point_stride));
Jacobian_part_stride = 0;

N_Jacobian_values = Jacobian_interp_point_stride * interp_grid.N_points;
CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): N_Jacobian_values=%d",
	   N_Jacobian_values);

typedef CCTK_REAL *CCTK_REAL_PTR;
Jacobian_pointer = new CCTK_REAL_PTR[N_output_arrays];
Jacobian_voidptr = new CCTK_POINTER [N_output_arrays];
Jacobian_offset = new CCTK_INT[N_output_arrays];
	for (int out = 0 ; out < N_output_arrays ; ++out)
	{
	Jacobian_pointer[out] = NULL;
	Jacobian_voidptr[out] = NULL;
	Jacobian_offset [out] = 0;
	}
Jacobian_pointer[0] = new CCTK_REAL[N_Jacobian_values];
Jacobian_voidptr[0] = static_cast<CCTK_POINTER>(Jacobian_pointer[0]);

//
// compute Jacobian_offset[0] such that
//	subscript(out=0, pt=0, molecule_min_m[]) = 0
//
Jacobian_offset[0] = 0;		// temporary value, needed so we can call
				// subscript() to compute permanent value
Jacobian_offset[0]
	= - subscript(0, 0,
		      molecule_min_m[0], molecule_min_m[1], molecule_min_m[2]);
CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::Jacobian(): Jacobian_offset[0]=%d",
	   Jacobian_offset[0]);
}

//******************************************************************************

//
// This function destroys a Jacobian object.
//
Jacobian::~Jacobian()
{
// this is safe even if we're not using the offset, since delete[]
// of a NULL pointer is (guaranteed by the C++ standard to be) a no-op
delete[] Jacobian_offset;

delete[] Jacobian_pointer[0];
delete[] Jacobian_voidptr;
delete[] Jacobian_pointer;

	for (int axis = N_dims-1 ; axis >= 0 ; --axis)
	{
	delete[] molecule_position_pointers[axis];
	}
}

//******************************************************************************

//
// This function sets up the Jacobian query entries in an interpolator
// parameter table.
//
void Jacobian::setup_interp_query(const grid& interp_grid,
				  int param_table_handle)
	const
{
int status, status1, status2, status3;

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::setup_interp_query(): setting molecule positions");
status = Util_TableSetPointerArray(param_table_handle,
				   N_dims, molecule_position_voidptrs,
				   "molecule_positions");
if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::setup_interp_query(): error %d when trying to set\n"
"                                   molecule_positions in parameter table!"
		   ,
		   status);					/*NOTREACHED*/

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::setup_interp_query(): setting Jacobian pointer");
status = Util_TableSetPointerArray(param_table_handle,
				   N_output_arrays, Jacobian_voidptr,
				   "Jacobian_pointer");
if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::setup_interp_query(): error %d when trying to set\n"
"                                   Jacobian_pointer in parameter table!"
		   ,
		   status);					/*NOTREACHED*/

if (Jacobian_offset != NULL)
   then {
	CCTK_VInfo(CCTK_THORNSTRING,
		   "Jacobian::setup_interp_query(): setting Jacobian offset");
	status = Util_TableSetIntArray(param_table_handle,
				       N_output_arrays, Jacobian_offset,
				       "Jacobian_offset");
	if (status < 0)
	   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::setup_interp_query(): error %d when trying to set\n"
"                                   Jacobian_offset in parameter table!"
			   ,
			   status);				/*NOTREACHED*/
	}

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::setup_interp_query(): setting Jacobian strides");
status1 = Util_TableSetInt(param_table_handle,
			   Jacobian_interp_point_stride,
			   "Jacobian_interp_point_stride");
status2 = Util_TableSetIntArray(param_table_handle,
				N_dims, Jacobian_m_strides,
				"Jacobian_m_strides");
status3 = Util_TableSetInt(param_table_handle,
			   Jacobian_part_stride,
			   "Jacobian_part_stride");
if ((status1 < 0) || (status2 < 0) || (status3 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::setup_interp_query(): error when trying to set\n"
"                                   Jacobian stride info in parameter table!"
"                                   error status1=%d status2=%d status3=%d\n"
		   ,
		   status1, status2, status3);			/*NOTREACHED*/
}

//******************************************************************************

//
// This function removes the Jacobian query entries from an interpolator
// parameter table.
//
void Jacobian::delete_interp_query(int param_table_handle)
	const
{
int status, status1, status2, status3;

CCTK_VInfo(CCTK_THORNSTRING,
	   "Jacobian::delete_interp_query(): deleting interpolator query");

status1 = Util_TableDeleteKey(param_table_handle,
			      "molecule_positions");
status2 = Util_TableDeleteKey(param_table_handle,
			      "Jacobian_pointer");
if ((status1 < 0) || (status2 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::delete_interp_query(): error when trying to delete\n"
"                                    molecule_positions and/or\n"
"                                    Jacobian_pointer from parameter table!\n"
"                                    error status1=%d status2=%d"
		   ,
		   status1, status2);				/*NOTREACHED*/

if (Jacobian_offset != NULL)
   then {
	status = Util_TableDeleteKey(param_table_handle,
				     "Jacobian_offset");
	if (status < 0)
	   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::delete_interp_query(): error %d when trying to delete\n"
"                                    Jacobian_offset from parameter table!"
			   ,
			   status);				/*NOTREACHED*/
	}

status1 = Util_TableDeleteKey(param_table_handle,
			      "Jacobian_interp_point_stride");
status2 = Util_TableDeleteKey(param_table_handle,
			      "Jacobian_m_strides");
status3 = Util_TableDeleteKey(param_table_handle,
			      "Jacobian_part_stride");
if ((status1 < 0) || (status2 < 0) || (status3 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   Jacobian::delete_interp_query(): error when trying to delete\n"
"                                   Jacobian stride info from parameter table!"
"                                   error status1=%d status2=%d status3=%d\n"
		   ,
		   status1, status2, status3);			/*NOTREACHED*/
}

//******************************************************************************

//
// This function computes a "linear hash" of all the Jacobian values,
//	hash = sum_k w[k]*Jac[k]/N_Jacobian_values
// where the w[k] are pseudorandom weights in the range
// [hash_weight_min, hash_weight_max).
//
fp Jacobian::hash(const grid& interp_grid,
		  fp hash_weight_min, fp hash_weight_max)
	const
{
const int out = 0;
fp sum = 0.0;

	for (int pt = 0 ; pt < interp_grid.N_points ; ++pt)
	{
	for (int mk = molecule_min_m[2] ; mk <= molecule_max_m[2] ; ++mk)
	{
	for (int mj = molecule_min_m[1] ; mj <= molecule_max_m[1] ; ++mj)
	{
	for (int mi = molecule_min_m[0] ; mi <= molecule_max_m[0] ; ++mi)
	{
	const int posn = subscript(out, pt, mi,mj,mk);
	assert(posn >= 0);
	assert(posn < N_Jacobian_values);

	sum += jtutil::drandom(hash_weight_min, hash_weight_max)
	       * Jacobian_pointer[out][posn];
	}
	}
	}
	}

return sum / fp(N_Jacobian_values);
}

//******************************************************************************

//
// This function determines whether or not two Jacobian objects have the
// same molecule size.
//
//static
  bool Jacobian::same_molecule_size(const Jacobian& A, const Jacobian& B)
{
return    (A.molecule_min_m[0] == B.molecule_min_m[0])
       && (A.molecule_min_m[1] == B.molecule_min_m[1])
       && (A.molecule_min_m[2] == B.molecule_min_m[2])
       && (A.molecule_max_m[0] == B.molecule_max_m[0])
       && (A.molecule_max_m[1] == B.molecule_max_m[1])
       && (A.molecule_max_m[2] == B.molecule_max_m[2]);
}
