// TestInterpBase.hh -- top-level header file for this thorn
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/TestInterpBase.hh,v 1.2 2003/04/15 17:02:48 jthorn Exp $

// what type of test function are we using?
enum	fn_type
	{
	fn_invalid = 0,	// invalid value to catch uninitialized-variable bugs
	fn_xy,		// f(x,y,z) = x*y
	fn_generic,	// generic trancendental fn
			// (should have all Taylor series terms nonzero)
        fn_series,	// all possible Taylor-series terms
			// up to and including specified order
			// (takes several additional parameters)
	fn_random	// quasirandom noise in [0,1)
			// (requires input and interpolation grids to be
			//  identical, and doesn't allow taking derivatives)
	};

// max order for fn_series
const int max_series_order = 6;

// codes to select datatypes
// ... we conventionally use an integer itype to iterate over these
enum	select_datatype
	{
	select_datatype_real = 0,	// CCTK_REAL
	select_datatype_real4,		// CCTK_REAL4 
	select_datatype_real8,		// CCTK_REAL8
	select_datatype_complex,	// CCTK_COMPLEX
	select_datatype_complex8,	// CCTK_COMPLEX8
	select_datatype_complex16,	// CCTK_COMPLEX16
	};

// codes to select various subfunctions of fn_series
enum	select_fn
	{
	select_fn_none = -1,
	select_fn_real,		// fn_real
	select_fn_real4,	// fn_real4
	select_fn_real8,	// fn_real8
	select_fn_complex_Re,	// fn_complex.Re
	select_fn_complex_Im,	// fn_complex.Im
	select_fn_complex8_Re,	// fn_complex8.Re
	select_fn_complex8_Im,	// fn_complex8.Im
	select_fn_complex16_Re,	// fn_complex16.Re
	select_fn_complex16_Im,	// fn_complex16.Im
	N_select_fn		// *** this must be last in enum ***
	};
const int N_real_fn = 3;	// fn_real, fn_complex.{Re,Im}
const int N_real4_fn = 3;	// fn_real4, fn_complex8.{Re,Im}
const int N_real8_fn = 3;	// fn_real8, fn_complex16.{Re,Im}
const int N_input_arrays = 6;
const int N_output_arrays = 6;

// max number of interpolation axes in a test
// ... we conventionally use an integer iaxis to iterate over these
const int max_N_interp_axes = 6;

// max number of derivative operations in a test
// ... we conventionally use an integer iop to iterate over these
const int max_N_operations = 10;

// max number of outputs from interpolator, including derivatives
const int max_N_output = N_output_arrays * max_N_operations;

// what type of interpolator will we test?
enum	interp_type
	{
	interp_invalid = 0,
	interp_InterpLocal,		// CCTK_InterpLocal()
	interp_InterpLocalUniform,	// CCTK_InterpLocalUniform()
	interp_InterpGV,		// CCTK_InterpGV()
	interp_InterpGridArrays		// CCTK_InterpGridArrays()
	};

const int max_N_derivs = 2;	// for compile-time array sizes etc
const int max_N_dims = 3;	// for compile-time array sizes etc

// make derivative codes look a bit nicer
#define DERIV(x)   x

#ifdef JACOBIAN_STUFF
// what type of Jacobian test should we do
enum	Jacobian_test
	{
	Jacobian_test_none,	// no Jacobian test
	Jacobian_test_query,	// query the interpolator for the Jacobian
	Jacobian_test_query_and_NP
				// query the interpolator for the Jacobian
				// and estimate the Jacobian by numerical
				// perturbation (and compare the query and NP
				// Jacobians)
	};
#endif

//******************************************************************************

//
// prototypes for stuff visible from multiple files
//

// forward declarations
class axis_list;
class input_grid;
class interp_grid;
#ifdef JACOBIAN_STUFF
class Jacobian;
#endif

// test_fn.cc
fp test_fn(enum fn_type test_fn_type,
	   enum select_fn test_fn_select, int test_fn_order, int operation_code,
	   fp x, fp y, fp z);

// misc.cc
bool is_prefix(const char *source, const char *prefix);
int split_cstring_into_words(const char cstring[],
			     int max_N_words, const char* words[]);
int find_cstring_in_cstring_array(const char cstring[],
				  int N, const char* cstring_array[]);
int find_int_in_int_array(int x,
			  int N, const int int_array[]);
fp delta_product(int derivative_code, const fp delta[]);
int decode_derivative_code(int derivative_code,
			   int derivative_axes[]);
int encode_derivative_code(int N_derivs, const int derivative_axes[]);
int get_interp_handle_from_name(const char interp_operator[]);
int get_interp_handle_from_order(int order);
int get_coord_system_handle(const char coord_system[]);
int setup_interp_param_table(const char interp_pars[], int interp_order);
void setup_derivs_in_param_table(int param_table_handle,
				 int N_operations,
				 const int opcodes[max_N_operations]);
void setup_output_arrays(const interp_grid& interp_grid,
			 int   output_array_type_codes[max_N_output],
			 void* output_arrays          [max_N_output]);

// InterpLocal.cc
void do_InterpLocal(const cGH *GH, int timer_handle,
		    const char coord_system[],
		    const input_grid& input_grid, interp_grid& interp_grid,
		    const char interp_operator[], int interp_order,
						  const char interp_pars[],
		    int N_operations, const int opcodes[max_N_operations]);

// InterpLocalUniform.cc
void do_InterpLocalUniform(const cGH *GH, int timer_handle,
			   const char coord_system[],
			   const input_grid& input_grid, interp_grid& interp_grid,
			   const char interp_operator[], int interp_order,
							 const char interp_pars[],
			   int N_operations, const int opcodes[max_N_operations]);

// InterpGV.cc
void do_InterpGV(const cGH *GH, int timer_handle,
		 const char coord_system[],
		 const input_grid& input_grid, interp_grid& interp_grid,
		 const char interp_operator[], int interp_order,
					       const char interp_pars[],
		 int N_operations, const int opcodes[max_N_operations]);

// InterpGridArrays.cc
void do_InterpGridArrays(const cGH *GH, int timer_handle,
			 const char coord_system[],
			 const input_grid& input_grid, interp_grid& interp_grid,
			 const char interp_operator[], int interp_order,
						       const char interp_pars[],
			 int N_operations, const int opcodes[max_N_operations]);
