// axis_list.hh -- class to represent list of axes
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/axis_list.hh,v 1.2 2003/01/17 15:23:13 jthorn Exp $

//
// An  axis_list  object represents a list of axes and its extension
// to (a unique list of axes in) max_N_dims dimensions.  (For example,
// given the axes list "ZX", the extension to 3-D would be "ZXY".)
// Thanks to Dennis Pollney for suggesting the terminology "extension" here.
//
// Note there are no non-const member functions, so an  axis_list  object
// is effectively always  const .
//
class	axis_list
	{
public:
	// number of dimensions of this list
	int N_dims() const { return N_dims_; }

	// string form of axis list for printing/debugging/etc
	const char *string() const { return string_; }

	// is our extended axis list just the contiguous list "XYZ"?
	bool is_prefix_of_xyz() const;

	// verify that our extended axis list
	// is indeed the contiguous list "XYZ"
	// ... no-op if so; CCTK_VWarn(-1, ...) if not
	void verify_contiguous() const;

	// is this an unextended or an extended axis
	bool axis_is_unextended(int axis) const { return axis <  N_dims_; }
	bool axis_is_extended  (int axis) const { return axis >= N_dims_; }

	// get individual axes of the extended axis list
	// argument is 0-origin integer to select axis
	// return value is a 0-origin integer
	int extended_axis(int axis) const
		{
		assert(axis >= 0);
		assert(axis < max_N_dims);
		return axis_list_[axis];
		}

	// return position in extended axis list containing a given axis,
	// i.e. return posn such that operator()(posn) == axis
	int where_is_axis(int axis) const
		{
		assert(axis >= 0);
		assert(axis < max_N_dims);
		return where_is_axis_[axis];
		}

	//
	// the following functions assume the samantics
	// our axes = xyz, result of extended_axis() = XYZ
	// i.e. construct with "ZX" means x=Z, y=X
	//

	// map xyz derivative opcode --> XYZ derivatives opcode
	int XYZ_of_xyz_deriv_opcode(int xyz_deriv_opcode) const;

public:
	// construct from string specifier: assuming max_N_dims == 3,
	// valid strings are
	//	"XYZ" "XZY" "YXZ" "YZX" "ZXY" "ZYX"
	//	"XY"  "XZ"  "YX"  "YZ"  "ZX"  "ZY"
	//	"X"   "Y"   "Z"
	// ... n.b. these MUST be in upper case!!
	// ... n.b. the pointed-to string must be valid for the entire
	//     life of the object (we're lazy and just save the pointer)!
	axis_list(const char string_in[]);

	// no destructor needed -- compiler-generated default is ok

private:
	// helper function to do the main work of the constructor
	void setup(int axis0, int axis1, int axis2);

private:
	// we forbid copying and passing by value
	// by declaring the copy constructor and assignment operator
	// private, but never defining them
	axis_list(const axis_list& rhs);
	axis_list& operator=(const axis_list& rhs);

private:
	int N_dims_;
	const char *string_;
	const char *extended_string_;
	// the extended axes list and its inverse (considered as a permutation)
	int axis_list_    [max_N_dims];
	int where_is_axis_[max_N_dims];
	};
