// driver.cc -- test interpolator in 3D
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/driver.cc,v 1.13 2003/04/15 17:02:48 jthorn Exp $
//
// <<<data structures local to this file>>>
// <<<prototypes>>>
// TestInterpBase - top-level driver
//
/// decode_test_function_type
/// decode_interpolator_type
/// decode_select_function
/// decode_Jacobian_test_type
///
/// set_dummy_values_in_Cactus_grid
/// setup_test_functions
/// do_interpolation
///
/// do_write_plot_file
///
#ifdef JACOBIAN_STUFF
/// check_Jacobians_same_size
/// write_Jacobians
#endif
///
/// hash_interp_results
/// gather_hashes
/// verify_all_hashes_identical
/// do_write_hash_file
///
/// accumulate_interp_error_norms
#ifdef JACOBIAN_STUFF
/// compute_Jacobian_error_norms
#endif
/// classify_and_print_error_norms
/// classify_and_print_error_norm_pair
///

#include <assert.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <float.h>

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"
#include "util_ErrorCodes.h"
#include "util_Table.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "axis_list.hh"
#include "permutation.hh"
#include "input_grid.hh"
#include "interp_grid.hh"
#ifdef JACOBIAN_STUFF
  #include "Jacobian.hh"
#endif

#undef DEBUG_ARRAY_INDEXING
#undef DEBUG_HASH_DIFFERENT_PRINTING

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// ***** data structures local to this file *****
//

// hashes of interpolation and Jacobian results
struct	hashes
	{
	// individual hashes of each function
	fp hash_of_real,
	   hash_of_real4,
	   hash_of_real8,
	   hash_of_complex_Re, hash_of_complex_Im,
	   hash_of_complex8_Re, hash_of_complex8_Im,
	   hash_of_complex16_Re, hash_of_complex16_Im;

	// default constructor
	hashes()
		{
		hash_of_real = 0.0;
		hash_of_real4 = 0.0;
		hash_of_real8 = 0.0;
		hash_of_complex_Re = 0.0;	hash_of_complex_Im = 0.0;
		hash_of_complex8_Re = 0.0;	hash_of_complex8_Im = 0.0;
		hash_of_complex16_Re = 0.0;	hash_of_complex16_Im = 0.0;
		}

	// forbid copying and passing by value
	// by declaring the copy constructor and assignment operator
	// private, but never defining them
private:
	hashes(const hashes& rhs);
	hashes& operator=(const hashes& rhs);
	};

typedef struct hashes hashes_array_t[max_N_interp_axes][max_N_operations];

// norms of interpolation and Jacobian errors
struct	error_norms
	{
	// individual norms of each function
	jtutil::norm<fp> norm_of_real,
			 norm_of_real4,
			 norm_of_real8,
			 norm_of_complex_Re, norm_of_complex_Im,
			 norm_of_complex8_Re, norm_of_complex8_Im,
			 norm_of_complex16_Re, norm_of_complex16_Im;

	// default constructor
	// ... jtutil::norm<> will initialize itself,
	//     so we don't no need to do so explicitly here
	error_norms() { }

	// forbid copying and passing by value
	// by declaring the copy constructor and assignment operator
	// private, but never defining them
private:
	error_norms(const error_norms& rhs);
	error_norms& operator=(const error_norms& rhs);
	};

struct	scaled_error_norm_thresholds
	{
	fp rms_low,  infinity_low;
	fp rms_high, infinity_high;
	};

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// ***** prototypes *****
//

// external entry points (called by Cactus flesh)
extern "C"
  void TestInterpBase(CCTK_ARGUMENTS);

namespace {		// the remaining functions are all local to this file

// decode string argument into internal enums
enum fn_type decode_test_function_type(const char test_fn_type[]);
void decode_interpolator_type
	(const char interpolator_type[],
	 enum interp_type &interp_type, bool & interp_is_local);
enum select_fn decode_select_function(const char select_function[]);
void decode_Jacobian_test_type(const char Jacobian_test_type[],
			       bool& do_query_Jacobian, bool& do_NP_Jacobian);

// setup for interpolation
void set_dummy_values_in_Cactus_grid(CCTK_ARGUMENTS,
				     const input_grid& input_grid);
void setup_test_functions(enum fn_type test_fn_type, int test_fn_order,
			  input_grid& input_grid);
void do_interpolation(enum interp_type interp_type,
		      const cGH* GH, int timer_handle,
		      const char coord_system[],
		      const input_grid& input_grid, interp_grid& interp_grid,
		      const char interp_operator[], int interp_order,
						    const char interp_pars[],
		      int N_operations, const int opcodes[max_N_operations]);

// write plot files
void do_write_plot_file
	(enum fn_type test_fn_type, int test_fn_order,
	 const interp_grid& interp_grid,
	 enum select_fn which_fn_to_plot, int which_iop_to_plot,
					  int which_deriv_to_plot,
	 bool multiprocessor_file_names,
	 int my_proc, const char plot_file_base_name[]);

#ifdef JACOBIAN_STUFF
// write Jacobian files
void check_Jacobians_same_size(const Jacobian& query_J, const Jacobian& NP_J);
void write_Jacobians(const grid& interp_grid,
		     const Jacobian* query_J, const Jacobian* NP_J,
		     const char Jacobian_file_name[]);
#endif

// hash files
void hash_interp_results(const interp_grid& interp_grid,
			 struct hashes hashes_array[max_N_operations],
			 fp hash_weight_min, fp hash_weight_max);
void gather_hashes(const cGH *GH,
		   const hashes_array_t my_hashes_array,
		   int N_procs, int my_proc, hashes_array_t all_hashes_array[]);
void verify_all_hashes_identical(int N_procs, hashes_array_t all_hashes_array[],
				 bool abort_if_hashes_differ);
void do_write_hash_files
  (int N_interp_axes, const char* const interp_axes_strarr[max_N_interp_axes],
   int N_operations, const int opcodes[max_N_operations],
   const hashes_array_t hashes_array,
   const char hash_files_base_name[]);

// error norms
void accumulate_interp_error_norms
	(enum fn_type test_fn_type, int test_fn_order,
	 const int opcodes[max_N_operations],
	 const fp input_grid_delta[max_N_dims],
	 const interp_grid& interp_grid,
	 struct error_norms& raw_error_norms,
	 struct error_norms& scaled_error_norms);
#ifdef JACOBIAN_STUFF
void compute_Jacobian_error_norms(const grid& interp_grid,
				  const Jacobian& query_J, const Jacobian& NP_J,
				  struct error_norms& error_norms);
#endif
void classify_and_print_error_norms
	(const struct scaled_error_norm_thresholds& thresholds,
	 const struct error_norms& raw_error_norms,
	 const struct error_norms& scaled_error_norms,
	 int& N_ok, int& N_marginal, int& N_bad);
void classify_and_print_error_norm_pair
	(const char type_string[],
	 const struct scaled_error_norm_thresholds& thresholds,
	 const jtutil::norm<fp>& raw_error_norm,
	 const jtutil::norm<fp>& scaled_error_norm,
	 int& N_ok, int& N_marginal, int& N_bad);

	  };	// unnamed namespace

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function is a driver to do the interpolation test.
//
// N.b. This function is called by the Cactus flesh, and must thus have
// a C-style interface.
//
// If we're testing a *local* interpolator...
//    We want each processor to have identical input grids and identical
//    interpolation grids.  Cactus doesn't have a notion of processor-local
//    coordinates, so we create these with a(n artificial) 0 origin and
//    the usual Cactus grid spacing and local grid size.  This works, but
//    it does require that the grid be set up such that each processor has
//    the same local size, eg by using driver::local_n[xyz] to set the grid
//    size.
//
// If we're testing a *global* interpolator...
//    The input grid should be this processor's chunk of the Cactus grid,
//    while the interp grid should span the whole global Cactus grid.
//    By having the interp grid identical on each processor, we ensure
//    that we (should) get identical interpolation results on each processor.
//
extern "C"
  void TestInterpBase(CCTK_ARGUMENTS)
{
DECLARE_CCTK_ARGUMENTS
DECLARE_CCTK_PARAMETERS

const int N_procs = CCTK_nProcs(cctkGH);
const int my_proc = CCTK_MyProc(cctkGH);
CCTK_VInfo(CCTK_THORNSTRING,
	   "this is processor #%d of %d",
	   my_proc, N_procs);

//
// decode some parameters into internal variables
//

// which interpolation axes to test
const char* interp_axes_strarr[max_N_interp_axes];
const int N_interp_axes
	= split_cstring_into_words(interpolation_axes,
				   max_N_interp_axes, interp_axes_strarr);

// what type of test function
const enum fn_type test_fn_type
	= decode_test_function_type(test_function_type);

// test_function_type == "random" is only valid when the
// input and interpolation grids are identical
if ( (test_fn_type == fn_random)
     && !((N_extra_interp_points == 0) && (data_to_interp_dxyz_ratio == 1)) )
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   TestInterpBase(): test_function_type == \"random\" requires\n"
"                     input and interpolation grids to be identical, i.e.\n"
"                     you must explicitly set the parameters\n"
"                     TestInterpBase::N_extra_interp_points = 0\n"
"                     TestInterpBase::data_to_interp_dxyz_ratio = 1\n");
								/*NOTREACHED*/

// what type of interpolator
enum interp_type interp_type;
bool interp_is_local;
decode_interpolator_type(interpolator_type,
			 interp_type, interp_is_local);

// which derivative operations to test
const char* operation_codes_strarr[max_N_operations];
const int N_operations
	= split_cstring_into_words(operation_codes,
				   max_N_operations, operation_codes_strarr);
int opcodes[max_N_operations];
	for (int iop = 0 ; iop < N_operations ; ++iop)
	{
	if (sscanf(operation_codes_strarr[iop], "%d", &opcodes[iop]) != 1)
	   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"TestInterpBase(): bad operation_codes specifier \"%s\"!",
		   operation_codes_strarr[iop]);		/*NOTREACHED*/
	}

// stuff for plot files
const enum select_fn which_fn_to_plot
	= decode_select_function(which_function_to_plot);
const int which_iaxis_to_plot
	= (strlen(which_interpolation_axes_to_plot) == 0)
	  ? 0
	  : find_cstring_in_cstring_array(which_interpolation_axes_to_plot,
					  N_interp_axes, interp_axes_strarr);
const int which_iop_to_plot
	= find_int_in_int_array(which_operation_to_plot,
				N_operations, opcodes);


//
// get global/local Cactus grid origin
//
// KLUDGE -- is this the right way to get this??
CCTK_REAL Cactus_global_origin[max_N_dims];
CCTK_REAL Cactus_local_origin[max_N_dims];
	  {
	for (int axis = 0 ; axis < max_N_dims ; ++axis)
	{
	Cactus_global_origin[axis] = cctk_origin_space[axis];
	Cactus_local_origin[axis] = cctk_origin_space[axis]
				    + cctk_lbnd[axis] * cctk_delta_space[axis];
	CCTK_VInfo(CCTK_THORNSTRING,
		   "Cactus _global_origin[%d]=%g _local_origin[%d]=%g",
		   axis, double(Cactus_global_origin[axis]),
		   axis, double(Cactus_local_origin[axis]));
	}
	  }


//
// fake local coordinate origin
// (we use 0.0 here, but any arbitrary values would do so long as
//  they were the same on each processor)
//
CCTK_REAL fake_local_origin[max_N_dims];
	  {
	for (int axis = 0 ; axis < max_N_dims ; ++axis)
	{
	fake_local_origin[axis] = 0.0;
	}
	  }

//
// do the interpolation test for each axis specifier
//
// ... to avoid having to split the file-writing operation up into
//     pieces, we keep a 2-D array of hashes (indexed by iaxis and iop)
//     and write them all out at the end
// ... we keep separate raw and scaled error norms, the latter scaled by
//     the epsilon/Delta_x factors during accumulation
//
hashes_array_t hashes_array;
const struct scaled_error_norm_thresholds thresholds
	= {
	  scaled_rms_error_norm_low_threshold,
	  scaled_infinity_error_norm_low_threshold,
	  scaled_rms_error_norm_high_threshold,
	  scaled_infinity_error_norm_high_threshold,
	  };
const int timer_handle = CCTK_TimerCreateI();
int N_ok = 0, N_marginal = 0, N_bad = 0;
	for (int iaxis = 0 ; iaxis < N_interp_axes ; ++iaxis)
	{
	CCTK_VInfo(CCTK_THORNSTRING,
		   "========================================");
	// setup input grid
	const CCTK_REAL* input_grid_origin
		= interp_is_local ? fake_local_origin
				  : Cactus_local_origin;
	class input_grid input_grid(cctkGH,
				    cctk_lsh,
				    input_grid_origin, cctk_delta_space,
				    interp_axes_strarr[iaxis]);
	set_dummy_values_in_Cactus_grid(CCTK_PASS_CTOC, input_grid);

	// in case we have  test_function_type == "random",
	// we save the random-number-generator seed now,
	// and restore it each time we recompute the test function
	const int test_function_random_seed = jtutil::get_random_seed();

	setup_test_functions(test_fn_type, test_function_order,
			     input_grid);

	// setup interpolation grid
	const CCTK_INT* interp_grid_size = interp_is_local ? cctk_lsh
							   : cctk_gsh;
	const CCTK_REAL* interp_grid_origin
		= interp_is_local ? fake_local_origin
				  : Cactus_global_origin;
	class interp_grid interp_grid(interp_grid_size, cctk_nghostzones,
				      interp_grid_origin, cctk_delta_space,
				      input_grid,
				      N_extra_interp_points,
				      data_to_interp_dxyz_ratio,
				      N_operations);
	if (randomly_permute_interp_points)
	   then interp_grid.randomly_permute_interp_points();

	// this does the actual interpolation
	do_interpolation(interp_type,
			 cctkGH, timer_handle, coordinate_system,
			 input_grid, interp_grid,
			 interpolation_operator, interpolation_order,
						 interpolator_pars,
			 N_operations, opcodes);

	// write a plot file
	if ( (write_plot_file != 0) && (iaxis == which_iaxis_to_plot)
	     && ((my_proc == 0) || (all_processors_write_plot_files != 0)) )
	   then {
		jtutil::set_random_seed(test_function_random_seed);
		do_write_plot_file(test_fn_type, test_function_order,
				   interp_grid,
				   which_fn_to_plot, which_iop_to_plot,
						     opcodes[which_iop_to_plot],
				   (all_processors_write_plot_files != 0),
				   my_proc, plot_file_base_name);
		}

	// compute hashes
	hash_interp_results(interp_grid,
			    hashes_array[iaxis],
			    hash_weight_min, hash_weight_max);

	// accumulate and print error norms
	jtutil::set_random_seed(test_function_random_seed);
	struct error_norms raw_error_norms, scaled_error_norms;
	accumulate_interp_error_norms(test_fn_type, test_function_order,
				      opcodes, input_grid.xyz_delta,
				      interp_grid,
				      raw_error_norms, scaled_error_norms);
	classify_and_print_error_norms(thresholds,
				       raw_error_norms, scaled_error_norms,
				       N_ok, N_marginal, N_bad);
	}

//
// if this is a multiprocessor run,
//    gather all the hashes to processor #0
//           (each processor has to participate in this)
//    and verify that they're identical
//        (only processor #0 has all the hashes, so only it does this)
//
if (N_procs > 1)
   then {
	hashes_array_t* const all_hashes_array = new hashes_array_t[N_procs];
	gather_hashes(cctkGH,
		      hashes_array,
		      N_procs, my_proc, all_hashes_array);
	if (my_proc == 0)
	   then verify_all_hashes_identical(N_procs, all_hashes_array,
					    (abort_if_hashes_differ != 0));
	}

//
// write the hash files (only on processor #0)
//
if ((write_hash_files != 0) && (my_proc == 0))
   then do_write_hash_files(N_interp_axes, interp_axes_strarr,
			    N_operations, opcodes,
			    hashes_array,
			    hash_files_base_name);

//
// print final statistics
//
CCTK_VInfo(CCTK_THORNSTRING,
	   "total times for %d interpolation%s:",
	   N_interp_axes, ((N_interp_axes == 1) ? "" : "s"));
CCTK_TimerPrintDataI(timer_handle, -1);
const int N_total = N_ok + N_marginal + N_bad;
if	(N_bad == 0)
   then CCTK_VInfo(CCTK_THORNSTRING,
		   "total of %d/%d norms ok, %d/%d norms marginal",
		   N_ok, N_total, N_marginal, N_total);
else if (abort_if_error_norm_gt_high_threshold != 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "total of %d/%d norms ok, %d/%d norms marginal, %d/%d norms BAD!",
		   N_ok, N_total, N_marginal, N_total, N_bad, N_total);
   else CCTK_VInfo(CCTK_THORNSTRING,
		   "*** total of %d/%d norms ok, %d/%d norms marginal, %d/%d norms BAD!",
		   N_ok, N_total, N_marginal, N_total, N_bad, N_total);
}

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function decodes the character-string specification of which
// test function we want to use, into an enum for internal use.
//
namespace {
enum fn_type decode_test_function_type(const char test_function_type[])
{
if      (CCTK_Equals(test_function_type, "xy"))
   then return fn_xy;
else if (CCTK_Equals(test_function_type, "generic"))
   then return fn_generic;
else if (CCTK_Equals(test_function_type, "series"))
   then return fn_series;
else if (CCTK_Equals(test_function_type, "random"))
   then return fn_random;
else	{
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "decode_test_fn_type(): bad test_function_type=\"%s\"!",
		   test_function_type);				/*NOTREACHED*/
	return fn_invalid;					// dummy return
	}
}
	  }

//******************************************************************************

//
// This function decodes the character-string specification of which
// interpolator we want to use, into an enum for internal use.
//
namespace {
void decode_interpolator_type
	(const char interpolator_type[],
	 enum interp_type &interp_type, bool & interp_is_local)
{
if	(CCTK_Equals(interpolator_type, "InterpLocal"))
   then {
	interp_type = interp_InterpLocal;
	interp_is_local = true;
	}
else if (CCTK_Equals(interpolator_type, "InterpLocalUniform"))
   then {
	interp_type = interp_InterpLocalUniform;
	interp_is_local = true;
	}
else if (CCTK_Equals(interpolator_type, "InterpGV"))
   then {
	interp_type = interp_InterpGV;
	interp_is_local = false;
	}
else if (CCTK_Equals(interpolator_type, "InterpGridArrays"))
   then {
	interp_type = interp_InterpGridArrays;
	interp_is_local = false;
	}
else	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "decode_interpolator_type(): bad interpolator_type=\"%s\"!",
		   interpolator_type);				/*NOTREACHED*/
}
	  }

//******************************************************************************

//
// This function decodes the character-string specification of a
// CCTK_* datatype, into an enum for internal use.
//
namespace {
enum select_fn decode_select_function(const char select_function[])
{
if	(CCTK_Equals(select_function, "real"))
   then return select_fn_real;
else if (CCTK_Equals(select_function, "real4"))
   then return select_fn_real4;
else if (CCTK_Equals(select_function, "real8"))
   then return select_fn_real8;
else if (CCTK_Equals(select_function, "complex.Re"))
   then return select_fn_complex_Re;
else if (CCTK_Equals(select_function, "complex.Im"))
   then return select_fn_complex_Im;
else if (CCTK_Equals(select_function, "complex8.Re"))
   then return select_fn_complex8_Re;
else if (CCTK_Equals(select_function, "complex8.Im"))
   then return select_fn_complex8_Im;
else if (CCTK_Equals(select_function, "complex16.Re"))
   then return select_fn_complex16_Re;
else if (CCTK_Equals(select_function, "complex16.Im"))
   then return select_fn_complex16_Im;
else	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "decode_select_function(): bad select_function=\"%s\"!",
		   select_function);				/*NOTREACHED*/
}
	  }

//******************************************************************************

//
// This function decodes the character-string specification of what
// Jacobian tests we want to do, into a pair of Boolean flags for
// internal use.
//
namespace {
void decode_Jacobian_test_type(const char Jacobian_test_type[],
			       bool& do_query_Jacobian, bool& do_NP_Jacobian)
{
if	(CCTK_Equals(Jacobian_test_type, "none"))
   then {
	do_query_Jacobian = false;
	do_NP_Jacobian = false;
	}
else if (CCTK_Equals(Jacobian_test_type, "query interpolator"))
   then {
	do_query_Jacobian = true;
	do_NP_Jacobian = false;
	}
else if (CCTK_Equals(Jacobian_test_type,
		     "query interpolator + numerical perturbation"))
   then {
	do_query_Jacobian = true;
	do_NP_Jacobian = true;
	}
else	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"decode_Jacobian_test_type(): bad Jacobian_test_type=\"%s\"!",
		   Jacobian_test_type);				/*NOTREACHED*/
}
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function sets all the Cactus grid arrays to dummy values
// in this processor's chunk of the grid.
//
// Note that while we get the Cactus data pointers on this processor
// from the input grid, we are setting the *entire* Cactus grid on this
// processor (not just the input grid) to the dummy values.
//
namespace {
void set_dummy_values_in_Cactus_grid(CCTK_ARGUMENTS,
				     const input_grid& input_grid)
{
DECLARE_CCTK_ARGUMENTS
const int timelevel = 0;
#define DUMMY_VALUE	999999.0

CCTK_VInfo(CCTK_THORNSTRING,
	   "setting dummy values in Cactus grid");

	for (int k = 0 ; k < cctk_lsh[2] ; ++k)
	{
	for (int j = 0 ; j < cctk_lsh[1] ; ++j)
	{
	for (int i = 0 ; i < cctk_lsh[0] ; ++i)
	{
	const int posn = CCTK_GFINDEX3D(cctkGH, i,j,k);
	input_grid.data_ptrs.ptr_real     [posn]    = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_real4    [posn]    = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_real8    [posn]    = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex  [posn].Re = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex  [posn].Im = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex8 [posn].Re = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex8 [posn].Im = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex16[posn].Re = DUMMY_VALUE;
	input_grid.data_ptrs.ptr_complex16[posn].Im = DUMMY_VALUE;
	}
	}
	}
}
	  }

//******************************************************************************

//
// This function sets up the test functions on the input grid.
//
namespace {
void setup_test_functions(enum fn_type test_fn_type, int test_fn_order,
			  struct input_grid& input_grid)
{
const int no_deriv = 0;		// operation code for "no derivative"

	for (int k = 0 ; k < input_grid.array_dims[2] ; ++k)
	{
	for (int j = 0 ; j < input_grid.array_dims[1] ; ++j)
	{
	for (int i = 0 ; i < input_grid.array_dims[0] ; ++i)
	{
	const fp x = input_grid.x_of_i(i);
	const fp y = input_grid.y_of_j(j);
	const fp z = input_grid.z_of_k(k);
	const int posn = input_grid.subscript(i,j,k);

#ifdef DEBUG_ARRAY_INDEXING
	if ((posn == 0) || (posn == 2) || (posn == 1) || (posn == 10))
	   then printf("### posn=%d ==> test function xyz=%g %g %g\n",
		       posn, double(x), double(y), double(z));
#endif

	input_grid.data_ptrs.ptr_real[posn]
		= test_fn(test_fn_type,
			  select_fn_real, test_fn_order, no_deriv,
			  x,y,z);

	input_grid.data_ptrs.ptr_real4[posn]
		= test_fn(test_fn_type,
			  select_fn_real4, test_fn_order, no_deriv,
			  x,y,z);

	input_grid.data_ptrs.ptr_real8[posn]
		= test_fn(test_fn_type,
			  select_fn_real8, test_fn_order, no_deriv,
			  x,y,z);

	input_grid.data_ptrs.ptr_complex[posn].Re
		= test_fn(test_fn_type,
			  select_fn_complex_Re, test_fn_order, no_deriv,
			  x,y,z);
	input_grid.data_ptrs.ptr_complex[posn].Im
		= test_fn(test_fn_type,
			  select_fn_complex_Im, test_fn_order, no_deriv,
			  x,y,z);

	input_grid.data_ptrs.ptr_complex8[posn].Re
		= test_fn(test_fn_type,
			  select_fn_complex8_Re, test_fn_order, no_deriv,
			  x,y,z);
	input_grid.data_ptrs.ptr_complex8[posn].Im
		= test_fn(test_fn_type,
			  select_fn_complex8_Im, test_fn_order, no_deriv,
			  x,y,z);

	input_grid.data_ptrs.ptr_complex16[posn].Re
		= test_fn(test_fn_type,
			  select_fn_complex16_Re, test_fn_order, no_deriv,
			  x,y,z);
	input_grid.data_ptrs.ptr_complex16[posn].Im
		= test_fn(test_fn_type,
			  select_fn_complex16_Im, test_fn_order, no_deriv,
			  x,y,z);
	}
	}
	}
}
	  }

//******************************************************************************

//
// This function does the actual interpolation.
//
namespace {
void do_interpolation(enum interp_type interp_type,
		      const cGH* GH, int timer_handle,
		      const char coord_system[],
		      const input_grid& input_grid, interp_grid& interp_grid,
		      const char interp_operator[], int interp_order,
						    const char interp_pars[],
		      int N_operations, const int opcodes[max_N_operations])
{
switch	(interp_type)
	{
case interp_InterpLocal:
	do_InterpLocal(GH, timer_handle,
		       coord_system,
		       input_grid, interp_grid,
		       interp_operator, interp_order, interp_pars,
		       N_operations, opcodes);
	break;
case interp_InterpLocalUniform:
	do_InterpLocalUniform(GH, timer_handle,
			      coord_system,
			      input_grid, interp_grid,
			      interp_operator, interp_order, interp_pars,
			      N_operations, opcodes);
	break;
case interp_InterpGV:
	do_InterpGV(GH, timer_handle,
		    coord_system,
		    input_grid, interp_grid,
		    interp_operator, interp_order, interp_pars,
		    N_operations, opcodes);
	break;
case interp_InterpGridArrays:
	do_InterpGridArrays(GH, timer_handle,
			    coord_system,
			    input_grid, interp_grid,
			    interp_operator, interp_order, interp_pars,
			    N_operations, opcodes);
	break;
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "TestInterpBase(): bad interp_type=(int)%d!",
		   int(interp_type));			/*NOTREACHED*/
	}
}
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function prints the interpolation results to an output file.
//
namespace {
void do_write_plot_file(enum fn_type test_fn_type, int test_fn_order,
			const interp_grid& interp_grid,
			enum select_fn which_fn_to_plot, int which_iop_to_plot,
							 int which_deriv_to_plot,
			bool multiprocessor_file_names,
			int my_proc, const char plot_file_base_name[])
{
const int N_dims = interp_grid.N_dims;
const int N_operations = interp_grid.N_operations;
const int N_output = N_output_arrays * N_operations;

const int N_filename_buffer = 200;
char filename_buffer[N_filename_buffer];
if (multiprocessor_file_names)
   then snprintf(filename_buffer, N_filename_buffer,
		 "%s.proc%d",
		 plot_file_base_name, my_proc);
   else snprintf(filename_buffer, N_filename_buffer,
		 "%s",
		 plot_file_base_name);

FILE *fileptr = fopen(filename_buffer, "w");
if (fileptr == NULL)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"do_write_plot_file(): can't open output plot file \"%s\"!",
		   filename_buffer);				/*NOTREACHED*/

CCTK_VInfo(CCTK_THORNSTRING,
	   "writing plot file \"%s\"",
	   filename_buffer);

switch	(N_dims)
	{
case 1:
	fprintf(fileptr, "# interp x\t");
	break;
case 2:
	fprintf(fileptr, "# interp x\tinterp y\t");
	break;
case 3:
	fprintf(fileptr, "# interp x\tinterp y\tinterp z\t");
	break;
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"do_write_plot_file(): wierd N_dims=%d while printing header!",
		   N_dims);					/*NOTREACHED*/
	}
fprintf(fileptr, "test function\tinterp result\tinterp error\n");

assert( which_fn_to_plot >= 0 );
assert( which_fn_to_plot < N_output_arrays );
assert( which_iop_to_plot >= 0 );
assert( which_iop_to_plot < N_operations );
const int out = which_fn_to_plot*N_operations + which_iop_to_plot;
assert( out >= 0 );
assert( out < N_output );


fp old_x = interp_grid.interp_x[0];
	for (int posn = 0 ; posn < interp_grid.N_points ; ++posn)
	{
	const fp x = interp_grid.interp_x[posn];
	const fp y = interp_grid.interp_y[posn];
	const fp z = interp_grid.interp_z[posn];

	const fp test_function = test_fn(test_fn_type, which_fn_to_plot,
					 test_fn_order, which_deriv_to_plot,
					 x,y,z);
	fp interp_result;
	switch	(which_fn_to_plot)
		{
	case select_fn_real:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_real[posn];
		break;

	case select_fn_real4:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_real4[posn];
		break;

	case select_fn_real8:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_real8[posn];
		break;

	case select_fn_complex_Re:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex[posn].Re;
		break;
	case select_fn_complex_Im:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex[posn].Im;
		break;

	case select_fn_complex8_Re:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex8[posn].Re;
		break;
	case select_fn_complex8_Im:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex8[posn].Im;
		break;

	case select_fn_complex16_Re:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex16[posn].Re;
		break;
	case select_fn_complex16_Im:
		interp_result = interp_grid.output_arrays[which_iop_to_plot]
					   .ptr_complex16[posn].Im;
		break;

	default:
		CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"do_write_plot_file(): bad which_fn_to_plot=(int)%d!",
			   int(which_fn_to_plot));		/*NOTREACHED*/
		}

	const fp interp_error = interp_result - test_function;

#ifdef DEBUG_ARRAY_INDEXING
	if ((posn == 0) || (posn == 1) || (posn == 2) || (posn == 10))
	   then printf("### posn=%d xyz=%g %g %g test_fn=%g interp_result=%g\n",
		       posn, double(x), double(y), double(z),
		       double(test_function), double(interp_result));
#endif

	if ((N_dims > 1) && (x < old_x))
	   then fprintf(fileptr, "\n");

	switch	(N_dims)
		{
	case 1:
		fprintf(fileptr, "%g\t", double(x));
		break;
	case 2:
		fprintf(fileptr, "%g\t%g\t", double(x), double(y));
		break;
	case 3:
		fprintf(fileptr, "%g\t%g\t%g\t", double(x), double(y), double(z));
		break;
	default:
		CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
	"do_write_plot_file(): wierd interp_grid.N_dims=%d inside loop!",
			   interp_grid.N_dims);			/*NOTREACHED*/
		}
	fprintf(fileptr, "%g\t%g\t%e\n",
		double(test_function), double(interp_result),
		double(interp_error));

	old_x = x;
	}

fclose(fileptr);
}
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

#ifdef JACOBIAN_STUFF
//
// This function verifies that the interpolator-query and
// numerical-perturbation Jacobians have the same sized molecules.
// If so, this function is a no-op.  If not, this function prints a
// suitable error message and aborts the run.
//
namespace {
void check_Jacobians_same_size(const Jacobian& query_J, const Jacobian& NP_J)
{
if (! Jacobian::same_molecule_size(query_J, NP_J))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   implementation restriction: query and NP Jacobians must have\n"
"                               the same molecule size!\n"
"                               query molecule_min_m[]=[%d,%d,%d]\n"
"                                     molecule_max_m[]=[%d,%d,%d]\n"
"                               NP    molecule_min_m[]=[%d,%d,%d]\n"
"                                     molecule_max_m[]=[%d,%d,%d]"
		   ,
		   int(query_J.molecule_min_m[0]),
		   int(query_J.molecule_min_m[1]),
		   int(query_J.molecule_min_m[2]),
		   int(query_J.molecule_max_m[0]),
		   int(query_J.molecule_max_m[1]),
		   int(query_J.molecule_max_m[2]),
		   int(NP_J.molecule_min_m[0]),
		   int(NP_J.molecule_min_m[1]),
		   int(NP_J.molecule_min_m[2]),
		   int(NP_J.molecule_max_m[0]),
		   int(NP_J.molecule_max_m[1]),
		   int(NP_J.molecule_max_m[2]));		/*NOTREACHED*/
}
	  }
#endif

//******************************************************************************

#ifdef JACOBIAN_STUFF
//
// This function prints the interpolator-query and/or numerical-perturbation
// Jacobians to an output file.  If both are printed, they must have the
// same sized molecules, cf  check_Jacobians_same_size() .
//
namespace {
void write_Jacobians(const grid& interp_grid,
		     const Jacobian* p_query_J, const Jacobian* p_NP_J,
		     const char Jacobian_file_name[])
{
if ((p_query_J == NULL) && (p_NP_J == NULL))
   then return;				// nothing to print!

// get a guaranteed-to-be-valid set of molecule sizes
const CCTK_INT *const molecule_min_m = (p_query_J != NULL)
				       ? p_query_J->molecule_min_m
				       :    p_NP_J->molecule_min_m;
const CCTK_INT *const molecule_max_m = (p_query_J != NULL)
				       ? p_query_J->molecule_max_m
				       :    p_NP_J->molecule_max_m;

FILE *fileptr = fopen(Jacobian_file_name, "w");
if (fileptr == NULL)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "write_Jacobians(): can't open output plot file \"%s\"!",
		   Jacobian_file_name);			/*NOTREACHED*/

fprintf(fileptr, "# pt\tmi mj mk\tposn");
if (p_query_J != NULL)
   then fprintf(fileptr, "\tquery_J");
if (p_NP_J != NULL)
   then fprintf(fileptr, "\t\tNP_J");
if ((p_query_J != NULL) && (p_NP_J != NULL))
   then fprintf(fileptr, "\t\tJacobian error");
fprintf(fileptr, "\n");

const int out = 0;
	for (int pt = 0 ; pt < interp_grid.N_points ; ++pt)
	{
	for (int mk = molecule_min_m[2] ; mk <= molecule_max_m[2] ; ++mk)
	{
	for (int mj = molecule_min_m[1] ; mj <= molecule_max_m[1] ; ++mj)
	{
	for (int mi = molecule_min_m[0] ; mi <= molecule_max_m[0] ; ++mi)
	{
	const int query_J_posn = (p_query_J != NULL)
				 ? p_query_J->subscript(out, pt, mi,mj,mk)
				 : 0;
	const int    NP_J_posn = (p_NP_J != NULL)
				 ? p_NP_J->subscript(out, pt, mi,mj,mk)
				 : 0;
	if ((p_query_J != NULL) && (p_NP_J != NULL))
	   then if (query_J_posn != NP_J_posn)
		   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   write_Jacobians(): query and NP Jacobians have different J posns!\n"
"                      pt=%d mijk=[%d,%d,%d]\n"
"                      query_J_posn=%d NP_J_posn=%d"
				   ,
				   pt, mi,mj,mk,
				   query_J_posn, NP_J_posn);	/*NOTREACHED*/
	const int posn = (p_query_J != NULL) ? query_J_posn : NP_J_posn;

	const fp query_Jac = (p_query_J != NULL)
			     ? p_query_J->Jacobian_pointer[out][posn]
			     : 0.0;
	const fp    NP_Jac = (   p_NP_J != NULL)
			     ?    p_NP_J->Jacobian_pointer[out][posn]
			     : 0.0;

	fprintf(fileptr, "%d\t%2d %2d %2d\t%d", pt, mi,mj,mk, posn);
	if (p_query_J != NULL)
	   then fprintf(fileptr, "\t%.10g", double(query_Jac));
	if (p_NP_J != NULL)
	   then fprintf(fileptr, "\t%.10g", double(NP_Jac));
	if ((p_query_J != NULL) && (p_NP_J != NULL))
	   then fprintf(fileptr, "\t%e", double(query_Jac - NP_Jac));
	fprintf(fileptr, "\n");
	}
	}
	}
	}

fclose(fileptr);
}
	  }
#endif

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function computes a "linear hash" of all the interpolation-grid
// data arrays.
//	hash = sum_k w[k]*array[k]/N_points
// where the w[k] are pseudorandom weights in the range
// [hash_weight_min, hash_weight_max).
//
namespace {
void hash_interp_results(const interp_grid& interp_grid,
			 struct hashes hashes_array[max_N_operations],
			 fp hash_weight_min, fp hash_weight_max)
{
CCTK_VInfo(CCTK_THORNSTRING,
	   "hashing interpolation results");

const double temp = 1.0 / double(interp_grid.N_points);
#define RANDOM()   jtutil::drandom(temp*hash_weight_min, temp*hash_weight_max)

    for (int posn = 0 ; posn < interp_grid.N_points ; ++posn)
    {
	for (int iop = 0 ; iop < interp_grid.N_operations ; ++iop)
	{
	struct hashes& h = hashes_array[iop];
	struct data_ptrs& dp = interp_grid.output_arrays[iop];
	h.hash_of_real         += RANDOM() * dp.ptr_real[posn];
	h.hash_of_real4        += RANDOM() * dp.ptr_real4[posn];
	h.hash_of_real8        += RANDOM() * dp.ptr_real8[posn];
	h.hash_of_complex_Re   += RANDOM() * dp.ptr_complex[posn].Re;
	h.hash_of_complex_Im   += RANDOM() * dp.ptr_complex[posn].Im;
	h.hash_of_complex8_Re  += RANDOM() * dp.ptr_complex8[posn].Re;
	h.hash_of_complex8_Im  += RANDOM() * dp.ptr_complex8[posn].Im;
	h.hash_of_complex16_Re += RANDOM() * dp.ptr_complex16[posn].Re;
	h.hash_of_complex16_Im += RANDOM() * dp.ptr_complex16[posn].Im;
	}
    }
}
	  }

//******************************************************************************

//
// This function must be called on each processor.  It gathers the
// interpolation hashes into a single processor (conventionally processor 0).
//
// Argumkents:
// N_procs = (in) The number of processors.
// my_proc = (in) My processor number.
// my_hashes = (in) The hashes to be gathered from this processor.
// all_hashes[N_procs]
//	= (out*) (Pointers to caller-supplied) array for results.
//		 On processor #0 this function will store the gathered
//		 data in this array.  On all other processors it will
//		 write (unspecified) garbage into this array.
//
namespace {
void gather_hashes(const cGH *GH,
		   const hashes_array_t my_hashes_array,
		   int N_procs, int my_proc, hashes_array_t all_hashes_array[])
{
CCTK_VInfo(CCTK_THORNSTRING,
	   "gathering hashes to processor #0");

//
// ***** implementation notes *****
//
// We implement the gather operation as follows:
// * set up the buffers as all zeros, except for the desired data-to-gather
//   in our own processor number's slot
// * do a sum-reduction to processor 0
// * unpack the results to the all_hashes[] array
// This is admittedly a major kludge... :(
// 
// Because Cactus only knows how to reduce arrays of Cactus datatypes,
// we implement each buffer as a 4-D array of the appropriate CCTK_*
// datatype, with axes
//	[proc]		# range [0, N_procs)
//	[iaxis]		# range [0, max_N_interp_axes)
//	[iop]		# range [0, max_N_operations)
//	[iht]		# range [0, N_hash_types)
// In turn, we implement this as a 1-D new[]-allocated array for the
// first axis, of 3-D C-style arrays for the remaining axes.
// 
// For simplicity, we treat each buffer as a 1-D array of size N_buffer
// for the reduction operation.
//

// these enums index positions within the buffers for the gather operation
enum	{
	enum_hash_real = 0,
	enum_hash_complex_Re, enum_hash_complex_Im,
	N_hash_types
	};
const int N_buffer = N_procs * max_N_interp_axes
		     * max_N_operations * N_hash_types;

//
// set up the buffers
//
typedef CCTK_REAL  buffer_real_t [max_N_interp_axes]
				 [max_N_operations]
				 [N_hash_types];
typedef CCTK_REAL4 buffer_real4_t[max_N_interp_axes]
				 [max_N_operations]
				 [N_hash_types];
typedef CCTK_REAL8 buffer_real8_t[max_N_interp_axes]
				 [max_N_operations]
				 [N_hash_types];

buffer_real_t*  buffer_real  = new buffer_real_t [N_procs];
buffer_real4_t* buffer_real4 = new buffer_real4_t[N_procs];
buffer_real8_t* buffer_real8 = new buffer_real8_t[N_procs];

	  {
	for (int proc = 0 ; proc < N_procs ; ++proc)
	{
		for (int iaxis = 0 ; iaxis < max_N_interp_axes ; ++iaxis)
		{
		for (int iop = 0 ; iop < max_N_operations ; ++iop)
		{
			for (int iht = 0 ; iht < N_hash_types ; ++iht)
			{
			buffer_real [proc][iaxis][iop][iht] = 0.0;
			buffer_real4[proc][iaxis][iop][iht] = 0.0;
			buffer_real8[proc][iaxis][iop][iht] = 0.0;
			}
		}
		}
	}
	  }

	  {
	for (int iaxis = 0 ; iaxis < max_N_interp_axes ; ++iaxis)
	{
	for (int iop = 0 ; iop < max_N_operations ; ++iop)
	{
	const struct hashes& hashes = my_hashes_array[iaxis][iop];
	CCTK_REAL*  buffer_real_ptr  = buffer_real [my_proc][iaxis][iop];
	CCTK_REAL4* buffer_real4_ptr = buffer_real4[my_proc][iaxis][iop];
	CCTK_REAL8* buffer_real8_ptr = buffer_real8[my_proc][iaxis][iop];

	buffer_real_ptr [enum_hash_real      ] = hashes.hash_of_real;
	buffer_real_ptr [enum_hash_complex_Re] = hashes.hash_of_complex_Re;
	buffer_real_ptr [enum_hash_complex_Im] = hashes.hash_of_complex_Im;

	buffer_real4_ptr[enum_hash_real      ] = hashes.hash_of_real4;
	buffer_real4_ptr[enum_hash_complex_Re] = hashes.hash_of_complex8_Re;
	buffer_real4_ptr[enum_hash_complex_Im] = hashes.hash_of_complex8_Im;

	buffer_real8_ptr[enum_hash_real      ] = hashes.hash_of_real8;
	buffer_real8_ptr[enum_hash_complex_Re] = hashes.hash_of_complex16_Re;
	buffer_real8_ptr[enum_hash_complex_Im] = hashes.hash_of_complex16_Im;
	}
	}
	  }

//
// sum-reduce the buffers
//
const int reduction_handle = CCTK_ReductionHandle("sum");
if (reduction_handle < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"gather_hashes(): error %d trying to get reduction operator handle!"
		   ,
		   reduction_handle);				/*NOTREACHED*/
const int status
   = CCTK_ReduceArray(GH,
		      0,		// processor which will get result
		      reduction_handle,
		      N_buffer,				// output array size
		      CCTK_VARIABLE_REAL,		// output datatype
		      static_cast<void *>(buffer_real),	// output array
		      1,				// N_dims
		      1,				// N_input_arrays
		      CCTK_VARIABLE_REAL,		// input datatype
		      N_buffer,				// input array size
		      static_cast<const void *>(buffer_real));	//input array
const int status4
   = CCTK_ReduceArray(GH,
		      0,		// processor which will get result
		      reduction_handle,
		      N_buffer,				// output array size
		      CCTK_VARIABLE_REAL4,		// output datatype
		      static_cast<void *>(buffer_real4),// output array
		      1,				// N_dims
		      1,				// N_input_arrays
		      CCTK_VARIABLE_REAL4,		// input datatype
		      N_buffer,				// input array size
		      static_cast<const void *>(buffer_real4));	//input array
const int status8
   = CCTK_ReduceArray(GH,
		      0,		// processor which will get result
		      reduction_handle,
		      N_buffer,				// output array size
		      CCTK_VARIABLE_REAL8,		// output datatype
		      static_cast<void *>(buffer_real8),// output array
		      1,				// N_dims
		      1,				// N_input_arrays
		      CCTK_VARIABLE_REAL8,		// input datatype
		      N_buffer,				// input array size
		      static_cast<const void *>(buffer_real8));	//input array
if ((status < 0) || (status4 < 0) || (status8 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   gather_hashes(): error return from one or more reduction operator(s)!\n"
"                    error status=%d status4=%d status8=%d"
		   ,
		   status, status4, status8);			/*NOTREACHED*/

//
// unpack the results from the buffers back into the all_hashes[] array
//
      {
    for (int proc = 0 ; proc < N_procs ; ++proc)
    {
	for (int iaxis = 0 ; iaxis < max_N_interp_axes ; ++iaxis)
	{
	for (int iop = 0 ; iop < max_N_operations ; ++iop)
	{
	struct hashes& hashes = all_hashes_array[proc][iaxis][iop];
	const CCTK_REAL*  buffer_real_ptr  = buffer_real [proc][iaxis][iop];
	const CCTK_REAL4* buffer_real4_ptr = buffer_real4[proc][iaxis][iop];
	const CCTK_REAL8* buffer_real8_ptr = buffer_real8[proc][iaxis][iop];

	hashes.hash_of_real         = buffer_real_ptr [enum_hash_real      ];
	hashes.hash_of_complex_Re   = buffer_real_ptr [enum_hash_complex_Re];
	hashes.hash_of_complex_Im   = buffer_real_ptr [enum_hash_complex_Im];

	hashes.hash_of_real4        = buffer_real4_ptr[enum_hash_real      ];
	hashes.hash_of_complex8_Re  = buffer_real4_ptr[enum_hash_complex_Re];
	hashes.hash_of_complex8_Im  = buffer_real4_ptr[enum_hash_complex_Im];

	hashes.hash_of_real8        = buffer_real8_ptr[enum_hash_real      ];
	hashes.hash_of_complex16_Re = buffer_real8_ptr[enum_hash_complex_Re];
	hashes.hash_of_complex16_Im = buffer_real8_ptr[enum_hash_complex_Im];
	}
	}
    }
      }

//
// free the buffers
//
delete[] buffer_real8;
delete[] buffer_real4;
delete[] buffer_real;
}
	  }

//******************************************************************************

//
// This function verifies that all the hashes are identical across
// processors.  If so, this function prints a message to this effect.
// If not, then
//    if abort_if_hashes_differ
//       then this function does a  CCTK_VWarn(-1, ...)  (and does not return);
//       otherwise this function prints a message describing the problem
//                 (and returns).
//
namespace {
void verify_all_hashes_identical(int N_procs, hashes_array_t all_hashes_array[],
				 bool abort_if_hashes_differ)
{
CCTK_VInfo(CCTK_THORNSTRING,
	   "verifying that hashes are identical on all processors");

    for (int proc = 1 ; proc < N_procs ; ++proc)
    {
	for (int iaxis = 0 ; iaxis < max_N_interp_axes ; ++iaxis)
	{
	for (int iop = 0 ; iop < max_N_operations ; ++iop)
	{
	const struct hashes& h0 = all_hashes_array[0   ][iaxis][iop];
	const struct hashes& hp = all_hashes_array[proc][iaxis][iop];
#ifdef DEBUG_HASH_DIFFERENT_PRINTING
	printf("*** faking different hashes on processor #0 and #1:\n");
	if (proc == 1)
#else
	if (    (h0.hash_of_real         != hp.hash_of_real)
	     || (h0.hash_of_real4        != hp.hash_of_real4)
	     || (h0.hash_of_real8        != hp.hash_of_real8)
	     || (h0.hash_of_complex_Re   != hp.hash_of_complex_Re)
	     || (h0.hash_of_complex_Im   != hp.hash_of_complex_Im)
	     || (h0.hash_of_complex8_Re  != hp.hash_of_complex8_Re)
	     || (h0.hash_of_complex8_Im  != hp.hash_of_complex8_Im)
	     || (h0.hash_of_complex16_Re != hp.hash_of_complex16_Re)
	     || (h0.hash_of_complex16_Im != hp.hash_of_complex16_Im)    )
#endif
	   then {
		const int N_buffer = 1000;
		char buffer[N_buffer];
		snprintf(buffer, N_buffer,
"\n"
"   hashes differ between processor #0 and #%d!\n"
"   (i.e. different processors got different results from interpolation!)\n"
"\t\t\t\tprocessor #0\t\tprocessor #%d\n"
"   hash_of_real\t\t\t%#.15g\t%#.15g\n"
"   hash_of_real4\t\t%#.7g\t\t%#.7g\n"
"   hash_of_real8\t\t%#.15g\t%#.15g\n"
"   hash_of_complex.Re\t\t%#.15g\t%#.15g\n"
"   hash_of_complex.Im\t\t%#.15g\t%#.15g\n"
"   hash_of_complex8.Re\t\t%#.7g\t\t%#.7g\n"
"   hash_of_complex8.Im\t\t%#.7g\t\t%#.7g\n"
"   hash_of_complex16.Re\t\t%#.15g\t%#.15g\n"
"   hash_of_complex16.Im\t\t%#.15g\t%#.15g"
   ,
   proc,
   proc,
   double(h0.hash_of_real)        , double(hp.hash_of_real),
   double(h0.hash_of_real4)       , double(hp.hash_of_real4),
   double(h0.hash_of_real8)       , double(hp.hash_of_real8),
   double(h0.hash_of_complex_Re)  , double(hp.hash_of_complex_Re),
   double(h0.hash_of_complex_Im)  , double(hp.hash_of_complex_Im),
   double(h0.hash_of_complex8_Re) , double(hp.hash_of_complex8_Re),
   double(h0.hash_of_complex8_Im) , double(hp.hash_of_complex8_Im),
   double(h0.hash_of_complex16_Re), double(hp.hash_of_complex16_Re),
   double(h0.hash_of_complex16_Im), double(hp.hash_of_complex16_Im));
		if (abort_if_hashes_differ)
		   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
				   "%s", buffer);		/*NOTREACHED*/
		   else {
			CCTK_VInfo(CCTK_THORNSTRING,
				   "*** %s", buffer);
			return;
			}
		}
	}
	}
    }
}
	  }

//******************************************************************************

//
// This function writes the hashes to the hash output files.
// The file format is as follows:  There are three files, one for each
// of real, real4, and real8.  Each file has the following format:
//	## column 1 = derivative operation
//	## column 2 = CCTK_REAL interpolation hash
//	## column 3 = CCTK_COMPLEX.Re interpolation hash
//	## column 4 = CCTK_COMPLEX.Im interpolation hash
//
//	# input_axes = "XYZ"
//	0	hash_real	hash_complex_Re		hash_complex_Im
//	1	hash_real	hash_complex_Re		hash_complex_Im
//	2	hash_real	hash_complex_Re		hash_complex_Im
//	...
//	33	hash_real	hash_complex_Re		hash_complex_Im
//
//	# input_axes = "XYZ"
//	...
//
namespace {
void do_write_hash_files
  (int N_interp_axes, const char* const interp_axes_strarr[max_N_interp_axes],
   int N_operations, const int opcodes[max_N_operations],
   const hashes_array_t hashes_array,
   const char hash_files_base_name[])
{
CCTK_VInfo(CCTK_THORNSTRING,
           "writing hash files");

const int N_file_name_buffer = 200;
char hash_file_name_real[N_file_name_buffer],
     hash_file_name_real4[N_file_name_buffer],
     hash_file_name_real8[N_file_name_buffer];

snprintf(hash_file_name_real, N_file_name_buffer,
	 "%s.hash",
	 hash_files_base_name);
snprintf(hash_file_name_real4, N_file_name_buffer,
	 "%s.hash4",
	 hash_files_base_name);
snprintf(hash_file_name_real8, N_file_name_buffer,
	 "%s.hash8",
	 hash_files_base_name);

FILE *fileptr_real  = fopen(hash_file_name_real, "w");
FILE *fileptr_real4 = fopen(hash_file_name_real4, "w");
FILE *fileptr_real8 = fopen(hash_file_name_real8, "w");

if ( (fileptr_real == NULL)
     || (fileptr_real4 == NULL) || (fileptr_real8 == NULL) )
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   do_write_hash_file(): can't open one or more of the hash output files(s)\n"
"                         \"%s\", \"%s\", and/or \"%s\"!"
		   ,
		   hash_file_name_real,
		   hash_file_name_real4, hash_file_name_real8);	/*NOTREACHED*/

fprintf(fileptr_real,  "## column 1 = derivative operation\n");
fprintf(fileptr_real,  "## column 2 = CCTK_REAL interpolation hash\n");
fprintf(fileptr_real,  "## column 3 = CCTK_COMPLEX.Re interpolation hash\n");
fprintf(fileptr_real,  "## column 4 = CCTK_COMPLEX.Im interpolation hash\n");
fprintf(fileptr_real,  "\n");

fprintf(fileptr_real4, "## column 1 = derivative operation\n");
fprintf(fileptr_real4, "## column 2 = CCTK_REAL4 interpolation hash\n");
fprintf(fileptr_real4, "## column 3 = CCTK_COMPLEX8.Re interpolation hash\n");
fprintf(fileptr_real4, "## column 4 = CCTK_COMPLEX8.Im interpolation hash\n");
fprintf(fileptr_real4, "\n");

fprintf(fileptr_real8, "## column 1 = derivative operation\n");
fprintf(fileptr_real8, "## column 2 = CCTK_REAL8 interpolation hash\n");
fprintf(fileptr_real8, "## column 3 = CCTK_COMPLEX16.Re interpolation hash\n");
fprintf(fileptr_real8, "## column 4 = CCTK_COMPLEX16.Im interpolation hash\n");
fprintf(fileptr_real8, "\n");

	for (int iaxis = 0 ; iaxis < N_interp_axes ; ++iaxis)
	{
	fprintf(fileptr_real , "interp_axes = \"%s\"\n",
		interp_axes_strarr[iaxis]);
	fprintf(fileptr_real4, "interp_axes = \"%s\"\n",
		interp_axes_strarr[iaxis]);
	fprintf(fileptr_real8, "interp_axes = \"%s\"\n",
		interp_axes_strarr[iaxis]);
		for (int iop = 0 ; iop < N_operations ; ++iop)
		{
		const struct hashes& hashes = hashes_array[iaxis][iop];
		fprintf(fileptr_real,  "%d\t%#.15g\t%#.15g\t%#.15g\n",
			opcodes[iop], double(hashes.hash_of_real),
				      double(hashes.hash_of_complex_Re),
				      double(hashes.hash_of_complex_Im));
		fprintf(fileptr_real4, "%d\t%#.7g\t%#.7g\t%#.7g\n",
			opcodes[iop], double(hashes.hash_of_real4),
				      double(hashes.hash_of_complex8_Re),
				      double(hashes.hash_of_complex8_Im));
		fprintf(fileptr_real8, "%d\t%#.15g\t%#.15g\t%#.15g\n",
			opcodes[iop], double(hashes.hash_of_real8),
				      double(hashes.hash_of_complex16_Re),
				      double(hashes.hash_of_complex16_Im));
		}
	fprintf(fileptr_real , "\n");
	fprintf(fileptr_real4, "\n");
	fprintf(fileptr_real8, "\n");
	}

fclose(fileptr_real);
fclose(fileptr_real4);
fclose(fileptr_real8);
}
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function accumulates the norms of the current set of interpolation
// results into the raw and scaled error norms.
//
namespace {
void accumulate_interp_error_norms
	(enum fn_type test_fn_type, int test_fn_order,
	 const int opcodes[max_N_operations],
	 const fp input_grid_delta[max_N_dims],
	 const interp_grid& interp_grid,
	 struct error_norms& raw_error_norms,
	 struct error_norms& scaled_error_norms)
{
CCTK_VInfo(CCTK_THORNSTRING,
	   "accumulating interpolation error norms");

const fp epsilon_real = (sizeof(CCTK_REAL) == sizeof(double))
			? DBL_EPSILON
			: 1.0;
const fp epsilon_real4 = (sizeof(CCTK_REAL4) == sizeof(float))
			 ? FLT_EPSILON
			 : 1.0;
const fp epsilon_real8 = (sizeof(CCTK_REAL8) == sizeof(double))
			 ? DBL_EPSILON
			 : 1.0;

//
// precompute the derivative scaling factors for each iop
// we expect the error to be on the order of epsilon / delta_x
//
fp deltas[max_N_operations];
	for (int iop = 0 ; iop < interp_grid.N_operations ; ++iop)
	{
	deltas[iop] = delta_product(opcodes[iop], input_grid_delta);
	}

//
// accumulate the error norms
//
    for (int posn = 0 ; posn < interp_grid.N_points ; ++posn)
    {
    const fp x = interp_grid.interp_x[posn];
    const fp y = interp_grid.interp_y[posn];
    const fp z = interp_grid.interp_z[posn];

	for (int iop = 0 ; iop < interp_grid.N_operations ; ++iop)
	{
	// compute all the interpolation errors at this point
	const struct data_ptrs& dp = interp_grid.output_arrays[iop];
	const fp error_real = dp.ptr_real[posn]
			      -
			      test_fn(test_fn_type, select_fn_real,
				      test_fn_order, opcodes[iop],
				      x,y,z);
	const fp error_real4 = dp.ptr_real4[posn]
			       -
			       test_fn(test_fn_type, select_fn_real4,
				       test_fn_order, opcodes[iop],
				       x,y,z);
	const fp error_real8 = dp.ptr_real8[posn]
			       -
			       test_fn(test_fn_type, select_fn_real8,
				       test_fn_order, opcodes[iop],
				       x,y,z);
	const fp error_complex_Re = dp.ptr_complex[posn].Re
				    -
				    test_fn(test_fn_type, select_fn_complex_Re,
					    test_fn_order, opcodes[iop],
					    x,y,z);
	const fp error_complex_Im = dp.ptr_complex[posn].Im
				    -
				    test_fn(test_fn_type, select_fn_complex_Im,
					    test_fn_order, opcodes[iop],
					    x,y,z);
	const fp error_complex8_Re = dp.ptr_complex8[posn].Re
				     -
				     test_fn(test_fn_type, select_fn_complex8_Re,
					     test_fn_order, opcodes[iop],
					     x,y,z);
	const fp error_complex8_Im = dp.ptr_complex8[posn].Im
				     -
				     test_fn(test_fn_type, select_fn_complex8_Im,
					     test_fn_order, opcodes[iop],
					     x,y,z);
	const fp error_complex16_Re = dp.ptr_complex16[posn].Re
				      -
				      test_fn(test_fn_type, select_fn_complex16_Re,
					      test_fn_order, opcodes[iop],
					      x,y,z);
	const fp error_complex16_Im = dp.ptr_complex16[posn].Im
				      -
				      test_fn(test_fn_type, select_fn_complex16_Im,
					      test_fn_order, opcodes[iop],
					      x,y,z);

	// accumulate the raw error norms
	raw_error_norms.norm_of_real        .data(error_real);
	raw_error_norms.norm_of_real4       .data(error_real4);
	raw_error_norms.norm_of_real8       .data(error_real8);
	raw_error_norms.norm_of_complex_Re  .data(error_complex_Re);
	raw_error_norms.norm_of_complex_Im  .data(error_complex_Im);
	raw_error_norms.norm_of_complex8_Re .data(error_complex8_Re);
	raw_error_norms.norm_of_complex8_Im .data(error_complex8_Im);
	raw_error_norms.norm_of_complex16_Re.data(error_complex16_Re);
	raw_error_norms.norm_of_complex16_Im.data(error_complex16_Im);

	// accumulate the scaled error norms
	const fp scale_real  = epsilon_real  / deltas[iop];
	const fp scale_real4 = epsilon_real4 / deltas[iop];
	const fp scale_real8 = epsilon_real8 / deltas[iop];
	scaled_error_norms.norm_of_real
		          .data(error_real         / scale_real );
	scaled_error_norms.norm_of_real4
			  .data(error_real4        / scale_real4);
	scaled_error_norms.norm_of_real8
			  .data(error_real8        / scale_real8);
	scaled_error_norms.norm_of_complex_Re
			  .data(error_complex_Re   / scale_real );
	scaled_error_norms.norm_of_complex_Im
			  .data(error_complex_Im   / scale_real );
	scaled_error_norms.norm_of_complex8_Re
			  .data(error_complex8_Re  / scale_real4);
	scaled_error_norms.norm_of_complex8_Im
			  .data(error_complex8_Im  / scale_real4);
	scaled_error_norms.norm_of_complex16_Re
			  .data(error_complex16_Re / scale_real8);
	scaled_error_norms.norm_of_complex16_Im
			  .data(error_complex16_Im / scale_real8);
	}
    }
}
	  }

//******************************************************************************

#ifdef JACOBIAN_STUFF
//
// This function computes the Jacobian error norms.
//
namespace {
void compute_Jacobian_error_norms(const grid& interp_grid,
				  const Jacobian& query_J, const Jacobian& NP_J,
				  struct error_norms& error_norms)
{
const Jacobian& J = query_J;
const int out = 0;
	for (int pt = 0 ; pt < interp_grid.N_points ; ++pt)
	{
	for (int mk = J.molecule_min_m[2] ; mk <= J.molecule_max_m[2] ; ++mk)
	{
	for (int mj = J.molecule_min_m[1] ; mj <= J.molecule_max_m[1] ; ++mj)
	{
	for (int mi = J.molecule_min_m[0] ; mi <= J.molecule_max_m[0] ; ++mi)
	{
	const int query_posn = query_J.subscript(out, pt, mi,mj,mk);
	const int    NP_posn =    NP_J.subscript(out, pt, mi,mj,mk);
	if (query_posn != NP_posn)
	   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   compute_Jacobian_error_norms(): query and NP Jacobians\n"
"                                   have different J posns!\n"
"                                   pt=%d mijk=[%d,%d,%d]\n"
"                                   query_posn=%d NP_posn=%d"
			   ,
			   pt, mi,mj,mk,
			   query_posn, NP_posn);		/*NOTREACHED*/

	const fp query_Jac = query_J.Jacobian_pointer[out][query_posn];
	const fp    NP_Jac =    NP_J.Jacobian_pointer[out][   NP_posn];
	const fp error = query_Jac - NP_Jac;

	error_norms.norm_of_Jacobian.data(error);
	}
	}
	}
	}
}
	  }
#endif

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function prints all the raw and scaled error norms, and also
// the ok/marginal/BAD classifications of the scaled error norms.
//
// Results:
// This function returns the number of "BAD" scaled error norms.
//
namespace {
void classify_and_print_error_norms
	(const struct scaled_error_norm_thresholds& thresholds,
	 const struct error_norms& raw_error_norms,
	 const struct error_norms& scaled_error_norms,
	 int& N_ok, int& N_marginal, int& N_bad)
{
CCTK_VInfo(CCTK_THORNSTRING,
	   "error norms:      ------raw------ -----scaled----");
CCTK_VInfo(CCTK_THORNSTRING,
	   "data type         --rms-- --inf-- --rms-- --inf-- status");

classify_and_print_error_norm_pair("CCTK_REAL", thresholds,
				      raw_error_norms.norm_of_real,
				   scaled_error_norms.norm_of_real,
				   N_ok, N_marginal, N_bad);
classify_and_print_error_norm_pair("CCTK_REAL4", thresholds,
				      raw_error_norms.norm_of_real4,
				   scaled_error_norms.norm_of_real4,
				   N_ok, N_marginal, N_bad);
classify_and_print_error_norm_pair("CCTK_REAL8", thresholds,
				      raw_error_norms.norm_of_real8,
				   scaled_error_norms.norm_of_real8,
				   N_ok, N_marginal, N_bad);

classify_and_print_error_norm_pair("CCTK_COMPLEX.Re", thresholds,
				      raw_error_norms.norm_of_complex_Re,
				   scaled_error_norms.norm_of_complex_Re,
				   N_ok, N_marginal, N_bad);
classify_and_print_error_norm_pair("CCTK_COMPLEX.Im", thresholds,
				      raw_error_norms.norm_of_complex_Im,
				   scaled_error_norms.norm_of_complex_Im,
				   N_ok, N_marginal, N_bad);

classify_and_print_error_norm_pair("CCTK_COMPLEX8.Re", thresholds,
				      raw_error_norms.norm_of_complex8_Re,
				   scaled_error_norms.norm_of_complex8_Re,
				   N_ok, N_marginal, N_bad);
classify_and_print_error_norm_pair("CCTK_COMPLEX8.Im", thresholds,
				      raw_error_norms.norm_of_complex8_Im,
				   scaled_error_norms.norm_of_complex8_Im,
				   N_ok, N_marginal, N_bad);

classify_and_print_error_norm_pair("CCTK_COMPLEX16.Re", thresholds,
				      raw_error_norms.norm_of_complex16_Re,
				   scaled_error_norms.norm_of_complex16_Re,
				   N_ok, N_marginal, N_bad);
classify_and_print_error_norm_pair("CCTK_COMPLEX16.Im", thresholds,
				      raw_error_norms.norm_of_complex16_Im,
				   scaled_error_norms.norm_of_complex16_Im,
				   N_ok, N_marginal, N_bad);
}
	  }

//******************************************************************************

//
// This function classifies and prints a raw/scaled pair of error norms.
//
// Arguments:
// type_string = The character-string name of the CCTK_* data type,
//		 eg "CCTK_REAL".
// scaled_rms_norm,scaled_infinity_norm = The scaled error norms.
// thresholds = Gives the low/high thresholds for the scaled error norms.
// {raw,scaled}_error_norm = The error norms to be printed.
// N_{ok,marginal,bad} = (in out) The appropriate one of these counters
//				  is incremented.
//
// Results:
// This function returns false normally, or true if the error norm is
// classified as "BAD", i.e. if the error norm exceeds the high threshold
// in either the rms- or the infinity-norm.
//
namespace {
void classify_and_print_error_norm_pair
	(const char type_string[],
	 const struct scaled_error_norm_thresholds& thresholds,
	 const jtutil::norm<fp>& raw_error_norm,
	 const jtutil::norm<fp>& scaled_error_norm,
	 int& N_ok, int& N_marginal, int& N_bad)
{
const fp scaled_error_rms      = scaled_error_norm.rms_norm();
const fp scaled_error_infinity = scaled_error_norm.infinity_norm();
const bool ok =    (scaled_error_rms      <= thresholds.rms_low)
		&& (scaled_error_infinity <= thresholds.infinity_low);
const bool ok_or_marginal
	      =    (scaled_error_rms      <= thresholds.rms_high)
		&& (scaled_error_infinity <= thresholds.infinity_high);
if	(ok)             then ++N_ok;
else if (ok_or_marginal) then ++N_marginal;
else                          ++N_bad;
const char* outcome = ok ? "  ok" : ok_or_marginal ? "marginal" : "  BAD";

CCTK_VInfo(CCTK_THORNSTRING,
	   "%-17s %.1e %.1e %.1e %.1e %s",
	   type_string,
	   double(raw_error_norm.rms_norm()),
	   double(raw_error_norm.infinity_norm()),
	   double(scaled_error_rms), double(scaled_error_infinity),
	   outcome);
}
	  }
