// function.cc -- test function for interpolator
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/function.cc,v 1.3 2003/04/15 17:02:48 jthorn Exp $
//
// <<<prototypes>>>
// test_fn_pars
//
// test_fn
/// test_fn_xy
/// test_fn_generic
/// test_fn_series
/// test_fn_random
///
/// dipow
///

#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "cctk.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"

//
// prototypes for functions local to this file
//
namespace {
fp test_fn_xy     (int operation_code, fp x, fp y, fp  );
fp test_fn_generic(int operation_code, fp x, fp y, fp z);
fp test_fn_series (enum select_fn select, int order,
		   int operation_code, fp x, fp y, fp z);
fp test_fn_random  (int operation_code, fp x, fp y, fp z);

fp dipow(fp x, int n);
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// this is the test function for the interpolator
//
fp test_fn(enum fn_type test_fn_type,
	   enum select_fn test_fn_select, int test_fn_order, int operation_code,
	   fp x, fp y, fp z)
{
switch	(test_fn_type)
	{
case fn_xy:
	return test_fn_xy     (operation_code, x,y,z);
case fn_generic:
	return test_fn_generic(operation_code, x,y,z);
case fn_series:
	return test_fn_series (test_fn_select, test_fn_order, operation_code,
			       x,y,z);
case fn_random:
	return test_fn_random  (operation_code, x,y,z);
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "test_fn(): unknown test_fn_type=(int)%d!",
		   int(test_fn_type));				/*NOTREACHED*/
	return 0.0;	// dummy return (control flow)
	}
}

//******************************************************************************
//******************************************************************************
//******************************************************************************

namespace {
fp test_fn_xy(int operation_code, fp x, fp y, fp z)
{
switch	(operation_code)
	{
case DERIV(0):	return x*y;
case DERIV(1):	return y;
case DERIV(2):	return x;
case DERIV(3):	return 0.0;
case DERIV(11): return 0.0;
case DERIV(12): return 1.0;
case DERIV(22): return 0.0;
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "test_fn_xy(): operation_code=%d not supported!",
		   operation_code);		/*NOTREACHED*/
	}
}
	  }

//******************************************************************************

namespace {
fp test_fn_generic(int operation_code, fp x, fp y, fp z)
{
if (operation_code != DERIV(0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "test_fn_generic(): operation_code=%d, only 0 supported!",
		   operation_code);				/*NOTREACHED*/

return   tanh(  sin( x + atan(y- cos(z)+0.314) )  )
       * atan(  cos( y -  tan(z+ sin(x)-0.159) )  )
       *  cos( atan( z +  sin(x-tanh(y)+0.265) )  );
}
	  }

//******************************************************************************

//
// This function computes the test functions which are selected terms
// of a generic Taylor series with "arbitrary" coefficients.
//
// The series is
//	$f = f(x,y,z) = \sum_{ijk} c_{ijk} x^i y^j z^k$
//
// Arguments:
// select = an enum which selects one of a number of different
//	    sets of coefficients $c_{ijk}$; we actually just use
//	    this as an integer >= 0
// order = the maximum order of terms to be included in the sum
// operation code = specifies a partial derivative
//
namespace {
fp test_fn_series (enum select_fn select, int order, int operation_code,
		   fp x, fp y, fp z)
{
assert(order <= max_series_order);
static fp coeff[N_select_fn]
	       [max_series_order+1][max_series_order+1][max_series_order+1];
static bool first_time = true;
if (first_time)
   then {
	// initialize the coefficients
	first_time = false;
		for (int s = 0 ; s < N_select_fn ; ++s)
		{
			for (int i = 0 ; i <= max_series_order ; ++i)
			{
			for (int j = 0 ; j <= max_series_order-i ; ++j)
			{
			for (int k = 0 ; k <= max_series_order-(i+j) ; ++k)
			{
			assert(i + j + k <= max_series_order);
			coeff[s][i][j][k] = jtutil::drandom(0.5, 1.5);
			}
			}
			}
		}
	}

fp sum = 0.0;
	for (int i = 0 ; i <= order ; ++i)
	{
	for (int j = 0 ; j <= order-i ; ++j)
	{
	for (int k = 0 ; k <= order-(i+j) ; ++k)
	{
	assert(i + j + k <= order);
	assert(int(select) < N_select_fn);
	const fp c_ijk = coeff[int(select)][i][j][k];

	const fp xi = dipow(x,i);
	const fp yj = dipow(y,j);
	const fp zk = dipow(z,k);

	fp temp = 0.0;
	switch	(operation_code)
		{
	case DERIV(0):
		temp = xi*yj*zk;
		break;

	case DERIV(1):
		if (i >= 1)
		   then temp = i*dipow(x,i-1) * yj * zk;
		break;
	case DERIV(2):	if (j >= 1)
		   then temp = xi * j*dipow(y,j-1) * zk;
		break;
	case DERIV(3):	if (k >= 1)
		   then temp = xi * yj * k*dipow(z,k-1);
		break;

	case DERIV(11):
		if (i >= 2)
		   then temp = i*(i-1)*dipow(x,i-2) * yj * zk;
		break;
	case DERIV(22):
		if (j >= 2)
		   then temp = xi * j*(j-1)*dipow(y,j-2) * zk;
		break;
	case DERIV(33):
		if (k >= 2)
		   then temp = xi * yj * k*(k-1)*dipow(z,k-2);
		break;
	case DERIV(12):
	case DERIV(21):
		if ((i >= 1) && (j >= 1))
		   then temp = i*dipow(x,i-1) * j*dipow(y,j-1) * zk;
		break;
	case DERIV(13):
	case DERIV(31):
		if ((i >= 1) && (k >= 1))
		   then temp = i*dipow(x,i-1) * yj * k*dipow(z,k-1);
		break;
	case DERIV(23):
	case DERIV(32):
		if ((j >= 1) && (k >= 1))
		   then temp = xi * j*dipow(y,j-1) * k*dipow(z,k-1);
		break;
	default:
		CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
			   "test_fn_series(): operation_code=%d not supported!",
			   operation_code);		/*NOTREACHED*/
		temp = 0.0;	// dummy assignment (control flow)
		}

	sum += c_ijk * temp;
	}
	}
	}
return sum;
}
	  }

//******************************************************************************

//
// Bugs:
// If the fmod() argument happens to fall right on an integer, then
// differences in floating-point rounding from one machine to another
// may result in our returning significantly different results, which
// will cause a spurious test failure.  The _right_ solution would be
// to find a machine-independent floating-point hash function... but
// I can't seem to come up with such a beast.  For now, we just live
// with the (small) chance of the floating-point problem occuring...
//
namespace {
fp test_fn_random(int operation_code, fp x, fp y, fp z)
{
if (operation_code != DERIV(0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "test_fn_random(): operation_code=%d, only 0 supported!",
		   operation_code);				/*NOTREACHED*/

return fmod(1.0e6*PI*fabs(test_fn_generic(operation_code, x, y, z)), 1.0);
}
	  }

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function computes pow(x,n) for small integer n.
// (It only covers those cases needed in this file; unrecognized cases
// give a CCTK_VWarn(-1, ...) error exit.)
//
namespace {
fp dipow(fp x, int n)
{
switch	(n)
	{
case 0:	return 1.0;
case 1:	return x;
case 2:	return jtutil::pow2(x);
case 3:	return jtutil::pow3(x);
case 4: return jtutil::pow4(x);
case 5: return jtutil::pow5(x);
case 6: return jtutil::pow6(x);
case 7: return jtutil::pow7(x);
case 8: return jtutil::pow8(x);
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "dipow(): wierd n=%d (x=%g)!",
		   n, double(x));				/*NOTREACHED*/
	return 0.0;	// dummy return (control flow)
	}
}
	  }
