// interp_grid.cc -- grid of interpolation points and output arrays
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/interp_grid.cc,v 1.6 2003/04/15 17:02:48 jthorn Exp $

//
// interp_grid::interp_grid
// interp_grid::~interp_grid
// interp_grid::compute_grid_shape
// interp_grid::set_interp_xyz
//
// interp_grid::randomly_permute_interp_points
//

#include <assert.h>
#include <math.h>

#include "cctk.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "axis_list.hh"
#include "permutation.hh"
#include "input_grid.hh"
#include "interp_grid.hh"

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function constructs an  interp_grid  object.
//
interp_grid::interp_grid
	(const CCTK_INT Cactus_grid_size[max_N_dims],
	 const CCTK_INT N_ghost_zones[max_N_dims],
	 const fp Cactus_grid_origin[], const fp Cactus_grid_delta[],
	 const input_grid& input_grid,
	 int N_extra_interp_points, int data_to_interp_dxyz_ratio,
	 int N_operations_in)
	: N_dims(input_grid.N_dims),
	  N_operations(N_operations_in)
{
if ((N_extra_interp_points == 0) && (data_to_interp_dxyz_ratio == 1))
   then CCTK_VInfo(CCTK_THORNSTRING,
		   "constructing interp grid identical to input grid");
else if (N_extra_interp_points >= 0)
   then {
	CCTK_VInfo(CCTK_THORNSTRING,
		   "constructing interp grid like input grid");
	CCTK_VInfo(CCTK_THORNSTRING,
		   "   but with %d extra points on each axis",
		   N_extra_interp_points);
	}
else if (data_to_interp_dxyz_ratio >= 1)
   then {
	CCTK_VInfo(CCTK_THORNSTRING,
		   "constructing interp grid like input grid");
	CCTK_VInfo(CCTK_THORNSTRING,
		   "   but with %d times finer grid spacing",
		   data_to_interp_dxyz_ratio);
	}
else	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   interp_grid::interp_grid(): must specify exactly one of\n"
"                               N_extra_interp_points and\n"
"                               data_to_interp_dxyz_ratio!");	/*NOTREACHED*/

compute_grid_shape(Cactus_grid_size, N_ghost_zones,
		  Cactus_grid_origin, Cactus_grid_delta,
		  input_grid,
		  N_extra_interp_points, data_to_interp_dxyz_ratio);

//
// allocate the arrays
//
CCTK_VInfo(CCTK_THORNSTRING,
	   "interp grid: allocating arrays for N_operations=%d",
	   N_operations);
interp_x = new fp[N_points];
interp_y = new fp[N_points];
interp_z = new fp[N_points];

output_arrays = new struct data_ptrs[N_operations];
	for (int iop = 0 ; iop < N_operations ; ++iop)
	{
	output_arrays[iop].ptr_real      = new CCTK_REAL     [N_points];
	output_arrays[iop].ptr_real4     = new CCTK_REAL4    [N_points];
	output_arrays[iop].ptr_real8     = new CCTK_REAL8    [N_points];
	output_arrays[iop].ptr_complex   = new CCTK_COMPLEX  [N_points];
	output_arrays[iop].ptr_complex8  = new CCTK_COMPLEX8 [N_points];
	output_arrays[iop].ptr_complex16 = new CCTK_COMPLEX16[N_points];
	}

//
// set the xyz coordinates of the interpolation points (in lexicographic order)
//
set_interp_xyz();
}

//******************************************************************************

//
// This function destroys an  interp_grid  object.
//
interp_grid::~interp_grid()
{
	for (int iop = N_operations-1 ; iop >= 0 ; --iop)
	{
	delete[] output_arrays[iop].ptr_complex16;
	delete[] output_arrays[iop].ptr_complex8;
	delete[] output_arrays[iop].ptr_complex;
	delete[] output_arrays[iop].ptr_real8;
	delete[] output_arrays[iop].ptr_real4;
	delete[] output_arrays[iop].ptr_real;
	}

delete[] output_arrays;

delete[] interp_z;
delete[] interp_y;
delete[] interp_x;
}

//******************************************************************************

//
// This function computes
//	N_points  
//	interp_grid_size[]
//	interp_grid_origin[]
//	interp_grid_delta[]
//
void interp_grid::compute_grid_shape
	(const CCTK_INT Cactus_global_grid_size[max_N_dims],
	 const CCTK_INT N_ghost_zones[max_N_dims],
	 const fp Cactus_global_origin[], const fp Cactus_delta[],
	 const input_grid& input_grid,  
	 int N_extra_interp_points, int data_to_interp_dxyz_ratio)
{
N_points = 1;
	for (int axis = 0 ; axis < max_N_dims ; ++axis)
	{
	const bool interp_in_this_axis = input_grid.input_axes
						   .axis_is_unextended(axis);
	const int Cactus_axis = input_grid.input_axes.extended_axis(axis);
	CCTK_VInfo(CCTK_THORNSTRING,
		   "interp_grid: axis=%d Cactus_axis=%d (%s axis)",
		   axis, Cactus_axis,
		   (interp_in_this_axis ? "interpolation" : "extended"));

	const int N_input = Cactus_global_grid_size[Cactus_axis]
			    - 2*N_ghost_zones[Cactus_axis];
	const int N_interp = interp_in_this_axis
			     ? ( (N_extra_interp_points >= 0)
				 ? N_input + N_extra_interp_points
				 : 1 + data_to_interp_dxyz_ratio * (N_input-1) )
			     : 1;
	interp_grid_size[axis] = N_interp;
	N_points *= N_interp;
	CCTK_VInfo(CCTK_THORNSTRING,
		   "                    interp_grid_size[%d]=%d points",
		   axis, N_interp);

	interp_grid_origin[axis]
		= Cactus_global_origin[Cactus_axis]
		  + N_ghost_zones[Cactus_axis] * Cactus_delta[Cactus_axis];
	CCTK_VInfo(CCTK_THORNSTRING,
		   "                    interp_grid_origin[%d]=%g",
		   axis, double(interp_grid_origin[axis]));

	interp_grid_delta[axis] = interp_in_this_axis
				  ? ( Cactus_delta[Cactus_axis]
				      * fp(N_input - 1) / fp(N_interp - 1) )
				  : Cactus_delta[Cactus_axis];
	CCTK_VInfo(CCTK_THORNSTRING,
		   "                    interp_grid_delta[%d]=%g",
		   axis, double(interp_grid_delta[axis]));
	CCTK_VInfo(CCTK_THORNSTRING,
		   "                    xyz_[min,max]=[%g,%g]",
		   double(interp_grid_origin[axis]),
		   double(interp_grid_origin[axis] + interp_grid_delta[axis]
						     *(N_interp-1)));
	}

CCTK_VInfo(CCTK_THORNSTRING,
	   "             N_points=%d total interp points",
	   N_points);
}

//******************************************************************************

//
// This function coputes the xyz coordinates of the interpolation points
// (in lexicographic order) and stores them in  interp_x[] ,  interp_y[] ,
// and  interp_z[] .  Note that x, y, and z are all stored, regardless of
// the actual dimensionality of the interpolation grid.
//
void interp_grid::set_interp_xyz()
{
int posn = 0;
	for (int k = 0 ; k < interp_grid_size[2] ; ++k)
	{
	for (int j = 0 ; j < interp_grid_size[1] ; ++j)
	{
	for (int i = 0 ; i < interp_grid_size[0] ; ++i)
	{
	const fp x = interp_grid_origin[0] + i*interp_grid_delta[0];
	const fp y = interp_grid_origin[1] + j*interp_grid_delta[1];
	const fp z = interp_grid_origin[2] + k*interp_grid_delta[2];

	assert(posn < N_points);
	interp_x[posn] = x;
	interp_y[posn] = y;
	interp_z[posn] = z;

	posn++;
	}
	}
	}

assert(posn == N_points);
}

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function randomly permutes the interpolation points.
//
void interp_grid::randomly_permute_interp_points()
{
permutation P(N_points);

P.permute(interp_x);
P.permute(interp_y);
P.permute(interp_z);
}
