// misc.cc -- misc support routines for this thorn
// $Header: /numrelcvs/AEIThorns/TestInterpBase/src/misc.cc,v 1.1 2003/01/17 15:23:13 jthorn Exp $
//
// is_prefix
// split_cstring_into_words
// find_cstring_in_cstring_array
// find_int_in_int_array
//
// delta_product
// decode_derivative_code
// encode_derivative_code
//
// setup_interp_param_table
// get_interp_handle_from_name
// get_interp_handle_from_order
// get_coord_system_handle
// setup_derivs_in_param_table
// setup_output_arrays
//

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "util_String.h"
#include "util_Table.h"
#include "cctk.h"

#include "stdc.h"
#include "fp.h"
#include "util.hh"

#include "TestInterpBase.hh"
#include "data_ptrs.hh"
#include "interp_grid.hh"

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function checks whether or not one string is a prefix of another.
//
// Arguments:
// source = The "big string".
// prefix = The string to check for being a prefix of  source .
//
// Results:
// The function returns true iff prefix is a prefix of source, i.e. if
// source = prefix || tail for some (possibly empty) string tail.
//
bool is_prefix(const char *source, const char *prefix)
{
	for (int i = 0 ; (prefix[i] != '\0'); ++i)
	{
	if (source[i] == '\0')
	   then return false;		// source shorter than prefix
					// ==> not a prefix
	if (source[i] != prefix[i])
	   then return false;		// mismatch ==> not a prefix
	}

return true;
}

//******************************************************************************

//
// This function splits a C-style string into whitespace-separated "words".
// It copies up to  max_N_words  successive words from  cstring[]  into
// newly malloc-ed storage, stores pointers to the them into words[] ,
// and returns the number of words as the function result.
//
// For example:
//	cstring[] = "peanut tofu rice"
// results (assuming  max_N_words >= 3) in the function returning 3 with
//	words[0] = "peanut"
//	words[1] = "tofu"
//	words[1] = "rice"
//
int split_cstring_into_words(const char cstring[],
			     int max_N_words, const char* words[])
{
const char* const whitespace = " \t\n\r\v";
char *const buffer = Util_Strdup(cstring);
if (buffer == NULL)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"split_cstring_into_words(): couldn't duplicate cstring[]!");	/*NOTREACHED*/

char *p = buffer;
p += strspn(p, whitespace);		// skip over leading whitespace
int i = 0;
	while (true)
	{
	if (*p == '\0')
	   then break;					// *** LOOP EXIT ***

	// p --> start of words[i]
	if (i >= max_N_words)
	   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"split_cstring_into_words(): got max_N_words=%d words before end of string!\n"
"                            cstring=\"%s\"\n"
"                            p      =\"%s\""
			   ,
			   max_N_words, buffer, p);		/*NOTREACHED*/
	words[i++] = p;
	p += strcspn(p, whitespace);	// skip over the word
	if (*p == '\0')
	   then break;					// *** LOOP EXIT ***
	*p++ = '\0';
	p += strspn(p, whitespace);	// skip over any further whitespace
	}

return i;
}

//******************************************************************************

//
// This function finds the first i for which  cstring[] == cstring_array[i] ,
// where == means string equality.  If no such  i  exists, this function
// does a  CCTK_VWarn(-1, ...)  (and does not return).
//
int find_cstring_in_cstring_array(const char cstring[],
				  int N, const char* cstring_array[])
{
	for (int i = 0 ; i < N ; ++i)
	{
	if (STRING_EQUAL(cstring, cstring_array[i]))
	   then return i;
	}

CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   find_cstring_in_cstring_array(): didn't find cstring=\"%s\"\n"
"                                    in array of N=%d strings!"
	   ,
	   cstring, N);						/*NOTREACHED*/
}

//******************************************************************************

//
// This function finds the first i for which  x == int_array[i] .
// If no such  i  exists, this function does a  CCTK_VWarn(-1, ...)
// (and does not return).
//
int find_int_in_int_array(int x,
			  int N, const int int_array[])
{
	for (int i = 0 ; i < N ; ++i)
	{
	if (x == int_array[i])
	   then return i;
	}

CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   find_cstring_in_cstring_array(): didn't find x=%d\n"
"                                    in array of N=%d integers!"
	   ,
	   x, N);						/*NOTREACHED*/
}

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function computes the product of the $Delta x$ factors in
// all the axes where we're differentiating.
//
// Arguments:
// derivative_code = (in) The derivative code.  (N.b. this is 1-origin!)
// delta[] = (out) An array of the $\Delta x$ factors.
//
// Results:
// The function returns the number of (not necessarily distinct)
// derivatives taken, i.e. the number of entries stored in
//  derivative_axes[] .
//
fp delta_product(int derivative_code, const fp delta[])
{
// n.b. these are ***1***-origin axes!
int derivative_axes[max_N_derivs];

const int N_derivs = decode_derivative_code(derivative_code, derivative_axes);

switch	(N_derivs)
	{
case 0: return 1.0;
case 1: return delta[derivative_axes[0]-1];
case 2: return delta[derivative_axes[0]-1] * delta[derivative_axes[1]-1];
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "encode_derivative_code(): unknown N_derivs=%d!",
		   N_derivs);					/*NOTREACHED*/
	return -1.0;						// dummy return
	}
}

//******************************************************************************

//
// This function decodes a derivative code into an array of integers
// giving the axes in which we are taking derivatives, e.g.
//	 1 --> [1]
//	11 --> [1, 1]
//	12 --> [1, 2]
//
// Arguments:
// derivative_code = (in) The derivative code to be decoded.
// derivative_axes[] = (out) An array of 1-origin axes along which
//			     derivatives are to be taken.
//
// Results:
// The function returns the number of (not necessarily distinct)
// derivatives taken, i.e. the number of entries stored in
//  derivative_axes[] .
//
int decode_derivative_code(int derivative_code,
			   int derivative_axes[])
{
switch	(derivative_code)
	{
case 0:
	return 0;
case 1: case 2: case 3:
	derivative_axes[0] = derivative_code;
	return 1;
case 11: case 12: case 13:
case 21: case 22: case 23:
case 31: case 32: case 33:
	derivative_axes[0] = derivative_code / 10;
	derivative_axes[1] = derivative_code % 10;
	return 2;
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "decode_derivative_code(): unknown derivative_code=%d!",
		   derivative_code);				/*NOTREACHED*/
	return -1;						// dummy return
	}
}

//******************************************************************************

//
// This function encodes an array of 1-origin axes along which derivatives
// are to be taken, into a derivative code.
//
// Arguments:
// N_derivs = (in) the number of (not necessarily distinct) derivatives
//		   being taken
// derivative_axes[0...N_derivs) = (out) The 1-origin axes along which
//					 the derivatives are to be taken
//
// Results:
// This function returns the derivative code.
//
int encode_derivative_code(int N_derivs, const int derivative_axes[])
{
switch	(N_derivs)
	{
case 0: return 0;
case 1: return derivative_axes[0];
case 2: return 10*derivative_axes[0] + derivative_axes[1];
default:
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "encode_derivative_code(): unknown N_derivs=%d!",
		   N_derivs);					/*NOTREACHED*/
	return -1;						// dummy return
	}
}

//******************************************************************************
//******************************************************************************
//******************************************************************************

//
// This function gets the local interpolator handle from the character-string
// interpolation operator name.
//
int get_interp_handle_from_name(const char interp_operator[])
{
const int interp_handle = CCTK_InterpHandle(interp_operator);
if (interp_handle < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   get_interp_handle(): can't get interpolator handle\n"
"                        for operator \"%s\"!"
		   ,
		   interp_handle);				/*NOTREACHED*/

return interp_handle;
}

//******************************************************************************

//
// This function gets a local interpolator handle from the interpolation
// order, assuming the uniform Cartesian interpolator.
//
int get_interp_handle_from_order(int order)
{
const int max_order = 3;
assert(order >= 0);
assert(order <= max_order);

const char* const names_table[max_order+1] =
	{
	NULL,
	"first-order uniform cartesian",
	"second-order uniform cartesian",
	"third-order uniform cartesian",
	};
const char* name = names_table[order];

CCTK_VInfo(CCTK_THORNSTRING,
	   "using interpolator \"%s\"",
           name);

const int handle = CCTK_InterpHandle(name);
if (handle < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
		   "***** error return %d from CCTK_InterpHandle()!",
		   handle);					/*NOTREACHED*/

return handle;
}

//******************************************************************************

//
// This function creates the interpolator paramter table and sets the
// interpolation order in it.  It returns the table handle.
//
int setup_interp_param_table(const char interp_pars[], int interp_order)
{
int status;

const int param_table_handle = Util_TableCreateFromString(interp_pars);
if (param_table_handle < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"setup_interp_param_table(): error %d from Util_TableCreateFromString()!",
		   param_table_handle);				/*NOTREACHED*/

status = Util_TableSetInt(param_table_handle, interp_order, "order");
if (status < 0)
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"setup_interp_param_table(): error %d from Util_TableSetInt(order)!",
		   status);					/*NOTREACHED*/

return param_table_handle;
}

//******************************************************************************

//
// This function gets the coordinate system handle from the character-string
// coordinate system name.
//
int get_coord_system_handle(const char coord_system[])
{
const int coord_system_handle = CCTK_CoordSystemHandle(coord_system);
if (coord_system_handle < 0)
   then {
	CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   get_coord_system_handle(): can't get coordinate system handle for\n"
"                              for coordinate system \"%s\"!"
		   ,
		   coord_system);				/*NOTREACHED*/
	}

return coord_system_handle;
}

//******************************************************************************

//
// This function sets up the derivatives in the parameter table.
// It sets the table entries
//	CCTK_INT operand_indices[]
//	CCTK_INT operation_codes[]
//
void setup_derivs_in_param_table(int param_table_handle,
				 int N_operations,
				 const int opcodes[max_N_operations])
{
int status1, status2;

const int N_output = N_output_arrays * N_operations;
CCTK_INT operand_indices[max_N_output];
CCTK_INT operation_codes[max_N_output];

int out = 0;
	for (int itype = 0 ; itype < N_output_arrays ; ++itype)
	{
		for (int iop = 0 ; iop < N_operations ; ++iop)
		{
		operand_indices[out] = itype;
		operation_codes[out] = opcodes[iop];
		out++;
		}
	}
assert( out == N_output );

status1 = Util_TableSetIntArray(param_table_handle,
				N_output, operand_indices,
				"operand_indices");
status2 = Util_TableSetIntArray(param_table_handle,
				N_output, operation_codes,
				"operation_codes");
if ((status1 < 0) || (status2 < 0))
   then CCTK_VWarn(-1, __LINE__, __FILE__, CCTK_THORNSTRING,
"\n"
"   setup_derivs_in_param_table(): error when trying to set derivative info\n"
"                                  in parameter table!\n"
"                                  error status1=%d status2=%d"
		   ,
		   status1, status2);				/*NOTREACHED*/
}

//******************************************************************************

//
// This function sets up the arrays
//	CCTK_INT output_array_type_codes[]
//	CCTK_INT output_arrays          []
//
void setup_output_arrays(const interp_grid& interp_grid,
			 int   output_array_type_codes[max_N_output],
			 void* output_arrays          [max_N_output])
{
const int N_output = N_output_arrays * interp_grid.N_operations;
CCTK_VInfo(CCTK_THORNSTRING,
	   "setup_output_arrays(): N_output=%d output arrays",
	   N_output);

const CCTK_INT output_type_codes_temp[N_output_arrays]
   = {
     CCTK_VARIABLE_REAL, CCTK_VARIABLE_REAL4, CCTK_VARIABLE_REAL8,
     CCTK_VARIABLE_COMPLEX, CCTK_VARIABLE_COMPLEX8, CCTK_VARIABLE_COMPLEX16,
     };

int out = 0;
	for (int itype = 0 ; itype < N_output_arrays ; ++itype)
	{
		for (int iop = 0 ; iop < interp_grid.N_operations ; ++iop)
		{
		output_array_type_codes[out] = output_type_codes_temp[itype];
		output_arrays          [out] = interp_grid.output_arrays[iop]
							  .select_ptr(itype);
		out++;
		}
	}
assert( out == N_output );
}
