// util.hh -- low-level utility routines
// $Id: util.hh,v 1.2 2003/01/17 15:23:13 jthorn Exp $
//
// jtutil::how_many_in_range - how many integers are in the interval [low,high]?
// jtutil::{min,max} - min/max templates
// jtutil::abs - absolute value template
// jtutil::pow* - raise floating point value to small-integer power
//
// jtutil::norm<fp> - compute 2-, rms-, and infinity-norms
//

//******************************************************************************

namespace jtutil
	  {

// how many integers are in the closed interval [low,high]
inline int how_many_in_range(int low, int high)
	{ return high - low + 1; }

//
// minimum/maximum templates (valid for both integer and floating-point types)
// FIXME: these are supposed to be in the STL,
//	  but in practice they often aren't (yet) :(
//	  eg the gcc 2.95.2 library doesn't have them :(
//
template <typename num>
  inline num min(num x, num y) { return x < y ? x : y; }
template <typename num>
  inline num max(num x, num y) { return x > y ? x : y; }

//
// absolute value template
// FIXME: this ought to be somewhere in the STL?
//
template <typename fp>
  inline fp abs(fp x) { return (x > 0.0) ? x : -x; }

//
// These functions raise their arguments (presumably floating-point)
// to various small-integer powers.
// FIXME: do we ever use these?
//
template <typename fp>
  inline fp pow2(fp x) { return x*x; }
template <typename fp>
  inline fp pow3(fp x) { return x*x*x; }
template <typename fp>
  inline fp pow4(fp x) { return pow2(pow2(x)); }
template <typename fp>
  inline fp pow5(fp x) { return x * pow4(x); }
template <typename fp>
  inline fp pow6(fp x) { return pow3(pow2(x)); }
template <typename fp>
  inline fp pow7(fp x) { return x * pow6(x); }
template <typename fp>
  inline fp pow8(fp x) { return pow2(pow2(pow2(x))); }

// machine-independent uniform random numbers
// ... low-quality, but adequate for our purposes :)
double drandom();			// in [0,1)
double drandom(double a, double b);	// in [a,b)
int get_random_seed();
int set_random_seed(int new_seed);	// returns old seed

// swap x <--> y, using temp
#define SWAP(x,y,temp)	\
	temp = x;	\
	x = y;		\
	y = temp; /* end macro */

	  };	// close namespace jtutil::

//******************************************************************************

//
// this template class computes 2-norms, rms-norms, and/or infinity-norms
//
// n.b. all members are inline, since some systems (eg MacOS X, Solaris)
//	have trouble coping with out-of-line templates when the .o files
//	are put in a library (as they are for a Cactus build)
//
namespace jtutil
	  {
template <typename fp>
class	norm
	{
public:
	// get norms
	fp two_norm() const		// sqrt(sum x_i^2)
		{ return sqrt(sum2_); }
	fp rms_norm() const		// sqrt(average of x_i^2)
		{ return sqrt(sum2_/fp(N_)); }
	fp infinity_norm() const	// max(|x_i|)
		{ return infinity_norm_; }

	// specify data point
	void data(fp x)
		{
		++N_;
		sum2_ += x*x;
		infinity_norm_ = jtutil::max<fp>(infinity_norm_,
						 jtutil::abs<fp>(x));
		}

	// how many data points have been specified?
	long N_points() const { return N_; }

	// have any data points been specified?
	bool is_empty()    const { return N_ == 0; }
	bool is_nonempty() const { return N_ > 0; }

	// reset ==> just like newly-constructed object
	void reset() { N_ = 0L; sum2_ = 0.0; infinity_norm_ = 0.0; }

	// constructor, destructor
	// ... compiler-generated no-op destructor is ok
	norm()
		: N_(0L), sum2_(0.0), infinity_norm_(0.0)
		{ }

private:
        // we forbid copying and passing by value
        // by declaring the copy constructor and assignment operator
        // private, but never defining them
        norm(const norm &rhs);
	norm& operator=(const norm &rhs);

private:
	long N_;		// # of data points
	fp sum2_;		// sum(data^2)
	fp infinity_norm_;	// max |data|
	};
	  };
