 /*@@
   @file    TriggerTermination.c
   @date    Wednesday 1 December 2004
   @author  Thomas Radke
   @desc
            Implements termination triggers for certain events.
   @enddesc
   @version $Id: TriggerTermination.c,v 1.2 2005/05/17 15:01:03 tradke Exp $
 @@*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pbs_error.h>
#include <pbs_ifl.h>

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Termination.h"


typedef struct
{
  int seconds, minutes, hours, total;
  char *string;
} walltime_t;


int TriggerTermination_CheckPBSwalltime (const cGH *GH)
{
  int server, retval;
  struct batch_status *status;
  const char *jobID;
  struct attrl *attribs;
  walltime_t given, used;
  const struct attrl resources_used_walltime = {NULL,
                                                "resources_used",
                                                "walltime",
                                                ""};
  const struct attrl resources = {&resources_used_walltime,
                                  "Resource_List",
                                  "walltime",
                                  ""};
  DECLARE_CCTK_PARAMETERS


  /* only one processor needs to query the elapsed runtime */
  if (CCTK_MyProc (GH) != 0)
  {
    return (0);
  }

  /* always need a connection to the PBS server first */
  server = pbs_connect (NULL);
  if (server < 0)
  {
    CCTK_WARN (1, "couldn't connect to PBS server");
    return (-1);
  }

  /* get this PBS job's id */
  jobID = getenv ("PBS_JOBID");
  if (! jobID)
  {
    CCTK_WARN (1, "couldn't get PBS_JOBID from environment settings");
    pbs_disconnect (server);
    return (-2);
  }

  /* query the PBS job's attributes */
  status = pbs_statjob (server, jobID, &resources, NULL);
  if (! status)
  {
    CCTK_VWarn (1, __LINE__, __FILE__, CCTK_THORNSTRING,
                "couldn't query status of job '%s'", jobID);
    pbs_disconnect (server);
    return (-3);
  }

  /* extract the walltime info from the set of attributes */
  attribs = status->attribs;
  given.string = used.string = NULL;
  while (attribs)
  {
#ifdef DEBUG
    printf ("%s.%s = %s\n", attribs->name, attribs->resource, attribs->value);
#endif
    if (! (strcmp (attribs->name, "Resource_List") ||
           strcmp (attribs->resource, "walltime")))
    {
      given.string = strdup (attribs->value);
    }
    if (! (strcmp (attribs->name, "resources_used") ||
           strcmp (attribs->resource, "walltime")))
    {
      /* work-around for a bug in PBS */
      if (strlen (attribs->value) > 8)
      {
        attribs->value[8] = 0;
      }
      used.string = strdup (attribs->value);
    }

    attribs = attribs->next;
  }

  /* don't need this anymore */
  pbs_statfree (status);
  pbs_disconnect (server);

  /* now parse the walltime info */
  if (! (given.string && used.string))
  {
    CCTK_WARN (2, "couldn't query given/used walltime information");
    retval = -4;
  }
  else if (sscanf (given.string, "%d:%d:%d", &given.hours,
                   &given.minutes, &given.seconds) != 3)
  {
    CCTK_WARN (2, "couldn't parse given walltime information");
    retval = -5;
  }
  else if (sscanf (used.string, "%d:%d:%d", &used.hours,
                   &used.minutes, &used.seconds) != 3)
  {
    CCTK_WARN (2, "couldn't parse used walltime information");
    retval = -6;
  }
  else
  {
    /* calculate the total elapsed time in minutes */
    used.total = used.hours*60 + used.minutes + used.seconds/60;
    given.total = given.hours*60 + given.minutes + given.seconds/60;

    /* now, is our time over already ? */
    used.total = given.total - used.total;
    if (used.total <= on_remaining_PBS_walltime)
    {
      CCTK_VInfo (CCTK_THORNSTRING, "Remaining wallclock time for your PBS job "
                  "is %d minutes. Triggering termination ...", used.total);
      CCTK_TerminateNext (GH);
    }

    retval = 0;
  }

  free (given.string);
  free (used.string);

  return (retval);
}
