 /*@@
   @file      Register1D.c
   @date      May 2000
   @author    Mark Miller
   @desc 
   Functions for registering and calling EOS_1D routines
   @enddesc 
 @@*/


#include <stdlib.h>

#include "cctk.h"
#include "cctk_FortranString.h"

#include "StoreHandledData.h"

#include "EOS_1D_Base.h"

static cHandledData *methods = NULL;

enum call_methods {none, c, f} ;

typedef struct 
{
  enum call_methods lang;

  CCTK_REAL (*f)(CCTK_REAL *);
  CCTK_REAL (*c)(CCTK_REAL);
} func_t;

typedef struct 
{
  func_t pressfromrho;
  func_t pressfromeps;
  func_t epsfrompress;
  func_t epsfromrho;
  func_t rhofrompress;
  func_t rhofromeps;
} method_t;

 /*@@
   @routine    EOS_1D_RegisterMethod
   @date       May 2000
   @author     Mark Miller
   @desc 
   Registers an EOS_1D method
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/
int EOS_1D_RegisterMethod(const char *name)
{
  int handle;

  method_t *method;

  method = (method_t *)malloc(sizeof(method_t));

  if(method)
  {
    method->pressfromrho.lang = none;
    method->pressfromeps.lang = none;
    method->epsfrompress.lang = none;
    method->epsfromrho.lang = none;
    method->rhofrompress.lang = none;
    method->rhofromeps.lang = none;
  
    handle = Util_NewHandle(&methods, name, (void *)method);
  }
  else
  {
    handle = -1;
  }

  return handle;
}

void CCTK_FCALL CCTK_FNAME(EOS_1D_RegisterMethod)(int *handle, ONE_FORTSTRING_ARG)
{
  ONE_FORTSTRING_CREATE(name)

  *handle = EOS_1D_RegisterMethod(name);

  free(name);
}


 /*@@
   @routine    EOS_1D_Handle
   @date       May 2000
   @author     Mark Miller
   @desc 
   Gets the handle associated with an EOS_1D method
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/
int EOS_1D_Handle(const char *name)
{
  int handle;
  method_t method;

  handle = Util_GetHandle(methods, name, (void *)&method);

  return handle;
}

void CCTK_FCALL CCTK_FNAME(EOS_1D_Handle)(int *handle, ONE_FORTSTRING_ARG)
{
  ONE_FORTSTRING_CREATE(name)

  *handle = EOS_1D_Handle(name);

  free(name);
}

#define REGISTER_FUNCTION(x) \
int EOS_1D_Register ## x (int handle, CCTK_REAL (*func)(CCTK_REAL)) \
{                                                                           \
  int retval;                                                               \
  method_t *method;                                                         \
                                                                            \
  method = (method_t *)Util_GetHandledData(methods, handle);                \
                                                                            \
  if(method)                                                                \
  {                                                                         \
    method->x.lang = c;                                                     \
    method->x.c = func;                                                     \
    retval = 0;                                                             \
  }                                                                         \
  else                                                                      \
  {                                                                         \
    retval = 1;                                                             \
  }                                                                         \
                                                                            \
  return retval;                                                            \
}

/* Functions to register C EOS functions */
REGISTER_FUNCTION(pressfromrho)
REGISTER_FUNCTION(pressfromeps)
REGISTER_FUNCTION(epsfrompress)
REGISTER_FUNCTION(epsfromrho)
REGISTER_FUNCTION(rhofrompress)
REGISTER_FUNCTION(rhofromeps)


#define REGISTER_FORTRAN_FUNCTION(x)\
(int *retval, int *handle, CCTK_REAL (*func)(CCTK_REAL *)) \
{                                                                      \
  method_t *method;                                                    \
                                                                       \
  method = (method_t *)Util_GetHandledData(methods, *handle);          \
                                                                       \
  if(method)                                                           \
  {                                                                    \
    method->x.lang = f;                                                \
    method->x.f = func;                                                \
    *retval = 0;                                                       \
  }                                                                    \
  else                                                                 \
  {                                                                    \
    *retval = 1;                                                       \
  }                                                                    \
                                                                       \
}

/* Functions to register Fortran EOS_1D functions */
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerpressfromrho) REGISTER_FORTRAN_FUNCTION(pressfromrho)
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerpressfromeps) REGISTER_FORTRAN_FUNCTION(pressfromeps)
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerepsfrompress) REGISTER_FORTRAN_FUNCTION(epsfrompress)
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerepsfromrho) REGISTER_FORTRAN_FUNCTION(epsfromrho)
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerrhofrompress) REGISTER_FORTRAN_FUNCTION(rhofrompress)
void CCTK_FCALL CCTK_FNAME(EOS_1D_Registerrhofromeps) REGISTER_FORTRAN_FUNCTION(rhofromeps)
#define CALL_FUNC(x) \
CCTK_REAL EOS_1D_ ## x (int handle, CCTK_REAL a) \
{                                                                           \
  CCTK_REAL retval;                                                         \
  method_t *method;                                                         \
                                                                            \
  method = (method_t *)Util_GetHandledData(methods, handle);                \
                                                                            \
  if(method && method->x.lang != none)                                      \
  {                                                                         \
    if(method->x.lang == c)                                                 \
    {                                                                       \
      retval = method->x.c(a);                                            \
    }                                                                       \
    else                                                                    \
    {                                                                       \
      retval = method->x.f(&a);                                         \
    }                                                                       \
  }                                                                         \
  else                                                                      \
  {                                                                         \
    retval = -1;                                                            \
  }                                                                         \
  return retval;                                                            \
}

/* Functions to call EOS_1D functions from C */
CALL_FUNC(pressfromrho)
CALL_FUNC(pressfromeps)
CALL_FUNC(epsfrompress)
CALL_FUNC(epsfromrho)
CALL_FUNC(rhofrompress)
CALL_FUNC(rhofromeps)

#define CALL_FORTRAN_FUNC(x) \
(int *handle, CCTK_REAL *a)                                   \
{                                                                           \
  CCTK_REAL retval;                                                         \
  method_t *method;                                                         \
                                                                            \
  method = (method_t *)Util_GetHandledData(methods, *handle);               \
                                                                            \
  if(method && method->x.lang != none)                                      \
  {                                                                         \
    if(method->x.lang == c)                                                 \
    {                                                                       \
      retval = method->x.c(*a);                                          \
    }                                                                       \
    else                                                                    \
    {                                                                       \
      retval = method->x.f(a);                                           \
    }                                                                       \
  }                                                                         \
  else                                                                      \
  {                                                                         \
    retval = -1;                                                            \
  }                                                                         \
  return retval;                                                            \
}

/* Functions to call EOS_1D functions from Fortran*/
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_pressfromrho) CALL_FORTRAN_FUNC(pressfromrho)
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_pressfromeps) CALL_FORTRAN_FUNC(pressfromeps)
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_epsfrompress) CALL_FORTRAN_FUNC(epsfrompress)
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_epsfromrho) CALL_FORTRAN_FUNC(epsfromrho)
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_rhofrompress) CALL_FORTRAN_FUNC(rhofrompress)
CCTK_REAL CCTK_FCALL CCTK_FNAME(EOS_1D_rhofromeps) CALL_FORTRAN_FUNC(rhofromeps)

