#include "definitions.h"
#include "MPIFS.h"
#include "Array.h"
#include "ArrayGroup.h"

extern MPIFS *MPIFS_global_obj;

ArrayGroup::ArrayGroup()
{
  do_init();
}

ArrayGroup::ArrayGroup(char *name)
{
  do_init();
  name_ = (char *)malloc(strlen(name)+1);
  strcpy(name_, name);
}


/* Function to initialize the state of the newly created object */
void ArrayGroup::do_init()
{ 
  num_of_arrays_ = 0;
  list_ = new List();
  io_strategy_ = SIMPLE_IO;
  interleaved_ = NO;
  common_layouts_ = NO;
  common_layout_rank_ = 0;
  compute_layout_ = NULL;
  compute_distribution_ = NULL;
  io_layout_ = NULL;
  io_distribution_ = NULL;
  group_io_count_ = 0;
  read_io_count_ =0;
  checkpoint_count_ = 1;
  simulate_ = NO;
  verify_ = NO;
  name_ = NULL;
}

void ArrayGroup::clear()
{
  if (name_) free(name_);
  if (compute_layout_ != NULL) delete compute_layout_;
  if (compute_distribution_ != NULL) delete compute_distribution_;
  if (io_layout_ != NULL) delete io_layout_;
  if (io_distribution_ != NULL) delete io_distribution_;
  if (list_) delete list_;
  name_ = NULL;
  compute_layout_ = NULL; compute_distribution_ = NULL;
  io_layout_ = NULL; io_distribution_ = NULL;
  list_ = new List();
}

/* Destructor function - Note  that we don't have to delete the *
 * arrays in the arraygroup over here. The arrays are deleted   *
 * by the user							*/
ArrayGroup::~ArrayGroup()
{
  if (name_) free(name_);
  if (compute_layout_ != NULL) delete compute_layout_;
  if (compute_distribution_ != NULL) delete compute_distribution_;
  if (io_layout_ != NULL) delete io_layout_;
  if (io_distribution_ != NULL) delete io_distribution_;
  if (list_) delete list_;
  name_ = NULL;
  compute_layout_ = NULL;
  io_layout_ = NULL;
  compute_distribution_ = NULL;
  io_distribution_ = NULL;
  list_ = NULL;
}

/* Function to delete the arrays in the arraygroup. This is used *
 * on the io node side to delete the arrays after the collective *
 * io operation has been completed. On the compute node side, the*
 * user explicitly deletes the arrays                            */
void ArrayGroup::delete_arrays()
{
   Cell* list_ptr = (list_ != NULL? list_->head_: NULL);
   Array* array_ptr;
   
   while(list_ptr)
   {
	array_ptr = (Array *) list_ptr->item();
        delete array_ptr;
        list_ptr = list_ptr->next();
   }
   if (list_) delete list_;
   list_ = NULL;
}

/* Assign id numbers to the arrays in the arraygroup. This function *
 * must be called at the start of each collective i/o operation.    */
void ArrayGroup::assign_id()
{
   Cell* list_ptr = (list_ != NULL? list_->head_: NULL);
   Array* array_ptr;
   int i=0;
   
   while(list_ptr)
   {
	array_ptr = (Array *) list_ptr->item();
        array_ptr->set_array_id(i);
#ifdef DEBUG
	printf("Assigned Id %d\n", i);
#endif
	i++;
        list_ptr = list_ptr->next();
   }
}

/* Insert a new array. Before inserting the array, check to  *
 * see if it has a common layout with the rest of the arrays */
void ArrayGroup::insert(Array *new_array)
{
   num_of_arrays_++;

   /* Has common layouts since this is the first array */
   if (num_of_arrays_ == 1)
   {
        common_layout_rank_ = new_array->rank();
	compute_layout_ = new ArrayLayout(new_array->layout(COMPUTE_NODE));
	compute_distribution_ = copy_distribution(common_layout_rank_, 
				new_array->distribution(COMPUTE_NODE));
	io_layout_ = new ArrayLayout(new_array->layout(IO_NODE));
        io_distribution_ = copy_distribution(common_layout_rank_,
				new_array->distribution(IO_NODE));
	common_layouts_ = YES;
    }
    else if (common_layouts_)
    {
         /* check to see if the array has the same layouts/dist */
         if ((common_layout_rank_ == new_array->rank()) &&
	     (compute_layout_->equal(new_array->layout(COMPUTE_NODE))) &&
             (io_layout_->equal(new_array->layout(IO_NODE))) &&
             (equal_distribution(common_layout_rank_, compute_distribution_,
				 new_array->distribution(COMPUTE_NODE))) &&
	     (equal_distribution(common_layout_rank_, io_distribution_,
				 new_array->distribution(IO_NODE))))
	 {
		common_layouts_ = YES;
	 }
         else
	 {
		common_layouts_ = NO;
		if (io_layout_) delete io_layout_;
  		if (compute_layout_) delete compute_layout_;
		io_layout_ = compute_layout_ = NULL;
		if (io_distribution_) free(io_distribution_);
		if (compute_distribution_) free(compute_distribution_);
		io_distribution_ = compute_distribution_ = NULL;
	  }
    }	

    list_->add_last(new_array);
}

/* This function is called on the compute node side at the start *
 * of each collective io operation. The information is packed    *
 * into an integer buffer. An assumption is made that the a buf  *
 * of 100 ints is sufficent for each array.                      */
void  ArrayGroup::pack(int** schema, int* schema_size)
{
  int *ptr, *head;
  int i, len;
  
  /* Assuming that schema size of Array is ~= 100 elts */
  ptr = (int *) malloc(sizeof(int)*100*(num_of_arrays_+1));
  head = ptr;
  
  /* Round about way and space inefficent way of storing a name */
  *ptr++ = io_strategy_;
  len = strlen(name_);
  *ptr++ = len;
  for(i=0; i<len;i++)
	*ptr++ = (int) name_[i];

  *ptr++ = num_of_arrays_;
  *ptr++ = (int) interleaved_;
  *ptr++ = (int) simulate_;
  *ptr++ = (int) verify_;
  *ptr++ = (int) common_layouts_;
  if (common_layouts_)
  {
	*ptr++ = common_layout_rank_ ;
	compute_layout_->pack(&ptr);
        pack_distribution(&ptr, common_layout_rank_, compute_distribution_);
	io_layout_->pack(&ptr);
	pack_distribution(&ptr, common_layout_rank_, io_distribution_);
  }
  *ptr++ = group_io_count_;
  *ptr++ = checkpoint_count_;
  *ptr++ = op_type_;

  pack_arrays(&ptr, common_layouts_);
  
  *schema_size = (int)(ptr - head);
  *schema = head;
}	

/* This function is called on the I/O node side. After receiving *
 * the collective io schema, the information is unpacked. The    *
 * arrays are unpacked seperately via a another function call    */	
void ArrayGroup::unpack(int **schema_ptr)
{
   int *ptr = *schema_ptr;
   int len;
 
   /* Unpack the name */
   io_strategy_ = *ptr++;
   len = *ptr++;
   name_ = (char *) malloc(len+1);
   for(int i=0; i< len; i++)
	name_[i] = (char) *ptr++;
   name_[len] = '\0';

   num_of_arrays_ = *ptr++;
   interleaved_ = (Boolean) *ptr++;
   simulate_ = (Boolean) *ptr++;
   verify_   = (Boolean) *ptr++;
   common_layouts_ = (Boolean) *ptr++;
   if (common_layouts_)
   {  
	common_layout_rank_ = *ptr++;
        compute_layout_ = new ArrayLayout(&ptr);
        compute_distribution_ = new_distribution(&ptr, common_layout_rank_);
	io_layout_ = new ArrayLayout(&ptr);
	io_distribution_ = new_distribution(&ptr, common_layout_rank_);
   }
   else
   {
	common_layout_rank_ = 0;
        compute_layout_ = io_layout_ = NULL;
        compute_distribution_ = io_distribution_ = NULL;
    }
    group_io_count_ = *ptr++;
    checkpoint_count_ = *ptr++;
    op_type_ = *ptr++;

    /* Arrays are being unpacked seperately */
    *schema_ptr = ptr;
}

void ArrayGroup::unpack_arrays(int **schema_buf)
{
  Array *array;
  int i, *ptr = *schema_buf;
  
  if (common_layouts_){
    for(i=0;i<num_of_arrays_;i++){
      array = new Array(&ptr, common_layouts_, compute_layout_,
			compute_distribution_, io_layout_, 
			io_distribution_);
      list_->add_last(array);
    }
  } else {
    for(i=0;i<num_of_arrays_;i++){
      array = new Array(&ptr, common_layouts_);
      list_->add_last(array);
    }
  }
  *schema_buf = ptr;
}


/* The collective io operation to write out the arrays. */
void ArrayGroup::timestep()
{
   int *schema, schema_size;
   int node_type = MPIFS_global_obj->node_type();
  /* Assign id numbers to each array */
  assign_id();

   op_type_ = TIMESTEP;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size, this);
   }
   /* Commented out for testing purposes */
//   group_io_count_++;
}

/* The collective io operation to write out the arrays. */
void ArrayGroup::general_write()
{
   int *schema, schema_size;
   int node_type = MPIFS_global_obj->node_type();
  /* Assign id numbers to each array */
  assign_id();

   op_type_ = GENERAL_WRITE;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size, this);
   }
   /* Commented out for testing purposes */
//   group_io_count_++;
}


/* The collective io operation to write out the arrays. */
void ArrayGroup::checkpoint()
{
   int *schema, schema_size;
   int node_type = MPIFS_global_obj->node_type();
  /* Assign id numbers to each array */
  assign_id();

   if (checkpoint_count_ == 0) 
     checkpoint_count_ = 1;
   else 
     checkpoint_count_ = 0;

   op_type_ = CHECKPOINT;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size, this);
   }
   
}



/* The collective io operation to read in the arrays from a *
 * checkpoint file. Currently (for testing purposes) this   *
 * does not happen.                                         */
void ArrayGroup::restart()
{
  int *schema, schema_size;
  int node_type = MPIFS_global_obj->node_type();

    /* Assign id numbers to each array */
    assign_id();

    op_type_ = RESTART;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size , this);
   }
}


void ArrayGroup::read_timestep()
{
  int *schema, schema_size;
  int node_type = MPIFS_global_obj->node_type();

    /* Assign id numbers to each array */
    assign_id();

    op_type_ = READ_TIMESTEP;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size , this);
   }
   /* Commented out for testing purposes */
//  read_io_count_++;
}


void ArrayGroup::general_read()
{
  int *schema, schema_size;
  int node_type = MPIFS_global_obj->node_type();

    /* Assign id numbers to each array */
    assign_id();

    op_type_ = GENERAL_READ;
   if (node_type == COMPUTE_NODE){
     MPIFS_global_obj->send_group_schema(this);
     MPIFS_global_obj->compute_node_io_loop(this);
   }
   else if (node_type == PART_TIME_COMPUTE)
     MPIFS_global_obj->compute_node_io_loop(this);
   else {
     pack(&schema, &schema_size);
     MPIFS_global_obj->part_time_io_node_loop(schema, schema_size , this);
   }
   /* Commented out for testing purposes */
//  read_io_count_++;
}



  
/* Given an array id find the array object in the array group *
 * The code caches the previous search value and starts the   *
 * search from there. This helps especially in the case of    *
 * when the arrays are accessed sequentially                  */
Array* ArrayGroup::find_array(int array_id)
{
  Cell* list_ptr = (list_ != NULL ? 
		(list_->old_search_val_ != NULL ? list_->old_search_val_ :
			 list_->head_)  
		: NULL);
  Array* array_ptr;
  
  while(list_ptr)
  { 
	array_ptr = (Array *) list_ptr->item();
	if (array_ptr->array_id() == array_id)
	{
	   list_->old_search_val_ = list_ptr->next();
	    return array_ptr;
 	}
	list_ptr = list_ptr->next();
  }
  
  list_ptr = list_->head_;
  while (list_->old_search_val_ && (list_ptr != list_->old_search_val_))
  { 
	array_ptr = (Array *) list_ptr->item();
	if (array_ptr->array_id() == array_id)
	{
	   list_->old_search_val_ = list_ptr->next();
	    return array_ptr;
 	}
	list_ptr = list_ptr->next();
  }
  return NULL;
}

/* Pack the arrays into an integer schema buffer. Assumes that the * 
 * data is already allocated.                                      */
void ArrayGroup::pack_arrays(int **schema_buf, Boolean common_layouts)
{
  Cell* list_ptr = (list_ != NULL ? list_->head_ : NULL);
  Array* array_ptr;
  
  while(list_ptr)
  { 
	array_ptr = (Array *) list_ptr->item();
        array_ptr->pack(schema_buf, common_layouts);
	list_ptr = list_ptr->next();
  }
  
}


int ArrayGroup::op_type(){return op_type_;}

void ArrayGroup::set_simulate(){simulate_ = YES;}

void ArrayGroup::reset_simulate(){simulate_ = NO;}

void ArrayGroup::set_simulate_mode(){simulate_ = YES;}

void ArrayGroup::reset_simulate_mode(){simulate_ = NO;}

Boolean ArrayGroup::simulate(){return simulate_;}


void ArrayGroup::set_verify(){verify_ = YES;}

void ArrayGroup::reset_verify(){verify_ = NO;}

void ArrayGroup::set_verify_mode(){verify_ = YES;}

void ArrayGroup::reset_verify_mode(){verify_ = NO;}

Boolean ArrayGroup::verify(){return verify_;}

/* This function is called on each compute node side and after all the
 * arrays have been assigned an id. For each array in the arraygroup,
 * the function computes the total number of bytes on the compute node
 */
void ArrayGroup::init_array_info(int *num_arrays, int **array_bytes_to_go)
{
  int *tmp_buf = (int *) malloc(sizeof(int)*num_of_arrays_);
  *array_bytes_to_go = tmp_buf;
  *num_arrays = num_of_arrays_;

  for(int i=0; i< num_of_arrays_; i++)
    tmp_buf[i] = find_array(i)->array_info();
}
 
int ArrayGroup::io_strategy(){
  return io_strategy_;
}

void ArrayGroup::set_io_strategy(int new_strategy){
  io_strategy_ = new_strategy;
}

int ArrayGroup::num_of_arrays()
{
  return num_of_arrays_;
}
