#include "definitions.h"
#include "MPIFS.h"
#include "Collective_IO.h"
#include "Simple_IO.h"
#include "Array.h"
#include "Chunk.h"
#include "message.h"
#define Max_Open_Files 1000


#include "CactusExternal/FlexIO/src/Arch.h"

extern "C" {
  IOFile IEEEopen(char *,char *);
  int IOclose(IOFile);
}

int BRANCHING_FACTOR=8;
int SUBCHUNK_SIZE = 1048576;
MPIFS* MPIFS_global_obj;

/* Notes,Hacks,Assumptions:
 *  -   io_app_info_ and app_info_ point to the same object on the I/O
 *      for the regular. This hack is to allow for code re-use in the 
 *      part-time I/O case.
 */

/* Constructor for the normal case - i.e no part-time I/O nodes */
MPIFS::MPIFS(int node_type, int app_num, int relative_rank, int app_size,
	int *world_ranks)
{
#ifdef DEBUG
  int abs_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &abs_rank);
  printf("%d:node_type=%d, app_num=%d , relative_rank=%d, app_size=%d\n",
	 abs_rank, node_type, app_num, relative_rank, app_size);
#endif
  do_init(node_type, app_num, relative_rank, app_size, world_ranks);
}

/* Constructor for the normal case - i.e no part-time I/O nodes */
MPIFS::MPIFS(int node_type, int app_num, int relative_rank, int app_size,
	int *world_ranks, Boolean shared_flag)
{
  do_init(node_type, app_num, relative_rank, app_size, world_ranks);
}


/* Constructor for part-time I/O nodes */
MPIFS::MPIFS(int node_type, int comp_rank, int comp_size, int *comp_world_ranks,
	int io_rank, int io_size, int *io_world_ranks)
{
#ifdef DEBUG
  int abs_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &abs_rank);
  printf("%d:node_type=%d, comp_rank=%d, comp_size=%d io_rank=%d io_size=%d\n",
	 abs_rank,node_type,  comp_rank, comp_size, io_rank, io_size);
#endif
  do_init(node_type, comp_rank, comp_size, comp_world_ranks, io_rank, io_size,
	io_world_ranks);
}

/* Initialize the file system object for the regular case (i.e no part-time I/O nodes) */
void MPIFS::do_init(int node_type, int app_num, int relative_rank, int app_size,
		int *world_ranks)
{
  MPI_Status status;
  int tag, tmp;

  MPIFS_global_obj = this;

  if ((node_type  != IO_NODE) && (node_type != COMPUTE_NODE))
  {
	printf("Invalid node type in MPIFS::do_init(int,int,int,int,int*)\n");
	exit(10);
  }

  /* Initialize the state */
  node_type_    = node_type;

  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank_);

  app_num_	= app_num;
  app_rank_	= relative_rank;
  app_size_ 	= app_size;
  app_info_	= new App_Info(app_num, app_size, world_ranks);
  comm_		= (MPI_Comm *) malloc(sizeof(MPI_Comm));
  MPI_Comm_split(MPI_COMM_WORLD, app_num_, app_rank_, comm_);
  num_apps_ = num_apps_alive_ = global_barrier_count_ = 0;
  current_max_app_num_ = -1;
  compute_apps_info_ = NULL;
  mem_buf_ = NULL;
  num_open_files_ = 0;
  for (int i=0; i<Max_Open_Files; i++) open_file_names_[i] = NULL;

  /* Part-time i/o stuff - unneeded in this case */
  io_app_num_ = io_app_rank_ = io_app_size_ = -1;
  io_comm_ = NULL; 
  io_app_info_ = NULL;

 /* Broadcast the rank of the master I/O node. The strategy is *
  * to send the info to node 0 and have it broadcast it        */
  master_io_node_ = -1;
  if (node_type_ == IO_NODE)
     master_io_node_ = app_info_->get_master();
  
  if (world_rank_ == 0)
  {
    if (master_io_node_ != 0)
      receive_message((void *)&master_io_node_, 1, MPI_INT, MPI_ANY_SOURCE,
		      1000, MPI_COMM_WORLD, &status);
  }
  else if (world_rank_ == master_io_node_)
  {
    send_message((void *)&master_io_node_, 1 , MPI_INT, 0, 1000, 
		 MPI_COMM_WORLD);
  }
  MPI_Bcast((void *)&master_io_node_, 1, MPI_INT, 0, MPI_COMM_WORLD);
  /* Now all nodes know who the master I/O node is */
  
  if (node_type_ == IO_NODE)
  {
        mem_buf_size_ = 2*SUBCHUNK_SIZE;  /* Factor of 2 - just to be safe */
	mem_buf_ = (char *) malloc(sizeof(char)*mem_buf_size_);
 
	io_app_info_ = app_info_;
	io_node_main_loop();
  }
  else if (node_type_ == COMPUTE_NODE)
  {
	/* Send the info about the compute application to the *
         * master I/O node (only master I/O node has to do it) */
	if (am_master_compute_node())
        {	
		tag = app_num_*100+APP_INFO*10+SPECIAL;
		send_message((void *)world_ranks, app_size, MPI_INT, 
			master_io_node_, tag, MPI_COMM_WORLD);
		receive_message((void *)&tmp, 1, MPI_INT, master_io_node_,
			tag, MPI_COMM_WORLD, &status);
	}
	MPI_Barrier(*comm_);

	/* Create an intra-comm with the I/O nodes. This stuff is *
	 * used only for implemneting barriers etc 		  */
	MPI_Comm *inter_comm = (MPI_Comm *) malloc(sizeof(MPI_Comm));
	MPI_Comm *intra_com = (MPI_Comm *) malloc(sizeof(MPI_Comm));
	MPI_Intercomm_create(*comm_, 0, MPI_COMM_WORLD,
			 master_io_node_, app_num, inter_comm);
	MPI_Intercomm_merge(*inter_comm, 1, intra_com);
  	app_info_->set_intra_comm(intra_com);
   }
   else
   {
	printf("Unsupported node type\n");
	exit(1);
   }
}

void MPIFS::do_init(int node_type, int comp_rank, int comp_size, int *comp_world_ranks,
		    int io_rank, int io_size, int *io_world_ranks)
{
  MPI_Group global_group, comp_group, io_group;
  

  MPIFS_global_obj = this;
  if ((node_type != PART_TIME_COMPUTE) && (node_type != PART_TIME_IO))
  {
	printf("Incorrect initialization for node_type %d\n", node_type);
	exit(10);
  }
  
  /* Convention that logical I/O app gets app_num=0 and compute app get app_num=1 */
  node_type_ = node_type;

  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank_);

  app_num_ = 1;
  app_rank_ = comp_rank;
  app_size_ = comp_size;
  master_io_node_ = io_world_ranks[0];
  comm_ = (MPI_Comm *)malloc(sizeof(MPI_Comm));
  MPI_Comm_group(MPI_COMM_WORLD, &global_group);
  MPI_Group_incl(global_group, comp_size, comp_world_ranks, &comp_group);
  MPI_Comm_create(MPI_COMM_WORLD, comp_group, comm_);
  app_info_ = new App_Info(1, app_size_, comp_world_ranks);

  num_apps_ = 1;
  num_apps_alive_ =1;
  current_max_app_num_=-1;
  global_barrier_count_ =0;
  compute_apps_info_ = NULL;
  mem_buf_ = NULL;

  io_app_num_  = 0;  /* By convention */
  io_app_rank_ = io_rank;
  io_app_size_ = io_size;
  io_app_info_ = NULL;
  
  /* Everyone in MPI_COMM_WORLD must make this call */
  io_comm_ = (MPI_Comm*) malloc(sizeof(MPI_Comm));
  MPI_Comm_group(MPI_COMM_WORLD, &global_group);
  MPI_Group_incl(global_group, io_size, io_world_ranks, &io_group);
  MPI_Comm_create(MPI_COMM_WORLD, io_group, io_comm_);

  if (node_type_ == PART_TIME_IO)
  {
        mem_buf_size_ = 2*SUBCHUNK_SIZE;  /* Factor of 2 - just to be safe */
	mem_buf_ = (char *) malloc(sizeof(char)*mem_buf_size_);
	App_Info *app = new App_Info(1, comp_size, comp_world_ranks);
	io_app_info_ = new App_Info(0, io_app_size_, io_world_ranks);
	insert_compute_app(1, app);
  }
  
}
		
MPIFS::~MPIFS()
{
   if (node_type_ == COMPUTE_NODE)
   {
	if (am_master_compute_node())
		send_message((void *)&app_num_, 1, MPI_INT, master_io_node_,
				QUIT, MPI_COMM_WORLD);
   }


   if (app_info_) delete app_info_;
   if (mem_buf_) free(mem_buf_);

   for (int i=0; i<num_open_files_; i++) {
     free(open_file_names_[i]);
     IOclose((IOFile)open_file_ptrs_[i]);
   }
     
   if (compute_apps_info_){
	 for(int i=0; i<=current_max_app_num_; i++)
		if (compute_apps_info_[i]) delete compute_apps_info_[i];
	 delete compute_apps_info_;
   }

   app_info_  = io_app_info_ = NULL;
   compute_apps_info_ =NULL;
   if (comm_) 
   {
	MPI_Comm_free(comm_);
	free(comm_);
	comm_= NULL;
   }
}


Boolean MPIFS::am_compute_node()
{
  if (node_type_ == IO_NODE) return NO;
  else return YES;
}


Boolean MPIFS::am_io_node()
{
  if ((node_type_ == IO_NODE) || (node_type_ == PART_TIME_IO))
	return YES;
  else  return NO;
}

Boolean MPIFS::am_master_io_node()
{
   if (am_io_node() && (world_rank_ == io_app_info_->get_master()))
	return YES;
   else return NO;
}

Boolean MPIFS::am_master_compute_node()
{
   if (am_compute_node() && (world_rank_ == app_info_->get_master()))
 	return YES;
   else return NO;
}





/* This is a highly restricted version of a broadcast function. The broadcast
 * is performed using tree-structured  communication, starting at relative 
 * node 0. The broadcast is implemented using tree-structured communication.
 */
void MPIFS::Broadcast(int node_type, void *buf, int count, 
	              MPI_Datatype datatype, int tag)
{
  App_Info *app;
  int my_rank = app_rank_, size;
  int low, high, i, dest;
 
  if (node_type == COMPUTE_NODE) app = app_info_;
  else app = io_app_info_;
  size = app->app_size();
 
  low = my_rank*BRANCHING_FACTOR+1;
  high = (my_rank+1)*BRANCHING_FACTOR+1;
  i = low;
 
#ifdef DEBUG
  printf("%d: Bcast low %d high %d size=%d\n", world_rank_,low, high,size);
#endif
  /* Can use asynchronous sends */
  while ((i<size) && (i<high)) {
    dest = app->world_rank(i);
    send_message(buf, count, datatype, dest, tag, MPI_COMM_WORLD);
    i++;
  }
}
	

void MPIFS::io_node_main_loop()
{
  MPI_Status  msg_status;
  int         msg_tag, msg_code, msg_src;

   while(1){
     wait_for_next_message(&msg_code, &msg_src, &msg_tag, &msg_status);
     switch(msg_code){
     
     case SPECIAL:
       process_io_special_message(msg_src, msg_tag, &msg_status);
       break;
       
     case ARRAYGROUP_SCHEMA:
       start_collective_io(msg_src, msg_tag, &msg_status);
       break;
     
     case ATTRIBUTE_SCHEMA:
       start_attribute_io(msg_src, msg_tag, &msg_status);
       break;
       
     case QUIT:
       if (received_quit_message(msg_src, msg_tag, &msg_status))
	 return;
       break;
       
     default:
       printf("Error - did not understand message code %d from %d with  tag %d\n",
	      msg_code, msg_src, msg_tag);
       break;
     }
   }
}
  
void MPIFS::wait_for_next_message(int *msg_code, int *msg_src, int *msg_tag, 
			MPI_Status *msg_status)
{
   MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, msg_status);
   *msg_tag = msg_status->MPI_TAG;
   *msg_src = msg_status->MPI_SOURCE;
   *msg_code = *msg_tag % 10;
   return;
}

void MPIFS::process_io_special_message(int msg_src, int msg_tag, 
					MPI_Status *status)
{
   int msg_code = (msg_tag / 10) % 10 ;  /* Extract the second digit */
   int msg_len, app_num, *world_ranks;
   MPI_Comm *inter_comm, *intra_com;
   App_Info *app;

   switch (msg_code)
   {
	case APP_INFO:
		MPI_Get_count(status, MPI_INT, &msg_len);
		world_ranks = (int *) malloc(sizeof(int)*msg_len);
		receive_message((void *)world_ranks, msg_len, MPI_INT, msg_src,
				msg_tag, MPI_COMM_WORLD, status);
		Broadcast(IO_NODE, world_ranks, msg_len, MPI_INT, msg_tag);
		app_num = msg_tag / 100;
		if (am_master_io_node())
 			send_message((void *) &app_num, 1, MPI_INT, msg_src, 
				msg_tag, MPI_COMM_WORLD);
		app = new App_Info(app_num, msg_len, world_ranks);
		insert_compute_app(app_num, app);
                /* Create the Intra communicator */
		inter_comm = (MPI_Comm *) malloc(sizeof(MPI_Comm));
		intra_com = (MPI_Comm *) malloc(sizeof(MPI_Comm));
		MPI_Intercomm_create(*comm_, 0, MPI_COMM_WORLD,
				 world_ranks[0], (msg_tag/100), inter_comm);
		MPI_Intercomm_merge(*inter_comm, 0, intra_com);
		app->set_intra_comm(intra_com);
		free(world_ranks);
		break;

	case APP_BARRIER:
		/* This should be used very carefully when there are more than
		 * 1 compute application running
		 */
		receive_message((void *)&app_num,1, MPI_INT, msg_src, msg_tag,
			  MPI_COMM_WORLD, status);
		Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
		MPI_Barrier(*(find_compute_app(app_num)->intra_comm()));
		break;

	case GLOBAL_BARRIER:
		receive_message((void *)&app_num,1, MPI_INT, msg_src, msg_tag,
			  MPI_COMM_WORLD, status);
		Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
		global_barrier_count_++;
		if (global_barrier_count_ == num_apps_alive_)
		{
			MPI_Barrier(MPI_COMM_WORLD);
			global_barrier_count_ = 0;
		}
		break;

       case CLEANFILES:	
		receive_message((void *)&app_num,1, MPI_INT, msg_src, msg_tag,
			  MPI_COMM_WORLD,status);
		Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
		cleanfiles(app_num);
		break;

       case FLUSHFILES:
		receive_message((void *)&app_num,1, MPI_INT, msg_src, msg_tag,
			  MPI_COMM_WORLD, status);
		Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
		if (world_rank_ == 0) flushfiles(app_num);
		
		break;
				
  
	case CREATEFILES:
		receive_message((void *)&app_num,1, MPI_INT, msg_src, msg_tag,
			  MPI_COMM_WORLD, status);
		Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
		createfiles(app_num);
		break;

	default:
		printf("Unknown message code %d in proces_io_special\n", msg_code);
		break;
    }
}


void MPIFS::cleanfiles(int app_num)
{
  char buf[64];
  sprintf(buf, "rm -rf %s%d\n", FILEPREFIX, app_num);
  if (world_rank_ == 0)  system(buf);
}

void MPIFS::createfiles(int app_num)
{
  char buf[64];
  sprintf(buf, "mkdir %s%d\n", FILEPREFIX, app_num);
  if (world_rank_ == 0) system(buf);
}

void MPIFS::flushfiles(int app_num)
{
  FILE *flushfp;
  char filename[64];

  sprintf(filename, "%s%d/%s.%d", FILEPREFIX, app_num,"flushfile",world_rank_);
  if ((flushfp = fopen(filename,"wb+"))==NULL)
    {
      printf("Cannot open flush file on io node\n");
      exit(1);
    };

#ifdef TARGETAIX
  int size = 4*1024*1024;  // on SP2
#else
  int size = 1*1024*1024;   // on bunny
#endif
  char * buffer = (char*) malloc(sizeof(char) * size);
#ifdef TARGETAIX
  for(int i=0; i < 32; i++){
   fwrite(buffer, sizeof(char), size, flushfp);
  }
#else
  fwrite (buffer, sizeof(char), size, flushfp);
#endif
  int filedesc;
  filedesc = fileno(flushfp);
  fsync(filedesc);
  fclose(flushfp);

  if ((flushfp = fopen(filename,"r"))==NULL)
    {
      printf("Cannot open flush file on io node\n");
      exit(1);
    };

#ifdef TARGETAIX
  for(i=0; i < 32; i++){
    fread(buffer, sizeof(char), size, flushfp);
  }
#else
  fread(buffer, sizeof(char), size, flushfp);
#endif
  
  fclose(flushfp);
  free(buffer);
}


void MPIFS::insert_compute_app(int app_num, App_Info *app)
{

   if (app_num > current_max_app_num_)
   {
     if (compute_apps_info_)
       compute_apps_info_ = (App_Info **)realloc(compute_apps_info_,
					       sizeof(App_Info*)*(app_num+1));
     else
       compute_apps_info_ = (App_Info**)malloc(sizeof(App_Info*)*
					       (app_num + 1));
							      
     for(int i=current_max_app_num_+1; i <= app_num ;i++){
       compute_apps_info_[i] = NULL;
     }
     current_max_app_num_ = app_num;
   }
   compute_apps_info_[app_num] = app;
   num_apps_++;
   num_apps_alive_++;
}

App_Info* MPIFS::find_compute_app(int app_num)
{
  return  compute_apps_info_[app_num];
}


Boolean MPIFS::received_quit_message(int msg_src, int msg_tag,
					MPI_Status *status)
{
   int app_num;
   receive_message((void *) &app_num, 1, MPI_INT, msg_src, msg_tag,
		MPI_COMM_WORLD, status);
   Broadcast(IO_NODE, (void *) &app_num, 1, MPI_INT, msg_tag);
   num_apps_alive_--;
   if (num_apps_alive_ == 0) return YES;
   else return NO;
}

IOFile MPIFS::open_file(char *name, int op_type)
{
  int i;
  IOFile fp;

  for (i=0; i<num_open_files_; i++) 
    if (!strcmp(name, open_file_names_[i])) { 
      fp = open_file_ptrs_[i];
      free(name);
      break;
    } 

  if (i == num_open_files_) {
    char name1[1000], fpfx[100];
    FILE *fp1;

    fp1 = fopen("FILEPREFIX", "r");
    fscanf(fp1, "%s", fpfx);
    fclose(fp1);
    sprintf(name1, "%s/%s", fpfx, name);

    open_file_names_[num_open_files_] = name;
    if ((op_type == RESTART) || (op_type == READ_TIMESTEP)) 
      fp = open_file_ptrs_[num_open_files_++] = IEEEopen(name1, "r");
    else if ((op_type == TIMESTEP) || (op_type == CHECKPOINT)) 
      fp = open_file_ptrs_[num_open_files_++] = IEEEopen(name1, "w");
    is_new_file_[i] = YES;

  } 

  return fp;
}

Boolean MPIFS::is_new_file(char *name) {
  char name1[100];
  if (node_type() == PART_TIME_COMPUTE || node_type() == COMPUTE_NODE) return NO;

  if (app_size(IO_NODE) > 1)
  {
    sprintf(name1, "%s.file_%d.ieee", name, my_rank(IO_NODE));
  }
  else
  {
    sprintf(name1, "%s.ieee", name);
  }
  for (int i=0; i<num_open_files_; i++) 
    if (!strcmp(name1, open_file_names_[i])) {
      if (is_new_file_[i] == YES) {
	is_new_file_[i] = NO;
	return YES;
      } else return NO;
    }
  printf("Can't find the file\n");
  exit(0);
  return NO;
}


void MPIFS::start_attribute_io(int msg_src, int msg_tag, MPI_Status *status)
{
  char *schema_buf;
  int msg_len, i;
  IOFile fp = 0;
  
  mpi_get_count(status, MPI_CHAR, &msg_len);
  schema_buf = (char *) malloc(sizeof(char)*msg_len);
  receive_message((void *)schema_buf, msg_len, MPI_CHAR, msg_src, 
		  msg_tag, MPI_COMM_WORLD,status);
  Broadcast(IO_NODE, (void *)schema_buf, msg_len, MPI_CHAR, msg_tag);

  char *ptr = schema_buf;
  union int_to_char tmp;
  for (i=0; i<4; i++) tmp.c[i] = *ptr++;
  int op_type = tmp.i;
  for (i=0; i<4; i++) tmp.c[i] = *ptr++;
  int len = tmp.i;
  char *fname = (char *)malloc(sizeof(char) * (len + 1));
  char *name1 = (char *)malloc(sizeof(char) * (len + 16));
  for (i=0; i<len; i++) fname[i] = *ptr++;
  fname[i] = '\0';
  if (app_size(IO_NODE) > 1)
  {
    sprintf(name1, "%s.file_%d.ieee", fname, world_rank_);
  }
  else
  {
    sprintf(name1, "%s.ieee", fname);
  }
  fp = open_file(name1, op_type);
  Attribute *attr = new Attribute(ptr, op_type);
  if (op_type == TIMESTEP) attr->write_data(fp);  
  else if (op_type == READ_TIMESTEP) {
    attr->read_data(fp);
    if (am_master_io_node()) send_attr_data(attr);
  }
  delete attr;
  free(schema_buf);
}

void MPIFS::start_collective_io(int msg_src, int msg_tag, MPI_Status *status)
{
  int *schema_buf, msg_len, comp_app_num = (msg_tag / 10);
  Collective_IO *new_io;
  IOFile fp = 0;
  
  mpi_get_count(status, MPI_INT, &msg_len);
  schema_buf = (int *) malloc(sizeof(int)*msg_len);
  receive_message((void *)schema_buf, msg_len, MPI_INT, msg_src, 
		  msg_tag, MPI_COMM_WORLD,status);
  Broadcast(IO_NODE, (void *)schema_buf, msg_len, MPI_INT, msg_tag);

  int i, len = schema_buf[2];
  char *name = (char *)malloc(sizeof(char) * (len + 1));
  char *name1 = (char *)malloc(sizeof(char) * (len + 16));
  for (i=0; i<len; i++) name[i] = schema_buf[3+i];
  name[i] = '\0';
  if (app_size(IO_NODE) > 1)
  {
    sprintf(name1, "%s.file_%d.ieee", name, world_rank_);
  }
  else
  {
    sprintf(name1, "%s.ieee", name);
  }
  fp = open_file(name1, schema_buf[1]);
  free(name);

  switch(schema_buf[0]){
    case SIMPLE_IO:
      new_io = new Simple_IO(schema_buf, msg_len,world_rank_, comp_app_num,
			     compute_apps_info_[comp_app_num]->app_size(),
			     compute_apps_info_[comp_app_num], fp);
      break;
    
    default:
      printf("Error in start_collective_io - undefined strategy\n");
      exit(1);
      break;
  } 

#ifdef DEBUG
  printf("Starting the collective IO for compute app %d\n",comp_app_num);
#endif
  new_io->start_to_finish(NO, NULL);
#ifdef DEBUG
  printf("Finished the collective I/O for compute app %d\n", comp_app_num);
#endif
  release_compute_nodes(comp_app_num);
}


void MPIFS::part_time_io_node_loop(int *schema_buf, int msg_len, 
				   Array *array)
{
  Collective_IO *new_io;
  IOFile fp = 0;

  int i, len = schema_buf[2];
  char *name = (char *)malloc(sizeof(char) * (len + 1));
  char *name1 = (char *)malloc(sizeof(char) * (len + 16));
  for (i=0; i<len; i++) name[i] = schema_buf[3+i];
  name[i] = '\0';
  if (app_size(IO_NODE) > 1)
  {
    sprintf(name1, "%s.file_%d.ieee", name, world_rank_);
  }
  else
  {
    sprintf(name1, "%s.ieee", name);
  }
  fp = open_file(name1, schema_buf[1]);
  free(name);

  switch(schema_buf[0]){
    case SIMPLE_IO:
      new_io  = new Simple_IO(schema_buf, msg_len, world_rank_,
			      1, app_size_, app_info_, fp);
      break;

    default:
      printf("Error in part_time_io_node_loop - undefined strategy\n");
      exit(1);
      break;
  } 
  new_io->start_to_finish(YES, array);
  release_compute_nodes(1);
}

void MPIFS::compute_node_io_loop(Array *array)
{
  Simple_IO *simple;

  switch(array->io_strategy()){
  case SIMPLE_IO:
    simple = new Simple_IO();
    simple->compute_node_io_loop(array);
    break;

  default:
    printf("Error in MPIFS::compute_node_io_loop - Undefined i/o strategy\n");
    exit(1);
    break;
  }
  compute_side_io_done();
}


int MPIFS::app_size(int node_type)
{
  if (node_type == COMPUTE_NODE)
  {
	if (node_type_ != IO_NODE)
		return app_size_;
	else { 
		printf("Error in MPIFS::app_size - wrong node_type\n");
		exit(10);
	     }
   }
   else if (node_type == IO_NODE)
   {
	if (node_type_ == IO_NODE) return app_size_;
	else if (node_type_ == PART_TIME_IO) return io_app_size_;
	else {
		printf("Error in MPIFS::app_size - wrong node_type\n");
		exit(10);
	     }
   }
   else 
   {
		printf("Error in MPIFS::app_size - wrong node_type\n");
		exit(10);
	     
   }
   return -1;
}

int MPIFS::my_rank(int node_type)
{
  if (node_type == COMPUTE_NODE)
  {
	if (node_type_ != IO_NODE)
		return app_rank_;
	else { 
		printf("Error in MPIFS::my_rank - wrong node_type\n");
		exit(10);
	     }
   }
   else if (node_type == IO_NODE)
   {
	if (node_type_ == IO_NODE) return app_rank_;
	else if (node_type_ == PART_TIME_IO) return io_app_rank_;
	else {
		printf("Error in MPIFS::my_rank - wrong node_type\n");
		exit(10);
	     }
   }
   else 
   {
		printf("Error in MPIFS::my_rank - wrong node_type\n");
		exit(10);
	     
   }
   return -1;
}

int MPIFS::node_type(){return node_type_;}

void MPIFS::send_array_schema(Array *array)
{
  int *schema, schema_size;
  if (am_master_compute_node()){
    array->pack(&schema, &schema_size);
    send_message((void *)schema, schema_size, MPI_INT, master_io_node_,
		 app_num_*10+ARRAYGROUP_SCHEMA, MPI_COMM_WORLD);
  }
}

void MPIFS::receive_attr_data(Attribute *attr)
{
  int msg_len, i;
  MPI_Status status;
  char *ptr;

  MPI_Probe(MPI_ANY_SOURCE, ATTRIBUTE_DATA, MPI_COMM_WORLD, &status); 
  mpi_get_count(&status, MPI_CHAR, &msg_len);
  void *data_buf = (void *) malloc(msg_len);
  receive_message((void *)data_buf, msg_len, MPI_CHAR, status.MPI_SOURCE,
		  ATTRIBUTE_DATA, MPI_COMM_WORLD, &status);
  Broadcast(COMPUTE_NODE, data_buf, msg_len, MPI_CHAR, ATTRIBUTE_DATA);

  ptr = (char *)data_buf;
  union int_to_char tmp;
  for (i=0; i<4; i++) tmp.c[i] = *ptr++;
  attr->set_esize(tmp.i); 
  for (i=0; i<4; i++) tmp.c[i] = *ptr++;
  attr->set_count(tmp.i); 
  attr->set_data_ptr(ptr);
}

void MPIFS::send_attr_data(Attribute *attr)
{ 
  int i;
  void *data_buf = (void *)malloc((attr->data_size() + 8));
  char *ptr = (char *)data_buf;
  union int_to_char tmp;
  tmp.i = attr->esize();
  for (i=0; i<4; i++) *ptr++ = tmp.c[i];
  tmp.i = attr->count();
  for (i=0; i<4; i++) *ptr++ = tmp.c[i];
  memcpy(ptr, attr->get_data_ptr(), attr->data_size());

  int master_comp_node = compute_apps_info_[1]->get_master();
  send_message(data_buf, attr->data_size()+8, MPI_CHAR, 
	       master_comp_node, ATTRIBUTE_DATA, MPI_COMM_WORLD);
  free(data_buf);
}

void MPIFS::receive_attr_schema()
{
  char *schema_buf;
  int msg_len;
  MPI_Status status;

  MPI_Probe(MPI_ANY_SOURCE, ATTRIBUTE_SCHEMA, MPI_COMM_WORLD, &status); 
  mpi_get_count(&status, MPI_CHAR, &msg_len);
  schema_buf = (char *) malloc(sizeof(char) * msg_len);
  receive_message((void *)schema_buf, msg_len, MPI_CHAR, status.MPI_SOURCE,
		  ATTRIBUTE_SCHEMA, MPI_COMM_WORLD, &status);
  Broadcast(IO_NODE, (void *)schema_buf, msg_len, MPI_CHAR, ATTRIBUTE_SCHEMA);
  free(schema_buf);
}

void MPIFS::send_attr_schema(Attribute *attr, char *fname, int op_type)
{
  char *schema;
  int schema_size;

  if (am_master_compute_node()){
    attr->pack(schema_size, schema, fname, op_type);
    send_message((void *)schema, schema_size, MPI_CHAR, master_io_node_,
                 ATTRIBUTE_SCHEMA, MPI_COMM_WORLD);
  }
  free(schema);
}

/* Called from the compute node side */
void MPIFS::user_commands(int cmd)
{
  if (node_type_ == COMPUTE_NODE)
  {
    int tag = cmd*10+SPECIAL;
    if (am_master_compute_node()){
      send_message((void *)&app_num_,  1, MPI_INT, master_io_node_, 
		   tag,	MPI_COMM_WORLD);
    }
	
    if (cmd == APP_BARRIER){
      MPI_Barrier(*(app_info_->intra_comm()));
    } else if (cmd == GLOBAL_BARRIER){
      MPI_Barrier(MPI_COMM_WORLD);
    }
  }
  else {
    /* Must be the part_time I/O case */
    switch(cmd){
    case APP_BARRIER:
      MPI_Barrier(*comm_);
      break;
    case GLOBAL_BARRIER:
      MPI_Barrier(MPI_COMM_WORLD);
      break;
    case CLEANFILES:
      if (node_type_ == PART_TIME_IO)
	cleanfiles(app_num_);
      break;
    case FLUSHFILES:
      if (node_type_ == PART_TIME_IO)
	flushfiles(app_num_);
      break;
    case CREATEFILES:
      if (node_type_ == PART_TIME_IO)
	createfiles(app_num_);
      break;
    default:
      printf("Unknown message code %d\n", cmd);
      break;
    }
  }
}    

/* This function is called by io nodes or part-time io nodes */
void MPIFS::release_compute_nodes(int app_num)
{
  App_Info *app;

  if (node_type_ == PART_TIME_IO){
    MPI_Barrier(MPI_COMM_WORLD);
  } else if ((node_type_ == IO_NODE) && (num_apps_ == 1)){
    MPI_Barrier(*comm_);
    if (am_master_io_node()){
      app = find_compute_app(app_num);
      send_message(&app_num, 1, MPI_INT, app->get_master(), COMP_QUIT,
		   MPI_COMM_WORLD);
    }
  } else {
    printf("Error in release_compute_nodes - wrong node type \n");
    exit(11);
  }
}

/* This function is called by the compute nodes after they have done their part */
void MPIFS::compute_side_io_done()
{
  int app_num;
  MPI_Status status;

  if (node_type_ == PART_TIME_COMPUTE){
    MPI_Barrier(MPI_COMM_WORLD);
  } else if (node_type_ == COMPUTE_NODE){
    if (am_master_compute_node())
      receive_message((void *)&app_num, 1, MPI_INT, master_io_node_, COMP_QUIT,
		      MPI_COMM_WORLD, &status);
    MPI_Barrier(*comm_);
  } else {
    printf("Error in compute_side_io_doen - wrong node type\n");
    exit(11);
  }
}

App_Info* MPIFS::io_app_info(){
  return io_app_info_;
}

int MPIFS::master_io_node(){
  return master_io_node_;
}

int MPIFS::mem_buf_size()
{
  return mem_buf_size_;
}

char *MPIFS::mem_buf()
{
  return mem_buf_;
}

void MPIFS::set_mem_buf_size(int size)
{
  mem_buf_size_ = size;
}

void MPIFS::set_mem_buf(char *buf)
{
  mem_buf_ = buf;
}
