/*@@
@file       Whisky_Dust.F90
@routine    Whisky_Dust_Initial
@date       Mon Sep 16 16:51:24 2002
@author     Luca Baiotti
@desc 
   Sets to zero the arrays. For safety. This is done in a separate subroutine 
   so that it is run only for initial data. 
@enddesc 
@calls     Whisky_Dust_Set_Eta
@calledby  scheduler 
@history 

@endhistory 
@@*/

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"

#include "cctk_Functions.h"

subroutine Whisky_Dust_Initial(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  character(len=100) :: infoline

  if (activate_dust_analysis) then
    if (.NOT.(CCTK_EQUALS(initial_data,"dust"))) then
      write(infoline,'(a75)') "If you want to run Dust Analysis, you also have to set initial_data to 'dust'"
      call CCTK_WARN(0,infoline)
    end if
  else if (CCTK_EQUALS(initial_data,"dust")) then
    write(infoline,'(a50)') "Setting dust initial data and not analysing them"
    call CCTK_WARN(1,infoline)
  end if

  !initialisations (only for initial data)
  dust_temp = 0.d0
  dust_integrand = 0.d0
  dust_chi = 0.d0
  dust_big_R = 0.d0
  dust_small_r = 0.d0
  dust_alpha = 0.d0
  dust_alpha_tilde = 0.d0
  dust_beta = 0.d0
  dust_beta_tilde = 0.d0
  dust_A = 0.d0
  dust_A_tilde = 0.d0
  dust_eta = 0.d0
  dust_big_T = 0.d0
  dust_r_tilde = 0.d0
  dbig_T_by_dsmall_t_tilde = 0.d0
  dtau_by_dsmall_t_tilde = 0.d0
  deta_by_dsmall_t_tilde = 0.d0
  dsmallr_tilde_by_dt_tilde = 0.d0

  dust_analysis_flag = 0!means that the part of code is for initial data. For analysis it would be 1.
  
  !The following two lines are used only for Maximal slicing
  dust_analysis_first_time = 2 !this integer selects the way dust_eta(1) is computed:
                               !2 -> given initial value (-Pi); 1 -> linear approximation in time  
  dust_analysis_evolve = 0 !indicates that dust_eta(1) must not be evolved when analysing inital data

  call  Whisky_Dust_Set_Eta(CCTK_PASS_FTOF)

end subroutine Whisky_Dust_Initial


/*@@
@file       Whisky_Dust.F90
@routine    Whisky_Dust_Set_Eta
@date       Wed Sep 18 11:15:48 2002
@author     Luca Baiotti
@desc 
   Sets the initial value: dust_eta(1), according to whether this is run at 
   initial data or analysis and whether there is an analytic or approximate method for computing it.
@enddesc 
@calls     Whisky_Dust_Integrate, Whisky_Dust_Integrate_Polar
@calledby  Whisky_Dust_Initial, Whisky_Dust_Analysis 
@history 

@endhistory 
@@*/

subroutine Whisky_Dust_Set_Eta(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_REAL :: Pi, dust_chi_surface
  character(len=100) :: infoline

  Pi = acos(-1.d0)

  if (dust_analysis_flag == 1) then
    !find the proper eta to compute the dust quantities at given cactus time t

!!$    if (dust_analysis_first_time == 0) then
!!$      !Second order advance, even if it does not change much. 
!!$      !This is not currently used.
!!$      !This is not called correctly as is.
!!$      dust_eta(1) = 2.d0*dust_eta_p - dust_eta_pp
!!$      dust_eta_pp = dust_eta_p
!!$      dust_eta_p = dust_eta(1)

    if (dust_analysis_first_time == 1) then
      !first order advance.
      dust_eta(1) = dust_eta(1)+deta_by_dsmall_t_tilde(1)*dust_alpha_c*CCTK_delta_time
      dust_eta_p = dust_eta(1)
      dust_analysis_evolve = 1

    else if (dust_analysis_first_time == 2) then
      !do not advance in time when analysis is run on initial data
      dust_eta(1) = -Pi !copy line below, for initial data
      dust_eta_pp = dust_eta(1)

    else if (dust_analysis_first_time /= -1) then
      write(infoline,'(a34)') "dust_analysis_first_time has a not allowed value in Whisky_Dust_Set_Eta"
      call CCTK_WARN(0,infoline)
    end if
  else if (dust_analysis_flag == 0) then 
    dust_eta(1) = -Pi !must be chosen so that -Pi < eta(surface) < -Pi/2.
  else
    write(infoline,'(a34)') "dust_analysis_flag has a not allowed value in Whisky_Dust_Set_Eta"
    call CCTK_WARN(0,infoline)
  end if

  !Now we call the routines for the integration of the initial data equations.
  !But first we can set here the internal radius variable chi, since it is the same for both gauges

  !start integrating from a small value, instead of zero
  dust_chi(1) = 1.d-80
  dust_chi_surface = asin(sqrt(2.d0*Dust_Mass/Dust_Max_Radius))

  !find which element of chi corresponds to the surface
  do i_surface = 2, Dust_Num_Radial
    dust_chi(i_surface) = dust_chi(i_surface-1) + dust_dchi
    if (dust_chi(i_surface) >= dust_chi_surface) exit
  end do
  i_surface = i_surface-1

  dust_a_0 = sqrt(Dust_Max_Radius**3/(8.d0*Dust_Mass))

  !Now we call the routines for intagration of the equations in Maximal or Polar slicing.

  if (CCTK_EQUALS(Dust_Gauge,"Maximal")) then
    call  Whisky_Dust_Integrate(CCTK_PASS_FTOF)
  else if  (CCTK_EQUALS(Dust_Gauge,"Polar")) then
    call Whisky_Dust_Integrate_Polar(CCTK_PASS_FTOF)
  else 
    write(infoline,'(a34)') "Dust_Gauge has a not allowed value in Whisky_Dust_Set_Eta"
    call CCTK_WARN(0,infoline)
  end if

  !Also the info printing can be done here for both gauges:

  write(infoline,'(a39,g16.7)') "The dust star has an ADM surface radius of ", dust_small_r(i_surface)
  call CCTK_INFO(infoline)

  write(infoline,'(a41,g16.7)') "The dust star has an areal surface radius of ", dust_big_R(i_surface)
  call CCTK_INFO(infoline)

  write(infoline,'(a36,g16.7)') "The grid boundary is at areal radius ", dust_big_R(Dust_Num_Radial)
  call CCTK_INFO(infoline)

end subroutine Whisky_Dust_Set_Eta


/*@@
@file       Whisky_Dust.F90
@routine    Whisky_Dust_Integrate
@date       Thu Aug  1 11:27:00 2002
@author     Luca Baiotti, Ian Hawke
@desc 
   Finds lapse, shift and metric at a given time for an Oppenheimer-Snyder collpase (dust star),
   following PETRICH SHAPIRO TEUKOLSKY, Phys. Rev. D 31, 2459 (1985)
   Integrates the ODEs using RK4.
   We rescale in another routine to match to cactus gridpoint positions.
@enddesc 
@calls     Dust_Source_RHS, Function_Integrate
@calledby  Whisky_Dust_Set_Eta 
@history 

@endhistory 
@@*/

subroutine Whisky_Dust_Integrate(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS

  CCTK_INT :: i, j
  CCTK_REAL, dimension(7) :: old_data, source_data, in_data, new_data, k1, k2, k3, k4
  CCTK_REAL :: dust_alpha_0, dust_A_c, dln_A_c_by_dsmall_t_tilde,&
       dust_B, dB_by_dsmall_t_tilde, dB_by_dsmall_t, dust_integr_pace
  CCTK_REAL :: Pi
  character(len=100) :: infoline

  Pi = acos(-1.d0)

  !compute some terms referring to the INTERIOR:
  
  ! ODEs: (3.18) (3.19) (3.20) (3.40 after deriving the log) (3.45) (3.48 after expanding the lhs)
  !Initial conditions (in the last paragraph of section III of the paper):
  !Except the condition on dust_eta which is in the previous subroutine

  dust_big_T(1) = -dust_chi(1)*cos(dust_eta(1)/2.d0)/sin(dust_eta(1)/2.d0)
  dbig_T_by_dsmall_t_tilde(1) = dust_chi(1)/(dust_a_0*(1.d0-cos(dust_eta(1)))**2)
  dtau_by_dsmall_t_tilde(1) = 1.d0
  dust_r_tilde(1) = dust_a_0*(1.d0-cos(dust_eta(1)))*dust_chi(1)
  deta_by_dsmall_t_tilde(1) = 1.d0/(dust_a_0*(1.d0-cos(dust_eta(1))))
  dsmallr_tilde_by_dt_tilde(1) = dust_chi(1)*cos(dust_eta(1)/2.d0)/sin(dust_eta(1)/2.d0)

  !solve with RK4

  do i = 1, i_surface-1

    old_data(1) = dbig_T_by_dsmall_t_tilde(i)
    old_data(2) = dtau_by_dsmall_t_tilde(i)
    old_data(3) = dust_big_T(i)
    old_data(4) = dust_eta(i)
    old_data(5) = dust_r_tilde(i)
    old_data(6) = deta_by_dsmall_t_tilde(i)
    old_data(7) = dsmallr_tilde_by_dt_tilde(i)

    in_data = old_data

    source_data = 0.d0

    call Dust_Source_RHS(dust_chi(i), in_data, source_data, dust_a_0)
    k1 = dust_dchi * source_data
    in_data = old_data + 0.5d0 * k1
    call Dust_Source_RHS(dust_chi(i)+0.5d0*dust_dchi, in_data, source_data, dust_a_0)
    k2 = dust_dchi * source_data
    in_data = old_data + 0.5d0 * k2
    call Dust_Source_RHS(dust_chi(i)+0.5d0*dust_dchi, in_data, source_data, dust_a_0)
    k3 = dust_dchi * source_data
    in_data = old_data + k3
    call Dust_Source_RHS(dust_chi(i)+dust_dchi, in_data, source_data, dust_a_0)
    k4 = dust_dchi * source_data

    new_data = old_data + (k1 + k4 + 2.d0 * (k2 + k3)) / 6.d0

    dbig_T_by_dsmall_t_tilde(i+1) = new_data(1)
    dtau_by_dsmall_t_tilde(i+1) = new_data(2)
    dust_big_T(i+1) = new_data(3)
    dust_eta(i+1) = new_data(4)
    dust_r_tilde(i+1) = new_data(5)
    deta_by_dsmall_t_tilde(i+1) = new_data(6)
    dsmallr_tilde_by_dt_tilde(i+1) = new_data(7)

  end do

  !INTERIOR: algebraic terms

  do i=1, i_surface

    dust_A_tilde(i) = dust_a_0*(1-cos(dust_eta(i)))*sin(dust_chi(i))/dust_r_tilde(i) ! (3.41)

    dust_alpha_tilde(i) = dtau_by_dsmall_t_tilde(i)/sqrt(1.d0-dust_big_T(i)**2) ! (3.46)

  end do

  !EXTERIOR & MATCHING

  dust_big_R(i_surface) = dust_a_0*(1.d0-cos(dust_eta(i_surface)))*sin(dust_chi(i_surface))! (3.1)

  dust_B = dust_big_R(i_surface)**2*cos(dust_chi(i_surface))*(dust_big_T(i_surface)+&
       tan(dust_chi(i_surface))*cos(dust_eta(i_surface)/2.d0)/sin(dust_eta(i_surface)/2.d0))/&
       sqrt(1.d0-dust_big_T(i_surface)**2) ! (5.4)

  dB_by_dsmall_t_tilde = (2.d0*dust_B*cos(dust_eta(i_surface)/2.d0)/sin(dust_eta(i_surface)/2.d0)- &
       dust_big_R(i_surface)**2*sin(dust_chi(i_surface))/ &
       (2.d0*(sin(dust_eta(i_surface)/2.d0))**2*sqrt(1.d0-dust_big_T(i_surface)**2)))*&
       deta_by_dsmall_t_tilde(i_surface)+ &
       (dust_big_R(i_surface)**2*cos(dust_chi(i_surface))*&
       (1.d0+dust_big_T(i_surface)*tan(dust_chi(i_surface))*cos(dust_eta(i_surface)/2.d0)/ &
       sin(dust_eta(i_surface)/2.d0))/(1.d0-dust_big_T(i_surface)**2)**(1.5d0))* &
       dbig_T_by_dsmall_t_tilde(i_surface) ! (5.6)

  do i=i_surface, Dust_Num_Radial 

    !temporary variables for computing function integral (5.9b)
    dust_integr_pace = (1.d0/dust_big_R(i))/(Dust_Num_Integrate-1)
    do j=1, Dust_Num_Integrate
      dust_temp(j) = (j-1)*dust_integr_pace
      dust_integrand(j) = -(1.d0-2.d0*dust_temp(j)+dust_B**2*dust_temp(j)**4)**(-1.5d0)
    end do
    !compute dust_alpha_r (5.9b)
    call Function_Integrate(Dust_Num_Integrate, dust_integr_pace, dust_integrand, dust_alpha_r(i))

    !temporary variables for computing function integral (4.11)
    ! integrate from 10^-80 instead of 0
    dust_integr_pace = (1.d0/dust_big_R(i)-1.d-80)/(Dust_Num_Integrate-1)
    do j=1, Dust_Num_Integrate
      dust_temp(j) = 1.d-80+(j-1)*dust_integr_pace
      dust_integrand(j) = ((1.d0-2.d0*dust_temp(j)+dust_B**2*dust_temp(j)**4)**(-0.5d0)-1.d0)/dust_temp(j)
    end do
    dust_integrand(1) = 1.d0 !otherwise it does not get the limit right
    ! compute dust_A (4.11)
    call Function_Integrate(Dust_Num_Integrate, dust_integr_pace, dust_integrand, dust_A(i))
    dust_A(i) = dexp(dust_A(i))

    dust_small_r(i) = dust_big_R(i)/dust_A(i) !from (4.9)

    dust_big_R(i+1) = dust_big_R(i)+dust_dR

  end do

  dust_alpha_0 = sqrt(1.d0-2.d0/dust_big_R(i_surface)+dust_B**2/dust_big_R(i_surface)**4) ! (5.9a)

  dust_alpha_c = (dust_alpha_tilde(i_surface)/dust_alpha_0-&
       dust_alpha_r(i_surface)*dB_by_dsmall_t_tilde)**(-1) ! (5.11)

  dust_A_c = dust_A(i_surface)/dust_A_tilde(i_surface) ! (5.13)

  dB_by_dsmall_t = dust_alpha_c*dB_by_dsmall_t_tilde ! (5.5)

  !EXTERIOR FINAL, AFTER MATCHING

  do i=i_surface, Dust_Num_Radial 

    dust_alpha(i) = sqrt(1.d0-2.d0/dust_big_R(i)+dust_B**2/dust_big_R(i)**4)*&
         (1.d0+dB_by_dsmall_t*dust_alpha_r(i)) ! (4.12)

    !temporary variables for computing function integral in (4.13) and (5.16)
    dust_integr_pace = (1.d0/dust_big_R(i))/(Dust_Num_Integrate-1)
    do j=1, Dust_Num_Integrate
      dust_temp(j) =(j-1)*dust_integr_pace
      dust_integrand(j) = dust_temp(j)**3*(1.d0-2.d0*dust_temp(j)+dust_B**2*dust_temp(j)**4)**(-1.5d0)
    end do
    call Function_Integrate(Dust_Num_Integrate, dust_integr_pace, &
         dust_integrand, dust_integral_for_shift(i))

    dust_beta(i) = -dust_B/(dust_A(i)*dust_big_R(i)**2)*(1.d0+dB_by_dsmall_t*(dust_alpha_r(i)+&
         dust_big_R(i)**3*dust_integral_for_shift(i))) ! (4.13) NB: alpha_r has a minus in the definition

  end do

  !INTERIOR FINAL, AFTER MATCHING

  dln_A_c_by_dsmall_t_tilde = &
       -(1.d0-2.d0/dust_big_R(i_surface)+dust_B**2/dust_big_R(i_surface)**4)**(-0.5d0)*&
       deta_by_dsmall_t_tilde(i_surface)*cos(dust_eta(i_surface)/2.d0)/sin(dust_eta(i_surface)/2.d0)-&
       dust_B*dB_by_dsmall_t_tilde*dust_integral_for_shift(i_surface)+&
       dsmallr_tilde_by_dt_tilde(i_surface)/dust_r_tilde(i_surface) ! (5.16)

  do i=1, i_surface

    dust_small_r(i) = dust_r_tilde(i)/dust_A_c ! (3.35)

    dust_A(i) = dust_A_c*dust_A_tilde(i) ! (3.42)

    dust_alpha(i) = dust_alpha_c*dust_alpha_tilde(i) ! (3.47)

    dust_beta_tilde(i) = -dsmallr_tilde_by_dt_tilde(i) - &
         dust_big_T(i)*dust_alpha_tilde(i)/dust_A_tilde(i) ! (3.49)

    dust_beta(i) = dust_alpha_c/dust_A_c*(dust_beta_tilde(i)+&
         dust_r_tilde(i)*dln_A_c_by_dsmall_t_tilde) ! (3.50)

  end do

  dust_big_R = dust_small_r*dust_A

end subroutine Whisky_Dust_Integrate


/*@@
@routine    Dust_Source_RHS
@date       Wed Jul 31 18:59:33 2002
@author     Luca Baiotti    
@desc 
   The source terms for the ODEs. These are equations: (3.18) (3.19) (3.20)
   (3.40 after deriving the log) (3.45) (3.48 after expanding the lhs).
   From PETRICH SHAPIRO TEUKOLSKY, Phys. Rev. D 31, 2459 (1985)
@enddesc 
@calls     
@calledby Whisky_Dust_Integrate  
@history 

@endhistory 
@@*/

subroutine Dust_Source_RHS(dust_chi_point, old_data, source_data, dust_a_0)

  implicit none

  CCTK_REAL, dimension(7) :: old_data, source_data
  CCTK_REAL :: dust_chi_point, dust_a_0

  !mine
!!$  source_data(1) = (2.d0*(5.d0+old_data(3)-2.d0*old_data(3)**2)* &
!!$       cos(old_data(4)/2.d0)/sin(old_data(4)/2.d0)*old_data(3)+&
!!$       2.d0*cos(dust_chi_point)/sin(dust_chi_point)*(3.d0*old_data(3)**2-1.d0))*old_data(1)+&
!!$       (1.5d0-old_data(3)**2)*(1.d0-old_data(3)**2)*old_data(2)/ &
!!$       ((sin(old_data(4)/2.d0))**2*dust_a_0*(1.d0-cos(old_data(4))))

  !as on paper
  source_data(1) = (6.d0*cos(old_data(4)/2.d0)/sin(old_data(4)/2.d0)*old_data(3)+&
       2.d0*cos(dust_chi_point)/sin(dust_chi_point)*(3.d0*old_data(3)**2-1.d0))*old_data(1)+&
       1.5d0*(1.d0-old_data(3)**2)*old_data(2)/ &
       ((sin(old_data(4)/2.d0))**2*dust_a_0*(1.d0-cos(old_data(4)))) ! (3.43)

  source_data(2) = old_data(3)*old_data(2)*cos(old_data(4)/2.d0)/sin(old_data(4)/2.d0)+&
       dust_a_0*(1.d0-cos(old_data(4)))*old_data(1) ! (3.44)

  !mine
!!$  source_data(3) = -(1-old_data(3)**2)*((3.d0+2.d0*old_data(3)**2) &
!!$       *cos(old_data(4)/2.d0)/sin(old_data(4)/2.d0)+&
!!$       2.d0*old_data(3)*cos(dust_chi_point)/sin(dust_chi_point))

  !as on paper
  source_data(3) = -(1-old_data(3)**2)*(3.d0*cos(old_data(4)/2.d0)/sin(old_data(4)/2.d0)+&
       2.d0*old_data(3)*cos(dust_chi_point)/sin(dust_chi_point)) ! (3.18)

  source_data(4) = old_data(3) ! (3.20)

  source_data(5) = old_data(5)*sqrt(1.d0-old_data(3)**2)/sin(dust_chi_point) ! (3.40)

  source_data(6) = old_data(1) ! (3.45)

  source_data(7) = (old_data(7)*sqrt(1.d0-old_data(3)**2)-&
       old_data(5)*old_data(3)*old_data(1)/sqrt(1.d0-old_data(3)**2))/sin(dust_chi_point) ! (3.48)

end subroutine Dust_Source_RHS


/*@@
@routine   Function_Integrate     
@date      Sat Aug 31 12:34:32 2002 
@author    Luca Baiotti
@desc 
   Gives the values of the integrals of specified arguments and ranges with the trapezoidal rule.
@enddesc 
@calls
@calledby    Whisky_Dust_Integrate
@history 

@endhistory 
@@*/

subroutine Function_Integrate(n, pace, integrand, integral_value)

  implicit none

  CCTK_INT :: n, j
  CCTK_REAL :: integral_value, sum, pace
  CCTK_REAL, dimension(1:n) :: integrand 

  sum=0.d0
  do j=2,n-1
    sum=sum+integrand(j)
  end do
  integral_value = pace * (0.5d0 * integrand(1) + sum + 0.5d0 * integrand(n))

end subroutine Function_Integrate


/*@@
@file       Whisky_Dust.F90
@routine    Whisky_Dust_Integrate_Polar
@date       Tue Jan 14 10:02:39 2002
@author     Luca Baiotti
@desc 
   Finds lapse, shift and metric at a given time for an Oppenheimer-Snyder collpase (dust star),
   following PETRICH SHAPIRO TEUKOLSKY, Phys. Rev. D 33, 2100 (1986). Formulae number of this 
   routine (and only these) refer to this paper.
   We rescale in another routine to match to cactus gridpoint positions.
@enddesc 
@calls     
@calledby  Whiskt_Dust_Set_Eta 
@history 

@endhistory 
@@*/

subroutine Whisky_Dust_Integrate_Polar(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS

  CCTK_INT :: i, j
  CCTK_REAL :: Pi,  dust_integr_pace
  character(len=100) :: infoline

  Pi = acos(-1.d0)

  !COMPUTE SPATIAL DEPENDENCE OF ETA

  do i=2, i_surface

    dust_eta(i) = 2.d0*acos(cos(dust_eta(1)/2.d0)/sqrt(cos(dust_chi(i)))) ! (3.21)

  end do

  !EXTERIOR

  dust_big_R(i_surface) = dust_a_0*(1.d0-cos(dust_eta(i_surface)))*sin(dust_chi(i_surface))! (3.1)

  do i=i_surface, Dust_Num_Radial 

    dust_small_r(i) = 0.5d0*(dust_big_R(i)-1+sqrt(dust_big_R(i)*(dust_big_R(i)-2.d0))) ! (3.17)

    dust_alpha(i) = sqrt(1-2.d0/dust_big_R(i)) ! (3.14)

    dust_A(i) = dust_big_R(i)/dust_small_r(i) ! (3.15)

    dust_beta(i) = 0.d0 ! (3.16)

    dust_big_R(i+1) = dust_big_R(i)+dust_dR

  end do

  !INTERIOR

  dust_alpha_c = -(cos(dust_chi(i_surface))**3 - cos(dust_eta(1)/2.d0)**2)*sin(dust_eta(1)/2.d0)/&
       (cos(dust_chi(i_surface)) - cos(dust_eta(1)/2.d0)**2)**1.5d0 ! (3.50)

  do i=1, i_surface

    dust_big_T(i) = tan(dust_chi(i))*cos(dust_eta(1)/2.d0)/ &
         sqrt(cos(dust_chi(i))-cos(dust_eta(1)/2.d0)**2) ! (3.36)

    dust_alpha(i) = (cos(dust_chi(i_surface))**3 - cos(dust_eta(1)/2.d0)**2)/&
         (cos(dust_chi(i_surface)) - cos(dust_eta(1)/2.d0)**2)**1.5d0 * &
         (cos(dust_chi(i)) - cos(dust_eta(1)/2.d0)**2)/ &
         sqrt(cos(dust_chi(i))**3 - cos(dust_eta(1)/2.d0)**2) ! (3.51)

    !temporary variables for computing function integral (3.54)
    dust_integr_pace = (dust_chi(i_surface)-dust_chi(i))/(Dust_Num_Integrate-1)
    do j=1, Dust_Num_Integrate
      dust_temp(j) = dust_chi(i)+(j-1)*dust_integr_pace
      dust_integrand(j) = sqrt(cos(dust_temp(j))**3 - cos(dust_eta(1)/2.d0)**2)/&
           (cos(dust_temp(j))*sin(dust_temp(j))*sqrt(cos(dust_temp(j)) - &
           cos(dust_eta(1)/2.d0)**2)) - cos(dust_temp(j))/sin(dust_temp(j))
    end do
    dust_integrand(1) = 1.d0 !otherwise it does not get the limit right
    call Function_Integrate(Dust_Num_Integrate, dust_integr_pace, dust_integrand, dust_A(i))

    dust_A(i) = dust_a_0*(1.d0-cos(dust_eta(i)))*sin(dust_chi(i_surface))/dust_small_r(i_surface) * &
         exp(dust_A(i)) ! (3.54)

    !temporary variables for computing function integral in (3.62)
    dust_integr_pace = (dust_chi(i_surface)-dust_chi(i))/(Dust_Num_Integrate-1)
    do j=1, Dust_Num_Integrate
      dust_temp(j) = dust_chi(i)+(j-1)*dust_integr_pace
      dust_integrand(j) = cos(dust_temp(j))**2*sin(dust_temp(j))/ &
           ((cos(dust_temp(j))**3-cos(dust_eta(1)/2.d0)**2)**1.5d0* &
           sqrt(cos(dust_temp(j))-cos(dust_eta(1)/2.d0)**2))
    end do
    call Function_Integrate(Dust_Num_Integrate, dust_integr_pace, &
         dust_integrand, dust_integral_for_shift(i))

    dust_beta(i) = -0.75d0*dust_alpha_c*dust_big_R(i)*sin(dust_eta(1))/ &
         (dust_A(i)* 2.d0*dust_a_0*sin(dust_eta(1)/2.d0))*dust_integral_for_shift(i) ! (3.62)

    dust_big_R(i) = dust_a_0*(1.d0-cos(dust_eta(i)))*sin(dust_chi(i)) ! (3.2)

    dust_small_r(i) = dust_big_R(i) / dust_A(i)

  end do

end subroutine Whisky_Dust_Integrate_Polar


/*@@
@file      Whisky_Dust
@date      Thu Aug  1 09:18:35 2002
@author    Luca Baiotti, Ian Hawke
@desc 
   Sets 3D hydro and spacetime variables for the Oppenheimer-Snyder (dust star) collpase initial data. 
@enddesc 
@calls     Interp_Dust_Isotropic, SpatialDeterminant, prim2conpolytype
@calledby  scheduler, Whiskt_Dust_Analysis
@history 

@endhistory 
@@*/

subroutine Whisky_Dust(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_INT :: i,j,k, ierr
  CCTK_REAL :: det, Dust_Atmosphere, dust_alpha_point, dust_A_point, &
       dust_beta_point, rho_point, dust_vel_r, dust_rho_0, computed_mass
!  CCTK_REAL :: dust_integr_pace
  CCTK_REAL :: Pi
  character(len=100) :: infoline

  Pi = acos(-1.d0)

  dust_rho_0 = 3.d0*Dust_mass**2/(4.d0*Pi*dust_a_0**2*(1.d0-cos(dust_eta(1)))**3)! (2.5) 
  ! this is rho at the center and is used only for computing the atmosphere factor

  if (rho_abs_min > 0.d0) then
    Dust_Atmosphere = rho_abs_min
  else
    Dust_Atmosphere = dust_rho_0 * rho_rel_min
  end if

  if (initial_rho_abs_min > 0.d0) then
    Dust_Atmosphere = initial_rho_abs_min
  else if (initial_rho_rel_min > 0.d0) then
    Dust_Atmosphere = dust_rho_0 * initial_rho_rel_min
  end if

  if (initial_atmosphere_factor > 0.0) then
    Dust_Atmosphere = Dust_Atmosphere * initial_atmosphere_factor
  end if

  if (dust_analysis_flag == 0) then ! for initial data

    kxx = 0.d0
    kxy = 0.d0
    kxz = 0.d0
    kyy = 0.d0
    kyz = 0.d0
    kzz = 0.d0
    gxy = 0.d0
    gxz = 0.d0
    gyz = 0.d0

    press = 0.d0
    eps = 0.d0

  end if

  do i = 1, cctk_lsh(1)
    do j = 1, cctk_lsh(2)
      do k = 1, cctk_lsh(3)
        
        rho_point = dust_rho_0

        call Interp_Dust_Isotropic(Dust_Num_Radial, Dust_Atmosphere, dust_a_0,&
             dust_alpha, dust_A, dust_beta, dust_big_T, dust_eta, &
             dust_small_r, r(i,j,k), dust_alpha_point, dust_A_point, &
             dust_beta_point, dust_vel_r, rho_point, i_surface, Dust_Mass)

        if (dust_analysis_flag == 1) then ! for analysis
          !minimal set of comparison variables
          dust_alp(i,j,k) = abs(dust_alpha_point) ! (2.7)
          dust_betax(i,j,k) = dust_beta_point*x(i,j,k)/r(i,j,k) 
          dust_gxx(i,j,k) = dust_A_point**2

          dust_rho(i,j,k) = rho_point

          dust_velx(i,j,k) = dust_vel_r*x(i,j,k)/r(i,j,k) !(6.5) + factor taking into account the position of
                                                          ! the point with respect to the axes
          dust_vely(i,j,k) = dust_vel_r*y(i,j,k)/r(i,j,k)
          dust_velz(i,j,k) = dust_vel_r*z(i,j,k)/r(i,j,k)
          dust_w_lorentz(i,j,k) = 1.d0/sqrt(1.d0-dust_gxx(i,j,k)*dust_vel_r**2)


         call SpatialDeterminant(dust_gxx(i,j,k), 0.d0, 0.d0,dust_gxx(i,j,k), 0.d0, dust_gxx(i,j,k),det)

!!$         call prim2conpolytype(whisky_eos_handle, dust_gxx, 0.d0, 0.d0, &
!!$             dust_gxx, 0.d0, dust_gxx, det, dust_dens(i,j,k), dust_sx(i,j,k),&
!!$             dust_sx(i,j,k),dust_sx(i,j,k),dust_tau(i,j,k),&
!!$             dust_rho(i,j,k),dust_velx(i,j,k),dust_vely(i,j,k),&
!!$             dust_velz(i,j,k), 0.d0, 0.d0, dust_w_lorentz(i,j,k)) 

         dust_dens(i,j,k) = sqrt(det) * dust_rho(i,j,k) * dust_w_lorentz(i,j,k) 
         dust_sx(i,j,k) = sqrt(det) * dust_rho(i,j,k) * dust_w_lorentz(i,j,k)**2 * &
              dust_gxx(i,j,k)*dust_velx(i,j,k) 
!!$         dust_sy(i,j,k)  = sqrt(det) * dust_rho(i,j,k) * dust_w_lorentz(i,j,k)**2 * &
!!$              dust_gxx(i,j,k)*dust_velx(i,j,k) 
!!$         dust_sz(i,j,k)  = sqrt(det) * dust_rho(i,j,k) * dust_w_lorentz(i,j,k)**2 * &
!!$              dust_gxx(i,j,k)*dust_velx(i,j,k) 
         dust_tau(i,j,k) = sqrt(det) * (dust_rho(i,j,k) * dust_w_lorentz(i,j,k) * &
              dust_w_lorentz(i,j,k)) - dust_dens(i,j,k)  

        else if (dust_analysis_flag == 0) then !for initail data
          alp(i,j,k) = abs(dust_alpha_point) ! (2.7)
!set shift exactly to zero, instead of the numerical zero from the ID routine (1.d-18 here)
          betax(i,j,k) = 0.d0
          betay(i,j,k) = 0.d0
          betaz(i,j,k) = 0.d0
          gxx(i,j,k) = dust_A_point**2
          gyy(i,j,k) = dust_A_point**2
          gzz(i,j,k) = dust_A_point**2

          rho(i,j,k) = rho_point
!set velocities exactly to zero, instead of the numerical zero from the ID routine
          velx(i,j,k) = 0.d0
          vely(i,j,k) = 0.d0
          velz(i,j,k) = 0.d0
          w_lorentz(i,j,k) = 1.d0

          call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k), &
               gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)
          call prim2conpolytype(whisky_eos_handle,gxx(i,j,k),gxy(i,j,k), &
               gxz(i,j,k),gyy(i,j,k),gyz(i,j,k),gzz(i,j,k), &
               det, dens(i,j,k),sx(i,j,k),sy(i,j,k),sz(i,j,k), &
               tau(i,j,k),rho(i,j,k),velx(i,j,k),vely(i,j,k),velz(i,j,k), &
               eps(i,j,k),press(i,j,k),w_lorentz(i,j,k))
        else
          write(infoline,'(a34)') "dust_analysis_flag has a not allowed value in Whisky_Dust"
          call CCTK_WARN(0,infoline)
        end if

        if (CCTK_EQUALS(Dust_Gauge,"Polar")) then

          alp(i,j,k) = abs(dust_alpha_point)

        end if

        ! shift is analytic for both gauges
        betax(i,j,k) = dust_beta_point*x(i,j,k)/r(i,j,k)
        betay(i,j,k) = dust_beta_point*y(i,j,k)/r(i,j,k) 
        betaz(i,j,k) = dust_beta_point*z(i,j,k)/r(i,j,k)

      end do
    end do
  end do

  call CartSymGN(ierr,cctkGH,"ADMBase::lapse")  
  call CartSymGN(ierr,cctkGH,"ADMBase::metric")  
  call CartSymGN(ierr,cctkGH,"ADMBase::curv")  
  call CartSymGN(ierr,cctkGH,"ADMBase::shift") 
  call CartSymGN(ierr,cctkGH,"Whisky::whisky_prim_var") 
  call CartSymGN(ierr,cctkGH,"Whisky::whisky_cons_var") 

  write(infoline,'(a24,g16.7)') "The dust star density is", rho(3,3,3)
  call CCTK_INFO(infoline)

  write(infoline,'(a19)') "Done interpolation."
  call CCTK_INFO(infoline)

end subroutine Whisky_Dust


/*@@
@routine    Interp_Dust_Isotropic
@date       Thu Aug  1 10:53:34 2002
@author     Ian Hawke, Luca Baiotti
@desc 
   Interpolates (linearly - hack) to find the isotropic radius and
   density at a given point.
@enddesc 
@calls     
@calledby   
@history 

@endhistory 
@var    
@vdesc  
@vtype  
@vio    
@vcomment 

@endvar 
@@*/

subroutine Interp_Dust_Isotropic(Dust_Num_Radial, Dust_Atmosphere, dust_a_0, &
     dust_alpha, dust_A, dust_beta, dust_big_T, dust_eta, dust_small_r, r, dust_alpha_point, &
     dust_A_point, dust_beta_point, dust_vel_r, rho_point, i_surface, Dust_Mass)

  implicit none

  CCTK_INT :: i, i_surface, Left_Point, Dust_Num_Radial
  CCTK_REAL, dimension(Dust_Num_Radial) ::  dust_alpha, dust_A, dust_beta, &
       dust_big_T, dust_eta, dust_small_r
  CCTK_REAL :: r, r_point, h, Dust_Atmosphere, dust_alpha_point, dust_A_point, &
       dust_beta_point, dust_big_T_point, dust_eta_point, dust_vel_r, rho_point, dust_a_0
  CCTK_REAL :: Pi, Dust_Mass

  Pi = acos(-1.d0)

  Left_Point = -1

  dust_alpha_point = 0.d0
  dust_beta_point = 0.d0
  dust_A_point = 0.d0
  dust_beta_point = 0.d0
  dust_big_T_point = 0.d0
  dust_eta_point = 0.d0
  dust_vel_r = 0.d0

  do i = 1, Dust_Num_Radial

    if (dust_small_r(i) > r) then
      Left_Point = i - 1
      exit
    end if

  end do

  if (Left_Point < 0) then
    call CCTK_WARN(0, "Grid not large enough.")
  else if (Left_Point .eq. 0) then
    call CCTK_WARN(1, "First point not found? Or avoid_origin = no")
    Left_Point = 1
  end if

  h = (r - dust_small_r(Left_Point))/(dust_small_r(Left_Point+1) - dust_small_r(Left_Point))

  dust_alpha_point = (1.d0 - h) * dust_alpha(Left_Point)+h * dust_alpha(Left_Point + 1)

  dust_A_point = (1.d0 - h) * dust_A(Left_Point)+h * dust_A(Left_Point + 1)

  dust_beta_point = (1.d0 - h) * dust_beta(Left_Point)+h * dust_beta(Left_Point + 1)

  dust_big_T_point = (1.d0 - h) * dust_big_T(Left_Point)+h * dust_big_T(Left_Point + 1)

  dust_eta_point = (1.d0 - h) * dust_eta(Left_Point)+h * dust_eta(Left_Point + 1) 

  dust_vel_r = - dust_beta_point - dust_alpha_point * dust_big_T_point / dust_A_point

!  rho_point = 48.d0*Pi*Dust_Mass**2/(dust_a_0**2*(1.d0-cos(dust_eta_point))**3*(1.d0-dust_big_T_point**2))  

  rho_point = 3.d0*Dust_Mass**2/(4.d0*Pi*dust_a_0**2*(1.d0-cos(dust_eta_point))**3*(1.d0-dust_big_T_point**2))

  if (i > i_surface) then

    rho_point = Dust_Atmosphere 
    dust_vel_r = 0.d0

  end if

end subroutine Interp_Dust_Isotropic

