 /*@@
   @file       Whisky_Dust_Evolve_Eta.F90
   @routine    Whisky_Dust_Evolve_Eta
   @date       Thu Sep 19 16:49:08 2002
   @author     Luca Baiotti
   @desc 
      Finds the value of dust_eta(1) at the proper time step.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 
@@*/

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"

subroutine Whisky_Dust_Evolve_Eta(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS

  CCTK_REAL :: Pi

  Pi = acos(-1.d0)

!second order advance

    if (dust_analysis_evolve == 1) then

      dust_eta(1) = 2.d0*deta_by_dsmall_t_tilde(1)*dust_alpha_c*CCTK_Delta_Time + dust_eta_pp
      dust_eta_pp = dust_eta_p
      dust_eta_p = dust_eta(1)

      dust_analysis_first_time = -1

    end if

end subroutine Whisky_Dust_Evolve_Eta


 /*@@
   @file       Whisky_Dust_Evolve_Eta.F90
   @routine    Whisky_Dust_Evolve_Eta_Polar
   @date       Tue Jan 14 11:01:40 2003
   @author     Luca Baiotti
   @desc 
      Finds the value of dust_eta(1) at the proper time step with numerical
      inversion of the analytic formula (II.3.68) of the second paper (1986, polar gauge).
   @enddesc 
   @calls    Dust_Root_Find  
   @calledby   
   @history 
 
   @endhistory 
@@*/

subroutine Whisky_Dust_Evolve_Eta_Polar(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS

  CCTK_REAL :: Pi, pass_time

  Pi = acos(-1.d0)

  pass_time = CCTK_TIME

  call Dust_Root_find(pass_time, dust_chi(i_surface), dust_eta(i_surface))

  dust_eta(1) = 2.d0*acos(cos(dust_eta(i_surface)/2.d0)*sqrt(cos(dust_chi(i_surface)))) ! (3.21)

  dust_analysis_first_time = -1

end subroutine Whisky_Dust_Evolve_Eta_Polar


 /*@@
   @file       Whisky_Dust_Evolve_Eta.F90
   @routine    Dust_Root_Find
   @date       Tue Jan 14 11:15:16 2003
   @author     Luca Baiotti
   @desc 
      Finds the numerical solution for eta_surface from (II.3.68) at a given time.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 
@@*/

subroutine Dust_Root_Find(t, dust_chi_on_surface, xmn)

  implicit none

  CCTK_INT :: m
  CCTK_REAL :: Pi, xmn, xmn1, argument_of_arctanh, dust_chi_on_surface, fmn, t
  character(len=100) :: infoline

  Pi = acos(-1.d0)

  xmn=-Pi !inital guess

  do m=1,1000

    xmn1=xmn

    argument_of_arctanh = tan(dust_chi_on_surface)*cos(xmn/2.d0)/sin(xmn/2.d0)

    ! arctanh x = 0.5*ln((1+x)/(1-x))
    xmn=xmn-&
         (xmn-sin(xmn)+2.d0*sin(dust_chi_on_surface)**2* &
         (xmn-tan(dust_chi_on_surface)* &
         log((1+argument_of_arctanh)/(1-argument_of_arctanh)))+&
         Pi*(1.d0+2.d0*sin(dust_chi_on_surface)**2) - &
         t * sin(dust_chi_on_surface)**3/cos(dust_chi_on_surface))/ &
         (2.d0*sin(xmn/2.d0)**4/(cos(dust_chi_on_surface)**2-cos(xmn/2.d0)**2)) 
    ! the last term is f/f' where f = (3.68) and f' = (3.67), 
    !I have gathered and simplified the overall factor (which so is only in front of t here)

    argument_of_arctanh = tan(dust_chi_on_surface)*cos(xmn/2.d0)/sin(xmn/2.d0)

    fmn=cos(dust_chi_on_surface)/sin(dust_chi_on_surface)**3* &
         (xmn-sin(xmn)+2.d0*sin(dust_chi_on_surface)**2* &
         (xmn-tan(dust_chi_on_surface)* &
         log((1+argument_of_arctanh)/(1-argument_of_arctanh)))+&
         Pi*(1.d0+2.d0*sin(dust_chi_on_surface)**2)) - t ! (3.68)

    if (fmn == 0.d0 .OR. abs(xmn-xmn1)<1.d-6) then
      exit
    end if

  end do

  if (m>=1000) then
    write(infoline,'(a34)') "exceeded 1000 iterations in root finding with Newton-Raphson without reaching the requested accuracy"
    call CCTK_WARN(0,infoline)
  end if

end subroutine Dust_Root_Find

