 /*@@
   @file      TOV.F
   @date      Thu May 23 01:59:34 2002
   @author    Ian Hawke
   @desc 
      TOV initial data. 
      This comes from the notes of Thomas Baumgarte (see Mexico 
      meeting page).
      Here I am arbitrarily enforcing a polytropic EOS, 
      P = K rho^Gamma, Gamma in [1,2].
      Note that the Baumgarte notes have the total energy density
      as rho; here rho_Baumgarte = rho(1+eps), eps the specific
      internal energy, found from the ideal gas EOS
      P = (Gamma - 1) rho eps. Instead we use mu = rho_Baumgarte.
   @enddesc 
   @version = $Header: /EUNetwork/EUHydro/Whisky_TOVSolver/src/TOV.F,v 1.23 2004/03/15 10:39:49 hawke Exp $
 @@*/

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"

/*@@
   @routine    TOV_Source_RHS
   @date       Thu May 23 02:30:04 2002
   @author     
   @desc 
      The source terms for the ODEs. These are equations (2), (3), (4)
      and (18) from the Baumgarte notes. 
      That is the vector in order is (P, m, phi, rbar).
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 
@@*/

       subroutine TOV_Source_RHS(r, K, Gamma, old_data, source_data)

       implicit none

       CCTK_REAL, dimension(4) :: old_data, source_data
       CCTK_REAL :: r, K, Gamma
       CCTK_REAL :: press, rho, eps, mu, m, phi
       CCTK_REAL :: log_rbar_over_r
       CCTK_REAL :: fourpi
       CCTK_REAL :: r_minus_two_m
       CCTK_REAL :: local_tiny

       local_tiny = 1.e-35
       fourpi = 16.d0 * atan(1.d0)

       press           = max(old_data(1),local_tiny)
       m               = old_data(2)
       phi             = old_data(3)
       log_rbar_over_r = old_data(4)

       rho = (press / K)**(1.d0 / Gamma)
       eps = press / (Gamma - 1.d0) / rho
       mu  = rho * (1.d0 + eps)

       r_minus_two_m = r - 2.d0 * m

       source_data(3) = (m + fourpi * r*r*r * press) / r_minus_two_m / r
!!$    skipped this simplification, because it introduces more error
!!$    than the (more compute-intensive) expression
!!$    source_data(1) = - (press + mu) * source_data(3)
       source_data(1) = - (press + mu) * (m + fourpi * r * r * r * press) /
     .                     r_minus_two_m / r
       source_data(2) = fourpi * r*r * mu
       source_data(4) = (sqrt(r)-sqrt(r_minus_two_m)) / r / sqrt(r_minus_two_m)

      end subroutine TOV_Source_RHS

 /*@@
   @routine    TOV_Integrate_RHS
   @date       Thu May 23 02:45:08 2002
   @author     Ian Hawke
   @desc 
      Integrates the ODEs using RK4.
      We rescale at the end to match to a Schwarzschild exterior.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 
@@*/

       subroutine TOV_Integrate_RHS(CCTK_ARGUMENTS)

       implicit none

       DECLARE_CCTK_ARGUMENTS
       DECLARE_CCTK_PARAMETERS

       CCTK_INT :: star, i, TOV_Surface_Index
       CCTK_REAL :: local_tiny
       CCTK_REAL, dimension(4) :: old_data, source_data, in_data, 
     .                           new_data, k1, k2, k3, k4
       CCTK_REAL :: Surface_Mass, TOV_Surface, factor, local_rho
       CCTK_REAL :: rho_central

       local_tiny = 1.d-20

       TOV_press_1d = 0.d0
       TOV_m_1d = 0.d0
       TOV_phi_1d = 0.d0
       TOV_rbar_1d = 0.d0
       TOV_r_1d = 0.d0

!!$    do it for all stars
       do star = 1, TOV_Num_TOVs

!!$    check for parameters
         if ((TOV_Rho_Central(star) .eq. 0.0d0) .and.
     .       (whisky_rho_central .gt. 0.0d0)) then
           rho_central=whisky_rho_central
         else
           rho_central=TOV_Rho_Central(star)
         end if

         if (rho_abs_min .gt. 0.d0) then
           TOV_Atmosphere(star) = rho_abs_min
         else
           TOV_Atmosphere(star) = rho_central * rho_rel_min
         end if

         if (initial_rho_abs_min .gt. 0.d0) then
           TOV_Atmosphere(star) = initial_rho_abs_min
         else
           if (initial_rho_rel_min .gt. 0.d0) then
             TOV_Atmosphere(star) = rho_central * initial_rho_rel_min
           end if
         end if

         if (initial_atmosphere_factor .gt. 0.0d0) then
           TOV_Atmosphere(star) = TOV_Atmosphere(star) *
     .                            initial_atmosphere_factor
         end if

!!$      set start values
         TOV_press_1d(1,star) = TOV_k(star) *
     .                           rho_central **TOV_Gamma(star)
         TOV_r_1d(1,star) = local_tiny
         TOV_rbar_1d(1,star) = local_tiny

!!$      build TOV_r_1d()
         do i = 2, TOV_Num_Radial
           TOV_r_1d(i,star) = TOV_r_1d(i-1,star) + TOV_dr(star)
         end do

         TOV_Surface = -1.d0

!!$      loop over all radii
         do i = 1, TOV_Num_Radial-1

           old_data(1) = TOV_press_1d(i,star)
           old_data(2) = TOV_m_1d(i,star)
           old_data(3) = TOV_phi_1d(i,star)
           old_data(4) = log(TOV_rbar_1d(i,star) / TOV_r_1d(i,star))

           in_data = old_data

           source_data = 0.d0

           call TOV_Source_RHS(TOV_r_1d(i,star),
     .                         TOV_K(star), TOV_Gamma(star),
     .                         in_data, source_data)
           k1 = TOV_dr(star) * source_data
           in_data = old_data + 0.5d0 * k1
           call TOV_Source_RHS(TOV_r_1d(i,star) + 0.5d0 * TOV_dr(star),
     .                         TOV_K(star), TOV_Gamma(star),
     .                         in_data, source_data)
           k2 = TOV_dr(star) * source_data
           in_data = old_data + 0.5d0 * k2
           call TOV_Source_RHS(TOV_r_1d(i,star) + 0.5d0 * TOV_dr(star),
     .                         TOV_K(star), TOV_Gamma(star),
     .                         in_data, source_data)
           k3 = TOV_dr(star) * source_data
           in_data = old_data + k3
           call TOV_Source_RHS(TOV_r_1d(i,star) + TOV_dr(star),
     .                         TOV_K(star), TOV_Gamma(star), 
     .                         in_data, source_data)
           k4 = TOV_dr(star) * source_data

           new_data = old_data + (k1 + k4 + 2.d0 * (k2 + k3)) / 6.d0

           TOV_press_1d(i+1,star) = new_data(1)
           TOV_m_1d    (i+1,star) = new_data(2)
           TOV_phi_1d  (i+1,star) = new_data(3)
           TOV_rbar_1d (i+1,star) = TOV_r_1d(i+1,star) *
     .                              exp(new_data(4))

!!$        otherwise the code may crash later
           if (TOV_press_1d(i+1,star) .lt. 0.0d0) then
               TOV_press_1d(i+1,star)=0.0d0
           endif

           local_rho = (TOV_press_1d(i+1,star) /
     .                  TOV_K(star))**(1.d0 / TOV_Gamma(star))

!!$        scan for the surface
           if ( (local_rho .le. TOV_Atmosphere(star)) .or.
     .          (TOV_press_1d(i+1,star) .le. 0.d0) ) then
             TOV_Surface = TOV_r_1d(i,star)
             TOV_R_Surface(star) = TOV_rbar_1d(i,star)
             TOV_Surface_Index = i
             exit
           end if
         end do

         if (TOV_Surface .lt. 0.d0) then
           call CCTK_WARN(0, 'Did not integrate out to surface of the star! Increase TOV_dr or TOV_Num_Radial and rerun')
         end if

         Surface_Mass = TOV_m_1d(TOV_Surface_Index,star)
         factor = 0.5d0*(sqrt(TOV_Surface*
     .                        (TOV_Surface - 2.d0*Surface_Mass))+
     .                   TOV_Surface - Surface_Mass) /
     .                   TOV_rbar_1d(TOV_Surface_Index,star)
 
         TOV_R_Surface(star) = TOV_R_Surface(star) * factor
         do i = 1, TOV_Num_Radial
           TOV_rbar_1d(i,star) = TOV_rbar_1d(i,star) * factor
           TOV_phi_1d(i,star)  = TOV_phi_1d(i,star) -
     .                           TOV_phi_1d(TOV_Surface_Index,star) + 
     .                           0.5d0 * log( 1.d0 - 2.d0 *
     .                                        Surface_Mass/TOV_Surface)
           if (i .gt. TOV_Surface_Index) then
             TOV_press_1d(i,star) = TOV_K(star)*
     .                              TOV_Atmosphere(star)**TOV_Gamma(star)
             TOV_rbar_1d (i,star) = 0.5d0 *
     .                              (sqrt(TOV_r_1d(i,star)*
     .                                    (TOV_r_1d(i,star)-
     .                                     2.d0 * Surface_Mass)) + 
     .                              TOV_r_1d(i,star) - Surface_Mass)
             TOV_m_1d(i,star)     = Surface_Mass
             TOV_phi_1d(i,star)   = 0.5d0 *
     .                              log((1.d0-2.d0*Surface_Mass /
     .                                  TOV_r_1d(i,star)))
           end if
         end do
       end do
       end subroutine TOV_Integrate_RHS

!!$ -------------------------------------------------------------------

/*@@
   @routine    Find_Index
   @date       Thu Oct 15 15:02:23 2002
   @author     Frank Loeffler
   @desc 
               - Finds Index i in TOV_rbar_1d with value fulfiling
                  TOV_rbar_1d(i) < r < TOV_rbar_1d(i+1)
               - i is returned as 'lower_index'
               - uses recursive 'part_in_two'-algorithm -> assumes
                 ordered TOV_rbar_1d
   @enddesc 
   @calls     
   @calledby   Interp_TOV_Isotropic
   @history 
 
   @endhistory 
   @var    
   @vdesc 
   @vtype  INT
   @vio    
   @vcomment 
 
   @endvar 
@@*/

       RECURSIVE subroutine Find_Index(array_size, array, goal,
     .                                 lower_index, upper_index)

       implicit none

       CCTK_INT :: array_size
       CCTK_REAL, dimension(array_size) :: array
       CCTK_REAL :: goal
       CCTK_INT :: lower_index, upper_index

       CCTK_INT :: middle_index

       if (lower_index .lt. (upper_index-1)) then

         middle_index = (lower_index + upper_index) /2;

         if (array(middle_index) .lt. goal) then
           call Find_Index(array_size, array, goal,
     .                     middle_index, upper_index)
           lower_index = middle_index
         else
           call Find_Index(array_size, array, goal,
     .                     lower_index, middle_index)
         endif

       endif

       end subroutine Find_Index

 /*@@
   @routine    Interp_TOV_Isotropic
   @date       Thu May 23 03:50:23 2002
   @author     Ian Hawke
   @desc 
      Interpolates to find the isotropic radius and
      density at a given point.
   @enddesc 
   @calls     Find_Index
   @calledby   
   @history 
 
   @endhistory 
   @var    :: r, rbar_point, press_point
   @vdesc  radius, isotropic radius and density at the point
   @vtype  REAL
   @vio    
   @vcomment 
 
   @endvar 
@@*/

       subroutine Interp_TOV_Isotropic(TOV_Num_Radial,
     .                                 TOV_press_1d, 
     .                                 TOV_phi_1d,
     .                                 TOV_rbar_1d,
     .                                 TOV_r_1d,
     .                                 r,
     .                                 press_point,
     .                                 phi_point, 
     .                                 r_point,
     .                                 TOV_Fast_Interpolation)

       implicit none
      
       CCTK_INT :: TOV_Num_Radial
       CCTK_REAL, dimension(TOV_Num_Radial) :: TOV_press_1d, 
     .                                         TOV_phi_1d,
     .                                         TOV_rbar_1d,
     .                                         TOV_r_1d
       CCTK_REAL :: r, phi_point, press_point, r_point
       CCTK_INT :: TOV_Fast_Interpolation

       CCTK_REAL :: h
       CCTK_INT :: i,Left_Index
       character*200 str
      
       if (r .lt. TOV_rbar_1d(1)) then
         r = TOV_rbar_1d(1) + 1.d-10
c$$$         call CCTK_WARN(0, 'First point not found.')
       end if
       if (r .gt. TOV_rbar_1d(TOV_Num_Radial)) then
         write(str, '(a36,g12.7)')
     .         'Grid not large enough - last point: ',
     .         TOV_rbar_1d(TOV_Num_Radial)
         call CCTK_WARN(0, str)
       end if

       if (TOV_Fast_Interpolation .ne. 0) then
!!$      fast routine
         Left_Index = 0
         call Find_Index(TOV_Num_Radial-1, TOV_rbar_1d, r, Left_Index,
     .                   TOV_Num_Radial-1)
       else
!!$      linear hack
         Left_Index = -1 
         do i = 2, TOV_Num_Radial
           if (TOV_rbar_1d(i) .gt. r) then
             Left_Index = i - 1
             exit
           end if
         end do
       endif

       h = (r - TOV_rbar_1d(Left_Index)) / 
     .     (TOV_rbar_1d(Left_Index+1) - TOV_rbar_1d(Left_Index))

       press_point = (1.d0 - h) * TOV_press_1d(Left_Index) + 
     .                       h  * TOV_press_1d(Left_Index + 1)
       phi_point   = (1.d0 - h) * TOV_phi_1d(Left_Index) + 
     .                       h  * TOV_phi_1d(Left_Index + 1)
       r_point     = (1.d0 - h) * TOV_r_1d(Left_Index) + 
     .                       h  * TOV_r_1d(Left_Index + 1)

      end subroutine Interp_TOV_Isotropic

/*@@
   @routine    TOV_Exact
   @date
   @author     Ian Hawke
   @desc
               Schedule routine for interpolation of 1D to 3D grid
   @enddesc 
   @calls     
   @calledby   
   @history 
   @endhistory 
@@*/ 
      
       subroutine TOV_Exact(CCTK_ARGUMENTS)

       implicit none

       DECLARE_CCTK_ARGUMENTS
       DECLARE_CCTK_PARAMETERS

       CCTK_INT :: i,j,k, star, star_i
       CCTK_REAL, dimension(TOV_Num_TOVs) :: press_point, rho_point,
     .                                       eps_point, mu_point, 
     .                                       phi_point, r_point,
     .                                       r_to_star
       CCTK_REAL :: g_diag, max_g_diag, max_rho, det
       CCTK_REAL :: D_h_w
       character*200 str
       CCTK_REAL :: PiByTwo
       CCTK_REAL :: local_tiny

       local_tiny = 1.e-14

       PiByTwo = 2.d0*atan(1.d0)
  
!!$    clear initial data
       if (TOV_Clear_Initial_Data .gt. 0) then
         kxx = 0.d0
         kxy = 0.d0
         kxz = 0.d0
         kyy = 0.d0
         kyz = 0.d0
         kzz = 0.d0
         gxy = 0.d0
         gxz = 0.d0
         gyz = 0.d0

         if (shift_state .ne. 0) then
           betax = 0.d0
           betay = 0.d0
           betaz = 0.d0
         end if

         if (conformal_state .ne. 0) then
           psi = 1.d0
           if (conformal_state .gt. 1) then
             psix = 0.d0
             psiy = 0.d0
             psiz = 0.d0
             if (conformal_state .gt. 2) then
               psixx = 0.d0
               psixy = 0.d0
               psixz = 0.d0
               psiyy = 0.d0
               psiyz = 0.d0
               psizz = 0.d0
             end if
           end if
         end if
       end if

!!$   do some info
       do i = 1, TOV_Num_TOVs
         write(str,'(a8,i2,a25,g12.7)')
     .         'The TOV ',i,' has a surface radius of ',TOV_R_Surface(i)
         call CCTK_INFO(str)
       end do
!!$    use the fast interpolation? only useful for testing this
       if (TOV_Fast_Interpolation .eq. 0 ) then
         write(str,'(a27)') 'Interpolating the slow way.'
         call CCTK_INFO('Interpolating the slow way.')
       endif

!!$    go over all 3D-grid points
       do i = 1, cctk_lsh(1)
         do j = 1, cctk_lsh(2)
           do k = 1, cctk_lsh(3)
             do star = 1, TOV_Num_TOVs
               r_to_star(star) =
     .           sqrt((x(i,j,k)-TOV_Position_x(star)) *
     .                (x(i,j,k)-TOV_Position_x(star)) +
     .                (y(i,j,k)-TOV_Position_y(star)) *
     .                (y(i,j,k)-TOV_Position_y(star)) +
     .                (z(i,j,k)-TOV_Position_z(star)) *
     .                (z(i,j,k)-TOV_Position_z(star)) )

!!$            do the actual interpolation
               call Interp_TOV_Isotropic(TOV_Num_Radial,
     .                                   TOV_press_1d(:,star),
     .                                   TOV_phi_1d(:,star),
     .                                   TOV_rbar_1d(:,star),
     .                                   TOV_r_1d(:,star),
     .                                   r_to_star(star), 
     .                                   press_point(star),
     .                                   phi_point(star),
     .                                   r_point(star),
     .                                   TOV_Fast_Interpolation)

!!$            is some perturbation wanted?
               if (Perturb(star) .eq. 0) then
                 rho_point(star) = (press_point(star) / TOV_K(star))**
     .                              (1.d0 / TOV_Gamma(star))
               else
                 rho_point(star) = (press_point(star) / TOV_K(star))**
     .                              (1.d0 / TOV_Gamma(star)) *
     .                             (1.d0 + 
     .                              Pert_Amplitude(star) *
     .                                cos(PiByTwo * r(i,j,k) /
     .                                TOV_R_Surface(star)))
               end if
               eps_point(star) = press_point(star) /
     .                           (TOV_Gamma(star) - 1.d0) /
     .                           rho_point(star)
               mu_point(star) = rho_point(star) * (1.d0+eps_point(star))
             end do
!!$          find out from which star we want to have the data
!!$          to do this, we use here simply the max of the gxx-value
             star=1
             max_g_diag = (r_point(1) / (r_to_star(1) + 1.0e-30)) *
     .                    (r_point(1) / (r_to_star(1) + 1.0e-30))
             max_rho = rho_point(1)
             do star_i = 1, TOV_Num_TOVs
               g_diag = (r_point(star_i) / (r_to_star(star_i) + 1.0e-30)) *
     .                  (r_point(star_i) / (r_to_star(star_i) + 1.0e-30))
               if ((g_diag - max_g_diag) .gt. local_tiny) then
                 max_g_diag=g_diag
                 star=star_i
               end if
               if ((rho_point(star_i) - max_rho) .gt. local_tiny) then
                 max_rho=rho_point(star_i)
                 star=star_i
               end if
             end do

             alp(i,j,k) = exp(phi_point(star))
             gxx(i,j,k) = max_g_diag
             gyy(i,j,k) = max_g_diag
             gzz(i,j,k) = max_g_diag

             velx(i,j,k) = TOV_Velocity_x(star)
             vely(i,j,k) = TOV_Velocity_y(star)
             velz(i,j,k) = TOV_Velocity_z(star)

             w_lorentz(i,j,k) = 1/sqrt(1.d0-
     .                             gxx(i,j,k) * velx(i,j,k) * velx(i,j,k)+
     .                             gyy(i,j,k) * vely(i,j,k) * vely(i,j,k)+
     .                             gzz(i,j,k) * velz(i,j,k) * velz(i,j,k)+
     .                           2*gxy(i,j,k) * velx(i,j,k) * vely(i,j,k)+
     .                           2*gxz(i,j,k) * velx(i,j,k) * velz(i,j,k)+
     .                           2*gyz(i,j,k) * vely(i,j,k) * velz(i,j,k))

            rho(i,j,k) = max_rho
            eps(i,j,k) = eps_point(star)
            press(i,j,k) = press_point(star)

            call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),
     .           gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)

            dens(i,j,k) = sqrt(det) * w_lorentz(i,j,k) * rho(i,j,k)
  
!!$         this variable is only for temporal storage
            D_h_w = dens(i,j,k) * (1 + eps(i,j,k) +
     .                             press(i,j,k)/rho(i,j,k)) *
     .                             w_lorentz(i,j,k)

            sx(i,j,k) = D_h_w * velx(i,j,k)
            sy(i,j,k) = D_h_w * vely(i,j,k)
            sz(i,j,k) = D_h_w * velz(i,j,k)

!!$         One could use D_h_w here, but it would introduce more error
            tau(i,j,k) = sqrt(det) *
     .                   ( rho(i,j,k)  *
     .                     w_lorentz(i,j,k)*w_lorentz(i,j,k) *
     .                     (1.0 + eps(i,j,k)) +
     .                     press(i,j,k)*
     .                      (w_lorentz(i,j,k)*w_lorentz(i,j,k)-1.0) ) -
     .                   dens(i,j,k);

            if (evolve_tracer .ne. 0) then
              tracer(i,j,k) = r(i,j,k) - TOV_R_Surface(1)
              if (rho(i,j,k) < whisky_rho_min) then
                cons_tracer(i,j,k) = sqrt(det) * whisky_rho_min * tracer(i,j,k)
              else
                cons_tracer(i,j,k) = sqrt(det) * rho(i,j,k) * tracer(i,j,k)
              end if
            end if

           end do
         end do
       end do

!!$      write(infoline,'(a19)') 'Done interpolation.'
!!$      call CCTK_INFO(infoline)
       call CCTK_INFO('Done interpolation.')
       
       end subroutine TOV_Exact

